/* Generated by Cython 0.29.21 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "glypy._c.count_table",
        "sources": [
            "glypy/_c/count_table.pyx"
        ]
    },
    "module_name": "glypy._c.count_table"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__glypy___c__count_table
#define __PYX_HAVE_API__glypy___c__count_table
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "glypy/_c/count_table.pyx",
  "stringsource",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator;
struct __pyx_obj_5glypy_2_c_11count_table_CountTable;
struct __pyx_t_5glypy_2_c_11count_table_count_table_bin_cell;
struct __pyx_t_5glypy_2_c_11count_table_count_table_bin;
struct __pyx_t_5glypy_2_c_11count_table_count_table;
struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_pop;
struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_get;

/* "glypy/_c/count_table.pxd":3
 * from cpython cimport PyObject
 * 
 * cdef struct count_table_bin_cell:             # <<<<<<<<<<<<<<
 *     long value
 *     PyObject* key
 */
struct __pyx_t_5glypy_2_c_11count_table_count_table_bin_cell {
  long value;
  PyObject *key;
};

/* "glypy/_c/count_table.pxd":8
 * 
 * 
 * cdef struct count_table_bin:             # <<<<<<<<<<<<<<
 *     count_table_bin_cell* cells
 *     size_t size
 */
struct __pyx_t_5glypy_2_c_11count_table_count_table_bin {
  struct __pyx_t_5glypy_2_c_11count_table_count_table_bin_cell *cells;
  size_t size;
  size_t used;
};

/* "glypy/_c/count_table.pxd":14
 * 
 * 
 * cdef struct count_table:             # <<<<<<<<<<<<<<
 *     count_table_bin* bins
 *     size_t size
 */
struct __pyx_t_5glypy_2_c_11count_table_count_table {
  struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *bins;
  size_t size;
};

/* "glypy/_c/count_table.pxd":99
 *     cpdef clear(self)
 *     cpdef setdefault(self, key, value)
 *     cpdef long pop(self, object key, object default=?) except *             # <<<<<<<<<<<<<<
 *     cpdef get(self, object key, object default=?)
 * 
 */
struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_pop {
  int __pyx_n;
  PyObject *__pyx_default;
};

/* "glypy/_c/count_table.pxd":100
 *     cpdef setdefault(self, key, value)
 *     cpdef long pop(self, object key, object default=?) except *
 *     cpdef get(self, object key, object default=?)             # <<<<<<<<<<<<<<
 * 
 *     cdef void increment(self, object key, long value)
 */
struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_get {
  int __pyx_n;
  PyObject *__pyx_default;
};

/* "glypy/_c/count_table.pxd":51
 * 
 * 
 * cdef class CountTableIterator(object):             # <<<<<<<<<<<<<<
 *     cdef :
 *         count_table* table
 */
struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator {
  PyObject_HEAD
  struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTableIterator *__pyx_vtab;
  struct __pyx_t_5glypy_2_c_11count_table_count_table *table;
  size_t bin_index;
  size_t cell_index;
  PyObject *next_key;
  long next_value;
};


/* "glypy/_c/count_table.pxd":68
 * 
 * 
 * cdef class CountTable(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         count_table* table
 */
struct __pyx_obj_5glypy_2_c_11count_table_CountTable {
  PyObject_HEAD
  struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *__pyx_vtab;
  struct __pyx_t_5glypy_2_c_11count_table_count_table *table;
};



/* "glypy/_c/count_table.pyx":348
 * 
 * 
 * cdef class CountTableIterator(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, table):
 */

struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTableIterator {
  struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *(*_create)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *);
  void (*advance)(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *);
  int (*has_more)(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *);
  int (*_locate_next_value)(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *, PyObject **, long *);
  int (*get_next_value)(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *, PyObject **, long *);
};
static struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTableIterator *__pyx_vtabptr_5glypy_2_c_11count_table_CountTableIterator;


/* "glypy/_c/count_table.pyx":462
 * 
 * 
 * cdef class CountTable(object):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable {
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *(*_create)(void);
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *(*_create_from)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *);
  int (*_initialize_table)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *);
  void (*_update_from_dict)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *);
  void (*_update_from_count_table)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *);
  void (*_add_from)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *);
  void (*_subtract_from)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *);
  int (*_add_from_dict)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *);
  int (*_subtract_from_dict)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *);
  void (*_scale_by)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, long);
  int (*equal_to)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *);
  PyObject *(*_to_dict)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*update)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *(*copy)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*items)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*values)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*keys)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*clear)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*setdefault)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  long (*pop)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_pop *__pyx_optional_args);
  PyObject *(*get)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_get *__pyx_optional_args);
  void (*increment)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, long);
  void (*decrement)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, long);
  long (*getitem)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *);
  int (*setitem)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, long);
  long (*delitem)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *);
};
static struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *__pyx_vtabptr_5glypy_2_c_11count_table_CountTable;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_f_5glypy_2_c_11count_table_18CountTableIterator__create(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_table); /* proto*/
static int __pyx_f_5glypy_2_c_11count_table_18CountTableIterator_has_more(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self); /* proto*/
static int __pyx_f_5glypy_2_c_11count_table_18CountTableIterator__locate_next_value(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self, PyObject **__pyx_v_key, long *__pyx_v_value); /* proto*/
static int __pyx_f_5glypy_2_c_11count_table_18CountTableIterator_get_next_value(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self, PyObject **__pyx_v_key, long *__pyx_v_value); /* proto*/
static void __pyx_f_5glypy_2_c_11count_table_18CountTableIterator_advance(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self); /* proto*/
static struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_f_5glypy_2_c_11count_table_10CountTable__create(void); /* proto*/
static struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_f_5glypy_2_c_11count_table_10CountTable__create_from(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_table); /* proto*/
static int __pyx_f_5glypy_2_c_11count_table_10CountTable__initialize_table(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_update(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_5glypy_2_c_11count_table_10CountTable__update_from_dict(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_d); /* proto*/
static void __pyx_f_5glypy_2_c_11count_table_10CountTable__update_from_count_table(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_other); /* proto*/
static struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_f_5glypy_2_c_11count_table_10CountTable_copy(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_5glypy_2_c_11count_table_10CountTable__add_from(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_other); /* proto*/
static int __pyx_f_5glypy_2_c_11count_table_10CountTable__add_from_dict(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other); /* proto*/
static void __pyx_f_5glypy_2_c_11count_table_10CountTable__subtract_from(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_other); /* proto*/
static int __pyx_f_5glypy_2_c_11count_table_10CountTable__subtract_from_dict(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other); /* proto*/
static void __pyx_f_5glypy_2_c_11count_table_10CountTable__scale_by(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, long __pyx_v_value); /* proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable__to_dict(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5glypy_2_c_11count_table_10CountTable_equal_to(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_other); /* proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_keys(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_values(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_items(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_clear(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_setdefault(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value, int __pyx_skip_dispatch); /* proto*/
static long __pyx_f_5glypy_2_c_11count_table_10CountTable_getitem(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key); /* proto*/
static int __pyx_f_5glypy_2_c_11count_table_10CountTable_setitem(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, long __pyx_v_value); /* proto*/
static long __pyx_f_5glypy_2_c_11count_table_10CountTable_delitem(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key); /* proto*/
static void __pyx_f_5glypy_2_c_11count_table_10CountTable_increment(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, long __pyx_v_value); /* proto*/
static void __pyx_f_5glypy_2_c_11count_table_10CountTable_decrement(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, long __pyx_v_value); /* proto*/
static long __pyx_f_5glypy_2_c_11count_table_10CountTable_pop(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_pop *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_get(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_get *__pyx_optional_args); /* proto*/

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'cython' */

/* Module declarations from 'glypy._c.count_table' */
static PyTypeObject *__pyx_ptype_5glypy_2_c_11count_table_CountTableIterator = 0;
static PyTypeObject *__pyx_ptype_5glypy_2_c_11count_table_CountTable = 0;
static int __pyx_f_5glypy_2_c_11count_table_initialize_count_table_bin(struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *, size_t); /*proto*/
static void __pyx_f_5glypy_2_c_11count_table_free_count_table_bin(struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *); /*proto*/
static int __pyx_f_5glypy_2_c_11count_table_count_table_bin_append(struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *, PyObject *, long); /*proto*/
static int __pyx_f_5glypy_2_c_11count_table_count_table_bin_find(struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *, PyObject *, Py_ssize_t *); /*proto*/
static struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_f_5glypy_2_c_11count_table_make_count_table(size_t, size_t); /*proto*/
static void __pyx_f_5glypy_2_c_11count_table_free_count_table(struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
static int __pyx_f_5glypy_2_c_11count_table_count_table_find_bin(struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, Py_ssize_t *); /*proto*/
static int __pyx_f_5glypy_2_c_11count_table_count_table_put(struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, long); /*proto*/
static int __pyx_f_5glypy_2_c_11count_table_count_table_del(struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, long *); /*proto*/
static int __pyx_f_5glypy_2_c_11count_table_count_table_get(struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, long *); /*proto*/
static int __pyx_f_5glypy_2_c_11count_table_count_table_increment(struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, long); /*proto*/
static int __pyx_f_5glypy_2_c_11count_table_count_table_decrement(struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, long); /*proto*/
static Py_ssize_t __pyx_f_5glypy_2_c_11count_table_count_table_count(struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_count_table_keys(struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_count_table_values(struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_count_table_items(struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
static void __pyx_f_5glypy_2_c_11count_table_count_table_add(struct __pyx_t_5glypy_2_c_11count_table_count_table *, struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
static void __pyx_f_5glypy_2_c_11count_table_count_table_subtract(struct __pyx_t_5glypy_2_c_11count_table_count_table *, struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
static void __pyx_f_5glypy_2_c_11count_table_count_table_scale(struct __pyx_t_5glypy_2_c_11count_table_count_table *, long); /*proto*/
static void __pyx_f_5glypy_2_c_11count_table_count_table_update(struct __pyx_t_5glypy_2_c_11count_table_count_table *, struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
static void __pyx_f_5glypy_2_c_11count_table_count_table_clear(struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
static struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_f_5glypy_2_c_11count_table_count_table_copy(struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
static int __pyx_f_5glypy_2_c_11count_table_count_table_equals(struct __pyx_t_5glypy_2_c_11count_table_count_table *, struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_count_table_to_dict(struct __pyx_t_5glypy_2_c_11count_table_count_table *); /*proto*/
#define __Pyx_MODULE_NAME "glypy._c.count_table"
extern int __pyx_module_is_main_glypy___c__count_table;
int __pyx_module_is_main_glypy___c__count_table = 0;

/* Implementation of 'glypy._c.count_table' */
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_NotImplemented;
static const char __pyx_k_end[] = "end";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_pint[] = "pint";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_spam[] = "spam";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_table[] = "table";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_main_2[] = "main";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_string[] = "string";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_Mapping[] = "Mapping";
static const char __pyx_k_content[] = "content";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_pytable[] = "pytable";
static const char __pyx_k_to_dict[] = "_to_dict";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_register[] = "register";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_CountTable[] = "CountTable";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_setdefault[] = "setdefault";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_MutableMapping[] = "MutableMapping";
static const char __pyx_k_NotImplemented[] = "NotImplemented";
static const char __pyx_k_collections_abc[] = "collections.abc";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_r_is_not_hashable[] = "%r is not hashable";
static const char __pyx_k_CountTableIterator[] = "CountTableIterator";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_r_must_be_a_number[] = "%r must be a number";
static const char __pyx_k_glypy__c_count_table[] = "glypy._c.count_table";
static const char __pyx_k_glypy__c_count_table_pyx[] = "glypy/_c/count_table.pyx";
static const char __pyx_k_self___class_____name___content[] = "{self.__class__.__name__}({content})";
static const char __pyx_k_self_next_key_self_table_cannot[] = "self.next_key,self.table cannot be converted to a Python object for pickling";
static PyObject *__pyx_n_s_CountTable;
static PyObject *__pyx_n_s_CountTableIterator;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_Mapping;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_MutableMapping;
static PyObject *__pyx_n_s_NotImplemented;
static PyObject *__pyx_n_s_StopIteration;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_clear;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_collections_abc;
static PyObject *__pyx_n_s_content;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_default;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_glypy__c_count_table;
static PyObject *__pyx_kp_s_glypy__c_count_table_pyx;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_main_2;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_pint;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_pytable;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_kp_s_r_is_not_hashable;
static PyObject *__pyx_kp_s_r_must_be_a_number;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_register;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self___class_____name___content;
static PyObject *__pyx_kp_s_self_next_key_self_table_cannot;
static PyObject *__pyx_n_s_setdefault;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_spam;
static PyObject *__pyx_n_s_string;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_table;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_to_dict;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_val;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_values;
static int __pyx_pf_5glypy_2_c_11count_table_18CountTableIterator___init__(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self, PyObject *__pyx_v_table); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_2__iter__(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_4__next__(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_6_test(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5glypy_2_c_11count_table_10CountTable___init__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_2update(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_4copy(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_6__reduce__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static void __pyx_pf_5glypy_2_c_11count_table_10CountTable_8__dealloc__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_10__getitem__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_5glypy_2_c_11count_table_10CountTable_12__setitem__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5glypy_2_c_11count_table_10CountTable_14__delitem__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static Py_ssize_t __pyx_pf_5glypy_2_c_11count_table_10CountTable_16__len__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static int __pyx_pf_5glypy_2_c_11count_table_10CountTable_18__contains__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_20__iter__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_22__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_24__iadd__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_26__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_28__isub__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_30__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_32__imul__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_34__neg__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_36_to_dict(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_38__repr__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_40__eq__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_42__ne__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static Py_hash_t __pyx_pf_5glypy_2_c_11count_table_10CountTable_44__hash__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_46keys(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_48values(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_50items(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_52clear(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_54setdefault(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_56pop(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_58get(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_5glypy_2_c_11count_table_main(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_tp_new_5glypy_2_c_11count_table_CountTableIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5glypy_2_c_11count_table_CountTable(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_255_2;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_11;
static PyObject *__pyx_int_12;
static PyObject *__pyx_int_99;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_codeobj__4;
/* Late includes */

/* "glypy/_c/count_table.pyx":15
 * 
 * 
 * cdef int initialize_count_table_bin(count_table_bin* bin, size_t size):             # <<<<<<<<<<<<<<
 *     bin.cells = <count_table_bin_cell*>PyMem_Malloc(sizeof(count_table_bin_cell) * size)
 *     if bin.cells == NULL:
 */

static int __pyx_f_5glypy_2_c_11count_table_initialize_count_table_bin(struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *__pyx_v_bin, size_t __pyx_v_size) {
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  __Pyx_RefNannySetupContext("initialize_count_table_bin", 0);

  /* "glypy/_c/count_table.pyx":16
 * 
 * cdef int initialize_count_table_bin(count_table_bin* bin, size_t size):
 *     bin.cells = <count_table_bin_cell*>PyMem_Malloc(sizeof(count_table_bin_cell) * size)             # <<<<<<<<<<<<<<
 *     if bin.cells == NULL:
 *         return 1
 */
  __pyx_v_bin->cells = ((struct __pyx_t_5glypy_2_c_11count_table_count_table_bin_cell *)PyMem_Malloc(((sizeof(struct __pyx_t_5glypy_2_c_11count_table_count_table_bin_cell)) * __pyx_v_size)));

  /* "glypy/_c/count_table.pyx":17
 * cdef int initialize_count_table_bin(count_table_bin* bin, size_t size):
 *     bin.cells = <count_table_bin_cell*>PyMem_Malloc(sizeof(count_table_bin_cell) * size)
 *     if bin.cells == NULL:             # <<<<<<<<<<<<<<
 *         return 1
 *     for i in range(size):
 */
  __pyx_t_1 = ((__pyx_v_bin->cells == NULL) != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":18
 *     bin.cells = <count_table_bin_cell*>PyMem_Malloc(sizeof(count_table_bin_cell) * size)
 *     if bin.cells == NULL:
 *         return 1             # <<<<<<<<<<<<<<
 *     for i in range(size):
 *         bin.cells[i].key = NULL
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":17
 * cdef int initialize_count_table_bin(count_table_bin* bin, size_t size):
 *     bin.cells = <count_table_bin_cell*>PyMem_Malloc(sizeof(count_table_bin_cell) * size)
 *     if bin.cells == NULL:             # <<<<<<<<<<<<<<
 *         return 1
 *     for i in range(size):
 */
  }

  /* "glypy/_c/count_table.pyx":19
 *     if bin.cells == NULL:
 *         return 1
 *     for i in range(size):             # <<<<<<<<<<<<<<
 *         bin.cells[i].key = NULL
 *     bin.used = 0
 */
  __pyx_t_2 = __pyx_v_size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glypy/_c/count_table.pyx":20
 *         return 1
 *     for i in range(size):
 *         bin.cells[i].key = NULL             # <<<<<<<<<<<<<<
 *     bin.used = 0
 *     bin.size = size
 */
    (__pyx_v_bin->cells[__pyx_v_i]).key = NULL;
  }

  /* "glypy/_c/count_table.pyx":21
 *     for i in range(size):
 *         bin.cells[i].key = NULL
 *     bin.used = 0             # <<<<<<<<<<<<<<
 *     bin.size = size
 * 
 */
  __pyx_v_bin->used = 0;

  /* "glypy/_c/count_table.pyx":22
 *         bin.cells[i].key = NULL
 *     bin.used = 0
 *     bin.size = size             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_bin->size = __pyx_v_size;

  /* "glypy/_c/count_table.pyx":15
 * 
 * 
 * cdef int initialize_count_table_bin(count_table_bin* bin, size_t size):             # <<<<<<<<<<<<<<
 *     bin.cells = <count_table_bin_cell*>PyMem_Malloc(sizeof(count_table_bin_cell) * size)
 *     if bin.cells == NULL:
 */

  /* function exit code */
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":25
 * 
 * 
 * cdef void free_count_table_bin(count_table_bin* bin):             # <<<<<<<<<<<<<<
 *     for i in range(bin.used):
 *         if bin.cells[i].key != NULL:
 */

static void __pyx_f_5glypy_2_c_11count_table_free_count_table_bin(struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *__pyx_v_bin) {
  size_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("free_count_table_bin", 0);

  /* "glypy/_c/count_table.pyx":26
 * 
 * cdef void free_count_table_bin(count_table_bin* bin):
 *     for i in range(bin.used):             # <<<<<<<<<<<<<<
 *         if bin.cells[i].key != NULL:
 *             Py_XDECREF(bin.cells[i].key)
 */
  __pyx_t_1 = __pyx_v_bin->used;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":27
 * cdef void free_count_table_bin(count_table_bin* bin):
 *     for i in range(bin.used):
 *         if bin.cells[i].key != NULL:             # <<<<<<<<<<<<<<
 *             Py_XDECREF(bin.cells[i].key)
 *             bin.cells[i].key = NULL
 */
    __pyx_t_4 = (((__pyx_v_bin->cells[__pyx_v_i]).key != NULL) != 0);
    if (__pyx_t_4) {

      /* "glypy/_c/count_table.pyx":28
 *     for i in range(bin.used):
 *         if bin.cells[i].key != NULL:
 *             Py_XDECREF(bin.cells[i].key)             # <<<<<<<<<<<<<<
 *             bin.cells[i].key = NULL
 *     PyMem_Free(bin.cells)
 */
      Py_XDECREF((__pyx_v_bin->cells[__pyx_v_i]).key);

      /* "glypy/_c/count_table.pyx":29
 *         if bin.cells[i].key != NULL:
 *             Py_XDECREF(bin.cells[i].key)
 *             bin.cells[i].key = NULL             # <<<<<<<<<<<<<<
 *     PyMem_Free(bin.cells)
 * 
 */
      (__pyx_v_bin->cells[__pyx_v_i]).key = NULL;

      /* "glypy/_c/count_table.pyx":27
 * cdef void free_count_table_bin(count_table_bin* bin):
 *     for i in range(bin.used):
 *         if bin.cells[i].key != NULL:             # <<<<<<<<<<<<<<
 *             Py_XDECREF(bin.cells[i].key)
 *             bin.cells[i].key = NULL
 */
    }
  }

  /* "glypy/_c/count_table.pyx":30
 *             Py_XDECREF(bin.cells[i].key)
 *             bin.cells[i].key = NULL
 *     PyMem_Free(bin.cells)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  PyMem_Free(__pyx_v_bin->cells);

  /* "glypy/_c/count_table.pyx":25
 * 
 * 
 * cdef void free_count_table_bin(count_table_bin* bin):             # <<<<<<<<<<<<<<
 *     for i in range(bin.used):
 *         if bin.cells[i].key != NULL:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":33
 * 
 * 
 * cdef int count_table_bin_append(count_table_bin* bin, PyObject* key, long value):             # <<<<<<<<<<<<<<
 *     if bin.used == bin.size - 1:
 *         bin.cells = <count_table_bin_cell*>PyMem_Realloc(bin.cells, sizeof(count_table_bin_cell) * bin.size * 2)
 */

static int __pyx_f_5glypy_2_c_11count_table_count_table_bin_append(struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *__pyx_v_bin, PyObject *__pyx_v_key, long __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("count_table_bin_append", 0);

  /* "glypy/_c/count_table.pyx":34
 * 
 * cdef int count_table_bin_append(count_table_bin* bin, PyObject* key, long value):
 *     if bin.used == bin.size - 1:             # <<<<<<<<<<<<<<
 *         bin.cells = <count_table_bin_cell*>PyMem_Realloc(bin.cells, sizeof(count_table_bin_cell) * bin.size * 2)
 *         if bin.cells == NULL:
 */
  __pyx_t_1 = ((__pyx_v_bin->used == (__pyx_v_bin->size - 1)) != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":35
 * cdef int count_table_bin_append(count_table_bin* bin, PyObject* key, long value):
 *     if bin.used == bin.size - 1:
 *         bin.cells = <count_table_bin_cell*>PyMem_Realloc(bin.cells, sizeof(count_table_bin_cell) * bin.size * 2)             # <<<<<<<<<<<<<<
 *         if bin.cells == NULL:
 *             return 1
 */
    __pyx_v_bin->cells = ((struct __pyx_t_5glypy_2_c_11count_table_count_table_bin_cell *)PyMem_Realloc(__pyx_v_bin->cells, (((sizeof(struct __pyx_t_5glypy_2_c_11count_table_count_table_bin_cell)) * __pyx_v_bin->size) * 2)));

    /* "glypy/_c/count_table.pyx":36
 *     if bin.used == bin.size - 1:
 *         bin.cells = <count_table_bin_cell*>PyMem_Realloc(bin.cells, sizeof(count_table_bin_cell) * bin.size * 2)
 *         if bin.cells == NULL:             # <<<<<<<<<<<<<<
 *             return 1
 *         bin.size *= 2
 */
    __pyx_t_1 = ((__pyx_v_bin->cells == NULL) != 0);
    if (__pyx_t_1) {

      /* "glypy/_c/count_table.pyx":37
 *         bin.cells = <count_table_bin_cell*>PyMem_Realloc(bin.cells, sizeof(count_table_bin_cell) * bin.size * 2)
 *         if bin.cells == NULL:
 *             return 1             # <<<<<<<<<<<<<<
 *         bin.size *= 2
 *     Py_XINCREF(key)
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":36
 *     if bin.used == bin.size - 1:
 *         bin.cells = <count_table_bin_cell*>PyMem_Realloc(bin.cells, sizeof(count_table_bin_cell) * bin.size * 2)
 *         if bin.cells == NULL:             # <<<<<<<<<<<<<<
 *             return 1
 *         bin.size *= 2
 */
    }

    /* "glypy/_c/count_table.pyx":38
 *         if bin.cells == NULL:
 *             return 1
 *         bin.size *= 2             # <<<<<<<<<<<<<<
 *     Py_XINCREF(key)
 *     bin.cells[bin.used].key = key
 */
    __pyx_v_bin->size = (__pyx_v_bin->size * 2);

    /* "glypy/_c/count_table.pyx":34
 * 
 * cdef int count_table_bin_append(count_table_bin* bin, PyObject* key, long value):
 *     if bin.used == bin.size - 1:             # <<<<<<<<<<<<<<
 *         bin.cells = <count_table_bin_cell*>PyMem_Realloc(bin.cells, sizeof(count_table_bin_cell) * bin.size * 2)
 *         if bin.cells == NULL:
 */
  }

  /* "glypy/_c/count_table.pyx":39
 *             return 1
 *         bin.size *= 2
 *     Py_XINCREF(key)             # <<<<<<<<<<<<<<
 *     bin.cells[bin.used].key = key
 *     bin.cells[bin.used].value = value
 */
  Py_XINCREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":40
 *         bin.size *= 2
 *     Py_XINCREF(key)
 *     bin.cells[bin.used].key = key             # <<<<<<<<<<<<<<
 *     bin.cells[bin.used].value = value
 *     bin.used += 1
 */
  (__pyx_v_bin->cells[__pyx_v_bin->used]).key = __pyx_v_key;

  /* "glypy/_c/count_table.pyx":41
 *     Py_XINCREF(key)
 *     bin.cells[bin.used].key = key
 *     bin.cells[bin.used].value = value             # <<<<<<<<<<<<<<
 *     bin.used += 1
 *     return 0
 */
  (__pyx_v_bin->cells[__pyx_v_bin->used]).value = __pyx_v_value;

  /* "glypy/_c/count_table.pyx":42
 *     bin.cells[bin.used].key = key
 *     bin.cells[bin.used].value = value
 *     bin.used += 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  __pyx_v_bin->used = (__pyx_v_bin->used + 1);

  /* "glypy/_c/count_table.pyx":43
 *     bin.cells[bin.used].value = value
 *     bin.used += 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":33
 * 
 * 
 * cdef int count_table_bin_append(count_table_bin* bin, PyObject* key, long value):             # <<<<<<<<<<<<<<
 *     if bin.used == bin.size - 1:
 *         bin.cells = <count_table_bin_cell*>PyMem_Realloc(bin.cells, sizeof(count_table_bin_cell) * bin.size * 2)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":46
 * 
 * 
 * cdef int count_table_bin_find(count_table_bin* bin, PyObject* query, Py_ssize_t* cell_index):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object query_obj
 */

static int __pyx_f_5glypy_2_c_11count_table_count_table_bin_find(struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *__pyx_v_bin, PyObject *__pyx_v_query, Py_ssize_t *__pyx_v_cell_index) {
  PyObject *__pyx_v_query_obj = 0;
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_bin_find", 0);

  /* "glypy/_c/count_table.pyx":49
 *     cdef:
 *         object query_obj
 *     query_obj = <object>query             # <<<<<<<<<<<<<<
 *     for i in range(bin.used):
 *         if bin.cells[i].key == NULL:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_query);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_query_obj = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glypy/_c/count_table.pyx":50
 *         object query_obj
 *     query_obj = <object>query
 *     for i in range(bin.used):             # <<<<<<<<<<<<<<
 *         if bin.cells[i].key == NULL:
 *             continue
 */
  __pyx_t_2 = __pyx_v_bin->used;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glypy/_c/count_table.pyx":51
 *     query_obj = <object>query
 *     for i in range(bin.used):
 *         if bin.cells[i].key == NULL:             # <<<<<<<<<<<<<<
 *             continue
 *         Py_XINCREF(bin.cells[i].key)
 */
    __pyx_t_5 = (((__pyx_v_bin->cells[__pyx_v_i]).key == NULL) != 0);
    if (__pyx_t_5) {

      /* "glypy/_c/count_table.pyx":52
 *     for i in range(bin.used):
 *         if bin.cells[i].key == NULL:
 *             continue             # <<<<<<<<<<<<<<
 *         Py_XINCREF(bin.cells[i].key)
 *         if (bin.cells[i].key == (<PyObject*>query_obj)) or ((<object>bin.cells[i].key) == (query_obj)):
 */
      goto __pyx_L3_continue;

      /* "glypy/_c/count_table.pyx":51
 *     query_obj = <object>query
 *     for i in range(bin.used):
 *         if bin.cells[i].key == NULL:             # <<<<<<<<<<<<<<
 *             continue
 *         Py_XINCREF(bin.cells[i].key)
 */
    }

    /* "glypy/_c/count_table.pyx":53
 *         if bin.cells[i].key == NULL:
 *             continue
 *         Py_XINCREF(bin.cells[i].key)             # <<<<<<<<<<<<<<
 *         if (bin.cells[i].key == (<PyObject*>query_obj)) or ((<object>bin.cells[i].key) == (query_obj)):
 *             Py_XDECREF(bin.cells[i].key)
 */
    Py_XINCREF((__pyx_v_bin->cells[__pyx_v_i]).key);

    /* "glypy/_c/count_table.pyx":54
 *             continue
 *         Py_XINCREF(bin.cells[i].key)
 *         if (bin.cells[i].key == (<PyObject*>query_obj)) or ((<object>bin.cells[i].key) == (query_obj)):             # <<<<<<<<<<<<<<
 *             Py_XDECREF(bin.cells[i].key)
 *             cell_index[0] = i
 */
    __pyx_t_6 = (((__pyx_v_bin->cells[__pyx_v_i]).key == ((PyObject *)__pyx_v_query_obj)) != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_1 = PyObject_RichCompare(((PyObject *)(__pyx_v_bin->cells[__pyx_v_i]).key), __pyx_v_query_obj, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 54, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __pyx_t_6;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_5) {

      /* "glypy/_c/count_table.pyx":55
 *         Py_XINCREF(bin.cells[i].key)
 *         if (bin.cells[i].key == (<PyObject*>query_obj)) or ((<object>bin.cells[i].key) == (query_obj)):
 *             Py_XDECREF(bin.cells[i].key)             # <<<<<<<<<<<<<<
 *             cell_index[0] = i
 *             return 0
 */
      Py_XDECREF((__pyx_v_bin->cells[__pyx_v_i]).key);

      /* "glypy/_c/count_table.pyx":56
 *         if (bin.cells[i].key == (<PyObject*>query_obj)) or ((<object>bin.cells[i].key) == (query_obj)):
 *             Py_XDECREF(bin.cells[i].key)
 *             cell_index[0] = i             # <<<<<<<<<<<<<<
 *             return 0
 *         Py_XDECREF(bin.cells[i].key)
 */
      (__pyx_v_cell_index[0]) = __pyx_v_i;

      /* "glypy/_c/count_table.pyx":57
 *             Py_XDECREF(bin.cells[i].key)
 *             cell_index[0] = i
 *             return 0             # <<<<<<<<<<<<<<
 *         Py_XDECREF(bin.cells[i].key)
 *     cell_index[0] = -1
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":54
 *             continue
 *         Py_XINCREF(bin.cells[i].key)
 *         if (bin.cells[i].key == (<PyObject*>query_obj)) or ((<object>bin.cells[i].key) == (query_obj)):             # <<<<<<<<<<<<<<
 *             Py_XDECREF(bin.cells[i].key)
 *             cell_index[0] = i
 */
    }

    /* "glypy/_c/count_table.pyx":58
 *             cell_index[0] = i
 *             return 0
 *         Py_XDECREF(bin.cells[i].key)             # <<<<<<<<<<<<<<
 *     cell_index[0] = -1
 *     return 0
 */
    Py_XDECREF((__pyx_v_bin->cells[__pyx_v_i]).key);
    __pyx_L3_continue:;
  }

  /* "glypy/_c/count_table.pyx":59
 *             return 0
 *         Py_XDECREF(bin.cells[i].key)
 *     cell_index[0] = -1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  (__pyx_v_cell_index[0]) = -1L;

  /* "glypy/_c/count_table.pyx":60
 *         Py_XDECREF(bin.cells[i].key)
 *     cell_index[0] = -1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":46
 * 
 * 
 * cdef int count_table_bin_find(count_table_bin* bin, PyObject* query, Py_ssize_t* cell_index):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object query_obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("glypy._c.count_table.count_table_bin_find", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_query_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":63
 * 
 * 
 * cdef count_table* make_count_table(size_t table_size, size_t bin_size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

static struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_f_5glypy_2_c_11count_table_make_count_table(size_t __pyx_v_table_size, size_t __pyx_v_bin_size) {
  size_t __pyx_v_i;
  struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table;
  struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  __Pyx_RefNannySetupContext("make_count_table", 0);

  /* "glypy/_c/count_table.pyx":67
 *         size_t i, j
 *         count_table* table
 *     table = <count_table*>PyMem_Malloc(sizeof(count_table))             # <<<<<<<<<<<<<<
 *     if table == NULL:
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
 */
  __pyx_v_table = ((struct __pyx_t_5glypy_2_c_11count_table_count_table *)PyMem_Malloc((sizeof(struct __pyx_t_5glypy_2_c_11count_table_count_table))));

  /* "glypy/_c/count_table.pyx":68
 *         count_table* table
 *     table = <count_table*>PyMem_Malloc(sizeof(count_table))
 *     if table == NULL:             # <<<<<<<<<<<<<<
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
 *         return NULL
 */
  __pyx_t_1 = ((__pyx_v_table == NULL) != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":69
 *     table = <count_table*>PyMem_Malloc(sizeof(count_table))
 *     if table == NULL:
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")             # <<<<<<<<<<<<<<
 *         return NULL
 *     table.bins = <count_table_bin*>PyMem_Malloc(sizeof(count_table_bin) * table_size)
 */
    PyErr_SetString(__pyx_builtin_MemoryError, ((char *)"Could not allocate memory for count_table"));

    /* "glypy/_c/count_table.pyx":70
 *     if table == NULL:
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
 *         return NULL             # <<<<<<<<<<<<<<
 *     table.bins = <count_table_bin*>PyMem_Malloc(sizeof(count_table_bin) * table_size)
 *     if table.bins == NULL:
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":68
 *         count_table* table
 *     table = <count_table*>PyMem_Malloc(sizeof(count_table))
 *     if table == NULL:             # <<<<<<<<<<<<<<
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
 *         return NULL
 */
  }

  /* "glypy/_c/count_table.pyx":71
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
 *         return NULL
 *     table.bins = <count_table_bin*>PyMem_Malloc(sizeof(count_table_bin) * table_size)             # <<<<<<<<<<<<<<
 *     if table.bins == NULL:
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
 */
  __pyx_v_table->bins = ((struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *)PyMem_Malloc(((sizeof(struct __pyx_t_5glypy_2_c_11count_table_count_table_bin)) * __pyx_v_table_size)));

  /* "glypy/_c/count_table.pyx":72
 *         return NULL
 *     table.bins = <count_table_bin*>PyMem_Malloc(sizeof(count_table_bin) * table_size)
 *     if table.bins == NULL:             # <<<<<<<<<<<<<<
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
 *         return NULL
 */
  __pyx_t_1 = ((__pyx_v_table->bins == NULL) != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":73
 *     table.bins = <count_table_bin*>PyMem_Malloc(sizeof(count_table_bin) * table_size)
 *     if table.bins == NULL:
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")             # <<<<<<<<<<<<<<
 *         return NULL
 *     for i in range(table_size):
 */
    PyErr_SetString(__pyx_builtin_MemoryError, ((char *)"Could not allocate memory for count_table"));

    /* "glypy/_c/count_table.pyx":74
 *     if table.bins == NULL:
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
 *         return NULL             # <<<<<<<<<<<<<<
 *     for i in range(table_size):
 *         initialize_count_table_bin(&(table.bins[i]), bin_size)
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":72
 *         return NULL
 *     table.bins = <count_table_bin*>PyMem_Malloc(sizeof(count_table_bin) * table_size)
 *     if table.bins == NULL:             # <<<<<<<<<<<<<<
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
 *         return NULL
 */
  }

  /* "glypy/_c/count_table.pyx":75
 *         PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
 *         return NULL
 *     for i in range(table_size):             # <<<<<<<<<<<<<<
 *         initialize_count_table_bin(&(table.bins[i]), bin_size)
 *     table.size = table_size
 */
  __pyx_t_2 = __pyx_v_table_size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glypy/_c/count_table.pyx":76
 *         return NULL
 *     for i in range(table_size):
 *         initialize_count_table_bin(&(table.bins[i]), bin_size)             # <<<<<<<<<<<<<<
 *     table.size = table_size
 *     return table
 */
    (void)(__pyx_f_5glypy_2_c_11count_table_initialize_count_table_bin((&(__pyx_v_table->bins[__pyx_v_i])), __pyx_v_bin_size));
  }

  /* "glypy/_c/count_table.pyx":77
 *     for i in range(table_size):
 *         initialize_count_table_bin(&(table.bins[i]), bin_size)
 *     table.size = table_size             # <<<<<<<<<<<<<<
 *     return table
 * 
 */
  __pyx_v_table->size = __pyx_v_table_size;

  /* "glypy/_c/count_table.pyx":78
 *         initialize_count_table_bin(&(table.bins[i]), bin_size)
 *     table.size = table_size
 *     return table             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_table;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":63
 * 
 * 
 * cdef count_table* make_count_table(size_t table_size, size_t bin_size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":81
 * 
 * 
 * cdef void free_count_table(count_table* table):             # <<<<<<<<<<<<<<
 *     for i in range(table.size):
 *         free_count_table_bin(&table.bins[i])
 */

static void __pyx_f_5glypy_2_c_11count_table_free_count_table(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table) {
  size_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  __Pyx_RefNannySetupContext("free_count_table", 0);

  /* "glypy/_c/count_table.pyx":82
 * 
 * cdef void free_count_table(count_table* table):
 *     for i in range(table.size):             # <<<<<<<<<<<<<<
 *         free_count_table_bin(&table.bins[i])
 *     PyMem_Free(table.bins)
 */
  __pyx_t_1 = __pyx_v_table->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":83
 * cdef void free_count_table(count_table* table):
 *     for i in range(table.size):
 *         free_count_table_bin(&table.bins[i])             # <<<<<<<<<<<<<<
 *     PyMem_Free(table.bins)
 *     PyMem_Free(table)
 */
    __pyx_f_5glypy_2_c_11count_table_free_count_table_bin((&(__pyx_v_table->bins[__pyx_v_i])));
  }

  /* "glypy/_c/count_table.pyx":84
 *     for i in range(table.size):
 *         free_count_table_bin(&table.bins[i])
 *     PyMem_Free(table.bins)             # <<<<<<<<<<<<<<
 *     PyMem_Free(table)
 * 
 */
  PyMem_Free(__pyx_v_table->bins);

  /* "glypy/_c/count_table.pyx":85
 *         free_count_table_bin(&table.bins[i])
 *     PyMem_Free(table.bins)
 *     PyMem_Free(table)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  PyMem_Free(__pyx_v_table);

  /* "glypy/_c/count_table.pyx":81
 * 
 * 
 * cdef void free_count_table(count_table* table):             # <<<<<<<<<<<<<<
 *     for i in range(table.size):
 *         free_count_table_bin(&table.bins[i])
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":89
 * 
 * @cython.cdivision(True)
 * cdef int count_table_find_bin(count_table* table, PyObject* query, Py_ssize_t* bin_index) except 1:             # <<<<<<<<<<<<<<
 *     try:
 *         bin_index[0] = PyObject_Hash(<object>query) % table.size
 */

static int __pyx_f_5glypy_2_c_11count_table_count_table_find_bin(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table, PyObject *__pyx_v_query, Py_ssize_t *__pyx_v_bin_index) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  long __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  char *__pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_find_bin", 0);

  /* "glypy/_c/count_table.pyx":90
 * @cython.cdivision(True)
 * cdef int count_table_find_bin(count_table* table, PyObject* query, Py_ssize_t* bin_index) except 1:
 *     try:             # <<<<<<<<<<<<<<
 *         bin_index[0] = PyObject_Hash(<object>query) % table.size
 *     except TypeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "glypy/_c/count_table.pyx":91
 * cdef int count_table_find_bin(count_table* table, PyObject* query, Py_ssize_t* bin_index) except 1:
 *     try:
 *         bin_index[0] = PyObject_Hash(<object>query) % table.size             # <<<<<<<<<<<<<<
 *     except TypeError:
 *         PyErr_SetString(TypeError, "%r is not hashable" % (<object>query, ))
 */
      __pyx_t_4 = PyObject_Hash(((PyObject *)__pyx_v_query)); if (unlikely(__pyx_t_4 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 91, __pyx_L3_error)
      (__pyx_v_bin_index[0]) = (__pyx_t_4 % __pyx_v_table->size);

      /* "glypy/_c/count_table.pyx":90
 * @cython.cdivision(True)
 * cdef int count_table_find_bin(count_table* table, PyObject* query, Py_ssize_t* bin_index) except 1:
 *     try:             # <<<<<<<<<<<<<<
 *         bin_index[0] = PyObject_Hash(<object>query) % table.size
 *     except TypeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "glypy/_c/count_table.pyx":92
 *     try:
 *         bin_index[0] = PyObject_Hash(<object>query) % table.size
 *     except TypeError:             # <<<<<<<<<<<<<<
 *         PyErr_SetString(TypeError, "%r is not hashable" % (<object>query, ))
 *         return 1
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_5) {
      __Pyx_AddTraceback("glypy._c.count_table.count_table_find_bin", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 92, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "glypy/_c/count_table.pyx":93
 *         bin_index[0] = PyObject_Hash(<object>query) % table.size
 *     except TypeError:
 *         PyErr_SetString(TypeError, "%r is not hashable" % (<object>query, ))             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
      __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 93, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(((PyObject *)__pyx_v_query));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_query));
      PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_query));
      __pyx_t_10 = __Pyx_PyString_Format(__pyx_kp_s_r_is_not_hashable, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 93, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_11 = __Pyx_PyObject_AsWritableString(__pyx_t_10); if (unlikely((!__pyx_t_11) && PyErr_Occurred())) __PYX_ERR(0, 93, __pyx_L5_except_error)
      PyErr_SetString(__pyx_builtin_TypeError, __pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "glypy/_c/count_table.pyx":94
 *     except TypeError:
 *         PyErr_SetString(TypeError, "%r is not hashable" % (<object>query, ))
 *         return 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
      __pyx_r = 1;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "glypy/_c/count_table.pyx":90
 * @cython.cdivision(True)
 * cdef int count_table_find_bin(count_table* table, PyObject* query, Py_ssize_t* bin_index) except 1:
 *     try:             # <<<<<<<<<<<<<<
 *         bin_index[0] = PyObject_Hash(<object>query) % table.size
 *     except TypeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "glypy/_c/count_table.pyx":95
 *         PyErr_SetString(TypeError, "%r is not hashable" % (<object>query, ))
 *         return 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":89
 * 
 * @cython.cdivision(True)
 * cdef int count_table_find_bin(count_table* table, PyObject* query, Py_ssize_t* bin_index) except 1:             # <<<<<<<<<<<<<<
 *     try:
 *         bin_index[0] = PyObject_Hash(<object>query) % table.size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("glypy._c.count_table.count_table_find_bin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":98
 * 
 * 
 * cdef int count_table_put(count_table* table, PyObject* key, long value) except 1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t bin_index, cell_index
 */

static int __pyx_f_5glypy_2_c_11count_table_count_table_put(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table, PyObject *__pyx_v_key, long __pyx_v_value) {
  Py_ssize_t __pyx_v_bin_index;
  Py_ssize_t __pyx_v_cell_index;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_put", 0);

  /* "glypy/_c/count_table.pyx":102
 *         Py_ssize_t bin_index, cell_index
 *         int status
 *     if key == NULL:             # <<<<<<<<<<<<<<
 *         return 1
 *     Py_XINCREF(key)
 */
  __pyx_t_1 = ((__pyx_v_key == NULL) != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":103
 *         int status
 *     if key == NULL:
 *         return 1             # <<<<<<<<<<<<<<
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":102
 *         Py_ssize_t bin_index, cell_index
 *         int status
 *     if key == NULL:             # <<<<<<<<<<<<<<
 *         return 1
 *     Py_XINCREF(key)
 */
  }

  /* "glypy/_c/count_table.pyx":104
 *     if key == NULL:
 *         return 1
 *     Py_XINCREF(key)             # <<<<<<<<<<<<<<
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:
 */
  Py_XINCREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":105
 *         return 1
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)             # <<<<<<<<<<<<<<
 *     if status != 0:
 *         return 1
 */
  __pyx_t_2 = __pyx_f_5glypy_2_c_11count_table_count_table_find_bin(__pyx_v_table, __pyx_v_key, (&__pyx_v_bin_index)); if (unlikely(__pyx_t_2 == ((int)1))) __PYX_ERR(0, 105, __pyx_L1_error)
  __pyx_v_status = __pyx_t_2;

  /* "glypy/_c/count_table.pyx":106
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 */
  __pyx_t_1 = ((__pyx_v_status != 0) != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":107
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:
 *         return 1             # <<<<<<<<<<<<<<
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":106
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 */
  }

  /* "glypy/_c/count_table.pyx":108
 *     if status != 0:
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)             # <<<<<<<<<<<<<<
 *     if cell_index == -1:
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 */
  __pyx_v_status = __pyx_f_5glypy_2_c_11count_table_count_table_bin_find((&(__pyx_v_table->bins[__pyx_v_bin_index])), __pyx_v_key, (&__pyx_v_cell_index));

  /* "glypy/_c/count_table.pyx":109
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:             # <<<<<<<<<<<<<<
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 *         Py_XDECREF(key)
 */
  __pyx_t_1 = ((__pyx_v_cell_index == -1L) != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":110
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)             # <<<<<<<<<<<<<<
 *         Py_XDECREF(key)
 *         if status != 0:
 */
    __pyx_v_status = __pyx_f_5glypy_2_c_11count_table_count_table_bin_append((&(__pyx_v_table->bins[__pyx_v_bin_index])), __pyx_v_key, __pyx_v_value);

    /* "glypy/_c/count_table.pyx":111
 *     if cell_index == -1:
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 *         Py_XDECREF(key)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             return 1
 */
    Py_XDECREF(__pyx_v_key);

    /* "glypy/_c/count_table.pyx":112
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 *         Py_XDECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 1
 *     else:
 */
    __pyx_t_1 = ((__pyx_v_status != 0) != 0);
    if (__pyx_t_1) {

      /* "glypy/_c/count_table.pyx":113
 *         Py_XDECREF(key)
 *         if status != 0:
 *             return 1             # <<<<<<<<<<<<<<
 *     else:
 *         table.bins[bin_index].cells[cell_index].value = value
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":112
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 *         Py_XDECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 1
 *     else:
 */
    }

    /* "glypy/_c/count_table.pyx":109
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:             # <<<<<<<<<<<<<<
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 *         Py_XDECREF(key)
 */
    goto __pyx_L5;
  }

  /* "glypy/_c/count_table.pyx":115
 *             return 1
 *     else:
 *         table.bins[bin_index].cells[cell_index].value = value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(key)
 *     return 0
 */
  /*else*/ {
    ((__pyx_v_table->bins[__pyx_v_bin_index]).cells[__pyx_v_cell_index]).value = __pyx_v_value;

    /* "glypy/_c/count_table.pyx":116
 *     else:
 *         table.bins[bin_index].cells[cell_index].value = value
 *         Py_XDECREF(key)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    Py_XDECREF(__pyx_v_key);
  }
  __pyx_L5:;

  /* "glypy/_c/count_table.pyx":117
 *         table.bins[bin_index].cells[cell_index].value = value
 *         Py_XDECREF(key)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":98
 * 
 * 
 * cdef int count_table_put(count_table* table, PyObject* key, long value) except 1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t bin_index, cell_index
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glypy._c.count_table.count_table_put", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":120
 * 
 * 
 * cdef int count_table_del(count_table* table, PyObject* key, long* value) except 1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t bin_index, cell_index
 */

static int __pyx_f_5glypy_2_c_11count_table_count_table_del(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table, PyObject *__pyx_v_key, long *__pyx_v_value) {
  Py_ssize_t __pyx_v_bin_index;
  Py_ssize_t __pyx_v_cell_index;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  long __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_del", 0);

  /* "glypy/_c/count_table.pyx":124
 *         Py_ssize_t bin_index, cell_index
 *         int status
 *     Py_XINCREF(key)             # <<<<<<<<<<<<<<
 *     value[0] = 0
 *     status = count_table_find_bin(table, key, &bin_index)
 */
  Py_XINCREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":125
 *         int status
 *     Py_XINCREF(key)
 *     value[0] = 0             # <<<<<<<<<<<<<<
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:
 */
  (__pyx_v_value[0]) = 0;

  /* "glypy/_c/count_table.pyx":126
 *     Py_XINCREF(key)
 *     value[0] = 0
 *     status = count_table_find_bin(table, key, &bin_index)             # <<<<<<<<<<<<<<
 *     if status != 0:
 *         return 1
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_find_bin(__pyx_v_table, __pyx_v_key, (&__pyx_v_bin_index)); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 126, __pyx_L1_error)
  __pyx_v_status = __pyx_t_1;

  /* "glypy/_c/count_table.pyx":127
 *     value[0] = 0
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 */
  __pyx_t_2 = ((__pyx_v_status != 0) != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":128
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:
 *         return 1             # <<<<<<<<<<<<<<
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":127
 *     value[0] = 0
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 */
  }

  /* "glypy/_c/count_table.pyx":129
 *     if status != 0:
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)             # <<<<<<<<<<<<<<
 *     if cell_index == -1:
 *         value[0] = 0
 */
  __pyx_v_status = __pyx_f_5glypy_2_c_11count_table_count_table_bin_find((&(__pyx_v_table->bins[__pyx_v_bin_index])), __pyx_v_key, (&__pyx_v_cell_index));

  /* "glypy/_c/count_table.pyx":130
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:             # <<<<<<<<<<<<<<
 *         value[0] = 0
 *         Py_XDECREF(key)
 */
  __pyx_t_2 = ((__pyx_v_cell_index == -1L) != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":131
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:
 *         value[0] = 0             # <<<<<<<<<<<<<<
 *         Py_XDECREF(key)
 *         return 0
 */
    (__pyx_v_value[0]) = 0;

    /* "glypy/_c/count_table.pyx":132
 *     if cell_index == -1:
 *         value[0] = 0
 *         Py_XDECREF(key)             # <<<<<<<<<<<<<<
 *         return 0
 *     else:
 */
    Py_XDECREF(__pyx_v_key);

    /* "glypy/_c/count_table.pyx":133
 *         value[0] = 0
 *         Py_XDECREF(key)
 *         return 0             # <<<<<<<<<<<<<<
 *     else:
 *         value[0] = table.bins[bin_index].cells[cell_index].value
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":130
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:             # <<<<<<<<<<<<<<
 *         value[0] = 0
 *         Py_XDECREF(key)
 */
  }

  /* "glypy/_c/count_table.pyx":135
 *         return 0
 *     else:
 *         value[0] = table.bins[bin_index].cells[cell_index].value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(table.bins[bin_index].cells[cell_index].key)
 *         table.bins[bin_index].cells[cell_index].key = NULL
 */
  /*else*/ {
    __pyx_t_3 = ((__pyx_v_table->bins[__pyx_v_bin_index]).cells[__pyx_v_cell_index]).value;
    (__pyx_v_value[0]) = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":136
 *     else:
 *         value[0] = table.bins[bin_index].cells[cell_index].value
 *         Py_XDECREF(table.bins[bin_index].cells[cell_index].key)             # <<<<<<<<<<<<<<
 *         table.bins[bin_index].cells[cell_index].key = NULL
 *         table.bins[bin_index].cells[cell_index].value = 0
 */
    Py_XDECREF(((__pyx_v_table->bins[__pyx_v_bin_index]).cells[__pyx_v_cell_index]).key);

    /* "glypy/_c/count_table.pyx":137
 *         value[0] = table.bins[bin_index].cells[cell_index].value
 *         Py_XDECREF(table.bins[bin_index].cells[cell_index].key)
 *         table.bins[bin_index].cells[cell_index].key = NULL             # <<<<<<<<<<<<<<
 *         table.bins[bin_index].cells[cell_index].value = 0
 *         Py_XDECREF(key)
 */
    ((__pyx_v_table->bins[__pyx_v_bin_index]).cells[__pyx_v_cell_index]).key = NULL;

    /* "glypy/_c/count_table.pyx":138
 *         Py_XDECREF(table.bins[bin_index].cells[cell_index].key)
 *         table.bins[bin_index].cells[cell_index].key = NULL
 *         table.bins[bin_index].cells[cell_index].value = 0             # <<<<<<<<<<<<<<
 *         Py_XDECREF(key)
 *         return 0
 */
    ((__pyx_v_table->bins[__pyx_v_bin_index]).cells[__pyx_v_cell_index]).value = 0;

    /* "glypy/_c/count_table.pyx":139
 *         table.bins[bin_index].cells[cell_index].key = NULL
 *         table.bins[bin_index].cells[cell_index].value = 0
 *         Py_XDECREF(key)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    Py_XDECREF(__pyx_v_key);

    /* "glypy/_c/count_table.pyx":140
 *         table.bins[bin_index].cells[cell_index].value = 0
 *         Py_XDECREF(key)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "glypy/_c/count_table.pyx":120
 * 
 * 
 * cdef int count_table_del(count_table* table, PyObject* key, long* value) except 1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t bin_index, cell_index
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glypy._c.count_table.count_table_del", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":143
 * 
 * 
 * cdef int count_table_get(count_table* table, PyObject* key, long* value) except 1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t bin_index, cell_index
 */

static int __pyx_f_5glypy_2_c_11count_table_count_table_get(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table, PyObject *__pyx_v_key, long *__pyx_v_value) {
  Py_ssize_t __pyx_v_bin_index;
  Py_ssize_t __pyx_v_cell_index;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  long __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_get", 0);

  /* "glypy/_c/count_table.pyx":147
 *         Py_ssize_t bin_index, cell_index
 *         int status
 *     value[0] = 0             # <<<<<<<<<<<<<<
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)
 */
  (__pyx_v_value[0]) = 0;

  /* "glypy/_c/count_table.pyx":148
 *         int status
 *     value[0] = 0
 *     Py_XINCREF(key)             # <<<<<<<<<<<<<<
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:
 */
  Py_XINCREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":149
 *     value[0] = 0
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)             # <<<<<<<<<<<<<<
 *     if status != 0:
 *         return 1
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_find_bin(__pyx_v_table, __pyx_v_key, (&__pyx_v_bin_index)); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 149, __pyx_L1_error)
  __pyx_v_status = __pyx_t_1;

  /* "glypy/_c/count_table.pyx":150
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 */
  __pyx_t_2 = ((__pyx_v_status != 0) != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":151
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:
 *         return 1             # <<<<<<<<<<<<<<
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     Py_XDECREF(key)
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":150
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 */
  }

  /* "glypy/_c/count_table.pyx":152
 *     if status != 0:
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)             # <<<<<<<<<<<<<<
 *     Py_XDECREF(key)
 *     if cell_index == -1:
 */
  __pyx_v_status = __pyx_f_5glypy_2_c_11count_table_count_table_bin_find((&(__pyx_v_table->bins[__pyx_v_bin_index])), __pyx_v_key, (&__pyx_v_cell_index));

  /* "glypy/_c/count_table.pyx":153
 *         return 1
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     Py_XDECREF(key)             # <<<<<<<<<<<<<<
 *     if cell_index == -1:
 *         value[0] = 0
 */
  Py_XDECREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":154
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     Py_XDECREF(key)
 *     if cell_index == -1:             # <<<<<<<<<<<<<<
 *         value[0] = 0
 *         return 0
 */
  __pyx_t_2 = ((__pyx_v_cell_index == -1L) != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":155
 *     Py_XDECREF(key)
 *     if cell_index == -1:
 *         value[0] = 0             # <<<<<<<<<<<<<<
 *         return 0
 *     else:
 */
    (__pyx_v_value[0]) = 0;

    /* "glypy/_c/count_table.pyx":156
 *     if cell_index == -1:
 *         value[0] = 0
 *         return 0             # <<<<<<<<<<<<<<
 *     else:
 *         value[0] = table.bins[bin_index].cells[cell_index].value
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":154
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     Py_XDECREF(key)
 *     if cell_index == -1:             # <<<<<<<<<<<<<<
 *         value[0] = 0
 *         return 0
 */
  }

  /* "glypy/_c/count_table.pyx":158
 *         return 0
 *     else:
 *         value[0] = table.bins[bin_index].cells[cell_index].value             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  /*else*/ {
    __pyx_t_3 = ((__pyx_v_table->bins[__pyx_v_bin_index]).cells[__pyx_v_cell_index]).value;
    (__pyx_v_value[0]) = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":159
 *     else:
 *         value[0] = table.bins[bin_index].cells[cell_index].value
 *         return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "glypy/_c/count_table.pyx":143
 * 
 * 
 * cdef int count_table_get(count_table* table, PyObject* key, long* value) except 1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t bin_index, cell_index
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glypy._c.count_table.count_table_get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":162
 * 
 * 
 * cdef int count_table_increment(count_table* table, PyObject* key, long value):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t bin_index, cell_index
 */

static int __pyx_f_5glypy_2_c_11count_table_count_table_increment(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table, PyObject *__pyx_v_key, long __pyx_v_value) {
  Py_ssize_t __pyx_v_bin_index;
  Py_ssize_t __pyx_v_cell_index;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_increment", 0);

  /* "glypy/_c/count_table.pyx":166
 *         Py_ssize_t bin_index, cell_index
 *         int status
 *     Py_XINCREF(key)             # <<<<<<<<<<<<<<
 *     status = count_table_find_bin(table, key, &bin_index)
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 */
  Py_XINCREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":167
 *         int status
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)             # <<<<<<<<<<<<<<
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_find_bin(__pyx_v_table, __pyx_v_key, (&__pyx_v_bin_index)); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 167, __pyx_L1_error)
  __pyx_v_status = __pyx_t_1;

  /* "glypy/_c/count_table.pyx":168
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)             # <<<<<<<<<<<<<<
 *     if cell_index == -1:
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 */
  __pyx_v_status = __pyx_f_5glypy_2_c_11count_table_count_table_bin_find((&(__pyx_v_table->bins[__pyx_v_bin_index])), __pyx_v_key, (&__pyx_v_cell_index));

  /* "glypy/_c/count_table.pyx":169
 *     status = count_table_find_bin(table, key, &bin_index)
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:             # <<<<<<<<<<<<<<
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 *         Py_XDECREF(key)
 */
  __pyx_t_2 = ((__pyx_v_cell_index == -1L) != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":170
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)             # <<<<<<<<<<<<<<
 *         Py_XDECREF(key)
 *         if status != 0:
 */
    __pyx_v_status = __pyx_f_5glypy_2_c_11count_table_count_table_bin_append((&(__pyx_v_table->bins[__pyx_v_bin_index])), __pyx_v_key, __pyx_v_value);

    /* "glypy/_c/count_table.pyx":171
 *     if cell_index == -1:
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 *         Py_XDECREF(key)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             return 1
 */
    Py_XDECREF(__pyx_v_key);

    /* "glypy/_c/count_table.pyx":172
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 *         Py_XDECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 1
 *     else:
 */
    __pyx_t_2 = ((__pyx_v_status != 0) != 0);
    if (__pyx_t_2) {

      /* "glypy/_c/count_table.pyx":173
 *         Py_XDECREF(key)
 *         if status != 0:
 *             return 1             # <<<<<<<<<<<<<<
 *     else:
 *         table.bins[bin_index].cells[cell_index].value += value
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":172
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 *         Py_XDECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 1
 *     else:
 */
    }

    /* "glypy/_c/count_table.pyx":169
 *     status = count_table_find_bin(table, key, &bin_index)
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:             # <<<<<<<<<<<<<<
 *         status = count_table_bin_append(&table.bins[bin_index], key, value)
 *         Py_XDECREF(key)
 */
    goto __pyx_L3;
  }

  /* "glypy/_c/count_table.pyx":175
 *             return 1
 *     else:
 *         table.bins[bin_index].cells[cell_index].value += value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(key)
 *     return 0
 */
  /*else*/ {
    __pyx_t_3 = __pyx_v_bin_index;
    __pyx_t_4 = __pyx_v_cell_index;
    ((__pyx_v_table->bins[__pyx_t_3]).cells[__pyx_t_4]).value = (((__pyx_v_table->bins[__pyx_t_3]).cells[__pyx_t_4]).value + __pyx_v_value);

    /* "glypy/_c/count_table.pyx":176
 *     else:
 *         table.bins[bin_index].cells[cell_index].value += value
 *         Py_XDECREF(key)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    Py_XDECREF(__pyx_v_key);
  }
  __pyx_L3:;

  /* "glypy/_c/count_table.pyx":177
 *         table.bins[bin_index].cells[cell_index].value += value
 *         Py_XDECREF(key)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":162
 * 
 * 
 * cdef int count_table_increment(count_table* table, PyObject* key, long value):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t bin_index, cell_index
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("glypy._c.count_table.count_table_increment", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":180
 * 
 * 
 * cdef int count_table_decrement(count_table* table, PyObject* key, long value):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t bin_index, cell_index
 */

static int __pyx_f_5glypy_2_c_11count_table_count_table_decrement(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table, PyObject *__pyx_v_key, long __pyx_v_value) {
  Py_ssize_t __pyx_v_bin_index;
  Py_ssize_t __pyx_v_cell_index;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_decrement", 0);

  /* "glypy/_c/count_table.pyx":184
 *         Py_ssize_t bin_index, cell_index
 *         int status
 *     Py_XINCREF(key)             # <<<<<<<<<<<<<<
 *     status = count_table_find_bin(table, key, &bin_index)
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 */
  Py_XINCREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":185
 *         int status
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)             # <<<<<<<<<<<<<<
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_find_bin(__pyx_v_table, __pyx_v_key, (&__pyx_v_bin_index)); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 185, __pyx_L1_error)
  __pyx_v_status = __pyx_t_1;

  /* "glypy/_c/count_table.pyx":186
 *     Py_XINCREF(key)
 *     status = count_table_find_bin(table, key, &bin_index)
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)             # <<<<<<<<<<<<<<
 *     if cell_index == -1:
 *         status = count_table_bin_append(&table.bins[bin_index], key, -value)
 */
  __pyx_v_status = __pyx_f_5glypy_2_c_11count_table_count_table_bin_find((&(__pyx_v_table->bins[__pyx_v_bin_index])), __pyx_v_key, (&__pyx_v_cell_index));

  /* "glypy/_c/count_table.pyx":187
 *     status = count_table_find_bin(table, key, &bin_index)
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:             # <<<<<<<<<<<<<<
 *         status = count_table_bin_append(&table.bins[bin_index], key, -value)
 *         Py_XDECREF(key)
 */
  __pyx_t_2 = ((__pyx_v_cell_index == -1L) != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":188
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:
 *         status = count_table_bin_append(&table.bins[bin_index], key, -value)             # <<<<<<<<<<<<<<
 *         Py_XDECREF(key)
 *         if status != 0:
 */
    __pyx_v_status = __pyx_f_5glypy_2_c_11count_table_count_table_bin_append((&(__pyx_v_table->bins[__pyx_v_bin_index])), __pyx_v_key, (-__pyx_v_value));

    /* "glypy/_c/count_table.pyx":189
 *     if cell_index == -1:
 *         status = count_table_bin_append(&table.bins[bin_index], key, -value)
 *         Py_XDECREF(key)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             return 1
 */
    Py_XDECREF(__pyx_v_key);

    /* "glypy/_c/count_table.pyx":190
 *         status = count_table_bin_append(&table.bins[bin_index], key, -value)
 *         Py_XDECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 1
 *     else:
 */
    __pyx_t_2 = ((__pyx_v_status != 0) != 0);
    if (__pyx_t_2) {

      /* "glypy/_c/count_table.pyx":191
 *         Py_XDECREF(key)
 *         if status != 0:
 *             return 1             # <<<<<<<<<<<<<<
 *     else:
 *         table.bins[bin_index].cells[cell_index].value -= value
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":190
 *         status = count_table_bin_append(&table.bins[bin_index], key, -value)
 *         Py_XDECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 1
 *     else:
 */
    }

    /* "glypy/_c/count_table.pyx":187
 *     status = count_table_find_bin(table, key, &bin_index)
 *     status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
 *     if cell_index == -1:             # <<<<<<<<<<<<<<
 *         status = count_table_bin_append(&table.bins[bin_index], key, -value)
 *         Py_XDECREF(key)
 */
    goto __pyx_L3;
  }

  /* "glypy/_c/count_table.pyx":193
 *             return 1
 *     else:
 *         table.bins[bin_index].cells[cell_index].value -= value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(key)
 *     return 0
 */
  /*else*/ {
    __pyx_t_3 = __pyx_v_bin_index;
    __pyx_t_4 = __pyx_v_cell_index;
    ((__pyx_v_table->bins[__pyx_t_3]).cells[__pyx_t_4]).value = (((__pyx_v_table->bins[__pyx_t_3]).cells[__pyx_t_4]).value - __pyx_v_value);

    /* "glypy/_c/count_table.pyx":194
 *     else:
 *         table.bins[bin_index].cells[cell_index].value -= value
 *         Py_XDECREF(key)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    Py_XDECREF(__pyx_v_key);
  }
  __pyx_L3:;

  /* "glypy/_c/count_table.pyx":195
 *         table.bins[bin_index].cells[cell_index].value -= value
 *         Py_XDECREF(key)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":180
 * 
 * 
 * cdef int count_table_decrement(count_table* table, PyObject* key, long value):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t bin_index, cell_index
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("glypy._c.count_table.count_table_decrement", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":198
 * 
 * 
 * cdef Py_ssize_t count_table_count(count_table* table):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t count
 */

static Py_ssize_t __pyx_f_5glypy_2_c_11count_table_count_table_count(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table) {
  Py_ssize_t __pyx_v_count;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  __Pyx_RefNannySetupContext("count_table_count", 0);

  /* "glypy/_c/count_table.pyx":202
 *         Py_ssize_t count
 *         size_t i, j
 *     count = 0             # <<<<<<<<<<<<<<
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 */
  __pyx_v_count = 0;

  /* "glypy/_c/count_table.pyx":203
 *         size_t i, j
 *     count = 0
 *     for i in range(table.size):             # <<<<<<<<<<<<<<
 *         for j in range(table.bins[i].used):
 *             count += table.bins[i].cells[j].key != NULL
 */
  __pyx_t_1 = __pyx_v_table->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":204
 *     count = 0
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):             # <<<<<<<<<<<<<<
 *             count += table.bins[i].cells[j].key != NULL
 *     return count
 */
    __pyx_t_4 = (__pyx_v_table->bins[__pyx_v_i]).used;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "glypy/_c/count_table.pyx":205
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 *             count += table.bins[i].cells[j].key != NULL             # <<<<<<<<<<<<<<
 *     return count
 * 
 */
      __pyx_v_count = (__pyx_v_count + (((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL));
    }
  }

  /* "glypy/_c/count_table.pyx":206
 *         for j in range(table.bins[i].used):
 *             count += table.bins[i].cells[j].key != NULL
 *     return count             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_count;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":198
 * 
 * 
 * cdef Py_ssize_t count_table_count(count_table* table):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t count
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":209
 * 
 * 
 * cdef list count_table_keys(count_table* table):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list keys
 */

static PyObject *__pyx_f_5glypy_2_c_11count_table_count_table_keys(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table) {
  PyObject *__pyx_v_keys = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  CYTHON_UNUSED long __pyx_v_count;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_keys", 0);

  /* "glypy/_c/count_table.pyx":213
 *         list keys
 *         size_t i, j
 *     keys = []             # <<<<<<<<<<<<<<
 *     count = 0
 *     for i in range(table.size):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_keys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/_c/count_table.pyx":214
 *         size_t i, j
 *     keys = []
 *     count = 0             # <<<<<<<<<<<<<<
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 */
  __pyx_v_count = 0;

  /* "glypy/_c/count_table.pyx":215
 *     keys = []
 *     count = 0
 *     for i in range(table.size):             # <<<<<<<<<<<<<<
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:
 */
  __pyx_t_2 = __pyx_v_table->size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glypy/_c/count_table.pyx":216
 *     count = 0
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):             # <<<<<<<<<<<<<<
 *             if table.bins[i].cells[j].key != NULL:
 *                 keys.append(<object>table.bins[i].cells[j].key)
 */
    __pyx_t_5 = (__pyx_v_table->bins[__pyx_v_i]).used;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_j = __pyx_t_7;

      /* "glypy/_c/count_table.pyx":217
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 keys.append(<object>table.bins[i].cells[j].key)
 *     return keys
 */
      __pyx_t_8 = ((((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_8) {

        /* "glypy/_c/count_table.pyx":218
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:
 *                 keys.append(<object>table.bins[i].cells[j].key)             # <<<<<<<<<<<<<<
 *     return keys
 * 
 */
        __pyx_t_1 = ((PyObject *)((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).key);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_keys, __pyx_t_1); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "glypy/_c/count_table.pyx":217
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 keys.append(<object>table.bins[i].cells[j].key)
 *     return keys
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":219
 *             if table.bins[i].cells[j].key != NULL:
 *                 keys.append(<object>table.bins[i].cells[j].key)
 *     return keys             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_keys);
  __pyx_r = __pyx_v_keys;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":209
 * 
 * 
 * cdef list count_table_keys(count_table* table):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list keys
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.count_table_keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":222
 * 
 * 
 * cdef list count_table_values(count_table* table):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list values
 */

static PyObject *__pyx_f_5glypy_2_c_11count_table_count_table_values(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table) {
  PyObject *__pyx_v_values = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  CYTHON_UNUSED long __pyx_v_count;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_values", 0);

  /* "glypy/_c/count_table.pyx":226
 *         list values
 *         size_t i, j
 *     values = []             # <<<<<<<<<<<<<<
 *     count = 0
 *     for i in range(table.size):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_values = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/_c/count_table.pyx":227
 *         size_t i, j
 *     values = []
 *     count = 0             # <<<<<<<<<<<<<<
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 */
  __pyx_v_count = 0;

  /* "glypy/_c/count_table.pyx":228
 *     values = []
 *     count = 0
 *     for i in range(table.size):             # <<<<<<<<<<<<<<
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:
 */
  __pyx_t_2 = __pyx_v_table->size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glypy/_c/count_table.pyx":229
 *     count = 0
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):             # <<<<<<<<<<<<<<
 *             if table.bins[i].cells[j].key != NULL:
 *                 values.append(PyInt_FromLong(table.bins[i].cells[j].value))
 */
    __pyx_t_5 = (__pyx_v_table->bins[__pyx_v_i]).used;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_j = __pyx_t_7;

      /* "glypy/_c/count_table.pyx":230
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 values.append(PyInt_FromLong(table.bins[i].cells[j].value))
 *     return values
 */
      __pyx_t_8 = ((((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_8) {

        /* "glypy/_c/count_table.pyx":231
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:
 *                 values.append(PyInt_FromLong(table.bins[i].cells[j].value))             # <<<<<<<<<<<<<<
 *     return values
 * 
 */
        __pyx_t_1 = PyInt_FromLong(((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_1); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 231, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "glypy/_c/count_table.pyx":230
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 values.append(PyInt_FromLong(table.bins[i].cells[j].value))
 *     return values
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":232
 *             if table.bins[i].cells[j].key != NULL:
 *                 values.append(PyInt_FromLong(table.bins[i].cells[j].value))
 *     return values             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_values);
  __pyx_r = __pyx_v_values;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":222
 * 
 * 
 * cdef list count_table_values(count_table* table):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list values
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.count_table_values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":235
 * 
 * 
 * cdef list count_table_items(count_table* table):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list items
 */

static PyObject *__pyx_f_5glypy_2_c_11count_table_count_table_items(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table) {
  PyObject *__pyx_v_items = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_items", 0);

  /* "glypy/_c/count_table.pyx":239
 *         list items
 *         size_t i, j
 *     items = []             # <<<<<<<<<<<<<<
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_items = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/_c/count_table.pyx":240
 *         size_t i, j
 *     items = []
 *     for i in range(table.size):             # <<<<<<<<<<<<<<
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:
 */
  __pyx_t_2 = __pyx_v_table->size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glypy/_c/count_table.pyx":241
 *     items = []
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):             # <<<<<<<<<<<<<<
 *             if table.bins[i].cells[j].key != NULL:
 *                 items.append(
 */
    __pyx_t_5 = (__pyx_v_table->bins[__pyx_v_i]).used;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_j = __pyx_t_7;

      /* "glypy/_c/count_table.pyx":242
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 items.append(
 *                     (<object>table.bins[i].cells[j].key,
 */
      __pyx_t_8 = ((((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_8) {

        /* "glypy/_c/count_table.pyx":245
 *                 items.append(
 *                     (<object>table.bins[i].cells[j].key,
 *                      PyInt_FromLong(table.bins[i].cells[j].value)))             # <<<<<<<<<<<<<<
 *     return items
 * 
 */
        __pyx_t_1 = PyInt_FromLong(((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "glypy/_c/count_table.pyx":244
 *             if table.bins[i].cells[j].key != NULL:
 *                 items.append(
 *                     (<object>table.bins[i].cells[j].key,             # <<<<<<<<<<<<<<
 *                      PyInt_FromLong(table.bins[i].cells[j].value)))
 *     return items
 */
        __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 244, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF(((PyObject *)((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).key));
        __Pyx_GIVEREF(((PyObject *)((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).key));
        PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).key));
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "glypy/_c/count_table.pyx":243
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:
 *                 items.append(             # <<<<<<<<<<<<<<
 *                     (<object>table.bins[i].cells[j].key,
 *                      PyInt_FromLong(table.bins[i].cells[j].value)))
 */
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_9); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 243, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "glypy/_c/count_table.pyx":242
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 items.append(
 *                     (<object>table.bins[i].cells[j].key,
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":246
 *                     (<object>table.bins[i].cells[j].key,
 *                      PyInt_FromLong(table.bins[i].cells[j].value)))
 *     return items             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_items);
  __pyx_r = __pyx_v_items;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":235
 * 
 * 
 * cdef list count_table_items(count_table* table):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list items
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("glypy._c.count_table.count_table_items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":249
 * 
 * 
 * cdef void count_table_add(count_table* table_a, count_table* table_b):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

static void __pyx_f_5glypy_2_c_11count_table_count_table_add(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table_a, struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table_b) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  long __pyx_v_value;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_add", 0);

  /* "glypy/_c/count_table.pyx":253
 *         size_t i, j
 *         long value
 *     for i in range(table_b.size):             # <<<<<<<<<<<<<<
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:
 */
  __pyx_t_1 = __pyx_v_table_b->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":254
 *         long value
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):             # <<<<<<<<<<<<<<
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 */
    __pyx_t_4 = (__pyx_v_table_b->bins[__pyx_v_i]).used;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "glypy/_c/count_table.pyx":255
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 *                 count_table_increment(table_a, table_b.bins[i].cells[j].key, value)
 */
      __pyx_t_7 = ((((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_7) {

        /* "glypy/_c/count_table.pyx":256
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)             # <<<<<<<<<<<<<<
 *                 count_table_increment(table_a, table_b.bins[i].cells[j].key, value)
 * 
 */
        __pyx_t_8 = __pyx_f_5glypy_2_c_11count_table_count_table_get(__pyx_v_table_b, ((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).key, (&__pyx_v_value)); if (unlikely(__pyx_t_8 == ((int)1))) __PYX_ERR(0, 256, __pyx_L1_error)

        /* "glypy/_c/count_table.pyx":257
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 *                 count_table_increment(table_a, table_b.bins[i].cells[j].key, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        (void)(__pyx_f_5glypy_2_c_11count_table_count_table_increment(__pyx_v_table_a, ((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).key, __pyx_v_value));

        /* "glypy/_c/count_table.pyx":255
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 *                 count_table_increment(table_a, table_b.bins[i].cells[j].key, value)
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":249
 * 
 * 
 * cdef void count_table_add(count_table* table_a, count_table* table_b):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("glypy._c.count_table.count_table_add", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":260
 * 
 * 
 * cdef void count_table_subtract(count_table* table_a, count_table* table_b):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

static void __pyx_f_5glypy_2_c_11count_table_count_table_subtract(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table_a, struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table_b) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  long __pyx_v_value;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_subtract", 0);

  /* "glypy/_c/count_table.pyx":264
 *         size_t i, j
 *         long value, temp
 *     for i in range(table_b.size):             # <<<<<<<<<<<<<<
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:
 */
  __pyx_t_1 = __pyx_v_table_b->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":265
 *         long value, temp
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):             # <<<<<<<<<<<<<<
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 */
    __pyx_t_4 = (__pyx_v_table_b->bins[__pyx_v_i]).used;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "glypy/_c/count_table.pyx":266
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 *                 count_table_decrement(table_a, table_b.bins[i].cells[j].key, value)
 */
      __pyx_t_7 = ((((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_7) {

        /* "glypy/_c/count_table.pyx":267
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)             # <<<<<<<<<<<<<<
 *                 count_table_decrement(table_a, table_b.bins[i].cells[j].key, value)
 * 
 */
        __pyx_t_8 = __pyx_f_5glypy_2_c_11count_table_count_table_get(__pyx_v_table_b, ((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).key, (&__pyx_v_value)); if (unlikely(__pyx_t_8 == ((int)1))) __PYX_ERR(0, 267, __pyx_L1_error)

        /* "glypy/_c/count_table.pyx":268
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 *                 count_table_decrement(table_a, table_b.bins[i].cells[j].key, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        (void)(__pyx_f_5glypy_2_c_11count_table_count_table_decrement(__pyx_v_table_a, ((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).key, __pyx_v_value));

        /* "glypy/_c/count_table.pyx":266
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 *                 count_table_decrement(table_a, table_b.bins[i].cells[j].key, value)
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":260
 * 
 * 
 * cdef void count_table_subtract(count_table* table_a, count_table* table_b):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("glypy._c.count_table.count_table_subtract", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":271
 * 
 * 
 * cdef void count_table_scale(count_table* table, long value):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

static void __pyx_f_5glypy_2_c_11count_table_count_table_scale(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table, long __pyx_v_value) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  __Pyx_RefNannySetupContext("count_table_scale", 0);

  /* "glypy/_c/count_table.pyx":274
 *     cdef:
 *         size_t i, j
 *     for i in range(table.size):             # <<<<<<<<<<<<<<
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:
 */
  __pyx_t_1 = __pyx_v_table->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":275
 *         size_t i, j
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):             # <<<<<<<<<<<<<<
 *             if table.bins[i].cells[j].key != NULL:
 *                 table.bins[i].cells[j].value *= value
 */
    __pyx_t_4 = (__pyx_v_table->bins[__pyx_v_i]).used;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "glypy/_c/count_table.pyx":276
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 table.bins[i].cells[j].value *= value
 * 
 */
      __pyx_t_7 = ((((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_7) {

        /* "glypy/_c/count_table.pyx":277
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:
 *                 table.bins[i].cells[j].value *= value             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __pyx_t_8 = __pyx_v_i;
        __pyx_t_9 = __pyx_v_j;
        ((__pyx_v_table->bins[__pyx_t_8]).cells[__pyx_t_9]).value = (((__pyx_v_table->bins[__pyx_t_8]).cells[__pyx_t_9]).value * __pyx_v_value);

        /* "glypy/_c/count_table.pyx":276
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 *             if table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 table.bins[i].cells[j].value *= value
 * 
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":271
 * 
 * 
 * cdef void count_table_scale(count_table* table, long value):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":280
 * 
 * 
 * cdef void count_table_update(count_table* table_a, count_table* table_b):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

static void __pyx_f_5glypy_2_c_11count_table_count_table_update(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table_a, struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table_b) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  long __pyx_v_value;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_update", 0);

  /* "glypy/_c/count_table.pyx":284
 *         size_t i, j
 *         long value
 *     for i in range(table_b.size):             # <<<<<<<<<<<<<<
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:
 */
  __pyx_t_1 = __pyx_v_table_b->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":285
 *         long value
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):             # <<<<<<<<<<<<<<
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 */
    __pyx_t_4 = (__pyx_v_table_b->bins[__pyx_v_i]).used;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "glypy/_c/count_table.pyx":286
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 *                 count_table_put(table_a, table_b.bins[i].cells[j].key, value)
 */
      __pyx_t_7 = ((((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_7) {

        /* "glypy/_c/count_table.pyx":287
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)             # <<<<<<<<<<<<<<
 *                 count_table_put(table_a, table_b.bins[i].cells[j].key, value)
 * 
 */
        __pyx_t_8 = __pyx_f_5glypy_2_c_11count_table_count_table_get(__pyx_v_table_b, ((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).key, (&__pyx_v_value)); if (unlikely(__pyx_t_8 == ((int)1))) __PYX_ERR(0, 287, __pyx_L1_error)

        /* "glypy/_c/count_table.pyx":288
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 *                 count_table_put(table_a, table_b.bins[i].cells[j].key, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __pyx_t_8 = __pyx_f_5glypy_2_c_11count_table_count_table_put(__pyx_v_table_a, ((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).key, __pyx_v_value); if (unlikely(__pyx_t_8 == ((int)1))) __PYX_ERR(0, 288, __pyx_L1_error)

        /* "glypy/_c/count_table.pyx":286
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
 *                 count_table_put(table_a, table_b.bins[i].cells[j].key, value)
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":280
 * 
 * 
 * cdef void count_table_update(count_table* table_a, count_table* table_b):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("glypy._c.count_table.count_table_update", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":291
 * 
 * 
 * cdef void count_table_clear(count_table* table):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

static void __pyx_f_5glypy_2_c_11count_table_count_table_clear(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  __Pyx_RefNannySetupContext("count_table_clear", 0);

  /* "glypy/_c/count_table.pyx":294
 *     cdef:
 *         size_t i, j
 *     for i in range(table.size):             # <<<<<<<<<<<<<<
 *         for j in range(table.bins[i].used):
 *             Py_XDECREF(table.bins[i].cells[j].key)
 */
  __pyx_t_1 = __pyx_v_table->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":295
 *         size_t i, j
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):             # <<<<<<<<<<<<<<
 *             Py_XDECREF(table.bins[i].cells[j].key)
 *             table.bins[i].cells[j].key = NULL
 */
    __pyx_t_4 = (__pyx_v_table->bins[__pyx_v_i]).used;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "glypy/_c/count_table.pyx":296
 *     for i in range(table.size):
 *         for j in range(table.bins[i].used):
 *             Py_XDECREF(table.bins[i].cells[j].key)             # <<<<<<<<<<<<<<
 *             table.bins[i].cells[j].key = NULL
 *             table.bins[i].cells[j].value = 0
 */
      Py_XDECREF(((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).key);

      /* "glypy/_c/count_table.pyx":297
 *         for j in range(table.bins[i].used):
 *             Py_XDECREF(table.bins[i].cells[j].key)
 *             table.bins[i].cells[j].key = NULL             # <<<<<<<<<<<<<<
 *             table.bins[i].cells[j].value = 0
 * 
 */
      ((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).key = NULL;

      /* "glypy/_c/count_table.pyx":298
 *             Py_XDECREF(table.bins[i].cells[j].key)
 *             table.bins[i].cells[j].key = NULL
 *             table.bins[i].cells[j].value = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      ((__pyx_v_table->bins[__pyx_v_i]).cells[__pyx_v_j]).value = 0;
    }
  }

  /* "glypy/_c/count_table.pyx":291
 * 
 * 
 * cdef void count_table_clear(count_table* table):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":301
 * 
 * 
 * cdef count_table* count_table_copy(count_table* table_a):             # <<<<<<<<<<<<<<
 *     cdef:
 *         count_table* dup
 */

static struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_f_5glypy_2_c_11count_table_count_table_copy(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table_a) {
  struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_dup;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("count_table_copy", 0);

  /* "glypy/_c/count_table.pyx":305
 *         count_table* dup
 *         size_t i, j
 *     dup = make_count_table(table_a.size, 2)             # <<<<<<<<<<<<<<
 *     for i in range(table_a.size):
 *         for j in range(table_a.bins[i].used):
 */
  __pyx_v_dup = __pyx_f_5glypy_2_c_11count_table_make_count_table(__pyx_v_table_a->size, 2);

  /* "glypy/_c/count_table.pyx":306
 *         size_t i, j
 *     dup = make_count_table(table_a.size, 2)
 *     for i in range(table_a.size):             # <<<<<<<<<<<<<<
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:
 */
  __pyx_t_1 = __pyx_v_table_a->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":307
 *     dup = make_count_table(table_a.size, 2)
 *     for i in range(table_a.size):
 *         for j in range(table_a.bins[i].used):             # <<<<<<<<<<<<<<
 *             if table_a.bins[i].cells[j].key != NULL:
 *                 count_table_bin_append(
 */
    __pyx_t_4 = (__pyx_v_table_a->bins[__pyx_v_i]).used;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "glypy/_c/count_table.pyx":308
 *     for i in range(table_a.size):
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_bin_append(
 *                     &dup.bins[i], table_a.bins[i].cells[j].key,
 */
      __pyx_t_7 = ((((__pyx_v_table_a->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_7) {

        /* "glypy/_c/count_table.pyx":309
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:
 *                 count_table_bin_append(             # <<<<<<<<<<<<<<
 *                     &dup.bins[i], table_a.bins[i].cells[j].key,
 *                     table_a.bins[i].cells[j].value)
 */
        (void)(__pyx_f_5glypy_2_c_11count_table_count_table_bin_append((&(__pyx_v_dup->bins[__pyx_v_i])), ((__pyx_v_table_a->bins[__pyx_v_i]).cells[__pyx_v_j]).key, ((__pyx_v_table_a->bins[__pyx_v_i]).cells[__pyx_v_j]).value));

        /* "glypy/_c/count_table.pyx":308
 *     for i in range(table_a.size):
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_bin_append(
 *                     &dup.bins[i], table_a.bins[i].cells[j].key,
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":312
 *                     &dup.bins[i], table_a.bins[i].cells[j].key,
 *                     table_a.bins[i].cells[j].value)
 *     return dup             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_dup;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":301
 * 
 * 
 * cdef count_table* count_table_copy(count_table* table_a):             # <<<<<<<<<<<<<<
 *     cdef:
 *         count_table* dup
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":315
 * 
 * 
 * cdef dict count_table_to_dict(count_table* table_a):             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict dup
 */

static PyObject *__pyx_f_5glypy_2_c_11count_table_count_table_to_dict(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table_a) {
  PyObject *__pyx_v_dup = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_to_dict", 0);

  /* "glypy/_c/count_table.pyx":319
 *         dict dup
 *         size_t i, j
 *     dup = dict()             # <<<<<<<<<<<<<<
 *     for i in range(table_a.size):
 *         for j in range(table_a.bins[i].used):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_dup = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/_c/count_table.pyx":320
 *         size_t i, j
 *     dup = dict()
 *     for i in range(table_a.size):             # <<<<<<<<<<<<<<
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:
 */
  __pyx_t_2 = __pyx_v_table_a->size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glypy/_c/count_table.pyx":321
 *     dup = dict()
 *     for i in range(table_a.size):
 *         for j in range(table_a.bins[i].used):             # <<<<<<<<<<<<<<
 *             if table_a.bins[i].cells[j].key != NULL:
 *                 PyDict_SetItem(
 */
    __pyx_t_5 = (__pyx_v_table_a->bins[__pyx_v_i]).used;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_j = __pyx_t_7;

      /* "glypy/_c/count_table.pyx":322
 *     for i in range(table_a.size):
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 PyDict_SetItem(
 *                     dup, <object>table_a.bins[i].cells[j].key,
 */
      __pyx_t_8 = ((((__pyx_v_table_a->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_8) {

        /* "glypy/_c/count_table.pyx":324
 *             if table_a.bins[i].cells[j].key != NULL:
 *                 PyDict_SetItem(
 *                     dup, <object>table_a.bins[i].cells[j].key,             # <<<<<<<<<<<<<<
 *                     PyInt_FromLong(table_a.bins[i].cells[j].value))
 *     return dup
 */
        __pyx_t_1 = ((PyObject *)((__pyx_v_table_a->bins[__pyx_v_i]).cells[__pyx_v_j]).key);
        __Pyx_INCREF(__pyx_t_1);

        /* "glypy/_c/count_table.pyx":325
 *                 PyDict_SetItem(
 *                     dup, <object>table_a.bins[i].cells[j].key,
 *                     PyInt_FromLong(table_a.bins[i].cells[j].value))             # <<<<<<<<<<<<<<
 *     return dup
 * 
 */
        __pyx_t_9 = PyInt_FromLong(((__pyx_v_table_a->bins[__pyx_v_i]).cells[__pyx_v_j]).value); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 325, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);

        /* "glypy/_c/count_table.pyx":323
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:
 *                 PyDict_SetItem(             # <<<<<<<<<<<<<<
 *                     dup, <object>table_a.bins[i].cells[j].key,
 *                     PyInt_FromLong(table_a.bins[i].cells[j].value))
 */
        __pyx_t_10 = PyDict_SetItem(__pyx_v_dup, __pyx_t_1, __pyx_t_9); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 323, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "glypy/_c/count_table.pyx":322
 *     for i in range(table_a.size):
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 PyDict_SetItem(
 *                     dup, <object>table_a.bins[i].cells[j].key,
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":326
 *                     dup, <object>table_a.bins[i].cells[j].key,
 *                     PyInt_FromLong(table_a.bins[i].cells[j].value))
 *     return dup             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_dup);
  __pyx_r = __pyx_v_dup;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":315
 * 
 * 
 * cdef dict count_table_to_dict(count_table* table_a):             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict dup
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("glypy._c.count_table.count_table_to_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dup);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":329
 * 
 * 
 * cdef bint count_table_equals(count_table* table_a, count_table* table_b):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

static int __pyx_f_5glypy_2_c_11count_table_count_table_equals(struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table_a, struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table_b) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  long __pyx_v_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_table_equals", 0);

  /* "glypy/_c/count_table.pyx":333
 *         size_t i, j
 *         long value
 *     for i in range(table_a.size):             # <<<<<<<<<<<<<<
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:
 */
  __pyx_t_1 = __pyx_v_table_a->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":334
 *         long value
 *     for i in range(table_a.size):
 *         for j in range(table_a.bins[i].used):             # <<<<<<<<<<<<<<
 *             if table_a.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_a.bins[i].cells[j].key, &value)
 */
    __pyx_t_4 = (__pyx_v_table_a->bins[__pyx_v_i]).used;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "glypy/_c/count_table.pyx":335
 *     for i in range(table_a.size):
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_get(table_b, table_a.bins[i].cells[j].key, &value)
 *                 if table_a.bins[i].cells[j].value != value:
 */
      __pyx_t_7 = ((((__pyx_v_table_a->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_7) {

        /* "glypy/_c/count_table.pyx":336
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_a.bins[i].cells[j].key, &value)             # <<<<<<<<<<<<<<
 *                 if table_a.bins[i].cells[j].value != value:
 *                     return False
 */
        __pyx_t_8 = __pyx_f_5glypy_2_c_11count_table_count_table_get(__pyx_v_table_b, ((__pyx_v_table_a->bins[__pyx_v_i]).cells[__pyx_v_j]).key, (&__pyx_v_value)); if (unlikely(__pyx_t_8 == ((int)1))) __PYX_ERR(0, 336, __pyx_L1_error)

        /* "glypy/_c/count_table.pyx":337
 *             if table_a.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_a.bins[i].cells[j].key, &value)
 *                 if table_a.bins[i].cells[j].value != value:             # <<<<<<<<<<<<<<
 *                     return False
 *     for i in range(table_b.size):
 */
        __pyx_t_7 = ((((__pyx_v_table_a->bins[__pyx_v_i]).cells[__pyx_v_j]).value != __pyx_v_value) != 0);
        if (__pyx_t_7) {

          /* "glypy/_c/count_table.pyx":338
 *                 count_table_get(table_b, table_a.bins[i].cells[j].key, &value)
 *                 if table_a.bins[i].cells[j].value != value:
 *                     return False             # <<<<<<<<<<<<<<
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):
 */
          __pyx_r = 0;
          goto __pyx_L0;

          /* "glypy/_c/count_table.pyx":337
 *             if table_a.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_b, table_a.bins[i].cells[j].key, &value)
 *                 if table_a.bins[i].cells[j].value != value:             # <<<<<<<<<<<<<<
 *                     return False
 *     for i in range(table_b.size):
 */
        }

        /* "glypy/_c/count_table.pyx":335
 *     for i in range(table_a.size):
 *         for j in range(table_a.bins[i].used):
 *             if table_a.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_get(table_b, table_a.bins[i].cells[j].key, &value)
 *                 if table_a.bins[i].cells[j].value != value:
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":339
 *                 if table_a.bins[i].cells[j].value != value:
 *                     return False
 *     for i in range(table_b.size):             # <<<<<<<<<<<<<<
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:
 */
  __pyx_t_1 = __pyx_v_table_b->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":340
 *                     return False
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):             # <<<<<<<<<<<<<<
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_a, table_b.bins[i].cells[j].key, &value)
 */
    __pyx_t_4 = (__pyx_v_table_b->bins[__pyx_v_i]).used;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "glypy/_c/count_table.pyx":341
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_get(table_a, table_b.bins[i].cells[j].key, &value)
 *                 if table_b.bins[i].cells[j].value != value:
 */
      __pyx_t_7 = ((((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_7) {

        /* "glypy/_c/count_table.pyx":342
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_a, table_b.bins[i].cells[j].key, &value)             # <<<<<<<<<<<<<<
 *                 if table_b.bins[i].cells[j].value != value:
 *                     return False
 */
        __pyx_t_8 = __pyx_f_5glypy_2_c_11count_table_count_table_get(__pyx_v_table_a, ((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).key, (&__pyx_v_value)); if (unlikely(__pyx_t_8 == ((int)1))) __PYX_ERR(0, 342, __pyx_L1_error)

        /* "glypy/_c/count_table.pyx":343
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_a, table_b.bins[i].cells[j].key, &value)
 *                 if table_b.bins[i].cells[j].value != value:             # <<<<<<<<<<<<<<
 *                     return False
 *     return True
 */
        __pyx_t_7 = ((((__pyx_v_table_b->bins[__pyx_v_i]).cells[__pyx_v_j]).value != __pyx_v_value) != 0);
        if (__pyx_t_7) {

          /* "glypy/_c/count_table.pyx":344
 *                 count_table_get(table_a, table_b.bins[i].cells[j].key, &value)
 *                 if table_b.bins[i].cells[j].value != value:
 *                     return False             # <<<<<<<<<<<<<<
 *     return True
 * 
 */
          __pyx_r = 0;
          goto __pyx_L0;

          /* "glypy/_c/count_table.pyx":343
 *             if table_b.bins[i].cells[j].key != NULL:
 *                 count_table_get(table_a, table_b.bins[i].cells[j].key, &value)
 *                 if table_b.bins[i].cells[j].value != value:             # <<<<<<<<<<<<<<
 *                     return False
 *     return True
 */
        }

        /* "glypy/_c/count_table.pyx":341
 *     for i in range(table_b.size):
 *         for j in range(table_b.bins[i].used):
 *             if table_b.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                 count_table_get(table_a, table_b.bins[i].cells[j].key, &value)
 *                 if table_b.bins[i].cells[j].value != value:
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":345
 *                 if table_b.bins[i].cells[j].value != value:
 *                     return False
 *     return True             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":329
 * 
 * 
 * cdef bint count_table_equals(count_table* table_a, count_table* table_b):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, j
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("glypy._c.count_table.count_table_equals", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":350
 * cdef class CountTableIterator(object):
 * 
 *     def __init__(self, table):             # <<<<<<<<<<<<<<
 *         self.table = (<CountTable>table).table
 *         self.bin_index = 0
 */

/* Python wrapper */
static int __pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_table = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_table,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_table)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 350, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_table = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 350, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glypy._c.count_table.CountTableIterator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_18CountTableIterator___init__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self), __pyx_v_table);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5glypy_2_c_11count_table_18CountTableIterator___init__(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self, PyObject *__pyx_v_table) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_t_1;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "glypy/_c/count_table.pyx":351
 * 
 *     def __init__(self, table):
 *         self.table = (<CountTable>table).table             # <<<<<<<<<<<<<<
 *         self.bin_index = 0
 *         self.cell_index = 0
 */
  __pyx_t_1 = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_table)->table;
  __pyx_v_self->table = __pyx_t_1;

  /* "glypy/_c/count_table.pyx":352
 *     def __init__(self, table):
 *         self.table = (<CountTable>table).table
 *         self.bin_index = 0             # <<<<<<<<<<<<<<
 *         self.cell_index = 0
 *         self.next_key = NULL
 */
  __pyx_v_self->bin_index = 0;

  /* "glypy/_c/count_table.pyx":353
 *         self.table = (<CountTable>table).table
 *         self.bin_index = 0
 *         self.cell_index = 0             # <<<<<<<<<<<<<<
 *         self.next_key = NULL
 *         self.next_value = 0
 */
  __pyx_v_self->cell_index = 0;

  /* "glypy/_c/count_table.pyx":354
 *         self.bin_index = 0
 *         self.cell_index = 0
 *         self.next_key = NULL             # <<<<<<<<<<<<<<
 *         self.next_value = 0
 *         self.advance()
 */
  __pyx_v_self->next_key = NULL;

  /* "glypy/_c/count_table.pyx":355
 *         self.cell_index = 0
 *         self.next_key = NULL
 *         self.next_value = 0             # <<<<<<<<<<<<<<
 *         self.advance()
 * 
 */
  __pyx_v_self->next_value = 0;

  /* "glypy/_c/count_table.pyx":356
 *         self.next_key = NULL
 *         self.next_value = 0
 *         self.advance()             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self->__pyx_vtab)->advance(__pyx_v_self);

  /* "glypy/_c/count_table.pyx":350
 * cdef class CountTableIterator(object):
 * 
 *     def __init__(self, table):             # <<<<<<<<<<<<<<
 *         self.table = (<CountTable>table).table
 *         self.bin_index = 0
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":359
 * 
 *     @staticmethod
 *     cdef CountTableIterator _create(CountTable table):             # <<<<<<<<<<<<<<
 *         cdef CountTableIterator it = CountTableIterator.__new__(CountTableIterator)
 *         it.table = table.table
 */

static struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_f_5glypy_2_c_11count_table_18CountTableIterator__create(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_table) {
  struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_it = 0;
  struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "glypy/_c/count_table.pyx":360
 *     @staticmethod
 *     cdef CountTableIterator _create(CountTable table):
 *         cdef CountTableIterator it = CountTableIterator.__new__(CountTableIterator)             # <<<<<<<<<<<<<<
 *         it.table = table.table
 *         it.bin_index = 0
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_5glypy_2_c_11count_table_CountTableIterator(((PyTypeObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTableIterator), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_it = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/_c/count_table.pyx":361
 *     cdef CountTableIterator _create(CountTable table):
 *         cdef CountTableIterator it = CountTableIterator.__new__(CountTableIterator)
 *         it.table = table.table             # <<<<<<<<<<<<<<
 *         it.bin_index = 0
 *         it.cell_index = 0
 */
  __pyx_t_2 = __pyx_v_table->table;
  __pyx_v_it->table = __pyx_t_2;

  /* "glypy/_c/count_table.pyx":362
 *         cdef CountTableIterator it = CountTableIterator.__new__(CountTableIterator)
 *         it.table = table.table
 *         it.bin_index = 0             # <<<<<<<<<<<<<<
 *         it.cell_index = 0
 *         it.next_key = NULL
 */
  __pyx_v_it->bin_index = 0;

  /* "glypy/_c/count_table.pyx":363
 *         it.table = table.table
 *         it.bin_index = 0
 *         it.cell_index = 0             # <<<<<<<<<<<<<<
 *         it.next_key = NULL
 *         it.next_value = 0
 */
  __pyx_v_it->cell_index = 0;

  /* "glypy/_c/count_table.pyx":364
 *         it.bin_index = 0
 *         it.cell_index = 0
 *         it.next_key = NULL             # <<<<<<<<<<<<<<
 *         it.next_value = 0
 *         it.advance()
 */
  __pyx_v_it->next_key = NULL;

  /* "glypy/_c/count_table.pyx":365
 *         it.cell_index = 0
 *         it.next_key = NULL
 *         it.next_value = 0             # <<<<<<<<<<<<<<
 *         it.advance()
 *         return it
 */
  __pyx_v_it->next_value = 0;

  /* "glypy/_c/count_table.pyx":366
 *         it.next_key = NULL
 *         it.next_value = 0
 *         it.advance()             # <<<<<<<<<<<<<<
 *         return it
 * 
 */
  ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_it->__pyx_vtab)->advance(__pyx_v_it);

  /* "glypy/_c/count_table.pyx":367
 *         it.next_value = 0
 *         it.advance()
 *         return it             # <<<<<<<<<<<<<<
 * 
 *     cdef bint has_more(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_it));
  __pyx_r = __pyx_v_it;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":359
 * 
 *     @staticmethod
 *     cdef CountTableIterator _create(CountTable table):             # <<<<<<<<<<<<<<
 *         cdef CountTableIterator it = CountTableIterator.__new__(CountTableIterator)
 *         it.table = table.table
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTableIterator._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_it);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":369
 *         return it
 * 
 *     cdef bint has_more(self):             # <<<<<<<<<<<<<<
 *         return self.next_key != NULL
 * 
 */

static int __pyx_f_5glypy_2_c_11count_table_18CountTableIterator_has_more(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_more", 0);

  /* "glypy/_c/count_table.pyx":370
 * 
 *     cdef bint has_more(self):
 *         return self.next_key != NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef int _locate_next_value(self, PyObject** key, long* value):
 */
  __pyx_r = (__pyx_v_self->next_key != NULL);
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":369
 *         return it
 * 
 *     cdef bint has_more(self):             # <<<<<<<<<<<<<<
 *         return self.next_key != NULL
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":372
 *         return self.next_key != NULL
 * 
 *     cdef int _locate_next_value(self, PyObject** key, long* value):             # <<<<<<<<<<<<<<
 *         if not self.bin_index < self.table.size:
 *             key[0] = NULL
 */

static int __pyx_f_5glypy_2_c_11count_table_18CountTableIterator__locate_next_value(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self, PyObject **__pyx_v_key, long *__pyx_v_value) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5;
  long __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  __Pyx_RefNannySetupContext("_locate_next_value", 0);

  /* "glypy/_c/count_table.pyx":373
 * 
 *     cdef int _locate_next_value(self, PyObject** key, long* value):
 *         if not self.bin_index < self.table.size:             # <<<<<<<<<<<<<<
 *             key[0] = NULL
 *             return 1
 */
  __pyx_t_1 = ((!((__pyx_v_self->bin_index < __pyx_v_self->table->size) != 0)) != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":374
 *     cdef int _locate_next_value(self, PyObject** key, long* value):
 *         if not self.bin_index < self.table.size:
 *             key[0] = NULL             # <<<<<<<<<<<<<<
 *             return 1
 *         # start from the previous iteration's bin
 */
    (__pyx_v_key[0]) = NULL;

    /* "glypy/_c/count_table.pyx":375
 *         if not self.bin_index < self.table.size:
 *             key[0] = NULL
 *             return 1             # <<<<<<<<<<<<<<
 *         # start from the previous iteration's bin
 *         for i in range(self.cell_index, self.table.bins[self.bin_index].used):
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":373
 * 
 *     cdef int _locate_next_value(self, PyObject** key, long* value):
 *         if not self.bin_index < self.table.size:             # <<<<<<<<<<<<<<
 *             key[0] = NULL
 *             return 1
 */
  }

  /* "glypy/_c/count_table.pyx":377
 *             return 1
 *         # start from the previous iteration's bin
 *         for i in range(self.cell_index, self.table.bins[self.bin_index].used):             # <<<<<<<<<<<<<<
 *             self.cell_index = i
 *             if self.table.bins[self.bin_index].cells[i].key != NULL:
 */
  __pyx_t_2 = (__pyx_v_self->table->bins[__pyx_v_self->bin_index]).used;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = __pyx_v_self->cell_index; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glypy/_c/count_table.pyx":378
 *         # start from the previous iteration's bin
 *         for i in range(self.cell_index, self.table.bins[self.bin_index].used):
 *             self.cell_index = i             # <<<<<<<<<<<<<<
 *             if self.table.bins[self.bin_index].cells[i].key != NULL:
 *                 key[0] = self.table.bins[self.bin_index].cells[i].key
 */
    __pyx_v_self->cell_index = __pyx_v_i;

    /* "glypy/_c/count_table.pyx":379
 *         for i in range(self.cell_index, self.table.bins[self.bin_index].used):
 *             self.cell_index = i
 *             if self.table.bins[self.bin_index].cells[i].key != NULL:             # <<<<<<<<<<<<<<
 *                 key[0] = self.table.bins[self.bin_index].cells[i].key
 *                 value[0] = self.table.bins[self.bin_index].cells[i].value
 */
    __pyx_t_1 = ((((__pyx_v_self->table->bins[__pyx_v_self->bin_index]).cells[__pyx_v_i]).key != NULL) != 0);
    if (__pyx_t_1) {

      /* "glypy/_c/count_table.pyx":380
 *             self.cell_index = i
 *             if self.table.bins[self.bin_index].cells[i].key != NULL:
 *                 key[0] = self.table.bins[self.bin_index].cells[i].key             # <<<<<<<<<<<<<<
 *                 value[0] = self.table.bins[self.bin_index].cells[i].value
 *                 self.cell_index += 1
 */
      __pyx_t_5 = ((__pyx_v_self->table->bins[__pyx_v_self->bin_index]).cells[__pyx_v_i]).key;
      (__pyx_v_key[0]) = __pyx_t_5;

      /* "glypy/_c/count_table.pyx":381
 *             if self.table.bins[self.bin_index].cells[i].key != NULL:
 *                 key[0] = self.table.bins[self.bin_index].cells[i].key
 *                 value[0] = self.table.bins[self.bin_index].cells[i].value             # <<<<<<<<<<<<<<
 *                 self.cell_index += 1
 *                 return 0
 */
      __pyx_t_6 = ((__pyx_v_self->table->bins[__pyx_v_self->bin_index]).cells[__pyx_v_i]).value;
      (__pyx_v_value[0]) = __pyx_t_6;

      /* "glypy/_c/count_table.pyx":382
 *                 key[0] = self.table.bins[self.bin_index].cells[i].key
 *                 value[0] = self.table.bins[self.bin_index].cells[i].value
 *                 self.cell_index += 1             # <<<<<<<<<<<<<<
 *                 return 0
 *         # move onto the next bins
 */
      __pyx_v_self->cell_index = (__pyx_v_self->cell_index + 1);

      /* "glypy/_c/count_table.pyx":383
 *                 value[0] = self.table.bins[self.bin_index].cells[i].value
 *                 self.cell_index += 1
 *                 return 0             # <<<<<<<<<<<<<<
 *         # move onto the next bins
 *         for i in range(self.bin_index + 1, self.table.size):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":379
 *         for i in range(self.cell_index, self.table.bins[self.bin_index].used):
 *             self.cell_index = i
 *             if self.table.bins[self.bin_index].cells[i].key != NULL:             # <<<<<<<<<<<<<<
 *                 key[0] = self.table.bins[self.bin_index].cells[i].key
 *                 value[0] = self.table.bins[self.bin_index].cells[i].value
 */
    }
  }

  /* "glypy/_c/count_table.pyx":385
 *                 return 0
 *         # move onto the next bins
 *         for i in range(self.bin_index + 1, self.table.size):             # <<<<<<<<<<<<<<
 *             self.bin_index = i
 *             for j in range(0, self.table.bins[i].used):
 */
  __pyx_t_2 = __pyx_v_self->table->size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = (__pyx_v_self->bin_index + 1); __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glypy/_c/count_table.pyx":386
 *         # move onto the next bins
 *         for i in range(self.bin_index + 1, self.table.size):
 *             self.bin_index = i             # <<<<<<<<<<<<<<
 *             for j in range(0, self.table.bins[i].used):
 *                 self.cell_index = j
 */
    __pyx_v_self->bin_index = __pyx_v_i;

    /* "glypy/_c/count_table.pyx":387
 *         for i in range(self.bin_index + 1, self.table.size):
 *             self.bin_index = i
 *             for j in range(0, self.table.bins[i].used):             # <<<<<<<<<<<<<<
 *                 self.cell_index = j
 *                 if self.table.bins[i].cells[j].key != NULL:
 */
    __pyx_t_7 = (__pyx_v_self->table->bins[__pyx_v_i]).used;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_j = __pyx_t_9;

      /* "glypy/_c/count_table.pyx":388
 *             self.bin_index = i
 *             for j in range(0, self.table.bins[i].used):
 *                 self.cell_index = j             # <<<<<<<<<<<<<<
 *                 if self.table.bins[i].cells[j].key != NULL:
 *                     key[0] = self.table.bins[self.bin_index].cells[j].key
 */
      __pyx_v_self->cell_index = __pyx_v_j;

      /* "glypy/_c/count_table.pyx":389
 *             for j in range(0, self.table.bins[i].used):
 *                 self.cell_index = j
 *                 if self.table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                     key[0] = self.table.bins[self.bin_index].cells[j].key
 *                     value[0] = self.table.bins[self.bin_index].cells[j].value
 */
      __pyx_t_1 = ((((__pyx_v_self->table->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_1) {

        /* "glypy/_c/count_table.pyx":390
 *                 self.cell_index = j
 *                 if self.table.bins[i].cells[j].key != NULL:
 *                     key[0] = self.table.bins[self.bin_index].cells[j].key             # <<<<<<<<<<<<<<
 *                     value[0] = self.table.bins[self.bin_index].cells[j].value
 *                     self.cell_index += 1
 */
        __pyx_t_5 = ((__pyx_v_self->table->bins[__pyx_v_self->bin_index]).cells[__pyx_v_j]).key;
        (__pyx_v_key[0]) = __pyx_t_5;

        /* "glypy/_c/count_table.pyx":391
 *                 if self.table.bins[i].cells[j].key != NULL:
 *                     key[0] = self.table.bins[self.bin_index].cells[j].key
 *                     value[0] = self.table.bins[self.bin_index].cells[j].value             # <<<<<<<<<<<<<<
 *                     self.cell_index += 1
 *                     return 0
 */
        __pyx_t_6 = ((__pyx_v_self->table->bins[__pyx_v_self->bin_index]).cells[__pyx_v_j]).value;
        (__pyx_v_value[0]) = __pyx_t_6;

        /* "glypy/_c/count_table.pyx":392
 *                     key[0] = self.table.bins[self.bin_index].cells[j].key
 *                     value[0] = self.table.bins[self.bin_index].cells[j].value
 *                     self.cell_index += 1             # <<<<<<<<<<<<<<
 *                     return 0
 *         self.bin_index = self.table.size
 */
        __pyx_v_self->cell_index = (__pyx_v_self->cell_index + 1);

        /* "glypy/_c/count_table.pyx":393
 *                     value[0] = self.table.bins[self.bin_index].cells[j].value
 *                     self.cell_index += 1
 *                     return 0             # <<<<<<<<<<<<<<
 *         self.bin_index = self.table.size
 *         key[0] = NULL
 */
        __pyx_r = 0;
        goto __pyx_L0;

        /* "glypy/_c/count_table.pyx":389
 *             for j in range(0, self.table.bins[i].used):
 *                 self.cell_index = j
 *                 if self.table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                     key[0] = self.table.bins[self.bin_index].cells[j].key
 *                     value[0] = self.table.bins[self.bin_index].cells[j].value
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":394
 *                     self.cell_index += 1
 *                     return 0
 *         self.bin_index = self.table.size             # <<<<<<<<<<<<<<
 *         key[0] = NULL
 *         return 1
 */
  __pyx_t_2 = __pyx_v_self->table->size;
  __pyx_v_self->bin_index = __pyx_t_2;

  /* "glypy/_c/count_table.pyx":395
 *                     return 0
 *         self.bin_index = self.table.size
 *         key[0] = NULL             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  (__pyx_v_key[0]) = NULL;

  /* "glypy/_c/count_table.pyx":396
 *         self.bin_index = self.table.size
 *         key[0] = NULL
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int get_next_value(self, PyObject** key, long* value):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":372
 *         return self.next_key != NULL
 * 
 *     cdef int _locate_next_value(self, PyObject** key, long* value):             # <<<<<<<<<<<<<<
 *         if not self.bin_index < self.table.size:
 *             key[0] = NULL
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":398
 *         return 1
 * 
 *     cdef int get_next_value(self, PyObject** key, long* value):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* next_key
 */

static int __pyx_f_5glypy_2_c_11count_table_18CountTableIterator_get_next_value(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self, PyObject **__pyx_v_key, long *__pyx_v_value) {
  PyObject *__pyx_v_next_key;
  long __pyx_v_next_value;
  CYTHON_UNUSED int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  long __pyx_t_3;
  __Pyx_RefNannySetupContext("get_next_value", 0);

  /* "glypy/_c/count_table.pyx":403
 *             long next_value
 *             int status
 *         if self.next_key == NULL:             # <<<<<<<<<<<<<<
 *             key[0] = NULL
 *             value[0] = 0
 */
  __pyx_t_1 = ((__pyx_v_self->next_key == NULL) != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":404
 *             int status
 *         if self.next_key == NULL:
 *             key[0] = NULL             # <<<<<<<<<<<<<<
 *             value[0] = 0
 *             return 1
 */
    (__pyx_v_key[0]) = NULL;

    /* "glypy/_c/count_table.pyx":405
 *         if self.next_key == NULL:
 *             key[0] = NULL
 *             value[0] = 0             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
    (__pyx_v_value[0]) = 0;

    /* "glypy/_c/count_table.pyx":406
 *             key[0] = NULL
 *             value[0] = 0
 *             return 1             # <<<<<<<<<<<<<<
 *         else:
 *             key[0] = self.next_key
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":403
 *             long next_value
 *             int status
 *         if self.next_key == NULL:             # <<<<<<<<<<<<<<
 *             key[0] = NULL
 *             value[0] = 0
 */
  }

  /* "glypy/_c/count_table.pyx":408
 *             return 1
 *         else:
 *             key[0] = self.next_key             # <<<<<<<<<<<<<<
 *             value[0] = self.next_value
 *             status = self._locate_next_value(&next_key, &next_value)
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_self->next_key;
    (__pyx_v_key[0]) = __pyx_t_2;

    /* "glypy/_c/count_table.pyx":409
 *         else:
 *             key[0] = self.next_key
 *             value[0] = self.next_value             # <<<<<<<<<<<<<<
 *             status = self._locate_next_value(&next_key, &next_value)
 *             self.next_key = next_key
 */
    __pyx_t_3 = __pyx_v_self->next_value;
    (__pyx_v_value[0]) = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":410
 *             key[0] = self.next_key
 *             value[0] = self.next_value
 *             status = self._locate_next_value(&next_key, &next_value)             # <<<<<<<<<<<<<<
 *             self.next_key = next_key
 *             self.next_value = next_value
 */
    __pyx_v_status = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self->__pyx_vtab)->_locate_next_value(__pyx_v_self, (&__pyx_v_next_key), (&__pyx_v_next_value));

    /* "glypy/_c/count_table.pyx":411
 *             value[0] = self.next_value
 *             status = self._locate_next_value(&next_key, &next_value)
 *             self.next_key = next_key             # <<<<<<<<<<<<<<
 *             self.next_value = next_value
 *         return 0
 */
    __pyx_v_self->next_key = __pyx_v_next_key;

    /* "glypy/_c/count_table.pyx":412
 *             status = self._locate_next_value(&next_key, &next_value)
 *             self.next_key = next_key
 *             self.next_value = next_value             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_v_self->next_value = __pyx_v_next_value;
  }

  /* "glypy/_c/count_table.pyx":413
 *             self.next_key = next_key
 *             self.next_value = next_value
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef void advance(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":398
 *         return 1
 * 
 *     cdef int get_next_value(self, PyObject** key, long* value):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* next_key
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":415
 *         return 0
 * 
 *     cdef void advance(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* next_key
 */

static void __pyx_f_5glypy_2_c_11count_table_18CountTableIterator_advance(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self) {
  PyObject *__pyx_v_next_key;
  long __pyx_v_next_value;
  CYTHON_UNUSED int __pyx_v_status;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("advance", 0);

  /* "glypy/_c/count_table.pyx":420
 *             long next_value
 *             int status
 *         status = self._locate_next_value(&next_key, &next_value)             # <<<<<<<<<<<<<<
 *         self.next_key = next_key
 *         self.next_value = next_value
 */
  __pyx_v_status = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self->__pyx_vtab)->_locate_next_value(__pyx_v_self, (&__pyx_v_next_key), (&__pyx_v_next_value));

  /* "glypy/_c/count_table.pyx":421
 *             int status
 *         status = self._locate_next_value(&next_key, &next_value)
 *         self.next_key = next_key             # <<<<<<<<<<<<<<
 *         self.next_value = next_value
 * 
 */
  __pyx_v_self->next_key = __pyx_v_next_key;

  /* "glypy/_c/count_table.pyx":422
 *         status = self._locate_next_value(&next_key, &next_value)
 *         self.next_key = next_key
 *         self.next_value = next_value             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_v_self->next_value = __pyx_v_next_value;

  /* "glypy/_c/count_table.pyx":415
 *         return 0
 * 
 *     cdef void advance(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* next_key
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":424
 *         self.next_value = next_value
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_3__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_2__iter__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_2__iter__(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "glypy/_c/count_table.pyx":425
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":424
 *         self.next_value = next_value
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":427
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* key
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_5__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_5__next__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_4__next__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_4__next__(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self) {
  PyObject *__pyx_v_key;
  PyObject *__pyx_v_next_key;
  long __pyx_v_value;
  long __pyx_v_next_value;
  CYTHON_UNUSED int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3;
  long __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "glypy/_c/count_table.pyx":433
 *             long value, next_value
 *             int status
 *         if self.next_key == NULL:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->next_key == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "glypy/_c/count_table.pyx":434
 *             int status
 *         if self.next_key == NULL:
 *             raise StopIteration()             # <<<<<<<<<<<<<<
 *         else:
 *             key = self.next_key
 */
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_StopIteration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 434, __pyx_L1_error)

    /* "glypy/_c/count_table.pyx":433
 *             long value, next_value
 *             int status
 *         if self.next_key == NULL:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         else:
 */
  }

  /* "glypy/_c/count_table.pyx":436
 *             raise StopIteration()
 *         else:
 *             key = self.next_key             # <<<<<<<<<<<<<<
 *             value = self.next_value
 *             status = self._locate_next_value(&next_key, &next_value)
 */
  /*else*/ {
    __pyx_t_3 = __pyx_v_self->next_key;
    __pyx_v_key = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":437
 *         else:
 *             key = self.next_key
 *             value = self.next_value             # <<<<<<<<<<<<<<
 *             status = self._locate_next_value(&next_key, &next_value)
 *             self.next_key = next_key
 */
    __pyx_t_4 = __pyx_v_self->next_value;
    __pyx_v_value = __pyx_t_4;

    /* "glypy/_c/count_table.pyx":438
 *             key = self.next_key
 *             value = self.next_value
 *             status = self._locate_next_value(&next_key, &next_value)             # <<<<<<<<<<<<<<
 *             self.next_key = next_key
 *             self.next_value = next_value
 */
    __pyx_v_status = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self->__pyx_vtab)->_locate_next_value(__pyx_v_self, (&__pyx_v_next_key), (&__pyx_v_next_value));

    /* "glypy/_c/count_table.pyx":439
 *             value = self.next_value
 *             status = self._locate_next_value(&next_key, &next_value)
 *             self.next_key = next_key             # <<<<<<<<<<<<<<
 *             self.next_value = next_value
 *         return (<object>key, value)
 */
    __pyx_v_self->next_key = __pyx_v_next_key;

    /* "glypy/_c/count_table.pyx":440
 *             status = self._locate_next_value(&next_key, &next_value)
 *             self.next_key = next_key
 *             self.next_value = next_value             # <<<<<<<<<<<<<<
 *         return (<object>key, value)
 * 
 */
    __pyx_v_self->next_value = __pyx_v_next_value;
  }

  /* "glypy/_c/count_table.pyx":441
 *             self.next_key = next_key
 *             self.next_value = next_value
 *         return (<object>key, value)             # <<<<<<<<<<<<<<
 * 
 *     def _test(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_v_key));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_key));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_key));
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":427
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glypy._c.count_table.CountTableIterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":443
 *         return (<object>key, value)
 * 
 *     def _test(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* key
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_7_test(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_7_test(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_6_test(((struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_6_test(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self) {
  PyObject *__pyx_v_key;
  long __pyx_v_count;
  int __pyx_v_pos;
  int __pyx_v_j;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test", 0);

  /* "glypy/_c/count_table.pyx":449
 *             int pos, j
 *             list result
 *         pos = 0             # <<<<<<<<<<<<<<
 *         j = 0
 *         result = []
 */
  __pyx_v_pos = 0;

  /* "glypy/_c/count_table.pyx":450
 *             list result
 *         pos = 0
 *         j = 0             # <<<<<<<<<<<<<<
 *         result = []
 *         while self.has_more():
 */
  __pyx_v_j = 0;

  /* "glypy/_c/count_table.pyx":451
 *         pos = 0
 *         j = 0
 *         result = []             # <<<<<<<<<<<<<<
 *         while self.has_more():
 *             pos = self.get_next_value(&key, &count)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/_c/count_table.pyx":452
 *         j = 0
 *         result = []
 *         while self.has_more():             # <<<<<<<<<<<<<<
 *             pos = self.get_next_value(&key, &count)
 *             if pos != 0:
 */
  while (1) {
    __pyx_t_2 = (((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self->__pyx_vtab)->has_more(__pyx_v_self) != 0);
    if (!__pyx_t_2) break;

    /* "glypy/_c/count_table.pyx":453
 *         result = []
 *         while self.has_more():
 *             pos = self.get_next_value(&key, &count)             # <<<<<<<<<<<<<<
 *             if pos != 0:
 *                 break
 */
    __pyx_v_pos = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self->__pyx_vtab)->get_next_value(__pyx_v_self, (&__pyx_v_key), (&__pyx_v_count));

    /* "glypy/_c/count_table.pyx":454
 *         while self.has_more():
 *             pos = self.get_next_value(&key, &count)
 *             if pos != 0:             # <<<<<<<<<<<<<<
 *                 break
 *             result.append((<object>key, count))
 */
    __pyx_t_2 = ((__pyx_v_pos != 0) != 0);
    if (__pyx_t_2) {

      /* "glypy/_c/count_table.pyx":455
 *             pos = self.get_next_value(&key, &count)
 *             if pos != 0:
 *                 break             # <<<<<<<<<<<<<<
 *             result.append((<object>key, count))
 *             j += 1
 */
      goto __pyx_L4_break;

      /* "glypy/_c/count_table.pyx":454
 *         while self.has_more():
 *             pos = self.get_next_value(&key, &count)
 *             if pos != 0:             # <<<<<<<<<<<<<<
 *                 break
 *             result.append((<object>key, count))
 */
    }

    /* "glypy/_c/count_table.pyx":456
 *             if pos != 0:
 *                 break
 *             result.append((<object>key, count))             # <<<<<<<<<<<<<<
 *             j += 1
 *         return result
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)__pyx_v_key));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_key));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_key));
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "glypy/_c/count_table.pyx":457
 *                 break
 *             result.append((<object>key, count))
 *             j += 1             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_v_j = (__pyx_v_j + 1);
  }
  __pyx_L4_break:;

  /* "glypy/_c/count_table.pyx":458
 *             result.append((<object>key, count))
 *             j += 1
 *         return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":443
 *         return (<object>key, value)
 * 
 *     def _test(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy._c.count_table.CountTableIterator._test", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_8__reduce_cython__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTableIterator.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_10__setstate_cython__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_18CountTableIterator_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTableIterator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":465
 * 
 *     @staticmethod
 *     cdef CountTable _create():             # <<<<<<<<<<<<<<
 *         cdef CountTable inst
 *         inst = CountTable.__new__(CountTable)
 */

static struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_f_5glypy_2_c_11count_table_10CountTable__create(void) {
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_inst = 0;
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "glypy/_c/count_table.pyx":467
 *     cdef CountTable _create():
 *         cdef CountTable inst
 *         inst = CountTable.__new__(CountTable)             # <<<<<<<<<<<<<<
 *         inst._initialize_table()
 *         return inst
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_5glypy_2_c_11count_table_CountTable(((PyTypeObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTable), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_inst = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/_c/count_table.pyx":468
 *         cdef CountTable inst
 *         inst = CountTable.__new__(CountTable)
 *         inst._initialize_table()             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_inst->__pyx_vtab)->_initialize_table(__pyx_v_inst); if (unlikely(__pyx_t_2 == ((int)1))) __PYX_ERR(0, 468, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":469
 *         inst = CountTable.__new__(CountTable)
 *         inst._initialize_table()
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":465
 * 
 *     @staticmethod
 *     cdef CountTable _create():             # <<<<<<<<<<<<<<
 *         cdef CountTable inst
 *         inst = CountTable.__new__(CountTable)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":472
 * 
 *     @staticmethod
 *     cdef CountTable _create_from(CountTable table):             # <<<<<<<<<<<<<<
 *         cdef CountTable inst
 *         inst = CountTable.__new__(CountTable)
 */

static struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_f_5glypy_2_c_11count_table_10CountTable__create_from(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_table) {
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_inst = 0;
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create_from", 0);

  /* "glypy/_c/count_table.pyx":474
 *     cdef CountTable _create_from(CountTable table):
 *         cdef CountTable inst
 *         inst = CountTable.__new__(CountTable)             # <<<<<<<<<<<<<<
 *         inst.table = count_table_copy(table.table)
 *         return inst
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_5glypy_2_c_11count_table_CountTable(((PyTypeObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTable), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_inst = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/_c/count_table.pyx":475
 *         cdef CountTable inst
 *         inst = CountTable.__new__(CountTable)
 *         inst.table = count_table_copy(table.table)             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_v_inst->table = __pyx_f_5glypy_2_c_11count_table_count_table_copy(__pyx_v_table->table);

  /* "glypy/_c/count_table.pyx":476
 *         inst = CountTable.__new__(CountTable)
 *         inst.table = count_table_copy(table.table)
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, obj=None, **kwargs):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":472
 * 
 *     @staticmethod
 *     cdef CountTable _create_from(CountTable table):             # <<<<<<<<<<<<<<
 *         cdef CountTable inst
 *         inst = CountTable.__new__(CountTable)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable._create_from", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":478
 *         return inst
 * 
 *     def __init__(self, obj=None, **kwargs):             # <<<<<<<<<<<<<<
 *         self._initialize_table()
 *         if obj is not None:
 */

/* Python wrapper */
static int __pyx_pw_5glypy_2_c_11count_table_10CountTable_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5glypy_2_c_11count_table_10CountTable_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 478, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 478, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable___init__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), __pyx_v_obj, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5glypy_2_c_11count_table_10CountTable___init__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "glypy/_c/count_table.pyx":479
 * 
 *     def __init__(self, obj=None, **kwargs):
 *         self._initialize_table()             # <<<<<<<<<<<<<<
 *         if obj is not None:
 *             self.update(obj)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_initialize_table(__pyx_v_self); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 479, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":480
 *     def __init__(self, obj=None, **kwargs):
 *         self._initialize_table()
 *         if obj is not None:             # <<<<<<<<<<<<<<
 *             self.update(obj)
 *         if kwargs:
 */
  __pyx_t_2 = (__pyx_v_obj != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "glypy/_c/count_table.pyx":481
 *         self._initialize_table()
 *         if obj is not None:
 *             self.update(obj)             # <<<<<<<<<<<<<<
 *         if kwargs:
 *             self.update(kwargs)
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->update(__pyx_v_self, __pyx_v_obj, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "glypy/_c/count_table.pyx":480
 *     def __init__(self, obj=None, **kwargs):
 *         self._initialize_table()
 *         if obj is not None:             # <<<<<<<<<<<<<<
 *             self.update(obj)
 *         if kwargs:
 */
  }

  /* "glypy/_c/count_table.pyx":482
 *         if obj is not None:
 *             self.update(obj)
 *         if kwargs:             # <<<<<<<<<<<<<<
 *             self.update(kwargs)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 482, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "glypy/_c/count_table.pyx":483
 *             self.update(obj)
 *         if kwargs:
 *             self.update(kwargs)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _initialize_table(self) except 1:
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->update(__pyx_v_self, __pyx_v_kwargs, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "glypy/_c/count_table.pyx":482
 *         if obj is not None:
 *             self.update(obj)
 *         if kwargs:             # <<<<<<<<<<<<<<
 *             self.update(kwargs)
 * 
 */
  }

  /* "glypy/_c/count_table.pyx":478
 *         return inst
 * 
 *     def __init__(self, obj=None, **kwargs):             # <<<<<<<<<<<<<<
 *         self._initialize_table()
 *         if obj is not None:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":485
 *             self.update(kwargs)
 * 
 *     cdef int _initialize_table(self) except 1:             # <<<<<<<<<<<<<<
 *         self.table = make_count_table(6, 2)
 *         if self.table == NULL:
 */

static int __pyx_f_5glypy_2_c_11count_table_10CountTable__initialize_table(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_initialize_table", 0);

  /* "glypy/_c/count_table.pyx":486
 * 
 *     cdef int _initialize_table(self) except 1:
 *         self.table = make_count_table(6, 2)             # <<<<<<<<<<<<<<
 *         if self.table == NULL:
 *             return 1
 */
  __pyx_v_self->table = __pyx_f_5glypy_2_c_11count_table_make_count_table(6, 2);

  /* "glypy/_c/count_table.pyx":487
 *     cdef int _initialize_table(self) except 1:
 *         self.table = make_count_table(6, 2)
 *         if self.table == NULL:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  __pyx_t_1 = ((__pyx_v_self->table == NULL) != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":488
 *         self.table = make_count_table(6, 2)
 *         if self.table == NULL:
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":487
 *     cdef int _initialize_table(self) except 1:
 *         self.table = make_count_table(6, 2)
 *         if self.table == NULL:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  }

  /* "glypy/_c/count_table.pyx":489
 *         if self.table == NULL:
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cpdef update(self, obj):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":485
 *             self.update(kwargs)
 * 
 *     cdef int _initialize_table(self) except 1:             # <<<<<<<<<<<<<<
 *         self.table = make_count_table(6, 2)
 *         if self.table == NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":491
 *         return 0
 * 
 *     cpdef update(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, CountTable):
 *             self._update_from_count_table(<CountTable>obj)
 */

static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_3update(PyObject *__pyx_v_self, PyObject *__pyx_v_obj); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_update(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_2_c_11count_table_10CountTable_3update)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_obj) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_obj);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/_c/count_table.pyx":492
 * 
 *     cpdef update(self, obj):
 *         if isinstance(obj, CountTable):             # <<<<<<<<<<<<<<
 *             self._update_from_count_table(<CountTable>obj)
 *         if isinstance(obj, dict):
 */
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_ptype_5glypy_2_c_11count_table_CountTable); 
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "glypy/_c/count_table.pyx":493
 *     cpdef update(self, obj):
 *         if isinstance(obj, CountTable):
 *             self._update_from_count_table(<CountTable>obj)             # <<<<<<<<<<<<<<
 *         if isinstance(obj, dict):
 *             self._update_from_dict(<dict>obj)
 */
    ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_update_from_count_table(__pyx_v_self, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_obj));

    /* "glypy/_c/count_table.pyx":492
 * 
 *     cpdef update(self, obj):
 *         if isinstance(obj, CountTable):             # <<<<<<<<<<<<<<
 *             self._update_from_count_table(<CountTable>obj)
 *         if isinstance(obj, dict):
 */
  }

  /* "glypy/_c/count_table.pyx":494
 *         if isinstance(obj, CountTable):
 *             self._update_from_count_table(<CountTable>obj)
 *         if isinstance(obj, dict):             # <<<<<<<<<<<<<<
 *             self._update_from_dict(<dict>obj)
 *         else:
 */
  __pyx_t_6 = PyDict_Check(__pyx_v_obj); 
  __pyx_t_5 = (__pyx_t_6 != 0);
  if (__pyx_t_5) {

    /* "glypy/_c/count_table.pyx":495
 *             self._update_from_count_table(<CountTable>obj)
 *         if isinstance(obj, dict):
 *             self._update_from_dict(<dict>obj)             # <<<<<<<<<<<<<<
 *         else:
 *             for k, v in obj.items():
 */
    ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_update_from_dict(__pyx_v_self, ((PyObject*)__pyx_v_obj));

    /* "glypy/_c/count_table.pyx":494
 *         if isinstance(obj, CountTable):
 *             self._update_from_count_table(<CountTable>obj)
 *         if isinstance(obj, dict):             # <<<<<<<<<<<<<<
 *             self._update_from_dict(<dict>obj)
 *         else:
 */
    goto __pyx_L4;
  }

  /* "glypy/_c/count_table.pyx":497
 *             self._update_from_dict(<dict>obj)
 *         else:
 *             for k, v in obj.items():             # <<<<<<<<<<<<<<
 *                 self[k] = v
 * 
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 497, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 497, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 497, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 497, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 497, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 497, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_2);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 497, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 497, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        #else
        __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 497, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 497, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 497, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
        index = 0; __pyx_t_3 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_3)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        index = 1; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 497, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L8_unpacking_done;
        __pyx_L7_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 497, __pyx_L1_error)
        __pyx_L8_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "glypy/_c/count_table.pyx":498
 *         else:
 *             for k, v in obj.items():
 *                 self[k] = v             # <<<<<<<<<<<<<<
 * 
 *     cdef void _update_from_dict(self, dict d):
 */
      if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_self), __pyx_v_k, __pyx_v_v) < 0)) __PYX_ERR(0, 498, __pyx_L1_error)

      /* "glypy/_c/count_table.pyx":497
 *             self._update_from_dict(<dict>obj)
 *         else:
 *             for k, v in obj.items():             # <<<<<<<<<<<<<<
 *                 self[k] = v
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "glypy/_c/count_table.pyx":491
 *         return 0
 * 
 *     cpdef update(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, CountTable):
 *             self._update_from_count_table(<CountTable>obj)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_3update(PyObject *__pyx_v_self, PyObject *__pyx_v_obj); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_3update(PyObject *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_2update(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_2update(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_10CountTable_update(__pyx_v_self, __pyx_v_obj, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":500
 *                 self[k] = v
 * 
 *     cdef void _update_from_dict(self, dict d):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *key
 */

static void __pyx_f_5glypy_2_c_11count_table_10CountTable__update_from_dict(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_v_key;
  PyObject *__pyx_v_value;
  Py_ssize_t __pyx_v_pos;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  long __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_update_from_dict", 0);

  /* "glypy/_c/count_table.pyx":505
 *             PyObject *value
 *             Py_ssize_t pos
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while PyDict_Next(d, &pos, &key, &value):
 *             self.setitem(<object>key, PyInt_AsLong(int(<object>value)))
 */
  __pyx_v_pos = 0;

  /* "glypy/_c/count_table.pyx":506
 *             Py_ssize_t pos
 *         pos = 0
 *         while PyDict_Next(d, &pos, &key, &value):             # <<<<<<<<<<<<<<
 *             self.setitem(<object>key, PyInt_AsLong(int(<object>value)))
 * 
 */
  while (1) {
    __pyx_t_1 = (PyDict_Next(__pyx_v_d, (&__pyx_v_pos), (&__pyx_v_key), (&__pyx_v_value)) != 0);
    if (!__pyx_t_1) break;

    /* "glypy/_c/count_table.pyx":507
 *         pos = 0
 *         while PyDict_Next(d, &pos, &key, &value):
 *             self.setitem(<object>key, PyInt_AsLong(int(<object>value)))             # <<<<<<<<<<<<<<
 * 
 *     cdef void _update_from_count_table(self, CountTable other):
 */
    __pyx_t_2 = __Pyx_PyNumber_Int(((PyObject *)__pyx_v_value)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyInt_AsLong(__pyx_t_2); if (unlikely(__pyx_t_3 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 507, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->setitem(__pyx_v_self, ((PyObject *)__pyx_v_key), __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)1))) __PYX_ERR(0, 507, __pyx_L1_error)
  }

  /* "glypy/_c/count_table.pyx":500
 *                 self[k] = v
 * 
 *     cdef void _update_from_dict(self, dict d):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *key
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("glypy._c.count_table.CountTable._update_from_dict", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":509
 *             self.setitem(<object>key, PyInt_AsLong(int(<object>value)))
 * 
 *     cdef void _update_from_count_table(self, CountTable other):             # <<<<<<<<<<<<<<
 *         count_table_update(self.table, other.table)
 * 
 */

static void __pyx_f_5glypy_2_c_11count_table_10CountTable__update_from_count_table(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_other) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_update_from_count_table", 0);

  /* "glypy/_c/count_table.pyx":510
 * 
 *     cdef void _update_from_count_table(self, CountTable other):
 *         count_table_update(self.table, other.table)             # <<<<<<<<<<<<<<
 * 
 *     cpdef CountTable copy(self):
 */
  __pyx_f_5glypy_2_c_11count_table_count_table_update(__pyx_v_self->table, __pyx_v_other->table);

  /* "glypy/_c/count_table.pyx":509
 *             self.setitem(<object>key, PyInt_AsLong(int(<object>value)))
 * 
 *     cdef void _update_from_count_table(self, CountTable other):             # <<<<<<<<<<<<<<
 *         count_table_update(self.table, other.table)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":512
 *         count_table_update(self.table, other.table)
 * 
 *     cpdef CountTable copy(self):             # <<<<<<<<<<<<<<
 *         cdef CountTable inst = CountTable._create_from(self)
 *         return inst
 */

static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_5copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_f_5glypy_2_c_11count_table_10CountTable_copy(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_inst = 0;
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_2_c_11count_table_10CountTable_5copy)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 512, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5glypy_2_c_11count_table_CountTable))))) __PYX_ERR(0, 512, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/_c/count_table.pyx":513
 * 
 *     cpdef CountTable copy(self):
 *         cdef CountTable inst = CountTable._create_from(self)             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5glypy_2_c_11count_table_10CountTable__create_from(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_inst = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/_c/count_table.pyx":514
 *     cpdef CountTable copy(self):
 *         cdef CountTable inst = CountTable._create_from(self)
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":512
 *         count_table_update(self.table, other.table)
 * 
 *     cpdef CountTable copy(self):             # <<<<<<<<<<<<<<
 *         cdef CountTable inst = CountTable._create_from(self)
 *         return inst
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_5copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_5copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_4copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_4copy(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5glypy_2_c_11count_table_10CountTable_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":516
 *         return inst
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self._to_dict(), )
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_7__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_7__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_6__reduce__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_6__reduce__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "glypy/_c/count_table.pyx":517
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self._to_dict(), )             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_to_dict(__pyx_v_self, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":516
 *         return inst
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self._to_dict(), )
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":519
 *         return self.__class__, (self._to_dict(), )
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free_count_table(self.table)
 * 
 */

/* Python wrapper */
static void __pyx_pw_5glypy_2_c_11count_table_10CountTable_9__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_5glypy_2_c_11count_table_10CountTable_9__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_5glypy_2_c_11count_table_10CountTable_8__dealloc__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5glypy_2_c_11count_table_10CountTable_8__dealloc__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "glypy/_c/count_table.pyx":520
 * 
 *     def __dealloc__(self):
 *         free_count_table(self.table)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, object key):
 */
  __pyx_f_5glypy_2_c_11count_table_free_count_table(__pyx_v_self->table);

  /* "glypy/_c/count_table.pyx":519
 *         return self.__class__, (self._to_dict(), )
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free_count_table(self.table)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":522
 *         free_count_table(self.table)
 * 
 *     def __getitem__(self, object key):             # <<<<<<<<<<<<<<
 *         cdef long value = self.getitem(key)
 *         return PyInt_FromLong(value)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_11__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_11__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_10__getitem__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_10__getitem__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key) {
  long __pyx_v_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "glypy/_c/count_table.pyx":523
 * 
 *     def __getitem__(self, object key):
 *         cdef long value = self.getitem(key)             # <<<<<<<<<<<<<<
 *         return PyInt_FromLong(value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_key); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 523, __pyx_L1_error)
  __pyx_v_value = __pyx_t_1;

  /* "glypy/_c/count_table.pyx":524
 *     def __getitem__(self, object key):
 *         cdef long value = self.getitem(key)
 *         return PyInt_FromLong(value)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, object key, object value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyInt_FromLong(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":522
 *         free_count_table(self.table)
 * 
 *     def __getitem__(self, object key):             # <<<<<<<<<<<<<<
 *         cdef long value = self.getitem(key)
 *         return PyInt_FromLong(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":526
 *         return PyInt_FromLong(value)
 * 
 *     def __setitem__(self, object key, object value):             # <<<<<<<<<<<<<<
 *         self.setitem(key, PyInt_AsLong(int(value)))
 * 
 */

/* Python wrapper */
static int __pyx_pw_5glypy_2_c_11count_table_10CountTable_13__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5glypy_2_c_11count_table_10CountTable_13__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_12__setitem__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), ((PyObject *)__pyx_v_key), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5glypy_2_c_11count_table_10CountTable_12__setitem__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "glypy/_c/count_table.pyx":527
 * 
 *     def __setitem__(self, object key, object value):
 *         self.setitem(key, PyInt_AsLong(int(value)))             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, object key):
 */
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_AsLong(__pyx_t_1); if (unlikely(__pyx_t_2 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->setitem(__pyx_v_self, __pyx_v_key, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)1))) __PYX_ERR(0, 527, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":526
 *         return PyInt_FromLong(value)
 * 
 *     def __setitem__(self, object key, object value):             # <<<<<<<<<<<<<<
 *         self.setitem(key, PyInt_AsLong(int(value)))
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":529
 *         self.setitem(key, PyInt_AsLong(int(value)))
 * 
 *     def __delitem__(self, object key):             # <<<<<<<<<<<<<<
 *         self.pop(key)
 * 
 */

/* Python wrapper */
static int __pyx_pw_5glypy_2_c_11count_table_10CountTable_15__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_5glypy_2_c_11count_table_10CountTable_15__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_14__delitem__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5glypy_2_c_11count_table_10CountTable_14__delitem__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "glypy/_c/count_table.pyx":530
 * 
 *     def __delitem__(self, object key):
 *         self.pop(key)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->pop(__pyx_v_self, __pyx_v_key, 0, NULL); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 530, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":529
 *         self.setitem(key, PyInt_AsLong(int(value)))
 * 
 *     def __delitem__(self, object key):             # <<<<<<<<<<<<<<
 *         self.pop(key)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":532
 *         self.pop(key)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return count_table_count(self.table)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_5glypy_2_c_11count_table_10CountTable_17__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_5glypy_2_c_11count_table_10CountTable_17__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_16__len__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_5glypy_2_c_11count_table_10CountTable_16__len__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "glypy/_c/count_table.pyx":533
 * 
 *     def __len__(self):
 *         return count_table_count(self.table)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, key):
 */
  __pyx_r = __pyx_f_5glypy_2_c_11count_table_count_table_count(__pyx_v_self->table);
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":532
 *         self.pop(key)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return count_table_count(self.table)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":535
 *         return count_table_count(self.table)
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         cdef long value = self.getitem(key)
 *         return value != 0
 */

/* Python wrapper */
static int __pyx_pw_5glypy_2_c_11count_table_10CountTable_19__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_5glypy_2_c_11count_table_10CountTable_19__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_18__contains__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5glypy_2_c_11count_table_10CountTable_18__contains__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key) {
  long __pyx_v_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "glypy/_c/count_table.pyx":536
 * 
 *     def __contains__(self, key):
 *         cdef long value = self.getitem(key)             # <<<<<<<<<<<<<<
 *         return value != 0
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_key); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 536, __pyx_L1_error)
  __pyx_v_value = __pyx_t_1;

  /* "glypy/_c/count_table.pyx":537
 *     def __contains__(self, key):
 *         cdef long value = self.getitem(key)
 *         return value != 0             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_r = (__pyx_v_value != 0);
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":535
 *         return count_table_count(self.table)
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         cdef long value = self.getitem(key)
 *         return value != 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":539
 *         return value != 0
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.keys())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_21__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_21__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_20__iter__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_20__iter__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "glypy/_c/count_table.pyx":540
 * 
 *     def __iter__(self):
 *         return iter(self.keys())             # <<<<<<<<<<<<<<
 * 
 *     def __add__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->keys(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":539
 *         return value != 0
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.keys())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":542
 *         return iter(self.keys())
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CountTable dup
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_23__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_23__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__add__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_22__add__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_22__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_dup = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__add__", 0);
  __Pyx_INCREF(__pyx_v_self);

  /* "glypy/_c/count_table.pyx":545
 *         cdef:
 *             CountTable dup
 *         if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *             if isinstance(self, CountTable):
 *                 dup = (<CountTable>self).copy()
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_5glypy_2_c_11count_table_CountTable); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":546
 *             CountTable dup
 *         if isinstance(other, CountTable):
 *             if isinstance(self, CountTable):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>self).copy()
 *                 dup._add_from(<CountTable>other)
 */
    __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_self, __pyx_ptype_5glypy_2_c_11count_table_CountTable); 
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "glypy/_c/count_table.pyx":547
 *         if isinstance(other, CountTable):
 *             if isinstance(self, CountTable):
 *                 dup = (<CountTable>self).copy()             # <<<<<<<<<<<<<<
 *                 dup._add_from(<CountTable>other)
 *                 return dup
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 547, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":548
 *             if isinstance(self, CountTable):
 *                 dup = (<CountTable>self).copy()
 *                 dup._add_from(<CountTable>other)             # <<<<<<<<<<<<<<
 *                 return dup
 *             elif isinstance(self, dict):
 */
      ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_add_from(__pyx_v_dup, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other));

      /* "glypy/_c/count_table.pyx":549
 *                 dup = (<CountTable>self).copy()
 *                 dup._add_from(<CountTable>other)
 *                 return dup             # <<<<<<<<<<<<<<
 *             elif isinstance(self, dict):
 *                 dup = (<CountTable>other).copy()
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_dup));
      __pyx_r = ((PyObject *)__pyx_v_dup);
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":546
 *             CountTable dup
 *         if isinstance(other, CountTable):
 *             if isinstance(self, CountTable):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>self).copy()
 *                 dup._add_from(<CountTable>other)
 */
    }

    /* "glypy/_c/count_table.pyx":550
 *                 dup._add_from(<CountTable>other)
 *                 return dup
 *             elif isinstance(self, dict):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>other).copy()
 *                 dup._add_from_dict(<dict>self)
 */
    __pyx_t_1 = PyDict_Check(__pyx_v_self); 
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "glypy/_c/count_table.pyx":551
 *                 return dup
 *             elif isinstance(self, dict):
 *                 dup = (<CountTable>other).copy()             # <<<<<<<<<<<<<<
 *                 dup._add_from_dict(<dict>self)
 *                 return dup
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":552
 *             elif isinstance(self, dict):
 *                 dup = (<CountTable>other).copy()
 *                 dup._add_from_dict(<dict>self)             # <<<<<<<<<<<<<<
 *                 return dup
 *             else:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_add_from_dict(__pyx_v_dup, ((PyObject*)__pyx_v_self)); if (unlikely(__pyx_t_4 == ((int)1))) __PYX_ERR(0, 552, __pyx_L1_error)

      /* "glypy/_c/count_table.pyx":553
 *                 dup = (<CountTable>other).copy()
 *                 dup._add_from_dict(<dict>self)
 *                 return dup             # <<<<<<<<<<<<<<
 *             else:
 *                 self = CountTable(self)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_dup));
      __pyx_r = ((PyObject *)__pyx_v_dup);
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":550
 *                 dup._add_from(<CountTable>other)
 *                 return dup
 *             elif isinstance(self, dict):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>other).copy()
 *                 dup._add_from_dict(<dict>self)
 */
    }

    /* "glypy/_c/count_table.pyx":555
 *                 return dup
 *             else:
 *                 self = CountTable(self)             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>other).copy()
 *                 dup._add_from(self)
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTable), __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 555, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_self, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":556
 *             else:
 *                 self = CountTable(self)
 *                 dup = (<CountTable>other).copy()             # <<<<<<<<<<<<<<
 *                 dup._add_from(self)
 *                 return dup
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 556, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":557
 *                 self = CountTable(self)
 *                 dup = (<CountTable>other).copy()
 *                 dup._add_from(self)             # <<<<<<<<<<<<<<
 *                 return dup
 *         else:
 */
      if (!(likely(__Pyx_TypeTest(__pyx_v_self, __pyx_ptype_5glypy_2_c_11count_table_CountTable)))) __PYX_ERR(0, 557, __pyx_L1_error)
      ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_add_from(__pyx_v_dup, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

      /* "glypy/_c/count_table.pyx":558
 *                 dup = (<CountTable>other).copy()
 *                 dup._add_from(self)
 *                 return dup             # <<<<<<<<<<<<<<
 *         else:
 *             if isinstance(other, dict):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_dup));
      __pyx_r = ((PyObject *)__pyx_v_dup);
      goto __pyx_L0;
    }

    /* "glypy/_c/count_table.pyx":545
 *         cdef:
 *             CountTable dup
 *         if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *             if isinstance(self, CountTable):
 *                 dup = (<CountTable>self).copy()
 */
  }

  /* "glypy/_c/count_table.pyx":560
 *                 return dup
 *         else:
 *             if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>self).copy()
 *                 dup._add_from_dict(<dict>other)
 */
  /*else*/ {
    __pyx_t_2 = PyDict_Check(__pyx_v_other); 
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "glypy/_c/count_table.pyx":561
 *         else:
 *             if isinstance(other, dict):
 *                 dup = (<CountTable>self).copy()             # <<<<<<<<<<<<<<
 *                 dup._add_from_dict(<dict>other)
 *                 return dup
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":562
 *             if isinstance(other, dict):
 *                 dup = (<CountTable>self).copy()
 *                 dup._add_from_dict(<dict>other)             # <<<<<<<<<<<<<<
 *                 return dup
 *             else:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_add_from_dict(__pyx_v_dup, ((PyObject*)__pyx_v_other)); if (unlikely(__pyx_t_4 == ((int)1))) __PYX_ERR(0, 562, __pyx_L1_error)

      /* "glypy/_c/count_table.pyx":563
 *                 dup = (<CountTable>self).copy()
 *                 dup._add_from_dict(<dict>other)
 *                 return dup             # <<<<<<<<<<<<<<
 *             else:
 *                 dup = (<CountTable>self).copy()
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_dup));
      __pyx_r = ((PyObject *)__pyx_v_dup);
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":560
 *                 return dup
 *         else:
 *             if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>self).copy()
 *                 dup._add_from_dict(<dict>other)
 */
    }

    /* "glypy/_c/count_table.pyx":565
 *                 return dup
 *             else:
 *                 dup = (<CountTable>self).copy()             # <<<<<<<<<<<<<<
 *                 dup._add_from(CountTable(other))
 *                 return dup
 */
    /*else*/ {
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":566
 *             else:
 *                 dup = (<CountTable>self).copy()
 *                 dup._add_from(CountTable(other))             # <<<<<<<<<<<<<<
 *                 return dup
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTable), __pyx_v_other); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_add_from(__pyx_v_dup, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3));
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":567
 *                 dup = (<CountTable>self).copy()
 *                 dup._add_from(CountTable(other))
 *                 return dup             # <<<<<<<<<<<<<<
 * 
 *     def __iadd__(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_dup));
      __pyx_r = ((PyObject *)__pyx_v_dup);
      goto __pyx_L0;
    }
  }

  /* "glypy/_c/count_table.pyx":542
 *         return iter(self.keys())
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CountTable dup
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dup);
  __Pyx_XDECREF(__pyx_v_self);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":569
 *                 return dup
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, CountTable):
 *             self._add_from(<CountTable>other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_25__iadd__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_25__iadd__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iadd__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_24__iadd__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_24__iadd__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iadd__", 0);

  /* "glypy/_c/count_table.pyx":570
 * 
 *     def __iadd__(self, other):
 *         if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *             self._add_from(<CountTable>other)
 *         elif isinstance(other, dict):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_5glypy_2_c_11count_table_CountTable); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":571
 *     def __iadd__(self, other):
 *         if isinstance(other, CountTable):
 *             self._add_from(<CountTable>other)             # <<<<<<<<<<<<<<
 *         elif isinstance(other, dict):
 *             self._add_from_dict(<dict>other)
 */
    ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_add_from(__pyx_v_self, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other));

    /* "glypy/_c/count_table.pyx":570
 * 
 *     def __iadd__(self, other):
 *         if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *             self._add_from(<CountTable>other)
 *         elif isinstance(other, dict):
 */
    goto __pyx_L3;
  }

  /* "glypy/_c/count_table.pyx":572
 *         if isinstance(other, CountTable):
 *             self._add_from(<CountTable>other)
 *         elif isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             self._add_from_dict(<dict>other)
 *         else:
 */
  __pyx_t_2 = PyDict_Check(__pyx_v_other); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":573
 *             self._add_from(<CountTable>other)
 *         elif isinstance(other, dict):
 *             self._add_from_dict(<dict>other)             # <<<<<<<<<<<<<<
 *         else:
 *             self._add_from(CountTable(other))
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_add_from_dict(__pyx_v_self, ((PyObject*)__pyx_v_other)); if (unlikely(__pyx_t_3 == ((int)1))) __PYX_ERR(0, 573, __pyx_L1_error)

    /* "glypy/_c/count_table.pyx":572
 *         if isinstance(other, CountTable):
 *             self._add_from(<CountTable>other)
 *         elif isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             self._add_from_dict(<dict>other)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "glypy/_c/count_table.pyx":575
 *             self._add_from_dict(<dict>other)
 *         else:
 *             self._add_from(CountTable(other))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTable), __pyx_v_other); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_add_from(__pyx_v_self, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_4));
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "glypy/_c/count_table.pyx":576
 *         else:
 *             self._add_from(CountTable(other))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __sub__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":569
 *                 return dup
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, CountTable):
 *             self._add_from(<CountTable>other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__iadd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":578
 *         return self
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CountTable dup
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_27__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_27__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sub__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_26__sub__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_26__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_dup = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__sub__", 0);
  __Pyx_INCREF(__pyx_v_self);

  /* "glypy/_c/count_table.pyx":581
 *         cdef:
 *             CountTable dup
 *         if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *             if isinstance(self, CountTable):
 *                 dup = (<CountTable>self).copy()
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_5glypy_2_c_11count_table_CountTable); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":582
 *             CountTable dup
 *         if isinstance(other, CountTable):
 *             if isinstance(self, CountTable):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>self).copy()
 *                 dup._subtract_from(<CountTable>other)
 */
    __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_self, __pyx_ptype_5glypy_2_c_11count_table_CountTable); 
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "glypy/_c/count_table.pyx":583
 *         if isinstance(other, CountTable):
 *             if isinstance(self, CountTable):
 *                 dup = (<CountTable>self).copy()             # <<<<<<<<<<<<<<
 *                 dup._subtract_from(<CountTable>other)
 *                 return dup
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 583, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":584
 *             if isinstance(self, CountTable):
 *                 dup = (<CountTable>self).copy()
 *                 dup._subtract_from(<CountTable>other)             # <<<<<<<<<<<<<<
 *                 return dup
 *             elif isinstance(self, dict):
 */
      ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_subtract_from(__pyx_v_dup, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other));

      /* "glypy/_c/count_table.pyx":585
 *                 dup = (<CountTable>self).copy()
 *                 dup._subtract_from(<CountTable>other)
 *                 return dup             # <<<<<<<<<<<<<<
 *             elif isinstance(self, dict):
 *                 dup = (<CountTable>other).copy()
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_dup));
      __pyx_r = ((PyObject *)__pyx_v_dup);
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":582
 *             CountTable dup
 *         if isinstance(other, CountTable):
 *             if isinstance(self, CountTable):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>self).copy()
 *                 dup._subtract_from(<CountTable>other)
 */
    }

    /* "glypy/_c/count_table.pyx":586
 *                 dup._subtract_from(<CountTable>other)
 *                 return dup
 *             elif isinstance(self, dict):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>other).copy()
 *                 dup._subtract_from_dict(<dict>self)
 */
    __pyx_t_1 = PyDict_Check(__pyx_v_self); 
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "glypy/_c/count_table.pyx":587
 *                 return dup
 *             elif isinstance(self, dict):
 *                 dup = (<CountTable>other).copy()             # <<<<<<<<<<<<<<
 *                 dup._subtract_from_dict(<dict>self)
 *                 return dup
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 587, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":588
 *             elif isinstance(self, dict):
 *                 dup = (<CountTable>other).copy()
 *                 dup._subtract_from_dict(<dict>self)             # <<<<<<<<<<<<<<
 *                 return dup
 *             else:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_subtract_from_dict(__pyx_v_dup, ((PyObject*)__pyx_v_self)); if (unlikely(__pyx_t_4 == ((int)1))) __PYX_ERR(0, 588, __pyx_L1_error)

      /* "glypy/_c/count_table.pyx":589
 *                 dup = (<CountTable>other).copy()
 *                 dup._subtract_from_dict(<dict>self)
 *                 return dup             # <<<<<<<<<<<<<<
 *             else:
 *                 self = CountTable(self)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_dup));
      __pyx_r = ((PyObject *)__pyx_v_dup);
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":586
 *                 dup._subtract_from(<CountTable>other)
 *                 return dup
 *             elif isinstance(self, dict):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>other).copy()
 *                 dup._subtract_from_dict(<dict>self)
 */
    }

    /* "glypy/_c/count_table.pyx":591
 *                 return dup
 *             else:
 *                 self = CountTable(self)             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>other).copy()
 *                 dup._subtract_from(self)
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTable), __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_self, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":592
 *             else:
 *                 self = CountTable(self)
 *                 dup = (<CountTable>other).copy()             # <<<<<<<<<<<<<<
 *                 dup._subtract_from(self)
 *                 return dup
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 592, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":593
 *                 self = CountTable(self)
 *                 dup = (<CountTable>other).copy()
 *                 dup._subtract_from(self)             # <<<<<<<<<<<<<<
 *                 return dup
 *         else:
 */
      if (!(likely(__Pyx_TypeTest(__pyx_v_self, __pyx_ptype_5glypy_2_c_11count_table_CountTable)))) __PYX_ERR(0, 593, __pyx_L1_error)
      ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_subtract_from(__pyx_v_dup, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

      /* "glypy/_c/count_table.pyx":594
 *                 dup = (<CountTable>other).copy()
 *                 dup._subtract_from(self)
 *                 return dup             # <<<<<<<<<<<<<<
 *         else:
 *             if isinstance(other, dict):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_dup));
      __pyx_r = ((PyObject *)__pyx_v_dup);
      goto __pyx_L0;
    }

    /* "glypy/_c/count_table.pyx":581
 *         cdef:
 *             CountTable dup
 *         if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *             if isinstance(self, CountTable):
 *                 dup = (<CountTable>self).copy()
 */
  }

  /* "glypy/_c/count_table.pyx":596
 *                 return dup
 *         else:
 *             if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>self).copy()
 *                 dup._subtract_from_dict(<dict>other)
 */
  /*else*/ {
    __pyx_t_2 = PyDict_Check(__pyx_v_other); 
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "glypy/_c/count_table.pyx":597
 *         else:
 *             if isinstance(other, dict):
 *                 dup = (<CountTable>self).copy()             # <<<<<<<<<<<<<<
 *                 dup._subtract_from_dict(<dict>other)
 *                 return dup
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":598
 *             if isinstance(other, dict):
 *                 dup = (<CountTable>self).copy()
 *                 dup._subtract_from_dict(<dict>other)             # <<<<<<<<<<<<<<
 *                 return dup
 *             else:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_subtract_from_dict(__pyx_v_dup, ((PyObject*)__pyx_v_other)); if (unlikely(__pyx_t_4 == ((int)1))) __PYX_ERR(0, 598, __pyx_L1_error)

      /* "glypy/_c/count_table.pyx":599
 *                 dup = (<CountTable>self).copy()
 *                 dup._subtract_from_dict(<dict>other)
 *                 return dup             # <<<<<<<<<<<<<<
 *             else:
 *                 dup = (<CountTable>self).copy()
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_dup));
      __pyx_r = ((PyObject *)__pyx_v_dup);
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":596
 *                 return dup
 *         else:
 *             if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *                 dup = (<CountTable>self).copy()
 *                 dup._subtract_from_dict(<dict>other)
 */
    }

    /* "glypy/_c/count_table.pyx":601
 *                 return dup
 *             else:
 *                 dup = (<CountTable>self).copy()             # <<<<<<<<<<<<<<
 *                 dup._subtract_from(CountTable(other))
 *                 return dup
 */
    /*else*/ {
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 601, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":602
 *             else:
 *                 dup = (<CountTable>self).copy()
 *                 dup._subtract_from(CountTable(other))             # <<<<<<<<<<<<<<
 *                 return dup
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTable), __pyx_v_other); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_subtract_from(__pyx_v_dup, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3));
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "glypy/_c/count_table.pyx":603
 *                 dup = (<CountTable>self).copy()
 *                 dup._subtract_from(CountTable(other))
 *                 return dup             # <<<<<<<<<<<<<<
 * 
 *     def __isub__(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(((PyObject *)__pyx_v_dup));
      __pyx_r = ((PyObject *)__pyx_v_dup);
      goto __pyx_L0;
    }
  }

  /* "glypy/_c/count_table.pyx":578
 *         return self
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CountTable dup
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__sub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dup);
  __Pyx_XDECREF(__pyx_v_self);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":605
 *                 return dup
 * 
 *     def __isub__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, CountTable):
 *             self._subtract_from(<CountTable>other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_29__isub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_29__isub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__isub__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_28__isub__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_28__isub__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__isub__", 0);

  /* "glypy/_c/count_table.pyx":606
 * 
 *     def __isub__(self, other):
 *         if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *             self._subtract_from(<CountTable>other)
 *         elif isinstance(other, dict):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_5glypy_2_c_11count_table_CountTable); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":607
 *     def __isub__(self, other):
 *         if isinstance(other, CountTable):
 *             self._subtract_from(<CountTable>other)             # <<<<<<<<<<<<<<
 *         elif isinstance(other, dict):
 *             self._subtract_from_dict(<dict>other)
 */
    ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_subtract_from(__pyx_v_self, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other));

    /* "glypy/_c/count_table.pyx":606
 * 
 *     def __isub__(self, other):
 *         if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *             self._subtract_from(<CountTable>other)
 *         elif isinstance(other, dict):
 */
    goto __pyx_L3;
  }

  /* "glypy/_c/count_table.pyx":608
 *         if isinstance(other, CountTable):
 *             self._subtract_from(<CountTable>other)
 *         elif isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             self._subtract_from_dict(<dict>other)
 *         else:
 */
  __pyx_t_2 = PyDict_Check(__pyx_v_other); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":609
 *             self._subtract_from(<CountTable>other)
 *         elif isinstance(other, dict):
 *             self._subtract_from_dict(<dict>other)             # <<<<<<<<<<<<<<
 *         else:
 *             self._subtract_from(CountTable(other))
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_subtract_from_dict(__pyx_v_self, ((PyObject*)__pyx_v_other)); if (unlikely(__pyx_t_3 == ((int)1))) __PYX_ERR(0, 609, __pyx_L1_error)

    /* "glypy/_c/count_table.pyx":608
 *         if isinstance(other, CountTable):
 *             self._subtract_from(<CountTable>other)
 *         elif isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             self._subtract_from_dict(<dict>other)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "glypy/_c/count_table.pyx":611
 *             self._subtract_from_dict(<dict>other)
 *         else:
 *             self._subtract_from(CountTable(other))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTable), __pyx_v_other); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_subtract_from(__pyx_v_self, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_4));
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "glypy/_c/count_table.pyx":612
 *         else:
 *             self._subtract_from(CountTable(other))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __mul__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":605
 *                 return dup
 * 
 *     def __isub__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, CountTable):
 *             self._subtract_from(<CountTable>other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__isub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":614
 *         return self
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CountTable dup
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_31__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_31__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__mul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_30__mul__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_30__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_dup = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  long __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__mul__", 0);

  /* "glypy/_c/count_table.pyx":617
 *         cdef:
 *             CountTable dup
 *         if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *             dup = (<CountTable>other).copy()
 *             dup._scale_by(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_5glypy_2_c_11count_table_CountTable); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":618
 *             CountTable dup
 *         if isinstance(other, CountTable):
 *             dup = (<CountTable>other).copy()             # <<<<<<<<<<<<<<
 *             dup._scale_by(self)
 *             return dup
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glypy/_c/count_table.pyx":619
 *         if isinstance(other, CountTable):
 *             dup = (<CountTable>other).copy()
 *             dup._scale_by(self)             # <<<<<<<<<<<<<<
 *             return dup
 *         elif isinstance(self, CountTable):
 */
    __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_v_self); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 619, __pyx_L1_error)
    ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_scale_by(__pyx_v_dup, __pyx_t_4);

    /* "glypy/_c/count_table.pyx":620
 *             dup = (<CountTable>other).copy()
 *             dup._scale_by(self)
 *             return dup             # <<<<<<<<<<<<<<
 *         elif isinstance(self, CountTable):
 *             dup = (<CountTable>self).copy()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_dup));
    __pyx_r = ((PyObject *)__pyx_v_dup);
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":617
 *         cdef:
 *             CountTable dup
 *         if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *             dup = (<CountTable>other).copy()
 *             dup._scale_by(self)
 */
  }

  /* "glypy/_c/count_table.pyx":621
 *             dup._scale_by(self)
 *             return dup
 *         elif isinstance(self, CountTable):             # <<<<<<<<<<<<<<
 *             dup = (<CountTable>self).copy()
 *             dup._scale_by(other)
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_self, __pyx_ptype_5glypy_2_c_11count_table_CountTable); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "glypy/_c/count_table.pyx":622
 *             return dup
 *         elif isinstance(self, CountTable):
 *             dup = (<CountTable>self).copy()             # <<<<<<<<<<<<<<
 *             dup._scale_by(other)
 *             return dup
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self)->__pyx_vtab)->copy(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_dup = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glypy/_c/count_table.pyx":623
 *         elif isinstance(self, CountTable):
 *             dup = (<CountTable>self).copy()
 *             dup._scale_by(other)             # <<<<<<<<<<<<<<
 *             return dup
 *         else:
 */
    __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_v_other); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 623, __pyx_L1_error)
    ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_dup->__pyx_vtab)->_scale_by(__pyx_v_dup, __pyx_t_4);

    /* "glypy/_c/count_table.pyx":624
 *             dup = (<CountTable>self).copy()
 *             dup._scale_by(other)
 *             return dup             # <<<<<<<<<<<<<<
 *         else:
 *             return NotImplemented
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_dup));
    __pyx_r = ((PyObject *)__pyx_v_dup);
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":621
 *             dup._scale_by(self)
 *             return dup
 *         elif isinstance(self, CountTable):             # <<<<<<<<<<<<<<
 *             dup = (<CountTable>self).copy()
 *             dup._scale_by(other)
 */
  }

  /* "glypy/_c/count_table.pyx":626
 *             return dup
 *         else:
 *             return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     def __imul__(self, other):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;
  }

  /* "glypy/_c/count_table.pyx":614
 *         return self
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CountTable dup
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__mul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dup);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":628
 *             return NotImplemented
 * 
 *     def __imul__(self, other):             # <<<<<<<<<<<<<<
 *         self._scale_by(other)
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_33__imul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_33__imul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__imul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_32__imul__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_32__imul__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__imul__", 0);

  /* "glypy/_c/count_table.pyx":629
 * 
 *     def __imul__(self, other):
 *         self._scale_by(other)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_other); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 629, __pyx_L1_error)
  ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_scale_by(__pyx_v_self, __pyx_t_1);

  /* "glypy/_c/count_table.pyx":630
 *     def __imul__(self, other):
 *         self._scale_by(other)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __neg__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":628
 *             return NotImplemented
 * 
 *     def __imul__(self, other):             # <<<<<<<<<<<<<<
 *         self._scale_by(other)
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__imul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":632
 *         return self
 * 
 *     def __neg__(self):             # <<<<<<<<<<<<<<
 *         return self._scale_by(-1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_35__neg__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_35__neg__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neg__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_34__neg__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_34__neg__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__neg__", 0);

  /* "glypy/_c/count_table.pyx":633
 * 
 *     def __neg__(self):
 *         return self._scale_by(-1)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _add_from(self, CountTable other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_scale_by(__pyx_v_self, -1L)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":632
 *         return self
 * 
 *     def __neg__(self):             # <<<<<<<<<<<<<<
 *         return self._scale_by(-1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__neg__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":635
 *         return self._scale_by(-1)
 * 
 *     cdef void _add_from(self, CountTable other):             # <<<<<<<<<<<<<<
 *         count_table_add(self.table, other.table)
 * 
 */

static void __pyx_f_5glypy_2_c_11count_table_10CountTable__add_from(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_other) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_add_from", 0);

  /* "glypy/_c/count_table.pyx":636
 * 
 *     cdef void _add_from(self, CountTable other):
 *         count_table_add(self.table, other.table)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _add_from_dict(self, dict other) except 1:
 */
  __pyx_f_5glypy_2_c_11count_table_count_table_add(__pyx_v_self->table, __pyx_v_other->table);

  /* "glypy/_c/count_table.pyx":635
 *         return self._scale_by(-1)
 * 
 *     cdef void _add_from(self, CountTable other):             # <<<<<<<<<<<<<<
 *         count_table_add(self.table, other.table)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":638
 *         count_table_add(self.table, other.table)
 * 
 *     cdef int _add_from_dict(self, dict other) except 1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *key
 */

static int __pyx_f_5glypy_2_c_11count_table_10CountTable__add_from_dict(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_key;
  PyObject *__pyx_v_value;
  Py_ssize_t __pyx_v_pos;
  long __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  long __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  char *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_add_from_dict", 0);

  /* "glypy/_c/count_table.pyx":644
 *             Py_ssize_t pos
 *             long v
 *         pos = 0             # <<<<<<<<<<<<<<
 *         try:
 *             while PyDict_Next(other, &pos, &key, &value):
 */
  __pyx_v_pos = 0;

  /* "glypy/_c/count_table.pyx":645
 *             long v
 *         pos = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while PyDict_Next(other, &pos, &key, &value):
 *                 v = PyInt_AsLong(int(<object>value))
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "glypy/_c/count_table.pyx":646
 *         pos = 0
 *         try:
 *             while PyDict_Next(other, &pos, &key, &value):             # <<<<<<<<<<<<<<
 *                 v = PyInt_AsLong(int(<object>value))
 *                 self.setitem(
 */
      while (1) {
        __pyx_t_4 = (PyDict_Next(__pyx_v_other, (&__pyx_v_pos), (&__pyx_v_key), (&__pyx_v_value)) != 0);
        if (!__pyx_t_4) break;

        /* "glypy/_c/count_table.pyx":647
 *         try:
 *             while PyDict_Next(other, &pos, &key, &value):
 *                 v = PyInt_AsLong(int(<object>value))             # <<<<<<<<<<<<<<
 *                 self.setitem(
 *                     <object>key,
 */
        __pyx_t_5 = __Pyx_PyNumber_Int(((PyObject *)__pyx_v_value)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 647, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyInt_AsLong(__pyx_t_5); if (unlikely(__pyx_t_6 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 647, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_v = __pyx_t_6;

        /* "glypy/_c/count_table.pyx":650
 *                 self.setitem(
 *                     <object>key,
 *                     v + self.getitem(<object>key))             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))
 */
        __pyx_t_6 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, ((PyObject *)__pyx_v_key)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 650, __pyx_L3_error)

        /* "glypy/_c/count_table.pyx":648
 *             while PyDict_Next(other, &pos, &key, &value):
 *                 v = PyInt_AsLong(int(<object>value))
 *                 self.setitem(             # <<<<<<<<<<<<<<
 *                     <object>key,
 *                     v + self.getitem(<object>key))
 */
        __pyx_t_7 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->setitem(__pyx_v_self, ((PyObject *)__pyx_v_key), (__pyx_v_v + __pyx_t_6)); if (unlikely(__pyx_t_7 == ((int)1))) __PYX_ERR(0, 648, __pyx_L3_error)
      }

      /* "glypy/_c/count_table.pyx":645
 *             long v
 *         pos = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while PyDict_Next(other, &pos, &key, &value):
 *                 v = PyInt_AsLong(int(<object>value))
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "glypy/_c/count_table.pyx":651
 *                     <object>key,
 *                     v + self.getitem(<object>key))
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))
 *             return 1
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("glypy._c.count_table.CountTable._add_from_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(0, 651, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_9);

      /* "glypy/_c/count_table.pyx":652
 *                     v + self.getitem(<object>key))
 *         except TypeError:
 *             PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
      __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 652, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(((PyObject *)__pyx_v_value));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_value));
      PyTuple_SET_ITEM(__pyx_t_10, 0, ((PyObject *)__pyx_v_value));
      __pyx_t_11 = __Pyx_PyString_Format(__pyx_kp_s_r_must_be_a_number, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 652, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_12 = __Pyx_PyObject_AsWritableString(__pyx_t_11); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 652, __pyx_L5_except_error)
      PyErr_SetString(__pyx_builtin_TypeError, __pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "glypy/_c/count_table.pyx":653
 *         except TypeError:
 *             PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
      __pyx_r = 1;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "glypy/_c/count_table.pyx":645
 *             long v
 *         pos = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while PyDict_Next(other, &pos, &key, &value):
 *                 v = PyInt_AsLong(int(<object>value))
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "glypy/_c/count_table.pyx":654
 *             PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef void _subtract_from(self, CountTable other):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":638
 *         count_table_add(self.table, other.table)
 * 
 *     cdef int _add_from_dict(self, dict other) except 1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable._add_from_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":656
 *         return 0
 * 
 *     cdef void _subtract_from(self, CountTable other):             # <<<<<<<<<<<<<<
 *         count_table_subtract(self.table, other.table)
 * 
 */

static void __pyx_f_5glypy_2_c_11count_table_10CountTable__subtract_from(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_other) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_subtract_from", 0);

  /* "glypy/_c/count_table.pyx":657
 * 
 *     cdef void _subtract_from(self, CountTable other):
 *         count_table_subtract(self.table, other.table)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _subtract_from_dict(self, dict other) except 1:
 */
  __pyx_f_5glypy_2_c_11count_table_count_table_subtract(__pyx_v_self->table, __pyx_v_other->table);

  /* "glypy/_c/count_table.pyx":656
 *         return 0
 * 
 *     cdef void _subtract_from(self, CountTable other):             # <<<<<<<<<<<<<<
 *         count_table_subtract(self.table, other.table)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":659
 *         count_table_subtract(self.table, other.table)
 * 
 *     cdef int _subtract_from_dict(self, dict other) except 1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *key
 */

static int __pyx_f_5glypy_2_c_11count_table_10CountTable__subtract_from_dict(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_key;
  PyObject *__pyx_v_value;
  Py_ssize_t __pyx_v_pos;
  long __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  long __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  char *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_subtract_from_dict", 0);

  /* "glypy/_c/count_table.pyx":665
 *             Py_ssize_t pos
 *             long v
 *         pos = 0             # <<<<<<<<<<<<<<
 *         try:
 *             while PyDict_Next(other, &pos, &key, &value):
 */
  __pyx_v_pos = 0;

  /* "glypy/_c/count_table.pyx":666
 *             long v
 *         pos = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while PyDict_Next(other, &pos, &key, &value):
 *                 v = PyInt_AsLong(int(<object>value))
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "glypy/_c/count_table.pyx":667
 *         pos = 0
 *         try:
 *             while PyDict_Next(other, &pos, &key, &value):             # <<<<<<<<<<<<<<
 *                 v = PyInt_AsLong(int(<object>value))
 *                 self.setitem(
 */
      while (1) {
        __pyx_t_4 = (PyDict_Next(__pyx_v_other, (&__pyx_v_pos), (&__pyx_v_key), (&__pyx_v_value)) != 0);
        if (!__pyx_t_4) break;

        /* "glypy/_c/count_table.pyx":668
 *         try:
 *             while PyDict_Next(other, &pos, &key, &value):
 *                 v = PyInt_AsLong(int(<object>value))             # <<<<<<<<<<<<<<
 *                 self.setitem(
 *                     <object>key,
 */
        __pyx_t_5 = __Pyx_PyNumber_Int(((PyObject *)__pyx_v_value)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 668, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyInt_AsLong(__pyx_t_5); if (unlikely(__pyx_t_6 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 668, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_v = __pyx_t_6;

        /* "glypy/_c/count_table.pyx":671
 *                 self.setitem(
 *                     <object>key,
 *                     v - self.getitem(<object>key))             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))
 */
        __pyx_t_6 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, ((PyObject *)__pyx_v_key)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 671, __pyx_L3_error)

        /* "glypy/_c/count_table.pyx":669
 *             while PyDict_Next(other, &pos, &key, &value):
 *                 v = PyInt_AsLong(int(<object>value))
 *                 self.setitem(             # <<<<<<<<<<<<<<
 *                     <object>key,
 *                     v - self.getitem(<object>key))
 */
        __pyx_t_7 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->setitem(__pyx_v_self, ((PyObject *)__pyx_v_key), (__pyx_v_v - __pyx_t_6)); if (unlikely(__pyx_t_7 == ((int)1))) __PYX_ERR(0, 669, __pyx_L3_error)
      }

      /* "glypy/_c/count_table.pyx":666
 *             long v
 *         pos = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while PyDict_Next(other, &pos, &key, &value):
 *                 v = PyInt_AsLong(int(<object>value))
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "glypy/_c/count_table.pyx":672
 *                     <object>key,
 *                     v - self.getitem(<object>key))
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))
 *             return 1
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("glypy._c.count_table.CountTable._subtract_from_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(0, 672, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_9);

      /* "glypy/_c/count_table.pyx":673
 *                     v - self.getitem(<object>key))
 *         except TypeError:
 *             PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
      __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 673, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(((PyObject *)__pyx_v_value));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_value));
      PyTuple_SET_ITEM(__pyx_t_10, 0, ((PyObject *)__pyx_v_value));
      __pyx_t_11 = __Pyx_PyString_Format(__pyx_kp_s_r_must_be_a_number, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 673, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_12 = __Pyx_PyObject_AsWritableString(__pyx_t_11); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 673, __pyx_L5_except_error)
      PyErr_SetString(__pyx_builtin_TypeError, __pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "glypy/_c/count_table.pyx":674
 *         except TypeError:
 *             PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
      __pyx_r = 1;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "glypy/_c/count_table.pyx":666
 *             long v
 *         pos = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while PyDict_Next(other, &pos, &key, &value):
 *                 v = PyInt_AsLong(int(<object>value))
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "glypy/_c/count_table.pyx":675
 *             PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef void _scale_by(self, long value):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":659
 *         count_table_subtract(self.table, other.table)
 * 
 *     cdef int _subtract_from_dict(self, dict other) except 1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable._subtract_from_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":677
 *         return 0
 * 
 *     cdef void _scale_by(self, long value):             # <<<<<<<<<<<<<<
 *         count_table_scale(self.table, value)
 * 
 */

static void __pyx_f_5glypy_2_c_11count_table_10CountTable__scale_by(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, long __pyx_v_value) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_scale_by", 0);

  /* "glypy/_c/count_table.pyx":678
 * 
 *     cdef void _scale_by(self, long value):
 *         count_table_scale(self.table, value)             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _to_dict(self):
 */
  __pyx_f_5glypy_2_c_11count_table_count_table_scale(__pyx_v_self->table, __pyx_v_value);

  /* "glypy/_c/count_table.pyx":677
 *         return 0
 * 
 *     cdef void _scale_by(self, long value):             # <<<<<<<<<<<<<<
 *         count_table_scale(self.table, value)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":680
 *         count_table_scale(self.table, value)
 * 
 *     cpdef dict _to_dict(self):             # <<<<<<<<<<<<<<
 *         return count_table_to_dict(self.table)
 * 
 */

static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_37_to_dict(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable__to_dict(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_to_dict", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_to_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_2_c_11count_table_10CountTable_37_to_dict)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 680, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 680, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/_c/count_table.pyx":681
 * 
 *     cpdef dict _to_dict(self):
 *         return count_table_to_dict(self.table)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_to_dict(__pyx_v_self->table); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":680
 *         count_table_scale(self.table, value)
 * 
 *     cpdef dict _to_dict(self):             # <<<<<<<<<<<<<<
 *         return count_table_to_dict(self.table)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable._to_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_37_to_dict(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_37_to_dict(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_to_dict (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_36_to_dict(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_36_to_dict(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_to_dict", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_10CountTable__to_dict(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable._to_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":683
 *         return count_table_to_dict(self.table)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({content})".format(
 *             self=self, content=self._to_dict())
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_39__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_39__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_38__repr__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_38__repr__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "glypy/_c/count_table.pyx":684
 * 
 *     def __repr__(self):
 *         return "{self.__class__.__name__}({content})".format(             # <<<<<<<<<<<<<<
 *             self=self, content=self._to_dict())
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_self___class_____name___content, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "glypy/_c/count_table.pyx":685
 *     def __repr__(self):
 *         return "{self.__class__.__name__}({content})".format(
 *             self=self, content=self._to_dict())             # <<<<<<<<<<<<<<
 * 
 *     cdef bint equal_to(self, CountTable other):
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 685, __pyx_L1_error)
  __pyx_t_3 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_to_dict(__pyx_v_self, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_content, __pyx_t_3) < 0) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glypy/_c/count_table.pyx":684
 * 
 *     def __repr__(self):
 *         return "{self.__class__.__name__}({content})".format(             # <<<<<<<<<<<<<<
 *             self=self, content=self._to_dict())
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":683
 *         return count_table_to_dict(self.table)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({content})".format(
 *             self=self, content=self._to_dict())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":687
 *             self=self, content=self._to_dict())
 * 
 *     cdef bint equal_to(self, CountTable other):             # <<<<<<<<<<<<<<
 *         return count_table_equals(self.table, other.table)
 * 
 */

static int __pyx_f_5glypy_2_c_11count_table_10CountTable_equal_to(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_other) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equal_to", 0);

  /* "glypy/_c/count_table.pyx":688
 * 
 *     cdef bint equal_to(self, CountTable other):
 *         return count_table_equals(self.table, other.table)             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __pyx_r = __pyx_f_5glypy_2_c_11count_table_count_table_equals(__pyx_v_self->table, __pyx_v_other->table);
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":687
 *             self=self, content=self._to_dict())
 * 
 *     cdef bint equal_to(self, CountTable other):             # <<<<<<<<<<<<<<
 *         return count_table_equals(self.table, other.table)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":690
 *         return count_table_equals(self.table, other.table)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(self, CountTable):
 *             if isinstance(other, CountTable):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_41__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_41__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_40__eq__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_40__eq__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "glypy/_c/count_table.pyx":691
 * 
 *     def __eq__(self, other):
 *         if isinstance(self, CountTable):             # <<<<<<<<<<<<<<
 *             if isinstance(other, CountTable):
 *                 return count_table_equals(self.table, (<CountTable>other).table)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_5glypy_2_c_11count_table_CountTable); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":692
 *     def __eq__(self, other):
 *         if isinstance(self, CountTable):
 *             if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *                 return count_table_equals(self.table, (<CountTable>other).table)
 *             else:
 */
    __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_5glypy_2_c_11count_table_CountTable); 
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "glypy/_c/count_table.pyx":693
 *         if isinstance(self, CountTable):
 *             if isinstance(other, CountTable):
 *                 return count_table_equals(self.table, (<CountTable>other).table)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self._to_dict() == other
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_f_5glypy_2_c_11count_table_count_table_equals(__pyx_v_self->table, ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_other)->table)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 693, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "glypy/_c/count_table.pyx":692
 *     def __eq__(self, other):
 *         if isinstance(self, CountTable):
 *             if isinstance(other, CountTable):             # <<<<<<<<<<<<<<
 *                 return count_table_equals(self.table, (<CountTable>other).table)
 *             else:
 */
    }

    /* "glypy/_c/count_table.pyx":695
 *                 return count_table_equals(self.table, (<CountTable>other).table)
 *             else:
 *                 return self._to_dict() == other             # <<<<<<<<<<<<<<
 *         else:
 *             return self == other._to_dict()
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->_to_dict(__pyx_v_self, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 695, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 695, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;
    }

    /* "glypy/_c/count_table.pyx":691
 * 
 *     def __eq__(self, other):
 *         if isinstance(self, CountTable):             # <<<<<<<<<<<<<<
 *             if isinstance(other, CountTable):
 *                 return count_table_equals(self.table, (<CountTable>other).table)
 */
  }

  /* "glypy/_c/count_table.pyx":697
 *                 return self._to_dict() == other
 *         else:
 *             return self == other._to_dict()             # <<<<<<<<<<<<<<
 * 
 *     def __ne__(self, other):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_to_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "glypy/_c/count_table.pyx":690
 *         return count_table_equals(self.table, other.table)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(self, CountTable):
 *             if isinstance(other, CountTable):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":699
 *             return self == other._to_dict()
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_43__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_43__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ne__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_42__ne__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_42__ne__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ne__", 0);

  /* "glypy/_c/count_table.pyx":700
 * 
 *     def __ne__(self, other):
 *         return not (self == other)             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":699
 *             return self == other._to_dict()
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__ne__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":702
 *         return not (self == other)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_hash_t hash_value
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_5glypy_2_c_11count_table_10CountTable_45__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_5glypy_2_c_11count_table_10CountTable_45__hash__(PyObject *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_44__hash__(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_5glypy_2_c_11count_table_10CountTable_44__hash__(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  Py_hash_t __pyx_v_hash_value;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_hash_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "glypy/_c/count_table.pyx":707
 *             size_t i, j
 * 
 *         hash_value = 0             # <<<<<<<<<<<<<<
 *         for i in range(self.table.size):
 *             for j in range(self.table.bins[i].used):
 */
  __pyx_v_hash_value = 0;

  /* "glypy/_c/count_table.pyx":708
 * 
 *         hash_value = 0
 *         for i in range(self.table.size):             # <<<<<<<<<<<<<<
 *             for j in range(self.table.bins[i].used):
 *                 if self.table.bins[i].cells[j].key != NULL:
 */
  __pyx_t_1 = __pyx_v_self->table->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glypy/_c/count_table.pyx":709
 *         hash_value = 0
 *         for i in range(self.table.size):
 *             for j in range(self.table.bins[i].used):             # <<<<<<<<<<<<<<
 *                 if self.table.bins[i].cells[j].key != NULL:
 *                     hash_value ^= hash(<object>self.table.bins[i].cells[j].key)
 */
    __pyx_t_4 = (__pyx_v_self->table->bins[__pyx_v_i]).used;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "glypy/_c/count_table.pyx":710
 *         for i in range(self.table.size):
 *             for j in range(self.table.bins[i].used):
 *                 if self.table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                     hash_value ^= hash(<object>self.table.bins[i].cells[j].key)
 *                     hash_value ^= self.table.bins[i].cells[j].value
 */
      __pyx_t_7 = ((((__pyx_v_self->table->bins[__pyx_v_i]).cells[__pyx_v_j]).key != NULL) != 0);
      if (__pyx_t_7) {

        /* "glypy/_c/count_table.pyx":711
 *             for j in range(self.table.bins[i].used):
 *                 if self.table.bins[i].cells[j].key != NULL:
 *                     hash_value ^= hash(<object>self.table.bins[i].cells[j].key)             # <<<<<<<<<<<<<<
 *                     hash_value ^= self.table.bins[i].cells[j].value
 *         hash_value ^= (hash_value >> 11) ^ (hash_value >> 25)
 */
        __pyx_t_8 = ((PyObject *)((__pyx_v_self->table->bins[__pyx_v_i]).cells[__pyx_v_j]).key);
        __Pyx_INCREF(__pyx_t_8);
        __pyx_t_9 = PyObject_Hash(__pyx_t_8); if (unlikely(__pyx_t_9 == ((Py_hash_t)-1))) __PYX_ERR(0, 711, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_v_hash_value = (__pyx_v_hash_value ^ __pyx_t_9);

        /* "glypy/_c/count_table.pyx":712
 *                 if self.table.bins[i].cells[j].key != NULL:
 *                     hash_value ^= hash(<object>self.table.bins[i].cells[j].key)
 *                     hash_value ^= self.table.bins[i].cells[j].value             # <<<<<<<<<<<<<<
 *         hash_value ^= (hash_value >> 11) ^ (hash_value >> 25)
 *         hash_value = hash_value * 69069 + 907133923
 */
        __pyx_v_hash_value = (__pyx_v_hash_value ^ ((__pyx_v_self->table->bins[__pyx_v_i]).cells[__pyx_v_j]).value);

        /* "glypy/_c/count_table.pyx":710
 *         for i in range(self.table.size):
 *             for j in range(self.table.bins[i].used):
 *                 if self.table.bins[i].cells[j].key != NULL:             # <<<<<<<<<<<<<<
 *                     hash_value ^= hash(<object>self.table.bins[i].cells[j].key)
 *                     hash_value ^= self.table.bins[i].cells[j].value
 */
      }
    }
  }

  /* "glypy/_c/count_table.pyx":713
 *                     hash_value ^= hash(<object>self.table.bins[i].cells[j].key)
 *                     hash_value ^= self.table.bins[i].cells[j].value
 *         hash_value ^= (hash_value >> 11) ^ (hash_value >> 25)             # <<<<<<<<<<<<<<
 *         hash_value = hash_value * 69069 + 907133923
 *         return hash_value
 */
  __pyx_v_hash_value = (__pyx_v_hash_value ^ ((__pyx_v_hash_value >> 11) ^ (__pyx_v_hash_value >> 25)));

  /* "glypy/_c/count_table.pyx":714
 *                     hash_value ^= self.table.bins[i].cells[j].value
 *         hash_value ^= (hash_value >> 11) ^ (hash_value >> 25)
 *         hash_value = hash_value * 69069 + 907133923             # <<<<<<<<<<<<<<
 *         return hash_value
 * 
 */
  __pyx_v_hash_value = ((__pyx_v_hash_value * 0x10DCD) + 0x3611C3E3);

  /* "glypy/_c/count_table.pyx":715
 *         hash_value ^= (hash_value >> 11) ^ (hash_value >> 25)
 *         hash_value = hash_value * 69069 + 907133923
 *         return hash_value             # <<<<<<<<<<<<<<
 * 
 *     cpdef list keys(self):
 */
  __pyx_r = __pyx_v_hash_value;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":702
 *         return not (self == other)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_hash_t hash_value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":717
 *         return hash_value
 * 
 *     cpdef list keys(self):             # <<<<<<<<<<<<<<
 *         return count_table_keys(self.table)
 * 
 */

static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_47keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_keys(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 717, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_2_c_11count_table_10CountTable_47keys)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 717, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 717, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/_c/count_table.pyx":718
 * 
 *     cpdef list keys(self):
 *         return count_table_keys(self.table)             # <<<<<<<<<<<<<<
 * 
 *     cpdef list values(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_keys(__pyx_v_self->table); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":717
 *         return hash_value
 * 
 *     cpdef list keys(self):             # <<<<<<<<<<<<<<
 *         return count_table_keys(self.table)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_47keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_47keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_46keys(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_46keys(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_10CountTable_keys(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":720
 *         return count_table_keys(self.table)
 * 
 *     cpdef list values(self):             # <<<<<<<<<<<<<<
 *         return count_table_values(self.table)
 * 
 */

static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_49values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_values(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_values); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_2_c_11count_table_10CountTable_49values)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 720, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 720, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/_c/count_table.pyx":721
 * 
 *     cpdef list values(self):
 *         return count_table_values(self.table)             # <<<<<<<<<<<<<<
 * 
 *     cpdef list items(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_values(__pyx_v_self->table); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":720
 *         return count_table_keys(self.table)
 * 
 *     cpdef list values(self):             # <<<<<<<<<<<<<<
 *         return count_table_values(self.table)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_49values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_49values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_48values(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_48values(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_10CountTable_values(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":723
 *         return count_table_values(self.table)
 * 
 *     cpdef list items(self):             # <<<<<<<<<<<<<<
 *         return count_table_items(self.table)
 * 
 */

static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_51items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_items(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("items", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 723, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_2_c_11count_table_10CountTable_51items)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 723, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 723, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/_c/count_table.pyx":724
 * 
 *     cpdef list items(self):
 *         return count_table_items(self.table)             # <<<<<<<<<<<<<<
 * 
 *     cpdef clear(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_items(__pyx_v_self->table); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":723
 *         return count_table_values(self.table)
 * 
 *     cpdef list items(self):             # <<<<<<<<<<<<<<
 *         return count_table_items(self.table)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_51items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_51items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_50items(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_50items(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("items", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_10CountTable_items(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":726
 *         return count_table_items(self.table)
 * 
 *     cpdef clear(self):             # <<<<<<<<<<<<<<
 *         count_table_clear(self.table)
 * 
 */

static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_53clear(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_clear(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clear); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_2_c_11count_table_10CountTable_53clear)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 726, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/_c/count_table.pyx":727
 * 
 *     cpdef clear(self):
 *         count_table_clear(self.table)             # <<<<<<<<<<<<<<
 * 
 *     cpdef setdefault(self, key, value):
 */
  __pyx_f_5glypy_2_c_11count_table_count_table_clear(__pyx_v_self->table);

  /* "glypy/_c/count_table.pyx":726
 *         return count_table_items(self.table)
 * 
 *     cpdef clear(self):             # <<<<<<<<<<<<<<
 *         count_table_clear(self.table)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_53clear(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_53clear(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_52clear(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_52clear(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_10CountTable_clear(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":729
 *         count_table_clear(self.table)
 * 
 *     cpdef setdefault(self, key, value):             # <<<<<<<<<<<<<<
 *         if self.getitem(key) == 0:
 *             self.setitem(key, value)
 */

static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_55setdefault(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_setdefault(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setdefault", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_setdefault); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 729, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_2_c_11count_table_10CountTable_55setdefault)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_key, __pyx_v_value};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 729, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_key, __pyx_v_value};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 729, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 729, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_key);
          __Pyx_GIVEREF(__pyx_v_key);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_key);
          __Pyx_INCREF(__pyx_v_value);
          __Pyx_GIVEREF(__pyx_v_value);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_value);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 729, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/_c/count_table.pyx":730
 * 
 *     cpdef setdefault(self, key, value):
 *         if self.getitem(key) == 0:             # <<<<<<<<<<<<<<
 *             self.setitem(key, value)
 * 
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_key); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 == 0) != 0);
  if (__pyx_t_8) {

    /* "glypy/_c/count_table.pyx":731
 *     cpdef setdefault(self, key, value):
 *         if self.getitem(key) == 0:
 *             self.setitem(key, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef long getitem(self, object key) except *:
 */
    __pyx_t_7 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_7 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 731, __pyx_L1_error)
    __pyx_t_5 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->setitem(__pyx_v_self, __pyx_v_key, __pyx_t_7); if (unlikely(__pyx_t_5 == ((int)1))) __PYX_ERR(0, 731, __pyx_L1_error)

    /* "glypy/_c/count_table.pyx":730
 * 
 *     cpdef setdefault(self, key, value):
 *         if self.getitem(key) == 0:             # <<<<<<<<<<<<<<
 *             self.setitem(key, value)
 * 
 */
  }

  /* "glypy/_c/count_table.pyx":729
 *         count_table_clear(self.table)
 * 
 *     cpdef setdefault(self, key, value):             # <<<<<<<<<<<<<<
 *         if self.getitem(key) == 0:
 *             self.setitem(key, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.setdefault", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_55setdefault(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_55setdefault(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setdefault (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_value,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setdefault", 1, 2, 2, 1); __PYX_ERR(0, 729, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setdefault") < 0)) __PYX_ERR(0, 729, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_key = values[0];
    __pyx_v_value = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setdefault", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 729, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.setdefault", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_54setdefault(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), __pyx_v_key, __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_54setdefault(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setdefault", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_10CountTable_setdefault(__pyx_v_self, __pyx_v_key, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.setdefault", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":733
 *             self.setitem(key, value)
 * 
 *     cdef long getitem(self, object key) except *:             # <<<<<<<<<<<<<<
 *         cdef long value
 *         cdef int status
 */

static long __pyx_f_5glypy_2_c_11count_table_10CountTable_getitem(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key) {
  long __pyx_v_value;
  int __pyx_v_status;
  PyObject *__pyx_v_pkey;
  long __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getitem", 0);

  /* "glypy/_c/count_table.pyx":736
 *         cdef long value
 *         cdef int status
 *         cdef PyObject* pkey = <PyObject*>key             # <<<<<<<<<<<<<<
 *         Py_INCREF(key)
 *         status = count_table_get(self.table, pkey, &value)
 */
  __pyx_v_pkey = ((PyObject *)__pyx_v_key);

  /* "glypy/_c/count_table.pyx":737
 *         cdef int status
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)             # <<<<<<<<<<<<<<
 *         status = count_table_get(self.table, pkey, &value)
 *         Py_DECREF(key)
 */
  Py_INCREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":738
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)
 *         status = count_table_get(self.table, pkey, &value)             # <<<<<<<<<<<<<<
 *         Py_DECREF(key)
 *         if status != 0:
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_get(__pyx_v_self->table, __pyx_v_pkey, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 738, __pyx_L1_error)
  __pyx_v_status = __pyx_t_1;

  /* "glypy/_c/count_table.pyx":739
 *         Py_INCREF(key)
 *         status = count_table_get(self.table, pkey, &value)
 *         Py_DECREF(key)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             return 0
 */
  Py_DECREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":740
 *         status = count_table_get(self.table, pkey, &value)
 *         Py_DECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         return value
 */
  __pyx_t_2 = ((__pyx_v_status != 0) != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":741
 *         Py_DECREF(key)
 *         if status != 0:
 *             return 0             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":740
 *         status = count_table_get(self.table, pkey, &value)
 *         Py_DECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         return value
 */
  }

  /* "glypy/_c/count_table.pyx":742
 *         if status != 0:
 *             return 0
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef int setitem(self, object key, long value) except 1:
 */
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":733
 *             self.setitem(key, value)
 * 
 *     cdef long getitem(self, object key) except *:             # <<<<<<<<<<<<<<
 *         cdef long value
 *         cdef int status
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.getitem", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":744
 *         return value
 * 
 *     cdef int setitem(self, object key, long value) except 1:             # <<<<<<<<<<<<<<
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)
 */

static int __pyx_f_5glypy_2_c_11count_table_10CountTable_setitem(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, long __pyx_v_value) {
  PyObject *__pyx_v_pkey;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem", 0);

  /* "glypy/_c/count_table.pyx":745
 * 
 *     cdef int setitem(self, object key, long value) except 1:
 *         cdef PyObject* pkey = <PyObject*>key             # <<<<<<<<<<<<<<
 *         Py_INCREF(key)
 *         cdef int status = count_table_put(self.table, pkey, value)
 */
  __pyx_v_pkey = ((PyObject *)__pyx_v_key);

  /* "glypy/_c/count_table.pyx":746
 *     cdef int setitem(self, object key, long value) except 1:
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)             # <<<<<<<<<<<<<<
 *         cdef int status = count_table_put(self.table, pkey, value)
 *         Py_DECREF(key)
 */
  Py_INCREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":747
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)
 *         cdef int status = count_table_put(self.table, pkey, value)             # <<<<<<<<<<<<<<
 *         Py_DECREF(key)
 *         if status != 0:
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_put(__pyx_v_self->table, __pyx_v_pkey, __pyx_v_value); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 747, __pyx_L1_error)
  __pyx_v_status = __pyx_t_1;

  /* "glypy/_c/count_table.pyx":748
 *         Py_INCREF(key)
 *         cdef int status = count_table_put(self.table, pkey, value)
 *         Py_DECREF(key)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             return 1
 */
  Py_DECREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":749
 *         cdef int status = count_table_put(self.table, pkey, value)
 *         Py_DECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  __pyx_t_2 = ((__pyx_v_status != 0) != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":750
 *         Py_DECREF(key)
 *         if status != 0:
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":749
 *         cdef int status = count_table_put(self.table, pkey, value)
 *         Py_DECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  }

  /* "glypy/_c/count_table.pyx":751
 *         if status != 0:
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef long delitem(self, object key) except *:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":744
 *         return value
 * 
 *     cdef int setitem(self, object key, long value) except 1:             # <<<<<<<<<<<<<<
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.setitem", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":753
 *         return 0
 * 
 *     cdef long delitem(self, object key) except *:             # <<<<<<<<<<<<<<
 *         cdef PyObject* pkey = <PyObject*>key
 *         cdef long value
 */

static long __pyx_f_5glypy_2_c_11count_table_10CountTable_delitem(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_pkey;
  long __pyx_v_value;
  int __pyx_v_status;
  long __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("delitem", 0);

  /* "glypy/_c/count_table.pyx":754
 * 
 *     cdef long delitem(self, object key) except *:
 *         cdef PyObject* pkey = <PyObject*>key             # <<<<<<<<<<<<<<
 *         cdef long value
 *         Py_INCREF(key)
 */
  __pyx_v_pkey = ((PyObject *)__pyx_v_key);

  /* "glypy/_c/count_table.pyx":756
 *         cdef PyObject* pkey = <PyObject*>key
 *         cdef long value
 *         Py_INCREF(key)             # <<<<<<<<<<<<<<
 *         cdef int status = count_table_del(self.table, pkey, &value)
 *         Py_DECREF(key)
 */
  Py_INCREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":757
 *         cdef long value
 *         Py_INCREF(key)
 *         cdef int status = count_table_del(self.table, pkey, &value)             # <<<<<<<<<<<<<<
 *         Py_DECREF(key)
 *         if status != 0:
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_del(__pyx_v_self->table, __pyx_v_pkey, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 757, __pyx_L1_error)
  __pyx_v_status = __pyx_t_1;

  /* "glypy/_c/count_table.pyx":758
 *         Py_INCREF(key)
 *         cdef int status = count_table_del(self.table, pkey, &value)
 *         Py_DECREF(key)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             return 0
 */
  Py_DECREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":759
 *         cdef int status = count_table_del(self.table, pkey, &value)
 *         Py_DECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         return value
 */
  __pyx_t_2 = ((__pyx_v_status != 0) != 0);
  if (__pyx_t_2) {

    /* "glypy/_c/count_table.pyx":760
 *         Py_DECREF(key)
 *         if status != 0:
 *             return 0             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":759
 *         cdef int status = count_table_del(self.table, pkey, &value)
 *         Py_DECREF(key)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         return value
 */
  }

  /* "glypy/_c/count_table.pyx":761
 *         if status != 0:
 *             return 0
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef void increment(self, object key, long value):
 */
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":753
 *         return 0
 * 
 *     cdef long delitem(self, object key) except *:             # <<<<<<<<<<<<<<
 *         cdef PyObject* pkey = <PyObject*>key
 *         cdef long value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.delitem", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":763
 *         return value
 * 
 *     cdef void increment(self, object key, long value):             # <<<<<<<<<<<<<<
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)
 */

static void __pyx_f_5glypy_2_c_11count_table_10CountTable_increment(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, long __pyx_v_value) {
  PyObject *__pyx_v_pkey;
  CYTHON_UNUSED int __pyx_v_status;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("increment", 0);

  /* "glypy/_c/count_table.pyx":764
 * 
 *     cdef void increment(self, object key, long value):
 *         cdef PyObject* pkey = <PyObject*>key             # <<<<<<<<<<<<<<
 *         Py_INCREF(key)
 *         cdef int status = count_table_increment(self.table, pkey, value)
 */
  __pyx_v_pkey = ((PyObject *)__pyx_v_key);

  /* "glypy/_c/count_table.pyx":765
 *     cdef void increment(self, object key, long value):
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)             # <<<<<<<<<<<<<<
 *         cdef int status = count_table_increment(self.table, pkey, value)
 *         Py_DECREF(key)
 */
  Py_INCREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":766
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)
 *         cdef int status = count_table_increment(self.table, pkey, value)             # <<<<<<<<<<<<<<
 *         Py_DECREF(key)
 * 
 */
  __pyx_v_status = __pyx_f_5glypy_2_c_11count_table_count_table_increment(__pyx_v_self->table, __pyx_v_pkey, __pyx_v_value);

  /* "glypy/_c/count_table.pyx":767
 *         Py_INCREF(key)
 *         cdef int status = count_table_increment(self.table, pkey, value)
 *         Py_DECREF(key)             # <<<<<<<<<<<<<<
 * 
 *     cdef void decrement(self, object key, long value):
 */
  Py_DECREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":763
 *         return value
 * 
 *     cdef void increment(self, object key, long value):             # <<<<<<<<<<<<<<
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":769
 *         Py_DECREF(key)
 * 
 *     cdef void decrement(self, object key, long value):             # <<<<<<<<<<<<<<
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)
 */

static void __pyx_f_5glypy_2_c_11count_table_10CountTable_decrement(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, long __pyx_v_value) {
  PyObject *__pyx_v_pkey;
  CYTHON_UNUSED int __pyx_v_status;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("decrement", 0);

  /* "glypy/_c/count_table.pyx":770
 * 
 *     cdef void decrement(self, object key, long value):
 *         cdef PyObject* pkey = <PyObject*>key             # <<<<<<<<<<<<<<
 *         Py_INCREF(key)
 *         cdef int status = count_table_decrement(self.table, pkey, value)
 */
  __pyx_v_pkey = ((PyObject *)__pyx_v_key);

  /* "glypy/_c/count_table.pyx":771
 *     cdef void decrement(self, object key, long value):
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)             # <<<<<<<<<<<<<<
 *         cdef int status = count_table_decrement(self.table, pkey, value)
 *         Py_DECREF(key)
 */
  Py_INCREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":772
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)
 *         cdef int status = count_table_decrement(self.table, pkey, value)             # <<<<<<<<<<<<<<
 *         Py_DECREF(key)
 * 
 */
  __pyx_v_status = __pyx_f_5glypy_2_c_11count_table_count_table_decrement(__pyx_v_self->table, __pyx_v_pkey, __pyx_v_value);

  /* "glypy/_c/count_table.pyx":773
 *         Py_INCREF(key)
 *         cdef int status = count_table_decrement(self.table, pkey, value)
 *         Py_DECREF(key)             # <<<<<<<<<<<<<<
 * 
 *     cpdef long pop(self, object key, object default=None) except *:
 */
  Py_DECREF(__pyx_v_key);

  /* "glypy/_c/count_table.pyx":769
 *         Py_DECREF(key)
 * 
 *     cdef void decrement(self, object key, long value):             # <<<<<<<<<<<<<<
 *         cdef PyObject* pkey = <PyObject*>key
 *         Py_INCREF(key)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "glypy/_c/count_table.pyx":775
 *         Py_DECREF(key)
 * 
 *     cpdef long pop(self, object key, object default=None) except *:             # <<<<<<<<<<<<<<
 *         cdef long value = self.delitem(key)
 *         if value == 0:
 */

static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_57pop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static long __pyx_f_5glypy_2_c_11count_table_10CountTable_pop(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_pop *__pyx_optional_args) {
  PyObject *__pyx_v_default = ((PyObject *)Py_None);
  long __pyx_v_value;
  long __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default = __pyx_optional_args->__pyx_default;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 775, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_2_c_11count_table_10CountTable_57pop)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_key, __pyx_v_default};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 775, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_key, __pyx_v_default};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 775, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 775, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_key);
          __Pyx_GIVEREF(__pyx_v_key);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_key);
          __Pyx_INCREF(__pyx_v_default);
          __Pyx_GIVEREF(__pyx_v_default);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_default);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 775, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_7 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 775, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/_c/count_table.pyx":776
 * 
 *     cpdef long pop(self, object key, object default=None) except *:
 *         cdef long value = self.delitem(key)             # <<<<<<<<<<<<<<
 *         if value == 0:
 *             return default
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->delitem(__pyx_v_self, __pyx_v_key); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L1_error)
  __pyx_v_value = __pyx_t_7;

  /* "glypy/_c/count_table.pyx":777
 *     cpdef long pop(self, object key, object default=None) except *:
 *         cdef long value = self.delitem(key)
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             return default
 *         return PyInt_FromLong(value)
 */
  __pyx_t_8 = ((__pyx_v_value == 0) != 0);
  if (__pyx_t_8) {

    /* "glypy/_c/count_table.pyx":778
 *         cdef long value = self.delitem(key)
 *         if value == 0:
 *             return default             # <<<<<<<<<<<<<<
 *         return PyInt_FromLong(value)
 * 
 */
    __pyx_t_7 = __Pyx_PyInt_As_long(__pyx_v_default); if (unlikely((__pyx_t_7 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 778, __pyx_L1_error)
    __pyx_r = __pyx_t_7;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":777
 *     cpdef long pop(self, object key, object default=None) except *:
 *         cdef long value = self.delitem(key)
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             return default
 *         return PyInt_FromLong(value)
 */
  }

  /* "glypy/_c/count_table.pyx":779
 *         if value == 0:
 *             return default
 *         return PyInt_FromLong(value)             # <<<<<<<<<<<<<<
 * 
 *     cpdef get(self, object key, object default=None):
 */
  __pyx_t_1 = PyInt_FromLong(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_7 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":775
 *         Py_DECREF(key)
 * 
 *     cpdef long pop(self, object key, object default=None) except *:             # <<<<<<<<<<<<<<
 *         cdef long value = self.delitem(key)
 *         if value == 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_57pop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_57pop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_default = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_default,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_default);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pop") < 0)) __PYX_ERR(0, 775, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_key = values[0];
    __pyx_v_default = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pop", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 775, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_56pop(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), __pyx_v_key, __pyx_v_default);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_56pop(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_default) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_pop __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.__pyx_default = __pyx_v_default;
  __pyx_t_1 = __pyx_vtabptr_5glypy_2_c_11count_table_CountTable->pop(__pyx_v_self, __pyx_v_key, 1, &__pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 775, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":781
 *         return PyInt_FromLong(value)
 * 
 *     cpdef get(self, object key, object default=None):             # <<<<<<<<<<<<<<
 *         cdef long value = self.getitem(key)
 *         if value == 0:
 */

static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_59get(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5glypy_2_c_11count_table_10CountTable_get(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_get *__pyx_optional_args) {
  PyObject *__pyx_v_default = ((PyObject *)Py_None);
  long __pyx_v_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default = __pyx_optional_args->__pyx_default;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 781, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_2_c_11count_table_10CountTable_59get)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_key, __pyx_v_default};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_key, __pyx_v_default};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 781, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_key);
          __Pyx_GIVEREF(__pyx_v_key);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_key);
          __Pyx_INCREF(__pyx_v_default);
          __Pyx_GIVEREF(__pyx_v_default);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_default);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/_c/count_table.pyx":782
 * 
 *     cpdef get(self, object key, object default=None):
 *         cdef long value = self.getitem(key)             # <<<<<<<<<<<<<<
 *         if value == 0:
 *             return default
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_key); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 782, __pyx_L1_error)
  __pyx_v_value = __pyx_t_7;

  /* "glypy/_c/count_table.pyx":783
 *     cpdef get(self, object key, object default=None):
 *         cdef long value = self.getitem(key)
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             return default
 *         return PyInt_FromLong(value)
 */
  __pyx_t_8 = ((__pyx_v_value == 0) != 0);
  if (__pyx_t_8) {

    /* "glypy/_c/count_table.pyx":784
 *         cdef long value = self.getitem(key)
 *         if value == 0:
 *             return default             # <<<<<<<<<<<<<<
 *         return PyInt_FromLong(value)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_default);
    __pyx_r = __pyx_v_default;
    goto __pyx_L0;

    /* "glypy/_c/count_table.pyx":783
 *     cpdef get(self, object key, object default=None):
 *         cdef long value = self.getitem(key)
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             return default
 *         return PyInt_FromLong(value)
 */
  }

  /* "glypy/_c/count_table.pyx":785
 *         if value == 0:
 *             return default
 *         return PyInt_FromLong(value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromLong(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/_c/count_table.pyx":781
 *         return PyInt_FromLong(value)
 * 
 *     cpdef get(self, object key, object default=None):             # <<<<<<<<<<<<<<
 *         cdef long value = self.getitem(key)
 *         if value == 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_59get(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5glypy_2_c_11count_table_10CountTable_59get(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_default = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_default,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_default);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get") < 0)) __PYX_ERR(0, 781, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_key = values[0];
    __pyx_v_default = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 781, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_10CountTable_58get(((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_v_self), __pyx_v_key, __pyx_v_default);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_10CountTable_58get(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_default) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_get __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.__pyx_default = __pyx_v_default;
  __pyx_t_1 = __pyx_vtabptr_5glypy_2_c_11count_table_CountTable->get(__pyx_v_self, __pyx_v_key, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy._c.count_table.CountTable.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/_c/count_table.pyx":792
 * 
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef long val
 *     cdef str string
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_2_c_11count_table_1main(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_5glypy_2_c_11count_table_1main = {"main", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_1main, METH_NOARGS, 0};
static PyObject *__pyx_pw_5glypy_2_c_11count_table_1main(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("main (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_2_c_11count_table_main(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_2_c_11count_table_main(CYTHON_UNUSED PyObject *__pyx_self) {
  long __pyx_v_val;
  PyObject *__pyx_v_string = 0;
  PyObject *__pyx_v_pint = 0;
  PyObject *__pyx_v_key;
  struct __pyx_t_5glypy_2_c_11count_table_count_table *__pyx_v_table;
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *__pyx_v_pytable = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("main", 0);

  /* "glypy/_c/count_table.pyx":797
 *     cdef object pint
 *     cdef PyObject* key
 *     string = "spam"             # <<<<<<<<<<<<<<
 *     pint = 12
 *     key = <PyObject*>string
 */
  __Pyx_INCREF(__pyx_n_s_spam);
  __pyx_v_string = __pyx_n_s_spam;

  /* "glypy/_c/count_table.pyx":798
 *     cdef PyObject* key
 *     string = "spam"
 *     pint = 12             # <<<<<<<<<<<<<<
 *     key = <PyObject*>string
 *     cdef count_table* table = make_count_table(6, 2)
 */
  __Pyx_INCREF(__pyx_int_12);
  __pyx_v_pint = __pyx_int_12;

  /* "glypy/_c/count_table.pyx":799
 *     string = "spam"
 *     pint = 12
 *     key = <PyObject*>string             # <<<<<<<<<<<<<<
 *     cdef count_table* table = make_count_table(6, 2)
 *     count_table_put(table, key, 42)
 */
  __pyx_v_key = ((PyObject *)__pyx_v_string);

  /* "glypy/_c/count_table.pyx":800
 *     pint = 12
 *     key = <PyObject*>string
 *     cdef count_table* table = make_count_table(6, 2)             # <<<<<<<<<<<<<<
 *     count_table_put(table, key, 42)
 *     count_table_get(table, key, &val)
 */
  __pyx_v_table = __pyx_f_5glypy_2_c_11count_table_make_count_table(6, 2);

  /* "glypy/_c/count_table.pyx":801
 *     key = <PyObject*>string
 *     cdef count_table* table = make_count_table(6, 2)
 *     count_table_put(table, key, 42)             # <<<<<<<<<<<<<<
 *     count_table_get(table, key, &val)
 *     assert val == 42
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_put(__pyx_v_table, __pyx_v_key, 42); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 801, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":802
 *     cdef count_table* table = make_count_table(6, 2)
 *     count_table_put(table, key, 42)
 *     count_table_get(table, key, &val)             # <<<<<<<<<<<<<<
 *     assert val == 42
 *     key = <PyObject*>pint
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_get(__pyx_v_table, __pyx_v_key, (&__pyx_v_val)); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 802, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":803
 *     count_table_put(table, key, 42)
 *     count_table_get(table, key, &val)
 *     assert val == 42             # <<<<<<<<<<<<<<
 *     key = <PyObject*>pint
 *     count_table_put(table, key, 67)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_val == 42) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 803, __pyx_L1_error)
    }
  }
  #endif

  /* "glypy/_c/count_table.pyx":804
 *     count_table_get(table, key, &val)
 *     assert val == 42
 *     key = <PyObject*>pint             # <<<<<<<<<<<<<<
 *     count_table_put(table, key, 67)
 *     count_table_del(table, key, &val)
 */
  __pyx_v_key = ((PyObject *)__pyx_v_pint);

  /* "glypy/_c/count_table.pyx":805
 *     assert val == 42
 *     key = <PyObject*>pint
 *     count_table_put(table, key, 67)             # <<<<<<<<<<<<<<
 *     count_table_del(table, key, &val)
 *     assert val == 67
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_put(__pyx_v_table, __pyx_v_key, 67); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 805, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":806
 *     key = <PyObject*>pint
 *     count_table_put(table, key, 67)
 *     count_table_del(table, key, &val)             # <<<<<<<<<<<<<<
 *     assert val == 67
 *     val = 1
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_del(__pyx_v_table, __pyx_v_key, (&__pyx_v_val)); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 806, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":807
 *     count_table_put(table, key, 67)
 *     count_table_del(table, key, &val)
 *     assert val == 67             # <<<<<<<<<<<<<<
 *     val = 1
 *     count_table_get(table, key, &val)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_val == 67) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 807, __pyx_L1_error)
    }
  }
  #endif

  /* "glypy/_c/count_table.pyx":808
 *     count_table_del(table, key, &val)
 *     assert val == 67
 *     val = 1             # <<<<<<<<<<<<<<
 *     count_table_get(table, key, &val)
 *     free_count_table(table)
 */
  __pyx_v_val = 1;

  /* "glypy/_c/count_table.pyx":809
 *     assert val == 67
 *     val = 1
 *     count_table_get(table, key, &val)             # <<<<<<<<<<<<<<
 *     free_count_table(table)
 *     pytable = CountTable()
 */
  __pyx_t_1 = __pyx_f_5glypy_2_c_11count_table_count_table_get(__pyx_v_table, __pyx_v_key, (&__pyx_v_val)); if (unlikely(__pyx_t_1 == ((int)1))) __PYX_ERR(0, 809, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":810
 *     val = 1
 *     count_table_get(table, key, &val)
 *     free_count_table(table)             # <<<<<<<<<<<<<<
 *     pytable = CountTable()
 *     Py_INCREF(<object>key)
 */
  __pyx_f_5glypy_2_c_11count_table_free_count_table(__pyx_v_table);

  /* "glypy/_c/count_table.pyx":811
 *     count_table_get(table, key, &val)
 *     free_count_table(table)
 *     pytable = CountTable()             # <<<<<<<<<<<<<<
 *     Py_INCREF(<object>key)
 *     pytable[<object>key] = 2
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTable)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_pytable = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glypy/_c/count_table.pyx":812
 *     free_count_table(table)
 *     pytable = CountTable()
 *     Py_INCREF(<object>key)             # <<<<<<<<<<<<<<
 *     pytable[<object>key] = 2
 *     pytable[int(12)] = 11
 */
  Py_INCREF(((PyObject *)__pyx_v_key));

  /* "glypy/_c/count_table.pyx":813
 *     pytable = CountTable()
 *     Py_INCREF(<object>key)
 *     pytable[<object>key] = 2             # <<<<<<<<<<<<<<
 *     pytable[int(12)] = 11
 *     pytable[255.2] = 99
 */
  if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_pytable), ((PyObject *)__pyx_v_key), __pyx_int_2) < 0)) __PYX_ERR(0, 813, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":814
 *     Py_INCREF(<object>key)
 *     pytable[<object>key] = 2
 *     pytable[int(12)] = 11             # <<<<<<<<<<<<<<
 *     pytable[255.2] = 99
 *     assert len(pytable) == 2
 */
  __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_int_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_pytable), __pyx_t_2, __pyx_int_11) < 0)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glypy/_c/count_table.pyx":815
 *     pytable[<object>key] = 2
 *     pytable[int(12)] = 11
 *     pytable[255.2] = 99             # <<<<<<<<<<<<<<
 *     assert len(pytable) == 2
 *     del pytable
 */
  if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_pytable), __pyx_float_255_2, __pyx_int_99) < 0)) __PYX_ERR(0, 815, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":816
 *     pytable[int(12)] = 11
 *     pytable[255.2] = 99
 *     assert len(pytable) == 2             # <<<<<<<<<<<<<<
 *     del pytable
 *     pytable = CountTable._create()
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_3 = PyObject_Length(((PyObject *)__pyx_v_pytable)); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 816, __pyx_L1_error)
    if (unlikely(!((__pyx_t_3 == 2) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 816, __pyx_L1_error)
    }
  }
  #endif

  /* "glypy/_c/count_table.pyx":817
 *     pytable[255.2] = 99
 *     assert len(pytable) == 2
 *     del pytable             # <<<<<<<<<<<<<<
 *     pytable = CountTable._create()
 *     print pytable
 */
  __Pyx_DECREF(((PyObject *)__pyx_v_pytable));
  __pyx_v_pytable = NULL;

  /* "glypy/_c/count_table.pyx":818
 *     assert len(pytable) == 2
 *     del pytable
 *     pytable = CountTable._create()             # <<<<<<<<<<<<<<
 *     print pytable
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_5glypy_2_c_11count_table_10CountTable__create()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_pytable = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glypy/_c/count_table.pyx":819
 *     del pytable
 *     pytable = CountTable._create()
 *     print pytable             # <<<<<<<<<<<<<<
 */
  if (__Pyx_PrintOne(0, ((PyObject *)__pyx_v_pytable)) < 0) __PYX_ERR(0, 819, __pyx_L1_error)

  /* "glypy/_c/count_table.pyx":792
 * 
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef long val
 *     cdef str string
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glypy._c.count_table.main", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_string);
  __Pyx_XDECREF(__pyx_v_pint);
  __Pyx_XDECREF((PyObject *)__pyx_v_pytable);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTableIterator __pyx_vtable_5glypy_2_c_11count_table_CountTableIterator;

static PyObject *__pyx_tp_new_5glypy_2_c_11count_table_CountTableIterator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *)o);
  p->__pyx_vtab = __pyx_vtabptr_5glypy_2_c_11count_table_CountTableIterator;
  return o;
}

static void __pyx_tp_dealloc_5glypy_2_c_11count_table_CountTableIterator(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_5__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {return __pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_5__next__(self);}

static PyMethodDef __pyx_methods_5glypy_2_c_11count_table_CountTableIterator[] = {
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_5__next__, METH_NOARGS|METH_COEXIST, 0},
  {"_test", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_7_test, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5glypy_2_c_11count_table_CountTableIterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "glypy._c.count_table.CountTableIterator", /*tp_name*/
  sizeof(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5glypy_2_c_11count_table_CountTableIterator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_3__iter__, /*tp_iter*/
  __pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_5__next__, /*tp_iternext*/
  __pyx_methods_5glypy_2_c_11count_table_CountTableIterator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5glypy_2_c_11count_table_18CountTableIterator_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5glypy_2_c_11count_table_CountTableIterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_5glypy_2_c_11count_table_CountTable __pyx_vtable_5glypy_2_c_11count_table_CountTable;

static PyObject *__pyx_tp_new_5glypy_2_c_11count_table_CountTable(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5glypy_2_c_11count_table_CountTable *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5glypy_2_c_11count_table_CountTable *)o);
  p->__pyx_vtab = __pyx_vtabptr_5glypy_2_c_11count_table_CountTable;
  return o;
}

static void __pyx_tp_dealloc_5glypy_2_c_11count_table_CountTable(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_5glypy_2_c_11count_table_10CountTable_9__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_5glypy_2_c_11count_table_CountTable(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_5glypy_2_c_11count_table_CountTable(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_5glypy_2_c_11count_table_10CountTable_13__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_5glypy_2_c_11count_table_10CountTable_15__delitem__(o, i);
  }
}

static PyObject *__pyx_tp_richcompare_5glypy_2_c_11count_table_CountTable(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_5glypy_2_c_11count_table_10CountTable_41__eq__(o1, o2);
    }
    case Py_NE: {
      return __pyx_pw_5glypy_2_c_11count_table_10CountTable_43__ne__(o1, o2);
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyMethodDef __pyx_methods_5glypy_2_c_11count_table_CountTable[] = {
  {"update", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_10CountTable_3update, METH_O, 0},
  {"copy", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_10CountTable_5copy, METH_NOARGS, 0},
  {"__reduce__", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_10CountTable_7__reduce__, METH_NOARGS, 0},
  {"_to_dict", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_10CountTable_37_to_dict, METH_NOARGS, 0},
  {"keys", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_10CountTable_47keys, METH_NOARGS, 0},
  {"values", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_10CountTable_49values, METH_NOARGS, 0},
  {"items", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_10CountTable_51items, METH_NOARGS, 0},
  {"clear", (PyCFunction)__pyx_pw_5glypy_2_c_11count_table_10CountTable_53clear, METH_NOARGS, 0},
  {"setdefault", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5glypy_2_c_11count_table_10CountTable_55setdefault, METH_VARARGS|METH_KEYWORDS, 0},
  {"pop", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5glypy_2_c_11count_table_10CountTable_57pop, METH_VARARGS|METH_KEYWORDS, 0},
  {"get", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5glypy_2_c_11count_table_10CountTable_59get, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static PyNumberMethods __pyx_tp_as_number_CountTable = {
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_23__add__, /*nb_add*/
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_27__sub__, /*nb_subtract*/
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_31__mul__, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_35__neg__, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  0, /*nb_nonzero*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_25__iadd__, /*nb_inplace_add*/
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_29__isub__, /*nb_inplace_subtract*/
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_33__imul__, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_CountTable = {
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_17__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_5glypy_2_c_11count_table_CountTable, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_19__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_CountTable = {
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_17__len__, /*mp_length*/
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_11__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_5glypy_2_c_11count_table_CountTable, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_5glypy_2_c_11count_table_CountTable = {
  PyVarObject_HEAD_INIT(0, 0)
  "glypy._c.count_table.CountTable", /*tp_name*/
  sizeof(struct __pyx_obj_5glypy_2_c_11count_table_CountTable), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5glypy_2_c_11count_table_CountTable, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_39__repr__, /*tp_repr*/
  &__pyx_tp_as_number_CountTable, /*tp_as_number*/
  &__pyx_tp_as_sequence_CountTable, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_CountTable, /*tp_as_mapping*/
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_45__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_tp_richcompare_5glypy_2_c_11count_table_CountTable, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_21__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5glypy_2_c_11count_table_CountTable, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5glypy_2_c_11count_table_10CountTable_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5glypy_2_c_11count_table_CountTable, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_count_table(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_count_table},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "count_table",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_CountTable, __pyx_k_CountTable, sizeof(__pyx_k_CountTable), 0, 0, 1, 1},
  {&__pyx_n_s_CountTableIterator, __pyx_k_CountTableIterator, sizeof(__pyx_k_CountTableIterator), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_Mapping, __pyx_k_Mapping, sizeof(__pyx_k_Mapping), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_MutableMapping, __pyx_k_MutableMapping, sizeof(__pyx_k_MutableMapping), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplemented, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_clear, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_collections_abc, __pyx_k_collections_abc, sizeof(__pyx_k_collections_abc), 0, 0, 1, 1},
  {&__pyx_n_s_content, __pyx_k_content, sizeof(__pyx_k_content), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_glypy__c_count_table, __pyx_k_glypy__c_count_table, sizeof(__pyx_k_glypy__c_count_table), 0, 0, 1, 1},
  {&__pyx_kp_s_glypy__c_count_table_pyx, __pyx_k_glypy__c_count_table_pyx, sizeof(__pyx_k_glypy__c_count_table_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_main_2, __pyx_k_main_2, sizeof(__pyx_k_main_2), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_pint, __pyx_k_pint, sizeof(__pyx_k_pint), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_pytable, __pyx_k_pytable, sizeof(__pyx_k_pytable), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_kp_s_r_is_not_hashable, __pyx_k_r_is_not_hashable, sizeof(__pyx_k_r_is_not_hashable), 0, 0, 1, 0},
  {&__pyx_kp_s_r_must_be_a_number, __pyx_k_r_must_be_a_number, sizeof(__pyx_k_r_must_be_a_number), 0, 0, 1, 0},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_register, __pyx_k_register, sizeof(__pyx_k_register), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self___class_____name___content, __pyx_k_self___class_____name___content, sizeof(__pyx_k_self___class_____name___content), 0, 0, 1, 0},
  {&__pyx_kp_s_self_next_key_self_table_cannot, __pyx_k_self_next_key_self_table_cannot, sizeof(__pyx_k_self_next_key_self_table_cannot), 0, 0, 1, 0},
  {&__pyx_n_s_setdefault, __pyx_k_setdefault, sizeof(__pyx_k_setdefault), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_spam, __pyx_k_spam, sizeof(__pyx_k_spam), 0, 0, 1, 1},
  {&__pyx_n_s_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_table, __pyx_k_table, sizeof(__pyx_k_table), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_to_dict, __pyx_k_to_dict, sizeof(__pyx_k_to_dict), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(0, 11, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 19, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 69, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 92, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 434, __pyx_L1_error)
  __pyx_builtin_NotImplemented = __Pyx_GetBuiltinName(__pyx_n_s_NotImplemented); if (!__pyx_builtin_NotImplemented) __PYX_ERR(0, 626, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_self_next_key_self_table_cannot); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.next_key,self.table cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_self_next_key_self_table_cannot); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "glypy/_c/count_table.pyx":792
 * 
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef long val
 *     cdef str string
 */
  __pyx_tuple__3 = PyTuple_Pack(6, __pyx_n_s_val, __pyx_n_s_string, __pyx_n_s_pint, __pyx_n_s_key, __pyx_n_s_table, __pyx_n_s_pytable); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);
  __pyx_codeobj__4 = (PyObject*)__Pyx_PyCode_New(0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__3, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_glypy__c_count_table_pyx, __pyx_n_s_main_2, 792, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__4)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_255_2 = PyFloat_FromDouble(255.2); if (unlikely(!__pyx_float_255_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_11 = PyInt_FromLong(11); if (unlikely(!__pyx_int_11)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_12 = PyInt_FromLong(12); if (unlikely(!__pyx_int_12)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_99 = PyInt_FromLong(99); if (unlikely(!__pyx_int_99)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("initialize_count_table_bin", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_initialize_count_table_bin, "int (struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("free_count_table_bin", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_free_count_table_bin, "void (struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_bin_append", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_bin_append, "int (struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *, PyObject *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_bin_find", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_bin_find, "int (struct __pyx_t_5glypy_2_c_11count_table_count_table_bin *, PyObject *, Py_ssize_t *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("make_count_table", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_make_count_table, "struct __pyx_t_5glypy_2_c_11count_table_count_table *(size_t, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("free_count_table", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_free_count_table, "void (struct __pyx_t_5glypy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_find_bin", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_find_bin, "int (struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, Py_ssize_t *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_put", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_put, "int (struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_del", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_del, "int (struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, long *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_get", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_get, "int (struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, long *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_increment", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_increment, "int (struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_decrement", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_decrement, "int (struct __pyx_t_5glypy_2_c_11count_table_count_table *, PyObject *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_count", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_count, "Py_ssize_t (struct __pyx_t_5glypy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_keys", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_keys, "PyObject *(struct __pyx_t_5glypy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_values", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_values, "PyObject *(struct __pyx_t_5glypy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_items", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_items, "PyObject *(struct __pyx_t_5glypy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_add", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_add, "void (struct __pyx_t_5glypy_2_c_11count_table_count_table *, struct __pyx_t_5glypy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_subtract", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_subtract, "void (struct __pyx_t_5glypy_2_c_11count_table_count_table *, struct __pyx_t_5glypy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_scale", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_scale, "void (struct __pyx_t_5glypy_2_c_11count_table_count_table *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_update", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_update, "void (struct __pyx_t_5glypy_2_c_11count_table_count_table *, struct __pyx_t_5glypy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_clear", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_clear, "void (struct __pyx_t_5glypy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_copy", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_copy, "struct __pyx_t_5glypy_2_c_11count_table_count_table *(struct __pyx_t_5glypy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("count_table_equals", (void (*)(void))__pyx_f_5glypy_2_c_11count_table_count_table_equals, "int (struct __pyx_t_5glypy_2_c_11count_table_count_table *, struct __pyx_t_5glypy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_5glypy_2_c_11count_table_CountTableIterator = &__pyx_vtable_5glypy_2_c_11count_table_CountTableIterator;
  __pyx_vtable_5glypy_2_c_11count_table_CountTableIterator._create = (struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *(*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *))__pyx_f_5glypy_2_c_11count_table_18CountTableIterator__create;
  __pyx_vtable_5glypy_2_c_11count_table_CountTableIterator.advance = (void (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *))__pyx_f_5glypy_2_c_11count_table_18CountTableIterator_advance;
  __pyx_vtable_5glypy_2_c_11count_table_CountTableIterator.has_more = (int (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *))__pyx_f_5glypy_2_c_11count_table_18CountTableIterator_has_more;
  __pyx_vtable_5glypy_2_c_11count_table_CountTableIterator._locate_next_value = (int (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *, PyObject **, long *))__pyx_f_5glypy_2_c_11count_table_18CountTableIterator__locate_next_value;
  __pyx_vtable_5glypy_2_c_11count_table_CountTableIterator.get_next_value = (int (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTableIterator *, PyObject **, long *))__pyx_f_5glypy_2_c_11count_table_18CountTableIterator_get_next_value;
  if (PyType_Ready(&__pyx_type_5glypy_2_c_11count_table_CountTableIterator) < 0) __PYX_ERR(0, 348, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5glypy_2_c_11count_table_CountTableIterator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5glypy_2_c_11count_table_CountTableIterator.tp_dictoffset && __pyx_type_5glypy_2_c_11count_table_CountTableIterator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5glypy_2_c_11count_table_CountTableIterator.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_5glypy_2_c_11count_table_CountTableIterator.tp_dict, __pyx_vtabptr_5glypy_2_c_11count_table_CountTableIterator) < 0) __PYX_ERR(0, 348, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CountTableIterator, (PyObject *)&__pyx_type_5glypy_2_c_11count_table_CountTableIterator) < 0) __PYX_ERR(0, 348, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5glypy_2_c_11count_table_CountTableIterator) < 0) __PYX_ERR(0, 348, __pyx_L1_error)
  __pyx_ptype_5glypy_2_c_11count_table_CountTableIterator = &__pyx_type_5glypy_2_c_11count_table_CountTableIterator;
  __pyx_vtabptr_5glypy_2_c_11count_table_CountTable = &__pyx_vtable_5glypy_2_c_11count_table_CountTable;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable._create = (struct __pyx_obj_5glypy_2_c_11count_table_CountTable *(*)(void))__pyx_f_5glypy_2_c_11count_table_10CountTable__create;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable._create_from = (struct __pyx_obj_5glypy_2_c_11count_table_CountTable *(*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *))__pyx_f_5glypy_2_c_11count_table_10CountTable__create_from;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable._initialize_table = (int (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *))__pyx_f_5glypy_2_c_11count_table_10CountTable__initialize_table;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable._update_from_dict = (void (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *))__pyx_f_5glypy_2_c_11count_table_10CountTable__update_from_dict;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable._update_from_count_table = (void (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *))__pyx_f_5glypy_2_c_11count_table_10CountTable__update_from_count_table;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable._add_from = (void (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *))__pyx_f_5glypy_2_c_11count_table_10CountTable__add_from;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable._subtract_from = (void (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *))__pyx_f_5glypy_2_c_11count_table_10CountTable__subtract_from;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable._add_from_dict = (int (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *))__pyx_f_5glypy_2_c_11count_table_10CountTable__add_from_dict;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable._subtract_from_dict = (int (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *))__pyx_f_5glypy_2_c_11count_table_10CountTable__subtract_from_dict;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable._scale_by = (void (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, long))__pyx_f_5glypy_2_c_11count_table_10CountTable__scale_by;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.equal_to = (int (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, struct __pyx_obj_5glypy_2_c_11count_table_CountTable *))__pyx_f_5glypy_2_c_11count_table_10CountTable_equal_to;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable._to_dict = (PyObject *(*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch))__pyx_f_5glypy_2_c_11count_table_10CountTable__to_dict;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.update = (PyObject *(*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5glypy_2_c_11count_table_10CountTable_update;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.copy = (struct __pyx_obj_5glypy_2_c_11count_table_CountTable *(*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch))__pyx_f_5glypy_2_c_11count_table_10CountTable_copy;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.items = (PyObject *(*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch))__pyx_f_5glypy_2_c_11count_table_10CountTable_items;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.values = (PyObject *(*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch))__pyx_f_5glypy_2_c_11count_table_10CountTable_values;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.keys = (PyObject *(*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch))__pyx_f_5glypy_2_c_11count_table_10CountTable_keys;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.clear = (PyObject *(*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch))__pyx_f_5glypy_2_c_11count_table_10CountTable_clear;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.setdefault = (PyObject *(*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5glypy_2_c_11count_table_10CountTable_setdefault;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.pop = (long (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_pop *__pyx_optional_args))__pyx_f_5glypy_2_c_11count_table_10CountTable_pop;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.get = (PyObject *(*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_2_c_11count_table_10CountTable_get *__pyx_optional_args))__pyx_f_5glypy_2_c_11count_table_10CountTable_get;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.increment = (void (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, long))__pyx_f_5glypy_2_c_11count_table_10CountTable_increment;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.decrement = (void (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, long))__pyx_f_5glypy_2_c_11count_table_10CountTable_decrement;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.getitem = (long (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *))__pyx_f_5glypy_2_c_11count_table_10CountTable_getitem;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.setitem = (int (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *, long))__pyx_f_5glypy_2_c_11count_table_10CountTable_setitem;
  __pyx_vtable_5glypy_2_c_11count_table_CountTable.delitem = (long (*)(struct __pyx_obj_5glypy_2_c_11count_table_CountTable *, PyObject *))__pyx_f_5glypy_2_c_11count_table_10CountTable_delitem;
  if (PyType_Ready(&__pyx_type_5glypy_2_c_11count_table_CountTable) < 0) __PYX_ERR(0, 462, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5glypy_2_c_11count_table_CountTable.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5glypy_2_c_11count_table_CountTable.tp_dictoffset && __pyx_type_5glypy_2_c_11count_table_CountTable.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5glypy_2_c_11count_table_CountTable.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_5glypy_2_c_11count_table_CountTable.tp_dict, __pyx_vtabptr_5glypy_2_c_11count_table_CountTable) < 0) __PYX_ERR(0, 462, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CountTable, (PyObject *)&__pyx_type_5glypy_2_c_11count_table_CountTable) < 0) __PYX_ERR(0, 462, __pyx_L1_error)
  __pyx_ptype_5glypy_2_c_11count_table_CountTable = &__pyx_type_5glypy_2_c_11count_table_CountTable;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(3, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(4, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcount_table(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcount_table(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_count_table(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_count_table(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_count_table(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'count_table' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_count_table(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("count_table", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_glypy___c__count_table) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "glypy._c.count_table")) {
      if (unlikely(PyDict_SetItemString(modules, "glypy._c.count_table", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "glypy/_c/count_table.pyx":1
 * import sys             # <<<<<<<<<<<<<<
 * cimport cython
 * from cpython cimport PyErr_SetString, PyObject_Hash
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glypy/_c/count_table.pyx":9
 * from cpython.dict cimport PyDict_Next, PyDict_SetItem
 * 
 * try:             # <<<<<<<<<<<<<<
 *     from collections import Mapping, MutableMapping
 * except ImportError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "glypy/_c/count_table.pyx":10
 * 
 * try:
 *     from collections import Mapping, MutableMapping             # <<<<<<<<<<<<<<
 * except ImportError:
 *     from collections.abc import Mapping, MutableMapping
 */
      __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_n_s_Mapping);
      __Pyx_GIVEREF(__pyx_n_s_Mapping);
      PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Mapping);
      __Pyx_INCREF(__pyx_n_s_MutableMapping);
      __Pyx_GIVEREF(__pyx_n_s_MutableMapping);
      PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_MutableMapping);
      __pyx_t_5 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, -1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 10, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_Mapping); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_Mapping, __pyx_t_1) < 0) __PYX_ERR(0, 10, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_MutableMapping, __pyx_t_1) < 0) __PYX_ERR(0, 10, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "glypy/_c/count_table.pyx":9
 * from cpython.dict cimport PyDict_Next, PyDict_SetItem
 * 
 * try:             # <<<<<<<<<<<<<<
 *     from collections import Mapping, MutableMapping
 * except ImportError:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L7_try_end;
    __pyx_L2_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "glypy/_c/count_table.pyx":11
 * try:
 *     from collections import Mapping, MutableMapping
 * except ImportError:             # <<<<<<<<<<<<<<
 *     from collections.abc import Mapping, MutableMapping
 * 
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("glypy._c.count_table", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_7) < 0) __PYX_ERR(0, 11, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_7);

      /* "glypy/_c/count_table.pyx":12
 *     from collections import Mapping, MutableMapping
 * except ImportError:
 *     from collections.abc import Mapping, MutableMapping             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_8 = PyList_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 12, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_n_s_Mapping);
      __Pyx_GIVEREF(__pyx_n_s_Mapping);
      PyList_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_Mapping);
      __Pyx_INCREF(__pyx_n_s_MutableMapping);
      __Pyx_GIVEREF(__pyx_n_s_MutableMapping);
      PyList_SET_ITEM(__pyx_t_8, 1, __pyx_n_s_MutableMapping);
      __pyx_t_9 = __Pyx_Import(__pyx_n_s_collections_abc, __pyx_t_8, -1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 12, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_9, __pyx_n_s_Mapping); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 12, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_Mapping, __pyx_t_8) < 0) __PYX_ERR(0, 12, __pyx_L4_except_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_9, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 12, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_MutableMapping, __pyx_t_8) < 0) __PYX_ERR(0, 12, __pyx_L4_except_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L3_exception_handled;
    }
    goto __pyx_L4_except_error;
    __pyx_L4_except_error:;

    /* "glypy/_c/count_table.pyx":9
 * from cpython.dict cimport PyDict_Next, PyDict_SetItem
 * 
 * try:             # <<<<<<<<<<<<<<
 *     from collections import Mapping, MutableMapping
 * except ImportError:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L3_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_L7_try_end:;
  }

  /* "glypy/_c/count_table.pyx":788
 * 
 * 
 * Mapping.register(CountTable)             # <<<<<<<<<<<<<<
 * MutableMapping.register(CountTable)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Mapping); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_register); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_1, ((PyObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTable)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "glypy/_c/count_table.pyx":789
 * 
 * Mapping.register(CountTable)
 * MutableMapping.register(CountTable)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_register); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_1, ((PyObject *)__pyx_ptype_5glypy_2_c_11count_table_CountTable)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "glypy/_c/count_table.pyx":792
 * 
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef long val
 *     cdef str string
 */
  __pyx_t_7 = PyCFunction_NewEx(&__pyx_mdef_5glypy_2_c_11count_table_1main, NULL, __pyx_n_s_glypy__c_count_table); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_main_2, __pyx_t_7) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "glypy/_c/count_table.pyx":1
 * import sys             # <<<<<<<<<<<<<<
 * cimport cython
 * from cpython cimport PyErr_SetString, PyObject_Hash
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_7) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init glypy._c.count_table", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init glypy._c.count_table");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* Print */
#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* PrintOne */
#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
