/* Generated by Cython 0.29.21 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "glypy/composition/compat.h"
        ],
        "include_dirs": [
            "./glypy/composition"
        ],
        "name": "glypy.composition.ccomposition",
        "sources": [
            "glypy/composition/ccomposition.pyx"
        ]
    },
    "module_name": "glypy.composition.ccomposition"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__glypy__composition__ccomposition
#define __PYX_HAVE_API__glypy__composition__ccomposition
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "pythread.h"
#include "compat.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "glypy/composition/ccomposition.pyx",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition;
struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass;
struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass;

/* "glypy/composition/ccomposition.pxd":6
 *     cdef tuple _mass_args
 *     cpdef CComposition clone(self)
 *     cpdef double calc_mass(self, int average=?, charge=?, dict mass_data=?) except -1             # <<<<<<<<<<<<<<
 *     cpdef _from_formula(self, str formula, dict mass_data)
 *     cpdef _from_dict(self, comp)
 */
struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass {
  int __pyx_n;
  int average;
  PyObject *charge;
  PyObject *mass_data;
};

/* "glypy/composition/ccomposition.pxd":31
 * 
 * 
 * cpdef double calculate_mass(CComposition composition=?, str formula=?, int average=?, charge=?, mass_data=?) except -1             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass {
  int __pyx_n;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  PyObject *formula;
  int average;
  PyObject *charge;
  PyObject *mass_data;
};

/* "glypy/composition/ccomposition.pxd":2
 * 
 * cdef class CComposition(dict):             # <<<<<<<<<<<<<<
 *     cdef object _mass
 *     cdef tuple _mass_args
 */
struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition {
  PyDictObject __pyx_base;
  struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *__pyx_vtab;
  PyObject *_mass;
  PyObject *_mass_args;
};



/* "glypy/composition/ccomposition.pyx":85
 * 
 * 
 * cdef class CComposition(dict):             # <<<<<<<<<<<<<<
 *     """A Composition object stores a chemical composition of a
 *     substance. Basically it is a dict object, in which keys are the names
 */

struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*clone)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, int __pyx_skip_dispatch);
  double (*calc_mass)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass *__pyx_optional_args);
  PyObject *(*_from_formula)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_from_dict)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, int __pyx_skip_dispatch);
  long (*getitem)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  void (*setitem)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, long);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*add_from)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*subtract_from)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*_create)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *);
};
static struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *__pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_5glypy_11composition_12ccomposition_12CComposition_add_from(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other); /* proto*/
static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_5glypy_11composition_12ccomposition_12CComposition_subtract_from(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other); /* proto*/
static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_5glypy_11composition_12ccomposition_12CComposition__create(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_inst); /* proto*/
static long __pyx_f_5glypy_11composition_12ccomposition_12CComposition_getitem(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_elem); /* proto*/
static void __pyx_f_5glypy_11composition_12ccomposition_12CComposition_setitem(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_elem, long __pyx_v_val); /* proto*/
static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_5glypy_11composition_12ccomposition_12CComposition_clone(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5glypy_11composition_12ccomposition_12CComposition__from_formula(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_formula, PyObject *__pyx_v_mass_data, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5glypy_11composition_12ccomposition_12CComposition__from_dict(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_comp, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_5glypy_11composition_12ccomposition_12CComposition_calc_mass(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass *__pyx_optional_args); /* proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'glypy.composition.compat' */

/* Module declarations from 'glypy.composition.ccomposition' */
static PyTypeObject *__pyx_ptype_5glypy_11composition_12ccomposition_CComposition = 0;
static PyObject *__pyx_v_5glypy_11composition_12ccomposition__atom = 0;
static PyObject *__pyx_v_5glypy_11composition_12ccomposition__formula = 0;
static PyObject *__pyx_v_5glypy_11composition_12ccomposition__isotope_string = 0;
static PyObject *__pyx_v_5glypy_11composition_12ccomposition_isotope_pattern = 0;
static PyObject *__pyx_v_5glypy_11composition_12ccomposition_formula_pattern = 0;
static PyObject *__pyx_f_5glypy_11composition_12ccomposition__parse_isotope_string(PyObject *, int *); /*proto*/
static PyObject *__pyx_f_5glypy_11composition_12ccomposition__make_isotope_string(PyObject *, int); /*proto*/
static double __pyx_f_5glypy_11composition_12ccomposition_calculate_mass(int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass *__pyx_optional_args); /*proto*/
static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_5glypy_11composition_12ccomposition_composition_sum(PyObject *, int __pyx_skip_dispatch); /*proto*/
#define __Pyx_MODULE_NAME "glypy.composition.ccomposition"
extern int __pyx_module_is_main_glypy__composition__ccomposition;
int __pyx_module_is_main_glypy__composition__ccomposition = 0;

/* Implementation of 'glypy.composition.ccomposition' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_round;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_id;
static const char __pyx_k_[] = "]";
static const char __pyx_k_H[] = "H+";
static const char __pyx_k__2[] = "[";
static const char __pyx_k__3[] = "";
static const char __pyx_k__4[] = "(";
static const char __pyx_k__5[] = ")";
static const char __pyx_k__6[] = "-";
static const char __pyx_k__9[] = "^({})*$";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_re[] = "re";
static const char __pyx_k_iso[] = "iso";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_len[] = "len";
static const char __pyx_k_s_d[] = "%s[%d]";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_atom[] = "_atom";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_elem[] = "elem";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_clone[] = "clone";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_label[] = "label";
static const char __pyx_k_match[] = "match";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_rfind[] = "rfind";
static const char __pyx_k_round[] = "round";
static const char __pyx_k_charge[] = "charge";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_sorted[] = "sorted";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_average[] = "average";
static const char __pyx_k_compile[] = "compile";
static const char __pyx_k_findall[] = "findall";
static const char __pyx_k_formula[] = "formula";
static const char __pyx_k_isdigit[] = "isdigit";
static const char __pyx_k_reverse[] = "reverse";
static const char __pyx_k_endswith[] = "endswith";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_A_Z_a_z_d[] = "^([A-Z][a-z+]*)(?:\\[(\\d+)\\])?$";
static const char __pyx_k_calc_mass[] = "calc_mass";
static const char __pyx_k_formula_2[] = "_formula";
static const char __pyx_k_from_dict[] = "_from_dict";
static const char __pyx_k_mass_data[] = "mass_data";
static const char __pyx_k_mass_dict[] = "mass_dict";
static const char __pyx_k_nist_mass[] = "nist_mass";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_A_Z_a_z_d_d[] = "([A-Z][a-z+]*)(?:\\[(\\d+)\\])?([+-]?\\d+)?";
static const char __pyx_k_Composition[] = "Composition({})";
static const char __pyx_k_composition[] = "composition";
static const char __pyx_k_CComposition[] = "CComposition";
static const char __pyx_k_from_formula[] = "_from_formula";
static const char __pyx_k_Composition_2[] = "Composition";
static const char __pyx_k_calculate_mass[] = "calculate_mass";
static const char __pyx_k_isotope_string[] = "_isotope_string";
static const char __pyx_k_Invalid_formula[] = "Invalid formula: ";
static const char __pyx_k_formula_pattern[] = "formula_pattern";
static const char __pyx_k_isotope_pattern[] = "isotope_pattern";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_composition_factory[] = "composition_factory";
static const char __pyx_k_from_formula_parens[] = "_from_formula_parens";
static const char __pyx_k_parse_isotope_string[] = "parse_isotope_string";
static const char __pyx_k_ChemicalCompositionError[] = "ChemicalCompositionError";
static const char __pyx_k_Unknown_chemical_element[] = "Unknown chemical element: ";
static const char __pyx_k_Badly_formed_isotope_number_s[] = "Badly-formed isotope number: %s";
static const char __pyx_k_Badly_formed_number_of_atoms_s[] = "Badly-formed number of atoms: %s";
static const char __pyx_k_glypy_composition_ccomposition[] = "glypy.composition.ccomposition";
static const char __pyx_k_Charge_is_specified_both_by_the[] = "Charge is specified both by the number of protons and parameters";
static const char __pyx_k_Unknown_chemical_element_in_the[] = "Unknown chemical element in the formula: %s";
static const char __pyx_k_Cannot_multiply_Composition_by_n[] = "Cannot multiply Composition by non-integer";
static const char __pyx_k_Could_not_create_a_Composition_o[] = "Could not create a Composition object from string: \"{}\": not a formula";
static const char __pyx_k_Must_provide_a_composition_or_fo[] = "Must provide a composition or formula argument";
static const char __pyx_k_glypy_composition_ccomposition_p[] = "glypy/composition/ccomposition.pyx";
static const char __pyx_k_Must_provide_a_composition_or_fo_2[] = "Must provide a composition or formula argument, but not both";
static PyObject *__pyx_kp_s_;
static PyObject *__pyx_kp_s_A_Z_a_z_d;
static PyObject *__pyx_kp_s_A_Z_a_z_d_d;
static PyObject *__pyx_kp_s_Badly_formed_isotope_number_s;
static PyObject *__pyx_kp_s_Badly_formed_number_of_atoms_s;
static PyObject *__pyx_n_s_CComposition;
static PyObject *__pyx_kp_s_Cannot_multiply_Composition_by_n;
static PyObject *__pyx_kp_s_Charge_is_specified_both_by_the;
static PyObject *__pyx_n_s_ChemicalCompositionError;
static PyObject *__pyx_kp_s_Composition;
static PyObject *__pyx_n_s_Composition_2;
static PyObject *__pyx_kp_s_Could_not_create_a_Composition_o;
static PyObject *__pyx_kp_s_H;
static PyObject *__pyx_kp_s_Invalid_formula;
static PyObject *__pyx_kp_s_Must_provide_a_composition_or_fo;
static PyObject *__pyx_kp_s_Must_provide_a_composition_or_fo_2;
static PyObject *__pyx_kp_s_Unknown_chemical_element;
static PyObject *__pyx_kp_s_Unknown_chemical_element_in_the;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s__2;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_kp_s__4;
static PyObject *__pyx_kp_s__5;
static PyObject *__pyx_kp_s__6;
static PyObject *__pyx_kp_s__9;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_atom;
static PyObject *__pyx_n_s_average;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_calc_mass;
static PyObject *__pyx_n_s_calculate_mass;
static PyObject *__pyx_n_s_charge;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_clone;
static PyObject *__pyx_n_s_compile;
static PyObject *__pyx_n_s_composition;
static PyObject *__pyx_n_s_composition_factory;
static PyObject *__pyx_n_s_elem;
static PyObject *__pyx_n_s_endswith;
static PyObject *__pyx_n_s_findall;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_formula;
static PyObject *__pyx_n_s_formula_2;
static PyObject *__pyx_n_s_formula_pattern;
static PyObject *__pyx_n_s_from_dict;
static PyObject *__pyx_n_s_from_formula;
static PyObject *__pyx_n_s_from_formula_parens;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_glypy_composition_ccomposition;
static PyObject *__pyx_kp_s_glypy_composition_ccomposition_p;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_isdigit;
static PyObject *__pyx_n_s_iso;
static PyObject *__pyx_n_s_isotope_pattern;
static PyObject *__pyx_n_s_isotope_string;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_label;
static PyObject *__pyx_n_s_len;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mass_data;
static PyObject *__pyx_n_s_mass_dict;
static PyObject *__pyx_n_s_match;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_nist_mass;
static PyObject *__pyx_n_s_parse_isotope_string;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_re;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_reverse;
static PyObject *__pyx_n_s_rfind;
static PyObject *__pyx_n_s_round;
static PyObject *__pyx_kp_s_s_d;
static PyObject *__pyx_n_s_sorted;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_parse_isotope_string(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition___str__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_2__repr__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_4__iadd__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_6__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_8__isub__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_10__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_12__reduce__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_14__getstate__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_16__setstate__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_d); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_18__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_20__imul__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_22__eq__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_24__ne__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_26__neg__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_28__missing__(CYTHON_UNUSED struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_30__setitem__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_32copy(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_34clone(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_36update(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_38_from_formula(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_formula, PyObject *__pyx_v_mass_data); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_40_from_formula_parens(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_formula, PyObject *__pyx_v_mass_data); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_42_from_dict(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_comp); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_44calc_mass(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, int __pyx_v_average, PyObject *__pyx_v_charge, PyObject *__pyx_v_mass_data); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_4mass___get__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self); /* proto */
static int __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_46__init__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_formula, PyObject *__pyx_v_mass_data, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_2composition_sum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compositions); /* proto */
static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_4calculate_mass(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_composition, PyObject *__pyx_v_formula, int __pyx_v_average, PyObject *__pyx_v_charge, PyObject *__pyx_v_mass_data); /* proto */
static PyObject *__pyx_tp_new_5glypy_11composition_12ccomposition_CComposition(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_k__8;
static PyObject *__pyx_slice__7;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_codeobj__11;
/* Late includes */

/* "glypy/composition/ccomposition.pyx":40
 * 
 * @cython.boundscheck(False)
 * cdef str _parse_isotope_string(str label, int* isotope_num):             # <<<<<<<<<<<<<<
 *     cdef:
 *         # int isotope_num = 0
 */

static PyObject *__pyx_f_5glypy_11composition_12ccomposition__parse_isotope_string(PyObject *__pyx_v_label, int *__pyx_v_isotope_num) {
  int __pyx_v_i;
  int __pyx_v_in_bracket;
  PyObject *__pyx_v_current = 0;
  PyObject *__pyx_v_name_parts = 0;
  PyObject *__pyx_v_num_parts = 0;
  PyObject *__pyx_v_element_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_isotope_string", 0);

  /* "glypy/composition/ccomposition.pyx":43
 *     cdef:
 *         # int isotope_num = 0
 *         int i = 0             # <<<<<<<<<<<<<<
 *         int in_bracket = False
 *         # str element_name
 */
  __pyx_v_i = 0;

  /* "glypy/composition/ccomposition.pyx":44
 *         # int isotope_num = 0
 *         int i = 0
 *         int in_bracket = False             # <<<<<<<<<<<<<<
 *         # str element_name
 *         str current
 */
  __pyx_v_in_bracket = 0;

  /* "glypy/composition/ccomposition.pyx":47
 *         # str element_name
 *         str current
 *         list name_parts = []             # <<<<<<<<<<<<<<
 *         list num_parts = []
 *         #Isotope result
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_name_parts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":48
 *         str current
 *         list name_parts = []
 *         list num_parts = []             # <<<<<<<<<<<<<<
 *         #Isotope result
 *     for i in range(len(label)):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_num_parts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":50
 *         list num_parts = []
 *         #Isotope result
 *     for i in range(len(label)):             # <<<<<<<<<<<<<<
 *         current = label[i]
 *         if in_bracket:
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_label); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glypy/composition/ccomposition.pyx":51
 *         #Isotope result
 *     for i in range(len(label)):
 *         current = label[i]             # <<<<<<<<<<<<<<
 *         if in_bracket:
 *             if current == "]":
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_label, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_current, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glypy/composition/ccomposition.pyx":52
 *     for i in range(len(label)):
 *         current = label[i]
 *         if in_bracket:             # <<<<<<<<<<<<<<
 *             if current == "]":
 *                 break
 */
    __pyx_t_5 = (__pyx_v_in_bracket != 0);
    if (__pyx_t_5) {

      /* "glypy/composition/ccomposition.pyx":53
 *         current = label[i]
 *         if in_bracket:
 *             if current == "]":             # <<<<<<<<<<<<<<
 *                 break
 *             num_parts.append(current)
 */
      __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_v_current, __pyx_kp_s_, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 53, __pyx_L1_error)
      __pyx_t_6 = (__pyx_t_5 != 0);
      if (__pyx_t_6) {

        /* "glypy/composition/ccomposition.pyx":54
 *         if in_bracket:
 *             if current == "]":
 *                 break             # <<<<<<<<<<<<<<
 *             num_parts.append(current)
 *         elif current == "[":
 */
        goto __pyx_L4_break;

        /* "glypy/composition/ccomposition.pyx":53
 *         current = label[i]
 *         if in_bracket:
 *             if current == "]":             # <<<<<<<<<<<<<<
 *                 break
 *             num_parts.append(current)
 */
      }

      /* "glypy/composition/ccomposition.pyx":55
 *             if current == "]":
 *                 break
 *             num_parts.append(current)             # <<<<<<<<<<<<<<
 *         elif current == "[":
 *             in_bracket = True
 */
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_num_parts, __pyx_v_current); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 55, __pyx_L1_error)

      /* "glypy/composition/ccomposition.pyx":52
 *     for i in range(len(label)):
 *         current = label[i]
 *         if in_bracket:             # <<<<<<<<<<<<<<
 *             if current == "]":
 *                 break
 */
      goto __pyx_L5;
    }

    /* "glypy/composition/ccomposition.pyx":56
 *                 break
 *             num_parts.append(current)
 *         elif current == "[":             # <<<<<<<<<<<<<<
 *             in_bracket = True
 *         else:
 */
    __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_v_current, __pyx_kp_s__2, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 56, __pyx_L1_error)
    __pyx_t_5 = (__pyx_t_6 != 0);
    if (__pyx_t_5) {

      /* "glypy/composition/ccomposition.pyx":57
 *             num_parts.append(current)
 *         elif current == "[":
 *             in_bracket = True             # <<<<<<<<<<<<<<
 *         else:
 *             name_parts.append(current)
 */
      __pyx_v_in_bracket = 1;

      /* "glypy/composition/ccomposition.pyx":56
 *                 break
 *             num_parts.append(current)
 *         elif current == "[":             # <<<<<<<<<<<<<<
 *             in_bracket = True
 *         else:
 */
      goto __pyx_L5;
    }

    /* "glypy/composition/ccomposition.pyx":59
 *             in_bracket = True
 *         else:
 *             name_parts.append(current)             # <<<<<<<<<<<<<<
 *     element_name = (''.join(name_parts))
 *     if len(num_parts) > 0:
 */
    /*else*/ {
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_name_parts, __pyx_v_current); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 59, __pyx_L1_error)
    }
    __pyx_L5:;
  }
  __pyx_L4_break:;

  /* "glypy/composition/ccomposition.pyx":60
 *         else:
 *             name_parts.append(current)
 *     element_name = (''.join(name_parts))             # <<<<<<<<<<<<<<
 *     if len(num_parts) > 0:
 *         isotope_num[0] = (int(''.join(num_parts)))
 */
  __pyx_t_1 = __Pyx_PyString_Join(__pyx_kp_s__3, __pyx_v_name_parts); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_element_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":61
 *             name_parts.append(current)
 *     element_name = (''.join(name_parts))
 *     if len(num_parts) > 0:             # <<<<<<<<<<<<<<
 *         isotope_num[0] = (int(''.join(num_parts)))
 *     else:
 */
  __pyx_t_2 = PyList_GET_SIZE(__pyx_v_num_parts); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_5) {

    /* "glypy/composition/ccomposition.pyx":62
 *     element_name = (''.join(name_parts))
 *     if len(num_parts) > 0:
 *         isotope_num[0] = (int(''.join(num_parts)))             # <<<<<<<<<<<<<<
 *     else:
 *         isotope_num[0] = 0
 */
    __pyx_t_1 = __Pyx_PyString_Join(__pyx_kp_s__3, __pyx_v_num_parts); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    (__pyx_v_isotope_num[0]) = __pyx_t_4;

    /* "glypy/composition/ccomposition.pyx":61
 *             name_parts.append(current)
 *     element_name = (''.join(name_parts))
 *     if len(num_parts) > 0:             # <<<<<<<<<<<<<<
 *         isotope_num[0] = (int(''.join(num_parts)))
 *     else:
 */
    goto __pyx_L7;
  }

  /* "glypy/composition/ccomposition.pyx":64
 *         isotope_num[0] = (int(''.join(num_parts)))
 *     else:
 *         isotope_num[0] = 0             # <<<<<<<<<<<<<<
 *     return element_name
 * 
 */
  /*else*/ {
    (__pyx_v_isotope_num[0]) = 0;
  }
  __pyx_L7:;

  /* "glypy/composition/ccomposition.pyx":65
 *     else:
 *         isotope_num[0] = 0
 *     return element_name             # <<<<<<<<<<<<<<
 * 
 * def parse_isotope_string(str label):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyString_CheckExact(__pyx_v_element_name))||((__pyx_v_element_name) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_element_name)->tp_name), 0))) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_INCREF(__pyx_v_element_name);
  __pyx_r = ((PyObject*)__pyx_v_element_name);
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":40
 * 
 * @cython.boundscheck(False)
 * cdef str _parse_isotope_string(str label, int* isotope_num):             # <<<<<<<<<<<<<<
 *     cdef:
 *         # int isotope_num = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glypy.composition.ccomposition._parse_isotope_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_current);
  __Pyx_XDECREF(__pyx_v_name_parts);
  __Pyx_XDECREF(__pyx_v_num_parts);
  __Pyx_XDECREF(__pyx_v_element_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":67
 *     return element_name
 * 
 * def parse_isotope_string(str label):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int iso
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_1parse_isotope_string(PyObject *__pyx_self, PyObject *__pyx_v_label); /*proto*/
static PyMethodDef __pyx_mdef_5glypy_11composition_12ccomposition_1parse_isotope_string = {"parse_isotope_string", (PyCFunction)__pyx_pw_5glypy_11composition_12ccomposition_1parse_isotope_string, METH_O, 0};
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_1parse_isotope_string(PyObject *__pyx_self, PyObject *__pyx_v_label) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parse_isotope_string (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_label), (&PyString_Type), 1, "label", 1))) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_parse_isotope_string(__pyx_self, ((PyObject*)__pyx_v_label));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_parse_isotope_string(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label) {
  int __pyx_v_iso;
  PyObject *__pyx_v_elem = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse_isotope_string", 0);

  /* "glypy/composition/ccomposition.pyx":71
 *         int iso
 *         str elem
 *     elem = _parse_isotope_string(label, &iso)             # <<<<<<<<<<<<<<
 *     return (elem, iso)
 * 
 */
  __pyx_t_1 = __pyx_f_5glypy_11composition_12ccomposition__parse_isotope_string(__pyx_v_label, (&__pyx_v_iso)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_elem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":72
 *         str elem
 *     elem = _parse_isotope_string(label, &iso)
 *     return (elem, iso)             # <<<<<<<<<<<<<<
 * 
 * cdef str _make_isotope_string(str element_name, int isotope_num):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_iso); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_elem);
  __Pyx_GIVEREF(__pyx_v_elem);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_elem);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":67
 *     return element_name
 * 
 * def parse_isotope_string(str label):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int iso
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glypy.composition.ccomposition.parse_isotope_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_elem);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":74
 *     return (elem, iso)
 * 
 * cdef str _make_isotope_string(str element_name, int isotope_num):             # <<<<<<<<<<<<<<
 *     """Form a string label for an isotope."""
 *     cdef:
 */

static PyObject *__pyx_f_5glypy_11composition_12ccomposition__make_isotope_string(PyObject *__pyx_v_element_name, int __pyx_v_isotope_num) {
  PyObject *__pyx_v_parts = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_isotope_string", 0);

  /* "glypy/composition/ccomposition.pyx":78
 *     cdef:
 *         tuple parts
 *     if isotope_num == 0:             # <<<<<<<<<<<<<<
 *         return element_name
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_isotope_num == 0) != 0);
  if (__pyx_t_1) {

    /* "glypy/composition/ccomposition.pyx":79
 *         tuple parts
 *     if isotope_num == 0:
 *         return element_name             # <<<<<<<<<<<<<<
 *     else:
 *         parts = (element_name, isotope_num)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_element_name);
    __pyx_r = __pyx_v_element_name;
    goto __pyx_L0;

    /* "glypy/composition/ccomposition.pyx":78
 *     cdef:
 *         tuple parts
 *     if isotope_num == 0:             # <<<<<<<<<<<<<<
 *         return element_name
 *     else:
 */
  }

  /* "glypy/composition/ccomposition.pyx":81
 *         return element_name
 *     else:
 *         parts = (element_name, isotope_num)             # <<<<<<<<<<<<<<
 *         return <str>PyStr_Format('%s[%d]', parts)
 * 
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_isotope_num); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_element_name);
    __Pyx_GIVEREF(__pyx_v_element_name);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_element_name);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_v_parts = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glypy/composition/ccomposition.pyx":82
 *     else:
 *         parts = (element_name, isotope_num)
 *         return <str>PyStr_Format('%s[%d]', parts)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyStr_Format(__pyx_kp_s_s_d, __pyx_v_parts); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject*)__pyx_t_3));
    __pyx_r = ((PyObject*)__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "glypy/composition/ccomposition.pyx":74
 *     return (elem, iso)
 * 
 * cdef str _make_isotope_string(str element_name, int isotope_num):             # <<<<<<<<<<<<<<
 *     """Form a string label for an isotope."""
 *     cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy.composition.ccomposition._make_isotope_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_parts);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":108
 *         value is :py:data:`nist_mass`). It is used for formulae parsing only.
 *         """
 *     def __str__(self):   # pragma: no cover             # <<<<<<<<<<<<<<
 *         return 'Composition({})'.format(dict.__repr__(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_1__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_1__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition___str__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition___str__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "glypy/composition/ccomposition.pyx":109
 *         """
 *     def __str__(self):   # pragma: no cover
 *         return 'Composition({})'.format(dict.__repr__(self))             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):  # pragma: no cover
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Composition, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyDict_Type)), __pyx_n_s_repr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, ((PyObject *)__pyx_v_self)) : __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_v_self));
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":108
 *         value is :py:data:`nist_mass`). It is used for formulae parsing only.
 *         """
 *     def __str__(self):   # pragma: no cover             # <<<<<<<<<<<<<<
 *         return 'Composition({})'.format(dict.__repr__(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":111
 *         return 'Composition({})'.format(dict.__repr__(self))
 * 
 *     def __repr__(self):  # pragma: no cover             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_3__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_2__repr__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_2__repr__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "glypy/composition/ccomposition.pyx":112
 * 
 *     def __repr__(self):  # pragma: no cover
 *         return str(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef CComposition add_from(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":111
 *         return 'Composition({})'.format(dict.__repr__(self))
 * 
 *     def __repr__(self):  # pragma: no cover             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":114
 *         return str(self)
 * 
 *     cdef CComposition add_from(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str elem
 */

static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_5glypy_11composition_12ccomposition_12CComposition_add_from(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_elem = 0;
  long __pyx_v_cnt;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pvalue;
  Py_ssize_t __pyx_v_ppos;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  long __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_from", 0);

  /* "glypy/composition/ccomposition.pyx":120
 *             PyObject *pkey
 *             PyObject *pvalue
 *             Py_ssize_t ppos = 0             # <<<<<<<<<<<<<<
 * 
 *         while(PyDict_Next(other, &ppos, &pkey, &pvalue)):
 */
  __pyx_v_ppos = 0;

  /* "glypy/composition/ccomposition.pyx":122
 *             Py_ssize_t ppos = 0
 * 
 *         while(PyDict_Next(other, &ppos, &pkey, &pvalue)):             # <<<<<<<<<<<<<<
 *             elem = <str>pkey
 *             cnt = self.getitem(elem)
 */
  while (1) {
    __pyx_t_1 = (PyDict_Next(__pyx_v_other, (&__pyx_v_ppos), (&__pyx_v_pkey), (&__pyx_v_pvalue)) != 0);
    if (!__pyx_t_1) break;

    /* "glypy/composition/ccomposition.pyx":123
 * 
 *         while(PyDict_Next(other, &ppos, &pkey, &pvalue)):
 *             elem = <str>pkey             # <<<<<<<<<<<<<<
 *             cnt = self.getitem(elem)
 *             self.setitem(elem, cnt + PyInt_AsLong(<object>pvalue))
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_pkey);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_elem, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "glypy/composition/ccomposition.pyx":124
 *         while(PyDict_Next(other, &ppos, &pkey, &pvalue)):
 *             elem = <str>pkey
 *             cnt = self.getitem(elem)             # <<<<<<<<<<<<<<
 *             self.setitem(elem, cnt + PyInt_AsLong(<object>pvalue))
 * 
 */
    __pyx_v_cnt = ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_elem);

    /* "glypy/composition/ccomposition.pyx":125
 *             elem = <str>pkey
 *             cnt = self.getitem(elem)
 *             self.setitem(elem, cnt + PyInt_AsLong(<object>pvalue))             # <<<<<<<<<<<<<<
 * 
 *         self._mass_args = None
 */
    __pyx_t_3 = PyInt_AsLong(((PyObject *)__pyx_v_pvalue)); if (unlikely(__pyx_t_3 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 125, __pyx_L1_error)
    ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->setitem(__pyx_v_self, __pyx_v_elem, (__pyx_v_cnt + __pyx_t_3));
  }

  /* "glypy/composition/ccomposition.pyx":127
 *             self.setitem(elem, cnt + PyInt_AsLong(<object>pvalue))
 * 
 *         self._mass_args = None             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_mass_args);
  __Pyx_DECREF(__pyx_v_self->_mass_args);
  __pyx_v_self->_mass_args = ((PyObject*)Py_None);

  /* "glypy/composition/ccomposition.pyx":128
 * 
 *         self._mass_args = None
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __iadd__(CComposition self, other):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":114
 *         return str(self)
 * 
 *     cdef CComposition add_from(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str elem
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.add_from", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_elem);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":130
 *         return self
 * 
 *     def __iadd__(CComposition self, other):             # <<<<<<<<<<<<<<
 *         self.add_from(other)
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_5__iadd__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_5__iadd__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iadd__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_4__iadd__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_4__iadd__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iadd__", 0);

  /* "glypy/composition/ccomposition.pyx":131
 * 
 *     def __iadd__(CComposition self, other):
 *         self.add_from(other)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->add_from(__pyx_v_self, __pyx_v_other)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":132
 *     def __iadd__(CComposition self, other):
 *         self.add_from(other)
 *         return self             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":130
 *         return self
 * 
 *     def __iadd__(CComposition self, other):             # <<<<<<<<<<<<<<
 *         self.add_from(other)
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__iadd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":135
 * 
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CComposition result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_7__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_7__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__add__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_6__add__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_6__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__add__", 0);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_INCREF(__pyx_v_other);

  /* "glypy/composition/ccomposition.pyx":138
 *         cdef:
 *             CComposition result
 *         if not isinstance(self, CComposition):             # <<<<<<<<<<<<<<
 *             other, self = self, other
 *         result = CComposition._create(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_self, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":139
 *             CComposition result
 *         if not isinstance(self, CComposition):
 *             other, self = self, other             # <<<<<<<<<<<<<<
 *         result = CComposition._create(self)
 *         result.add_from(other)
 */
    __pyx_t_3 = __pyx_v_self;
    __pyx_t_4 = __pyx_v_other;
    __pyx_v_other = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_v_self = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "glypy/composition/ccomposition.pyx":138
 *         cdef:
 *             CComposition result
 *         if not isinstance(self, CComposition):             # <<<<<<<<<<<<<<
 *             other, self = self, other
 *         result = CComposition._create(self)
 */
  }

  /* "glypy/composition/ccomposition.pyx":140
 *         if not isinstance(self, CComposition):
 *             other, self = self, other
 *         result = CComposition._create(self)             # <<<<<<<<<<<<<<
 *         result.add_from(other)
 *         return result
 */
  if (!(likely(((__pyx_v_self) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_self, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(0, 140, __pyx_L1_error)
  __pyx_t_5 = ((PyObject *)__pyx_f_5glypy_11composition_12ccomposition_12CComposition__create(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_result = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "glypy/composition/ccomposition.pyx":141
 *             other, self = self, other
 *         result = CComposition._create(self)
 *         result.add_from(other)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_result->__pyx_vtab)->add_from(__pyx_v_result, __pyx_v_other)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "glypy/composition/ccomposition.pyx":142
 *         result = CComposition._create(self)
 *         result.add_from(other)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef CComposition subtract_from(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":135
 * 
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CComposition result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_self);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":144
 *         return result
 * 
 *     cdef CComposition subtract_from(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str elem
 */

static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_5glypy_11composition_12ccomposition_12CComposition_subtract_from(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_elem = 0;
  long __pyx_v_cnt;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pvalue;
  Py_ssize_t __pyx_v_ppos;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  long __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("subtract_from", 0);

  /* "glypy/composition/ccomposition.pyx":150
 *             PyObject *pkey
 *             PyObject *pvalue
 *             Py_ssize_t ppos = 0             # <<<<<<<<<<<<<<
 * 
 *         while(PyDict_Next(other, &ppos, &pkey, &pvalue)):
 */
  __pyx_v_ppos = 0;

  /* "glypy/composition/ccomposition.pyx":152
 *             Py_ssize_t ppos = 0
 * 
 *         while(PyDict_Next(other, &ppos, &pkey, &pvalue)):             # <<<<<<<<<<<<<<
 *             elem = <str>pkey
 *             cnt = self.getitem(elem)
 */
  while (1) {
    __pyx_t_1 = (PyDict_Next(__pyx_v_other, (&__pyx_v_ppos), (&__pyx_v_pkey), (&__pyx_v_pvalue)) != 0);
    if (!__pyx_t_1) break;

    /* "glypy/composition/ccomposition.pyx":153
 * 
 *         while(PyDict_Next(other, &ppos, &pkey, &pvalue)):
 *             elem = <str>pkey             # <<<<<<<<<<<<<<
 *             cnt = self.getitem(elem)
 *             self.setitem(elem, cnt - PyInt_AsLong(<object>pvalue))
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_pkey);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_elem, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "glypy/composition/ccomposition.pyx":154
 *         while(PyDict_Next(other, &ppos, &pkey, &pvalue)):
 *             elem = <str>pkey
 *             cnt = self.getitem(elem)             # <<<<<<<<<<<<<<
 *             self.setitem(elem, cnt - PyInt_AsLong(<object>pvalue))
 * 
 */
    __pyx_v_cnt = ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_elem);

    /* "glypy/composition/ccomposition.pyx":155
 *             elem = <str>pkey
 *             cnt = self.getitem(elem)
 *             self.setitem(elem, cnt - PyInt_AsLong(<object>pvalue))             # <<<<<<<<<<<<<<
 * 
 *         self._mass_args = None
 */
    __pyx_t_3 = PyInt_AsLong(((PyObject *)__pyx_v_pvalue)); if (unlikely(__pyx_t_3 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 155, __pyx_L1_error)
    ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->setitem(__pyx_v_self, __pyx_v_elem, (__pyx_v_cnt - __pyx_t_3));
  }

  /* "glypy/composition/ccomposition.pyx":157
 *             self.setitem(elem, cnt - PyInt_AsLong(<object>pvalue))
 * 
 *         self._mass_args = None             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_mass_args);
  __Pyx_DECREF(__pyx_v_self->_mass_args);
  __pyx_v_self->_mass_args = ((PyObject*)Py_None);

  /* "glypy/composition/ccomposition.pyx":158
 * 
 *         self._mass_args = None
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __isub__(self, other):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":144
 *         return result
 * 
 *     cdef CComposition subtract_from(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str elem
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.subtract_from", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_elem);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":160
 *         return self
 * 
 *     def __isub__(self, other):             # <<<<<<<<<<<<<<
 *         return self.subtract_from(other)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_9__isub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_9__isub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__isub__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_8__isub__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_8__isub__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__isub__", 0);

  /* "glypy/composition/ccomposition.pyx":161
 * 
 *     def __isub__(self, other):
 *         return self.subtract_from(other)             # <<<<<<<<<<<<<<
 * 
 *     def __sub__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->subtract_from(__pyx_v_self, __pyx_v_other)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":160
 *         return self
 * 
 *     def __isub__(self, other):             # <<<<<<<<<<<<<<
 *         return self.subtract_from(other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__isub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":163
 *         return self.subtract_from(other)
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CComposition result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_11__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_11__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sub__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_10__sub__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_10__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__sub__", 0);
  __Pyx_INCREF(__pyx_v_self);

  /* "glypy/composition/ccomposition.pyx":166
 *         cdef:
 *             CComposition result
 *         if not isinstance(self, CComposition):             # <<<<<<<<<<<<<<
 *             self = CComposition(self)
 *         result = CComposition._create(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_self, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":167
 *             CComposition result
 *         if not isinstance(self, CComposition):
 *             self = CComposition(self)             # <<<<<<<<<<<<<<
 *         result = CComposition._create(self)
 *         result.subtract_from(other)
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5glypy_11composition_12ccomposition_CComposition), __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_self, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "glypy/composition/ccomposition.pyx":166
 *         cdef:
 *             CComposition result
 *         if not isinstance(self, CComposition):             # <<<<<<<<<<<<<<
 *             self = CComposition(self)
 *         result = CComposition._create(self)
 */
  }

  /* "glypy/composition/ccomposition.pyx":168
 *         if not isinstance(self, CComposition):
 *             self = CComposition(self)
 *         result = CComposition._create(self)             # <<<<<<<<<<<<<<
 *         result.subtract_from(other)
 *         return result
 */
  if (!(likely(((__pyx_v_self) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_self, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(0, 168, __pyx_L1_error)
  __pyx_t_3 = ((PyObject *)__pyx_f_5glypy_11composition_12ccomposition_12CComposition__create(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "glypy/composition/ccomposition.pyx":169
 *             self = CComposition(self)
 *         result = CComposition._create(self)
 *         result.subtract_from(other)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_result->__pyx_vtab)->subtract_from(__pyx_v_result, __pyx_v_other)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glypy/composition/ccomposition.pyx":170
 *         result = CComposition._create(self)
 *         result.subtract_from(other)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":163
 *         return self.subtract_from(other)
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CComposition result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__sub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_self);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":172
 *         return result
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return composition_factory, (list(self),), self.__getstate__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_13__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_13__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_12__reduce__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_12__reduce__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "glypy/composition/ccomposition.pyx":173
 * 
 *     def __reduce__(self):
 *         return composition_factory, (list(self),), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_composition_factory); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":172
 *         return result
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return composition_factory, (list(self),), self.__getstate__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":175
 *         return composition_factory, (list(self),), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return dict(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_15__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_15__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_14__getstate__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_14__getstate__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "glypy/composition/ccomposition.pyx":176
 * 
 *     def __getstate__(self):
 *         return dict(self)             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":175
 *         return composition_factory, (list(self),), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return dict(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":178
 *         return dict(self)
 * 
 *     def __setstate__(self, d):             # <<<<<<<<<<<<<<
 *         self._from_dict(d)
 *         self._mass = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_17__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_17__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_16__setstate__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), ((PyObject *)__pyx_v_d));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_16__setstate__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "glypy/composition/ccomposition.pyx":179
 * 
 *     def __setstate__(self, d):
 *         self._from_dict(d)             # <<<<<<<<<<<<<<
 *         self._mass = None
 *         self._mass_args = None
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->_from_dict(__pyx_v_self, __pyx_v_d, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":180
 *     def __setstate__(self, d):
 *         self._from_dict(d)
 *         self._mass = None             # <<<<<<<<<<<<<<
 *         self._mass_args = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_mass);
  __Pyx_DECREF(__pyx_v_self->_mass);
  __pyx_v_self->_mass = Py_None;

  /* "glypy/composition/ccomposition.pyx":181
 *         self._from_dict(d)
 *         self._mass = None
 *         self._mass_args = None             # <<<<<<<<<<<<<<
 * 
 *     def __mul__(self, other):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_mass_args);
  __Pyx_DECREF(__pyx_v_self->_mass_args);
  __pyx_v_self->_mass_args = ((PyObject*)Py_None);

  /* "glypy/composition/ccomposition.pyx":178
 *         return dict(self)
 * 
 *     def __setstate__(self, d):             # <<<<<<<<<<<<<<
 *         self._from_dict(d)
 *         self._mass = None
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":183
 *         self._mass_args = None
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CComposition prod
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_19__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_19__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__mul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_18__mul__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_18__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_prod = 0;
  int __pyx_v_rep;
  int __pyx_v_v;
  PyObject *__pyx_v_k = 0;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pvalue;
  Py_ssize_t __pyx_v_ppos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  long __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__mul__", 0);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_INCREF(__pyx_v_other);

  /* "glypy/composition/ccomposition.pyx":190
 *             PyObject *pkey
 *             PyObject *pvalue
 *             Py_ssize_t ppos = 0             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(other, CComposition):
 */
  __pyx_v_ppos = 0;

  /* "glypy/composition/ccomposition.pyx":192
 *             Py_ssize_t ppos = 0
 * 
 *         if isinstance(other, CComposition):             # <<<<<<<<<<<<<<
 *             self, other = other, self
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":193
 * 
 *         if isinstance(other, CComposition):
 *             self, other = other, self             # <<<<<<<<<<<<<<
 * 
 *         # prod = self.clone()
 */
    __pyx_t_3 = __pyx_v_other;
    __pyx_t_4 = __pyx_v_self;
    __pyx_v_self = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_v_other = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "glypy/composition/ccomposition.pyx":192
 *             Py_ssize_t ppos = 0
 * 
 *         if isinstance(other, CComposition):             # <<<<<<<<<<<<<<
 *             self, other = other, self
 * 
 */
  }

  /* "glypy/composition/ccomposition.pyx":197
 *         # prod = self.clone()
 *         # prod = CComposition()
 *         prod = CComposition._create(None)             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(other, int):
 */
  __pyx_t_5 = ((PyObject *)__pyx_f_5glypy_11composition_12ccomposition_12CComposition__create(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_prod = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "glypy/composition/ccomposition.pyx":199
 *         prod = CComposition._create(None)
 * 
 *         if not isinstance(other, int):             # <<<<<<<<<<<<<<
 *             raise ChemicalCompositionError(
 *                 'Cannot multiply Composition by non-integer', other)
 */
  __pyx_t_2 = PyInt_Check(__pyx_v_other); 
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "glypy/composition/ccomposition.pyx":200
 * 
 *         if not isinstance(other, int):
 *             raise ChemicalCompositionError(             # <<<<<<<<<<<<<<
 *                 'Cannot multiply Composition by non-integer', other)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "glypy/composition/ccomposition.pyx":201
 *         if not isinstance(other, int):
 *             raise ChemicalCompositionError(
 *                 'Cannot multiply Composition by non-integer', other)             # <<<<<<<<<<<<<<
 * 
 *         rep = PyInt_AsLong(other)
 */
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_kp_s_Cannot_multiply_Composition_by_n, __pyx_v_other};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_kp_s_Cannot_multiply_Composition_by_n, __pyx_v_other};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_Cannot_multiply_Composition_by_n);
      __Pyx_GIVEREF(__pyx_kp_s_Cannot_multiply_Composition_by_n);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_kp_s_Cannot_multiply_Composition_by_n);
      __Pyx_INCREF(__pyx_v_other);
      __Pyx_GIVEREF(__pyx_v_other);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_other);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 200, __pyx_L1_error)

    /* "glypy/composition/ccomposition.pyx":199
 *         prod = CComposition._create(None)
 * 
 *         if not isinstance(other, int):             # <<<<<<<<<<<<<<
 *             raise ChemicalCompositionError(
 *                 'Cannot multiply Composition by non-integer', other)
 */
  }

  /* "glypy/composition/ccomposition.pyx":203
 *                 'Cannot multiply Composition by non-integer', other)
 * 
 *         rep = PyInt_AsLong(other)             # <<<<<<<<<<<<<<
 *         while(PyDict_Next(self, &ppos, &pkey, &pvalue)):
 *             k = <str>pkey
 */
  __pyx_t_10 = PyInt_AsLong(__pyx_v_other); if (unlikely(__pyx_t_10 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 203, __pyx_L1_error)
  __pyx_v_rep = __pyx_t_10;

  /* "glypy/composition/ccomposition.pyx":204
 * 
 *         rep = PyInt_AsLong(other)
 *         while(PyDict_Next(self, &ppos, &pkey, &pvalue)):             # <<<<<<<<<<<<<<
 *             k = <str>pkey
 *             v = PyInt_AsLong(<object>pvalue)
 */
  while (1) {
    __pyx_t_1 = (PyDict_Next(__pyx_v_self, (&__pyx_v_ppos), (&__pyx_v_pkey), (&__pyx_v_pvalue)) != 0);
    if (!__pyx_t_1) break;

    /* "glypy/composition/ccomposition.pyx":205
 *         rep = PyInt_AsLong(other)
 *         while(PyDict_Next(self, &ppos, &pkey, &pvalue)):
 *             k = <str>pkey             # <<<<<<<<<<<<<<
 *             v = PyInt_AsLong(<object>pvalue)
 *             prod.setitem(k, v * rep)
 */
    __pyx_t_5 = ((PyObject *)__pyx_v_pkey);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_k, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "glypy/composition/ccomposition.pyx":206
 *         while(PyDict_Next(self, &ppos, &pkey, &pvalue)):
 *             k = <str>pkey
 *             v = PyInt_AsLong(<object>pvalue)             # <<<<<<<<<<<<<<
 *             prod.setitem(k, v * rep)
 * 
 */
    __pyx_t_10 = PyInt_AsLong(((PyObject *)__pyx_v_pvalue)); if (unlikely(__pyx_t_10 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 206, __pyx_L1_error)
    __pyx_v_v = __pyx_t_10;

    /* "glypy/composition/ccomposition.pyx":207
 *             k = <str>pkey
 *             v = PyInt_AsLong(<object>pvalue)
 *             prod.setitem(k, v * rep)             # <<<<<<<<<<<<<<
 * 
 *         return prod
 */
    ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_prod->__pyx_vtab)->setitem(__pyx_v_prod, __pyx_v_k, (__pyx_v_v * __pyx_v_rep));
  }

  /* "glypy/composition/ccomposition.pyx":209
 *             prod.setitem(k, v * rep)
 * 
 *         return prod             # <<<<<<<<<<<<<<
 * 
 *     def __imul__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_prod));
  __pyx_r = ((PyObject *)__pyx_v_prod);
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":183
 *         self._mass_args = None
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CComposition prod
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__mul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_prod);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_self);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":211
 *         return prod
 * 
 *     def __imul__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str k
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_21__imul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_21__imul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__imul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_20__imul__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_20__imul__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_k = 0;
  int __pyx_v_v;
  int __pyx_v_rep;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__imul__", 0);

  /* "glypy/composition/ccomposition.pyx":215
 *             str k
 *             int v, rep
 *         if not isinstance(other, int):             # <<<<<<<<<<<<<<
 *             raise ChemicalCompositionError(
 *                 'Cannot multiply Composition by non-integer',
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_other); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "glypy/composition/ccomposition.pyx":216
 *             int v, rep
 *         if not isinstance(other, int):
 *             raise ChemicalCompositionError(             # <<<<<<<<<<<<<<
 *                 'Cannot multiply Composition by non-integer',
 *                 other)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "glypy/composition/ccomposition.pyx":218
 *             raise ChemicalCompositionError(
 *                 'Cannot multiply Composition by non-integer',
 *                 other)             # <<<<<<<<<<<<<<
 *         rep = other
 *         for k, v in list(self.items()):
 */
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_Cannot_multiply_Composition_by_n, __pyx_v_other};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_Cannot_multiply_Composition_by_n, __pyx_v_other};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 216, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_Cannot_multiply_Composition_by_n);
      __Pyx_GIVEREF(__pyx_kp_s_Cannot_multiply_Composition_by_n);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_kp_s_Cannot_multiply_Composition_by_n);
      __Pyx_INCREF(__pyx_v_other);
      __Pyx_GIVEREF(__pyx_v_other);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_other);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 216, __pyx_L1_error)

    /* "glypy/composition/ccomposition.pyx":215
 *             str k
 *             int v, rep
 *         if not isinstance(other, int):             # <<<<<<<<<<<<<<
 *             raise ChemicalCompositionError(
 *                 'Cannot multiply Composition by non-integer',
 */
  }

  /* "glypy/composition/ccomposition.pyx":219
 *                 'Cannot multiply Composition by non-integer',
 *                 other)
 *         rep = other             # <<<<<<<<<<<<<<
 *         for k, v in list(self.items()):
 *             self.setitem(k, v * rep)
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_other); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 219, __pyx_L1_error)
  __pyx_v_rep = __pyx_t_6;

  /* "glypy/composition/ccomposition.pyx":220
 *                 other)
 *         rep = other
 *         for k, v in list(self.items()):             # <<<<<<<<<<<<<<
 *             self.setitem(k, v * rep)
 *         return self
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_t_4; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
    #else
    __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 220, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_7 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      index = 1; __pyx_t_5 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 220, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 220, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    if (!(likely(PyString_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_7)->tp_name), 0))) __PYX_ERR(0, 220, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_k, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;
    __pyx_v_v = __pyx_t_6;

    /* "glypy/composition/ccomposition.pyx":221
 *         rep = other
 *         for k, v in list(self.items()):
 *             self.setitem(k, v * rep)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->setitem(__pyx_v_self, __pyx_v_k, (__pyx_v_v * __pyx_v_rep));

    /* "glypy/composition/ccomposition.pyx":220
 *                 other)
 *         rep = other
 *         for k, v in list(self.items()):             # <<<<<<<<<<<<<<
 *             self.setitem(k, v * rep)
 *         return self
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glypy/composition/ccomposition.pyx":222
 *         for k, v in list(self.items()):
 *             self.setitem(k, v * rep)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":211
 *         return prod
 * 
 *     def __imul__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str k
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__imul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":224
 *         return self
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_ssize_t self_size, other_size
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_23__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_23__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_22__eq__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_22__eq__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other) {
  Py_ssize_t __pyx_v_self_size;
  Py_ssize_t __pyx_v_other_size;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pvalue;
  PyObject *__pyx_v_pinterm;
  long __pyx_v_self_value;
  long __pyx_v_other_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  long __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "glypy/composition/ccomposition.pyx":231
 *             PyObject *pinterm
 *             long self_value, other_value
 *         if not isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             return False
 *         self_size = PyDict_Size(self)
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_other); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":232
 *             long self_value, other_value
 *         if not isinstance(other, dict):
 *             return False             # <<<<<<<<<<<<<<
 *         self_size = PyDict_Size(self)
 *         other_size = PyDict_Size(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "glypy/composition/ccomposition.pyx":231
 *             PyObject *pinterm
 *             long self_value, other_value
 *         if not isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             return False
 *         self_size = PyDict_Size(self)
 */
  }

  /* "glypy/composition/ccomposition.pyx":233
 *         if not isinstance(other, dict):
 *             return False
 *         self_size = PyDict_Size(self)             # <<<<<<<<<<<<<<
 *         other_size = PyDict_Size(other)
 * 
 */
  __pyx_t_3 = PyDict_Size(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 233, __pyx_L1_error)
  __pyx_v_self_size = __pyx_t_3;

  /* "glypy/composition/ccomposition.pyx":234
 *             return False
 *         self_size = PyDict_Size(self)
 *         other_size = PyDict_Size(other)             # <<<<<<<<<<<<<<
 * 
 *         # if one dict is larger, it may contain more keys which may be zero,
 */
  __pyx_t_3 = PyDict_Size(__pyx_v_other); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 234, __pyx_L1_error)
  __pyx_v_other_size = __pyx_t_3;

  /* "glypy/composition/ccomposition.pyx":238
 *         # if one dict is larger, it may contain more keys which may be zero,
 *         # which will be ignored.
 *         if other_size > self_size:             # <<<<<<<<<<<<<<
 *             other_size = 0
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
 */
  __pyx_t_2 = ((__pyx_v_other_size > __pyx_v_self_size) != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":239
 *         # which will be ignored.
 *         if other_size > self_size:
 *             other_size = 0             # <<<<<<<<<<<<<<
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
 *                 other_value = PyInt_AsLong(<object>pvalue)
 */
    __pyx_v_other_size = 0;

    /* "glypy/composition/ccomposition.pyx":240
 *         if other_size > self_size:
 *             other_size = 0
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):             # <<<<<<<<<<<<<<
 *                 other_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(self, <object>pkey)
 */
    while (1) {
      __pyx_t_2 = (PyDict_Next(__pyx_v_other, (&__pyx_v_other_size), (&__pyx_v_pkey), (&__pyx_v_pvalue)) != 0);
      if (!__pyx_t_2) break;

      /* "glypy/composition/ccomposition.pyx":241
 *             other_size = 0
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
 *                 other_value = PyInt_AsLong(<object>pvalue)             # <<<<<<<<<<<<<<
 *                 pinterm = PyDict_GetItem(self, <object>pkey)
 *                 if pinterm == NULL:
 */
      __pyx_t_4 = PyInt_AsLong(((PyObject *)__pyx_v_pvalue)); if (unlikely(__pyx_t_4 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 241, __pyx_L1_error)
      __pyx_v_other_value = __pyx_t_4;

      /* "glypy/composition/ccomposition.pyx":242
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
 *                 other_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(self, <object>pkey)             # <<<<<<<<<<<<<<
 *                 if pinterm == NULL:
 *                     self_value = 0
 */
      __pyx_v_pinterm = PyDict_GetItem(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_pkey));

      /* "glypy/composition/ccomposition.pyx":243
 *                 other_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(self, <object>pkey)
 *                 if pinterm == NULL:             # <<<<<<<<<<<<<<
 *                     self_value = 0
 *                 else:
 */
      __pyx_t_2 = ((__pyx_v_pinterm == NULL) != 0);
      if (__pyx_t_2) {

        /* "glypy/composition/ccomposition.pyx":244
 *                 pinterm = PyDict_GetItem(self, <object>pkey)
 *                 if pinterm == NULL:
 *                     self_value = 0             # <<<<<<<<<<<<<<
 *                 else:
 *                     self_value = PyInt_AsLong(<object>pinterm)
 */
        __pyx_v_self_value = 0;

        /* "glypy/composition/ccomposition.pyx":243
 *                 other_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(self, <object>pkey)
 *                 if pinterm == NULL:             # <<<<<<<<<<<<<<
 *                     self_value = 0
 *                 else:
 */
        goto __pyx_L7;
      }

      /* "glypy/composition/ccomposition.pyx":246
 *                     self_value = 0
 *                 else:
 *                     self_value = PyInt_AsLong(<object>pinterm)             # <<<<<<<<<<<<<<
 *                 if self_value != other_value:
 *                     return False
 */
      /*else*/ {
        __pyx_t_4 = PyInt_AsLong(((PyObject *)__pyx_v_pinterm)); if (unlikely(__pyx_t_4 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L1_error)
        __pyx_v_self_value = __pyx_t_4;
      }
      __pyx_L7:;

      /* "glypy/composition/ccomposition.pyx":247
 *                 else:
 *                     self_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:             # <<<<<<<<<<<<<<
 *                     return False
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_self_value != __pyx_v_other_value) != 0);
      if (__pyx_t_2) {

        /* "glypy/composition/ccomposition.pyx":248
 *                     self_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:
 *                     return False             # <<<<<<<<<<<<<<
 *         else:
 *             self_size = 0
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L0;

        /* "glypy/composition/ccomposition.pyx":247
 *                 else:
 *                     self_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:             # <<<<<<<<<<<<<<
 *                     return False
 *         else:
 */
      }
    }

    /* "glypy/composition/ccomposition.pyx":238
 *         # if one dict is larger, it may contain more keys which may be zero,
 *         # which will be ignored.
 *         if other_size > self_size:             # <<<<<<<<<<<<<<
 *             other_size = 0
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
 */
    goto __pyx_L4;
  }

  /* "glypy/composition/ccomposition.pyx":250
 *                     return False
 *         else:
 *             self_size = 0             # <<<<<<<<<<<<<<
 *             while(PyDict_Next(self, &self_size, &pkey, &pvalue)):
 *                 self_value = PyInt_AsLong(<object>pvalue)
 */
  /*else*/ {
    __pyx_v_self_size = 0;

    /* "glypy/composition/ccomposition.pyx":251
 *         else:
 *             self_size = 0
 *             while(PyDict_Next(self, &self_size, &pkey, &pvalue)):             # <<<<<<<<<<<<<<
 *                 self_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(other, <object>pkey)
 */
    while (1) {
      __pyx_t_2 = (PyDict_Next(((PyObject *)__pyx_v_self), (&__pyx_v_self_size), (&__pyx_v_pkey), (&__pyx_v_pvalue)) != 0);
      if (!__pyx_t_2) break;

      /* "glypy/composition/ccomposition.pyx":252
 *             self_size = 0
 *             while(PyDict_Next(self, &self_size, &pkey, &pvalue)):
 *                 self_value = PyInt_AsLong(<object>pvalue)             # <<<<<<<<<<<<<<
 *                 pinterm = PyDict_GetItem(other, <object>pkey)
 *                 if pinterm == NULL:
 */
      __pyx_t_4 = PyInt_AsLong(((PyObject *)__pyx_v_pvalue)); if (unlikely(__pyx_t_4 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 252, __pyx_L1_error)
      __pyx_v_self_value = __pyx_t_4;

      /* "glypy/composition/ccomposition.pyx":253
 *             while(PyDict_Next(self, &self_size, &pkey, &pvalue)):
 *                 self_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(other, <object>pkey)             # <<<<<<<<<<<<<<
 *                 if pinterm == NULL:
 *                     other_value = 0
 */
      __pyx_v_pinterm = PyDict_GetItem(__pyx_v_other, ((PyObject *)__pyx_v_pkey));

      /* "glypy/composition/ccomposition.pyx":254
 *                 self_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(other, <object>pkey)
 *                 if pinterm == NULL:             # <<<<<<<<<<<<<<
 *                     other_value = 0
 *                 else:
 */
      __pyx_t_2 = ((__pyx_v_pinterm == NULL) != 0);
      if (__pyx_t_2) {

        /* "glypy/composition/ccomposition.pyx":255
 *                 pinterm = PyDict_GetItem(other, <object>pkey)
 *                 if pinterm == NULL:
 *                     other_value = 0             # <<<<<<<<<<<<<<
 *                 else:
 *                     other_value = PyInt_AsLong(<object>pinterm)
 */
        __pyx_v_other_value = 0;

        /* "glypy/composition/ccomposition.pyx":254
 *                 self_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(other, <object>pkey)
 *                 if pinterm == NULL:             # <<<<<<<<<<<<<<
 *                     other_value = 0
 *                 else:
 */
        goto __pyx_L11;
      }

      /* "glypy/composition/ccomposition.pyx":257
 *                     other_value = 0
 *                 else:
 *                     other_value = PyInt_AsLong(<object>pinterm)             # <<<<<<<<<<<<<<
 *                 if self_value != other_value:
 *                     return False
 */
      /*else*/ {
        __pyx_t_4 = PyInt_AsLong(((PyObject *)__pyx_v_pinterm)); if (unlikely(__pyx_t_4 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 257, __pyx_L1_error)
        __pyx_v_other_value = __pyx_t_4;
      }
      __pyx_L11:;

      /* "glypy/composition/ccomposition.pyx":258
 *                 else:
 *                     other_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:             # <<<<<<<<<<<<<<
 *                     return False
 *         return True
 */
      __pyx_t_2 = ((__pyx_v_self_value != __pyx_v_other_value) != 0);
      if (__pyx_t_2) {

        /* "glypy/composition/ccomposition.pyx":259
 *                     other_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:
 *                     return False             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L0;

        /* "glypy/composition/ccomposition.pyx":258
 *                 else:
 *                     other_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:             # <<<<<<<<<<<<<<
 *                     return False
 *         return True
 */
      }
    }
  }
  __pyx_L4:;

  /* "glypy/composition/ccomposition.pyx":260
 *                 if self_value != other_value:
 *                     return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     def __ne__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":224
 *         return self
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_ssize_t self_size, other_size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":262
 *         return True
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_ssize_t self_size, other_size
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_25__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_25__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ne__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_24__ne__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_24__ne__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_other) {
  Py_ssize_t __pyx_v_self_size;
  Py_ssize_t __pyx_v_other_size;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pvalue;
  PyObject *__pyx_v_pinterm;
  long __pyx_v_self_value;
  long __pyx_v_other_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  long __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ne__", 0);

  /* "glypy/composition/ccomposition.pyx":269
 *             PyObject *pinterm
 *             long self_value, other_value
 *         if not isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             return True
 *         self_size = PyDict_Size(self)
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_other); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":270
 *             long self_value, other_value
 *         if not isinstance(other, dict):
 *             return True             # <<<<<<<<<<<<<<
 *         self_size = PyDict_Size(self)
 *         other_size = PyDict_Size(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "glypy/composition/ccomposition.pyx":269
 *             PyObject *pinterm
 *             long self_value, other_value
 *         if not isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             return True
 *         self_size = PyDict_Size(self)
 */
  }

  /* "glypy/composition/ccomposition.pyx":271
 *         if not isinstance(other, dict):
 *             return True
 *         self_size = PyDict_Size(self)             # <<<<<<<<<<<<<<
 *         other_size = PyDict_Size(other)
 *         if other_size > self_size:
 */
  __pyx_t_3 = PyDict_Size(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 271, __pyx_L1_error)
  __pyx_v_self_size = __pyx_t_3;

  /* "glypy/composition/ccomposition.pyx":272
 *             return True
 *         self_size = PyDict_Size(self)
 *         other_size = PyDict_Size(other)             # <<<<<<<<<<<<<<
 *         if other_size > self_size:
 *             other_size = 0
 */
  __pyx_t_3 = PyDict_Size(__pyx_v_other); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 272, __pyx_L1_error)
  __pyx_v_other_size = __pyx_t_3;

  /* "glypy/composition/ccomposition.pyx":273
 *         self_size = PyDict_Size(self)
 *         other_size = PyDict_Size(other)
 *         if other_size > self_size:             # <<<<<<<<<<<<<<
 *             other_size = 0
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
 */
  __pyx_t_2 = ((__pyx_v_other_size > __pyx_v_self_size) != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":274
 *         other_size = PyDict_Size(other)
 *         if other_size > self_size:
 *             other_size = 0             # <<<<<<<<<<<<<<
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
 *                 other_value = PyInt_AsLong(<object>pvalue)
 */
    __pyx_v_other_size = 0;

    /* "glypy/composition/ccomposition.pyx":275
 *         if other_size > self_size:
 *             other_size = 0
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):             # <<<<<<<<<<<<<<
 *                 other_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(self, <object>pkey)
 */
    while (1) {
      __pyx_t_2 = (PyDict_Next(__pyx_v_other, (&__pyx_v_other_size), (&__pyx_v_pkey), (&__pyx_v_pvalue)) != 0);
      if (!__pyx_t_2) break;

      /* "glypy/composition/ccomposition.pyx":276
 *             other_size = 0
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
 *                 other_value = PyInt_AsLong(<object>pvalue)             # <<<<<<<<<<<<<<
 *                 pinterm = PyDict_GetItem(self, <object>pkey)
 *                 if pinterm == NULL:
 */
      __pyx_t_4 = PyInt_AsLong(((PyObject *)__pyx_v_pvalue)); if (unlikely(__pyx_t_4 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 276, __pyx_L1_error)
      __pyx_v_other_value = __pyx_t_4;

      /* "glypy/composition/ccomposition.pyx":277
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
 *                 other_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(self, <object>pkey)             # <<<<<<<<<<<<<<
 *                 if pinterm == NULL:
 *                     self_value = 0
 */
      __pyx_v_pinterm = PyDict_GetItem(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_pkey));

      /* "glypy/composition/ccomposition.pyx":278
 *                 other_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(self, <object>pkey)
 *                 if pinterm == NULL:             # <<<<<<<<<<<<<<
 *                     self_value = 0
 *                 else:
 */
      __pyx_t_2 = ((__pyx_v_pinterm == NULL) != 0);
      if (__pyx_t_2) {

        /* "glypy/composition/ccomposition.pyx":279
 *                 pinterm = PyDict_GetItem(self, <object>pkey)
 *                 if pinterm == NULL:
 *                     self_value = 0             # <<<<<<<<<<<<<<
 *                 else:
 *                     self_value = PyInt_AsLong(<object>pinterm)
 */
        __pyx_v_self_value = 0;

        /* "glypy/composition/ccomposition.pyx":278
 *                 other_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(self, <object>pkey)
 *                 if pinterm == NULL:             # <<<<<<<<<<<<<<
 *                     self_value = 0
 *                 else:
 */
        goto __pyx_L7;
      }

      /* "glypy/composition/ccomposition.pyx":281
 *                     self_value = 0
 *                 else:
 *                     self_value = PyInt_AsLong(<object>pinterm)             # <<<<<<<<<<<<<<
 *                 if self_value != other_value:
 *                     return True
 */
      /*else*/ {
        __pyx_t_4 = PyInt_AsLong(((PyObject *)__pyx_v_pinterm)); if (unlikely(__pyx_t_4 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 281, __pyx_L1_error)
        __pyx_v_self_value = __pyx_t_4;
      }
      __pyx_L7:;

      /* "glypy/composition/ccomposition.pyx":282
 *                 else:
 *                     self_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:             # <<<<<<<<<<<<<<
 *                     return True
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_self_value != __pyx_v_other_value) != 0);
      if (__pyx_t_2) {

        /* "glypy/composition/ccomposition.pyx":283
 *                     self_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:
 *                     return True             # <<<<<<<<<<<<<<
 *         else:
 *             self_size = 0
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_True);
        __pyx_r = Py_True;
        goto __pyx_L0;

        /* "glypy/composition/ccomposition.pyx":282
 *                 else:
 *                     self_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:             # <<<<<<<<<<<<<<
 *                     return True
 *         else:
 */
      }
    }

    /* "glypy/composition/ccomposition.pyx":273
 *         self_size = PyDict_Size(self)
 *         other_size = PyDict_Size(other)
 *         if other_size > self_size:             # <<<<<<<<<<<<<<
 *             other_size = 0
 *             while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
 */
    goto __pyx_L4;
  }

  /* "glypy/composition/ccomposition.pyx":285
 *                     return True
 *         else:
 *             self_size = 0             # <<<<<<<<<<<<<<
 *             while(PyDict_Next(self, &self_size, &pkey, &pvalue)):
 *                 self_value = PyInt_AsLong(<object>pvalue)
 */
  /*else*/ {
    __pyx_v_self_size = 0;

    /* "glypy/composition/ccomposition.pyx":286
 *         else:
 *             self_size = 0
 *             while(PyDict_Next(self, &self_size, &pkey, &pvalue)):             # <<<<<<<<<<<<<<
 *                 self_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(other, <object>pkey)
 */
    while (1) {
      __pyx_t_2 = (PyDict_Next(((PyObject *)__pyx_v_self), (&__pyx_v_self_size), (&__pyx_v_pkey), (&__pyx_v_pvalue)) != 0);
      if (!__pyx_t_2) break;

      /* "glypy/composition/ccomposition.pyx":287
 *             self_size = 0
 *             while(PyDict_Next(self, &self_size, &pkey, &pvalue)):
 *                 self_value = PyInt_AsLong(<object>pvalue)             # <<<<<<<<<<<<<<
 *                 pinterm = PyDict_GetItem(other, <object>pkey)
 *                 if pinterm == NULL:
 */
      __pyx_t_4 = PyInt_AsLong(((PyObject *)__pyx_v_pvalue)); if (unlikely(__pyx_t_4 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 287, __pyx_L1_error)
      __pyx_v_self_value = __pyx_t_4;

      /* "glypy/composition/ccomposition.pyx":288
 *             while(PyDict_Next(self, &self_size, &pkey, &pvalue)):
 *                 self_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(other, <object>pkey)             # <<<<<<<<<<<<<<
 *                 if pinterm == NULL:
 *                     other_value = 0
 */
      __pyx_v_pinterm = PyDict_GetItem(__pyx_v_other, ((PyObject *)__pyx_v_pkey));

      /* "glypy/composition/ccomposition.pyx":289
 *                 self_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(other, <object>pkey)
 *                 if pinterm == NULL:             # <<<<<<<<<<<<<<
 *                     other_value = 0
 *                 else:
 */
      __pyx_t_2 = ((__pyx_v_pinterm == NULL) != 0);
      if (__pyx_t_2) {

        /* "glypy/composition/ccomposition.pyx":290
 *                 pinterm = PyDict_GetItem(other, <object>pkey)
 *                 if pinterm == NULL:
 *                     other_value = 0             # <<<<<<<<<<<<<<
 *                 else:
 *                     other_value = PyInt_AsLong(<object>pinterm)
 */
        __pyx_v_other_value = 0;

        /* "glypy/composition/ccomposition.pyx":289
 *                 self_value = PyInt_AsLong(<object>pvalue)
 *                 pinterm = PyDict_GetItem(other, <object>pkey)
 *                 if pinterm == NULL:             # <<<<<<<<<<<<<<
 *                     other_value = 0
 *                 else:
 */
        goto __pyx_L11;
      }

      /* "glypy/composition/ccomposition.pyx":292
 *                     other_value = 0
 *                 else:
 *                     other_value = PyInt_AsLong(<object>pinterm)             # <<<<<<<<<<<<<<
 *                 if self_value != other_value:
 *                     return True
 */
      /*else*/ {
        __pyx_t_4 = PyInt_AsLong(((PyObject *)__pyx_v_pinterm)); if (unlikely(__pyx_t_4 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 292, __pyx_L1_error)
        __pyx_v_other_value = __pyx_t_4;
      }
      __pyx_L11:;

      /* "glypy/composition/ccomposition.pyx":293
 *                 else:
 *                     other_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:             # <<<<<<<<<<<<<<
 *                     return True
 *         return False
 */
      __pyx_t_2 = ((__pyx_v_self_value != __pyx_v_other_value) != 0);
      if (__pyx_t_2) {

        /* "glypy/composition/ccomposition.pyx":294
 *                     other_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:
 *                     return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_True);
        __pyx_r = Py_True;
        goto __pyx_L0;

        /* "glypy/composition/ccomposition.pyx":293
 *                 else:
 *                     other_value = PyInt_AsLong(<object>pinterm)
 *                 if self_value != other_value:             # <<<<<<<<<<<<<<
 *                     return True
 *         return False
 */
      }
    }
  }
  __pyx_L4:;

  /* "glypy/composition/ccomposition.pyx":295
 *                 if self_value != other_value:
 *                     return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __neg__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":262
 *         return True
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_ssize_t self_size, other_size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__ne__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":297
 *         return False
 * 
 *     def __neg__(self):             # <<<<<<<<<<<<<<
 *         return self * -1
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_27__neg__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_27__neg__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neg__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_26__neg__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_26__neg__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__neg__", 0);

  /* "glypy/composition/ccomposition.pyx":298
 * 
 *     def __neg__(self):
 *         return self * -1             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Multiply(((PyObject *)__pyx_v_self), __pyx_int_neg_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":297
 *         return False
 * 
 *     def __neg__(self):             # <<<<<<<<<<<<<<
 *         return self * -1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__neg__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":301
 * 
 *     @staticmethod
 *     cdef CComposition _create(CComposition inst):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CComposition result
 */

static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_5glypy_11composition_12ccomposition_12CComposition__create(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_inst) {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_result = 0;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "glypy/composition/ccomposition.pyx":305
 *             CComposition result
 * 
 *         result = CComposition.__new__(CComposition)             # <<<<<<<<<<<<<<
 *         result._mass = None
 *         result._mass_args = None
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_5glypy_11composition_12ccomposition_CComposition(((PyTypeObject *)__pyx_ptype_5glypy_11composition_12ccomposition_CComposition), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_result = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":306
 * 
 *         result = CComposition.__new__(CComposition)
 *         result._mass = None             # <<<<<<<<<<<<<<
 *         result._mass_args = None
 *         if inst is None:
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_result->_mass);
  __Pyx_DECREF(__pyx_v_result->_mass);
  __pyx_v_result->_mass = Py_None;

  /* "glypy/composition/ccomposition.pyx":307
 *         result = CComposition.__new__(CComposition)
 *         result._mass = None
 *         result._mass_args = None             # <<<<<<<<<<<<<<
 *         if inst is None:
 *             return result
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_result->_mass_args);
  __Pyx_DECREF(__pyx_v_result->_mass_args);
  __pyx_v_result->_mass_args = ((PyObject*)Py_None);

  /* "glypy/composition/ccomposition.pyx":308
 *         result._mass = None
 *         result._mass_args = None
 *         if inst is None:             # <<<<<<<<<<<<<<
 *             return result
 *         PyDict_Update(result, inst)
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_inst) == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "glypy/composition/ccomposition.pyx":309
 *         result._mass_args = None
 *         if inst is None:
 *             return result             # <<<<<<<<<<<<<<
 *         PyDict_Update(result, inst)
 *         return result
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;

    /* "glypy/composition/ccomposition.pyx":308
 *         result._mass = None
 *         result._mass_args = None
 *         if inst is None:             # <<<<<<<<<<<<<<
 *             return result
 *         PyDict_Update(result, inst)
 */
  }

  /* "glypy/composition/ccomposition.pyx":310
 *         if inst is None:
 *             return result
 *         PyDict_Update(result, inst)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_4 = PyDict_Update(((PyObject *)__pyx_v_result), ((PyObject *)__pyx_v_inst)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 310, __pyx_L1_error)

  /* "glypy/composition/ccomposition.pyx":311
 *             return result
 *         PyDict_Update(result, inst)
 *         return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":301
 * 
 *     @staticmethod
 *     cdef CComposition _create(CComposition inst):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CComposition result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":316
 *     # Override the default behavior, if a key is not present
 *     # do not initialize it to 0.
 *     def __missing__(self, str key):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_29__missing__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_29__missing__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__missing__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyString_Type), 1, "key", 1))) __PYX_ERR(0, 316, __pyx_L1_error)
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_28__missing__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), ((PyObject*)__pyx_v_key));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_28__missing__(CYTHON_UNUSED struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__missing__", 0);

  /* "glypy/composition/ccomposition.pyx":317
 *     # do not initialize it to 0.
 *     def __missing__(self, str key):
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, str key, object value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":316
 *     # Override the default behavior, if a key is not present
 *     # do not initialize it to 0.
 *     def __missing__(self, str key):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":319
 *         return 0
 * 
 *     def __setitem__(self, str key, object value):             # <<<<<<<<<<<<<<
 *         cdef long int_value = PyInt_AsLong(round(value))
 *         if int_value:  # Will not occur on 0 as 0 is falsey AND an integer
 */

/* Python wrapper */
static int __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_31__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_31__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyString_Type), 1, "key", 1))) __PYX_ERR(0, 319, __pyx_L1_error)
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_30__setitem__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), ((PyObject*)__pyx_v_key), ((PyObject *)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_30__setitem__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  long __pyx_v_int_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "glypy/composition/ccomposition.pyx":320
 * 
 *     def __setitem__(self, str key, object value):
 *         cdef long int_value = PyInt_AsLong(round(value))             # <<<<<<<<<<<<<<
 *         if int_value:  # Will not occur on 0 as 0 is falsey AND an integer
 *             self.setitem(key, int_value)
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_round, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_AsLong(__pyx_t_1); if (unlikely(__pyx_t_2 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_int_value = __pyx_t_2;

  /* "glypy/composition/ccomposition.pyx":321
 *     def __setitem__(self, str key, object value):
 *         cdef long int_value = PyInt_AsLong(round(value))
 *         if int_value:  # Will not occur on 0 as 0 is falsey AND an integer             # <<<<<<<<<<<<<<
 *             self.setitem(key, int_value)
 *         elif key in self:
 */
  __pyx_t_3 = (__pyx_v_int_value != 0);
  if (__pyx_t_3) {

    /* "glypy/composition/ccomposition.pyx":322
 *         cdef long int_value = PyInt_AsLong(round(value))
 *         if int_value:  # Will not occur on 0 as 0 is falsey AND an integer
 *             self.setitem(key, int_value)             # <<<<<<<<<<<<<<
 *         elif key in self:
 *             del self[key]
 */
    ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->setitem(__pyx_v_self, __pyx_v_key, __pyx_v_int_value);

    /* "glypy/composition/ccomposition.pyx":321
 *     def __setitem__(self, str key, object value):
 *         cdef long int_value = PyInt_AsLong(round(value))
 *         if int_value:  # Will not occur on 0 as 0 is falsey AND an integer             # <<<<<<<<<<<<<<
 *             self.setitem(key, int_value)
 *         elif key in self:
 */
    goto __pyx_L3;
  }

  /* "glypy/composition/ccomposition.pyx":323
 *         if int_value:  # Will not occur on 0 as 0 is falsey AND an integer
 *             self.setitem(key, int_value)
 *         elif key in self:             # <<<<<<<<<<<<<<
 *             del self[key]
 *         self._mass_args = None
 */
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, ((PyObject *)__pyx_v_self), Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 323, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "glypy/composition/ccomposition.pyx":324
 *             self.setitem(key, int_value)
 *         elif key in self:
 *             del self[key]             # <<<<<<<<<<<<<<
 *         self._mass_args = None
 * 
 */
    if (unlikely(PyObject_DelItem(((PyObject *)__pyx_v_self), __pyx_v_key) < 0)) __PYX_ERR(0, 324, __pyx_L1_error)

    /* "glypy/composition/ccomposition.pyx":323
 *         if int_value:  # Will not occur on 0 as 0 is falsey AND an integer
 *             self.setitem(key, int_value)
 *         elif key in self:             # <<<<<<<<<<<<<<
 *             del self[key]
 *         self._mass_args = None
 */
  }
  __pyx_L3:;

  /* "glypy/composition/ccomposition.pyx":325
 *         elif key in self:
 *             del self[key]
 *         self._mass_args = None             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_mass_args);
  __Pyx_DECREF(__pyx_v_self->_mass_args);
  __pyx_v_self->_mass_args = ((PyObject*)Py_None);

  /* "glypy/composition/ccomposition.pyx":319
 *         return 0
 * 
 *     def __setitem__(self, str key, object value):             # <<<<<<<<<<<<<<
 *         cdef long int_value = PyInt_AsLong(round(value))
 *         if int_value:  # Will not occur on 0 as 0 is falsey AND an integer
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":327
 *         self._mass_args = None
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef CComposition dup
 *         dup = CComposition._create(None)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_33copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_33copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_32copy(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_32copy(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self) {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_dup = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "glypy/composition/ccomposition.pyx":329
 *     def copy(self):
 *         cdef CComposition dup
 *         dup = CComposition._create(None)             # <<<<<<<<<<<<<<
 *         PyDict_Update(dup, self)
 *         return dup
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5glypy_11composition_12ccomposition_12CComposition__create(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_dup = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":330
 *         cdef CComposition dup
 *         dup = CComposition._create(None)
 *         PyDict_Update(dup, self)             # <<<<<<<<<<<<<<
 *         return dup
 * 
 */
  __pyx_t_2 = PyDict_Update(((PyObject *)__pyx_v_dup), ((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 330, __pyx_L1_error)

  /* "glypy/composition/ccomposition.pyx":331
 *         dup = CComposition._create(None)
 *         PyDict_Update(dup, self)
 *         return dup             # <<<<<<<<<<<<<<
 * 
 *     cdef long getitem(self, str elem):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_dup));
  __pyx_r = ((PyObject *)__pyx_v_dup);
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":327
 *         self._mass_args = None
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef CComposition dup
 *         dup = CComposition._create(None)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dup);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":333
 *         return dup
 * 
 *     cdef long getitem(self, str elem):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* resobj
 */

static long __pyx_f_5glypy_11composition_12ccomposition_12CComposition_getitem(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_elem) {
  PyObject *__pyx_v_resobj;
  long __pyx_v_count;
  long __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  long __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getitem", 0);

  /* "glypy/composition/ccomposition.pyx":337
 *             PyObject* resobj
 *             long count
 *         resobj = PyDict_GetItem(self, elem)             # <<<<<<<<<<<<<<
 *         if (resobj == NULL):
 *             return 0
 */
  __pyx_v_resobj = PyDict_GetItem(((PyObject *)__pyx_v_self), __pyx_v_elem);

  /* "glypy/composition/ccomposition.pyx":338
 *             long count
 *         resobj = PyDict_GetItem(self, elem)
 *         if (resobj == NULL):             # <<<<<<<<<<<<<<
 *             return 0
 *         count = PyInt_AsLong(<object>resobj)
 */
  __pyx_t_1 = ((__pyx_v_resobj == NULL) != 0);
  if (__pyx_t_1) {

    /* "glypy/composition/ccomposition.pyx":339
 *         resobj = PyDict_GetItem(self, elem)
 *         if (resobj == NULL):
 *             return 0             # <<<<<<<<<<<<<<
 *         count = PyInt_AsLong(<object>resobj)
 *         return count
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "glypy/composition/ccomposition.pyx":338
 *             long count
 *         resobj = PyDict_GetItem(self, elem)
 *         if (resobj == NULL):             # <<<<<<<<<<<<<<
 *             return 0
 *         count = PyInt_AsLong(<object>resobj)
 */
  }

  /* "glypy/composition/ccomposition.pyx":340
 *         if (resobj == NULL):
 *             return 0
 *         count = PyInt_AsLong(<object>resobj)             # <<<<<<<<<<<<<<
 *         return count
 * 
 */
  __pyx_t_2 = PyInt_AsLong(((PyObject *)__pyx_v_resobj)); if (unlikely(__pyx_t_2 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 340, __pyx_L1_error)
  __pyx_v_count = __pyx_t_2;

  /* "glypy/composition/ccomposition.pyx":341
 *             return 0
 *         count = PyInt_AsLong(<object>resobj)
 *         return count             # <<<<<<<<<<<<<<
 * 
 *     cdef void setitem(self, str elem, long val):
 */
  __pyx_r = __pyx_v_count;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":333
 *         return dup
 * 
 *     cdef long getitem(self, str elem):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* resobj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("glypy.composition.ccomposition.CComposition.getitem", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":343
 *         return count
 * 
 *     cdef void setitem(self, str elem, long val):             # <<<<<<<<<<<<<<
 *         PyDict_SetItem(self, elem, val)
 *         self._mass_args = None
 */

static void __pyx_f_5glypy_11composition_12ccomposition_12CComposition_setitem(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_elem, long __pyx_v_val) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem", 0);

  /* "glypy/composition/ccomposition.pyx":344
 * 
 *     cdef void setitem(self, str elem, long val):
 *         PyDict_SetItem(self, elem, val)             # <<<<<<<<<<<<<<
 *         self._mass_args = None
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyDict_SetItem(((PyObject *)__pyx_v_self), __pyx_v_elem, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":345
 *     cdef void setitem(self, str elem, long val):
 *         PyDict_SetItem(self, elem, val)
 *         self._mass_args = None             # <<<<<<<<<<<<<<
 * 
 *     cpdef CComposition clone(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_mass_args);
  __Pyx_DECREF(__pyx_v_self->_mass_args);
  __pyx_v_self->_mass_args = ((PyObject*)Py_None);

  /* "glypy/composition/ccomposition.pyx":343
 *         return count
 * 
 *     cdef void setitem(self, str elem, long val):             # <<<<<<<<<<<<<<
 *         PyDict_SetItem(self, elem, val)
 *         self._mass_args = None
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("glypy.composition.ccomposition.CComposition.setitem", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "glypy/composition/ccomposition.pyx":347
 *         self._mass_args = None
 * 
 *     cpdef CComposition clone(self):             # <<<<<<<<<<<<<<
 *         return CComposition._create(self)
 * 
 */

static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_35clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_5glypy_11composition_12ccomposition_12CComposition_clone(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_35clone)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 347, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(0, 347, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/composition/ccomposition.pyx":348
 * 
 *     cpdef CComposition clone(self):
 *         return CComposition._create(self)             # <<<<<<<<<<<<<<
 * 
 *     def update(self, *args, **kwargs):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((PyObject *)__pyx_f_5glypy_11composition_12ccomposition_12CComposition__create(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":347
 *         self._mass_args = None
 * 
 *     cpdef CComposition clone(self):             # <<<<<<<<<<<<<<
 *         return CComposition._create(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_35clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_35clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_34clone(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_34clone(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5glypy_11composition_12ccomposition_12CComposition_clone(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":350
 *         return CComposition._create(self)
 * 
 *     def update(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         dict.update(self, *args, **kwargs)
 *         self._mass_args = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_37update(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_37update(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "update", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = PyDict_Copy(__pyx_kwds); if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = NULL;
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_36update(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_36update(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 0);

  /* "glypy/composition/ccomposition.pyx":351
 * 
 *     def update(self, *args, **kwargs):
 *         dict.update(self, *args, **kwargs)             # <<<<<<<<<<<<<<
 *         self._mass_args = None
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyDict_Type)), __pyx_n_s_update); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self));
  __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glypy/composition/ccomposition.pyx":352
 *     def update(self, *args, **kwargs):
 *         dict.update(self, *args, **kwargs)
 *         self._mass_args = None             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_mass_args);
  __Pyx_DECREF(__pyx_v_self->_mass_args);
  __pyx_v_self->_mass_args = ((PyObject*)Py_None);

  /* "glypy/composition/ccomposition.pyx":350
 *         return CComposition._create(self)
 * 
 *     def update(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         dict.update(self, *args, **kwargs)
 *         self._mass_args = None
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":355
 * 
 *     @cython.boundscheck(False)
 *     cpdef _from_formula(self, str formula, dict mass_data):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str elem, isotope, number
 */

static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_39_from_formula(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5glypy_11composition_12ccomposition_12CComposition__from_formula(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_formula, PyObject *__pyx_v_mass_data, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_elem = 0;
  PyObject *__pyx_v_isotope = 0;
  PyObject *__pyx_v_number = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_from_formula", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_from_formula); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_39_from_formula)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_formula, __pyx_v_mass_data};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 355, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_formula, __pyx_v_mass_data};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 355, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 355, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_formula);
          __Pyx_GIVEREF(__pyx_v_formula);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_formula);
          __Pyx_INCREF(__pyx_v_mass_data);
          __Pyx_GIVEREF(__pyx_v_mass_data);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_mass_data);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 355, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/composition/ccomposition.pyx":358
 *         cdef:
 *             str elem, isotope, number
 *         if '(' in formula:             # <<<<<<<<<<<<<<
 *             self._from_formula_parens(formula, mass_data)
 *         elif not formula_pattern.match(formula):
 */
  __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_kp_s__4, __pyx_v_formula, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 358, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "glypy/composition/ccomposition.pyx":359
 *             str elem, isotope, number
 *         if '(' in formula:
 *             self._from_formula_parens(formula, mass_data)             # <<<<<<<<<<<<<<
 *         elif not formula_pattern.match(formula):
 *             raise ChemicalCompositionError('Invalid formula: ' + formula)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_from_formula_parens); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 359, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_formula, __pyx_v_mass_data};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 359, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_formula, __pyx_v_mass_data};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 359, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 359, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_formula);
      __Pyx_GIVEREF(__pyx_v_formula);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_formula);
      __Pyx_INCREF(__pyx_v_mass_data);
      __Pyx_GIVEREF(__pyx_v_mass_data);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_mass_data);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 359, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glypy/composition/ccomposition.pyx":358
 *         cdef:
 *             str elem, isotope, number
 *         if '(' in formula:             # <<<<<<<<<<<<<<
 *             self._from_formula_parens(formula, mass_data)
 *         elif not formula_pattern.match(formula):
 */
    goto __pyx_L3;
  }

  /* "glypy/composition/ccomposition.pyx":360
 *         if '(' in formula:
 *             self._from_formula_parens(formula, mass_data)
 *         elif not formula_pattern.match(formula):             # <<<<<<<<<<<<<<
 *             raise ChemicalCompositionError('Invalid formula: ' + formula)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_5glypy_11composition_12ccomposition_formula_pattern, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_v_formula) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_formula);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = ((!__pyx_t_8) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "glypy/composition/ccomposition.pyx":361
 *             self._from_formula_parens(formula, mass_data)
 *         elif not formula_pattern.match(formula):
 *             raise ChemicalCompositionError('Invalid formula: ' + formula)             # <<<<<<<<<<<<<<
 *         else:
 *             for elem, isotope, number in re.findall(_atom, formula):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyNumber_Add(__pyx_kp_s_Invalid_formula, __pyx_v_formula); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 361, __pyx_L1_error)

    /* "glypy/composition/ccomposition.pyx":360
 *         if '(' in formula:
 *             self._from_formula_parens(formula, mass_data)
 *         elif not formula_pattern.match(formula):             # <<<<<<<<<<<<<<
 *             raise ChemicalCompositionError('Invalid formula: ' + formula)
 *         else:
 */
  }

  /* "glypy/composition/ccomposition.pyx":363
 *             raise ChemicalCompositionError('Invalid formula: ' + formula)
 *         else:
 *             for elem, isotope, number in re.findall(_atom, formula):             # <<<<<<<<<<<<<<
 *                 if not elem in mass_data:
 *                     raise ChemicalCompositionError('Unknown chemical element: ' + elem)
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_findall); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_5glypy_11composition_12ccomposition__atom, __pyx_v_formula};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_5glypy_11composition_12ccomposition__atom, __pyx_v_formula};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_5glypy_11composition_12ccomposition__atom);
      __Pyx_GIVEREF(__pyx_v_5glypy_11composition_12ccomposition__atom);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_5, __pyx_v_5glypy_11composition_12ccomposition__atom);
      __Pyx_INCREF(__pyx_v_formula);
      __Pyx_GIVEREF(__pyx_v_formula);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_5, __pyx_v_formula);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_6 = __pyx_t_1; __Pyx_INCREF(__pyx_t_6); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 363, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 363, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_6, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 363, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_6, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_10(__pyx_t_6);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 363, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 3)) {
          if (size > 3) __Pyx_RaiseTooManyValuesError(3);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 363, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
        } else {
          __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
        }
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        #else
        __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_11 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = Py_TYPE(__pyx_t_11)->tp_iternext;
        index = 0; __pyx_t_3 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        index = 1; __pyx_t_2 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 2; __pyx_t_4 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_11), 3) < 0) __PYX_ERR(0, 363, __pyx_L1_error)
        __pyx_t_12 = NULL;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L7_unpacking_done;
        __pyx_L6_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_12 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 363, __pyx_L1_error)
        __pyx_L7_unpacking_done:;
      }
      if (!(likely(PyString_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 363, __pyx_L1_error)
      if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 363, __pyx_L1_error)
      if (!(likely(PyString_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_elem, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_isotope, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_number, ((PyObject*)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "glypy/composition/ccomposition.pyx":364
 *         else:
 *             for elem, isotope, number in re.findall(_atom, formula):
 *                 if not elem in mass_data:             # <<<<<<<<<<<<<<
 *                     raise ChemicalCompositionError('Unknown chemical element: ' + elem)
 *                 self[_make_isotope_string(elem, int(isotope) if isotope else 0)
 */
      if (unlikely(__pyx_v_mass_data == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 364, __pyx_L1_error)
      }
      __pyx_t_7 = (__Pyx_PyDict_ContainsTF(__pyx_v_elem, __pyx_v_mass_data, Py_NE)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 364, __pyx_L1_error)
      __pyx_t_8 = (__pyx_t_7 != 0);
      if (unlikely(__pyx_t_8)) {

        /* "glypy/composition/ccomposition.pyx":365
 *             for elem, isotope, number in re.findall(_atom, formula):
 *                 if not elem in mass_data:
 *                     raise ChemicalCompositionError('Unknown chemical element: ' + elem)             # <<<<<<<<<<<<<<
 *                 self[_make_isotope_string(elem, int(isotope) if isotope else 0)
 *                         ] += int(number) if number else 1
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 365, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = PyNumber_Add(__pyx_kp_s_Unknown_chemical_element, __pyx_v_elem); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 365, __pyx_L1_error)

        /* "glypy/composition/ccomposition.pyx":364
 *         else:
 *             for elem, isotope, number in re.findall(_atom, formula):
 *                 if not elem in mass_data:             # <<<<<<<<<<<<<<
 *                     raise ChemicalCompositionError('Unknown chemical element: ' + elem)
 *                 self[_make_isotope_string(elem, int(isotope) if isotope else 0)
 */
      }

      /* "glypy/composition/ccomposition.pyx":366
 *                 if not elem in mass_data:
 *                     raise ChemicalCompositionError('Unknown chemical element: ' + elem)
 *                 self[_make_isotope_string(elem, int(isotope) if isotope else 0)             # <<<<<<<<<<<<<<
 *                         ] += int(number) if number else 1
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_isotope); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 366, __pyx_L1_error)
      if (__pyx_t_8) {
        __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_isotope); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 366, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_5 = __pyx_t_13;
      } else {
        __pyx_t_5 = 0;
      }
      __pyx_t_1 = __pyx_f_5glypy_11composition_12ccomposition__make_isotope_string(__pyx_v_elem, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "glypy/composition/ccomposition.pyx":367
 *                     raise ChemicalCompositionError('Unknown chemical element: ' + elem)
 *                 self[_make_isotope_string(elem, int(isotope) if isotope else 0)
 *                         ] += int(number) if number else 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(True)
 */
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_number); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 367, __pyx_L1_error)
      if (__pyx_t_8) {
        __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_v_number); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 367, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __pyx_t_3;
        __pyx_t_3 = 0;
      } else {
        __Pyx_INCREF(__pyx_int_1);
        __pyx_t_2 = __pyx_int_1;
      }

      /* "glypy/composition/ccomposition.pyx":366
 *                 if not elem in mass_data:
 *                     raise ChemicalCompositionError('Unknown chemical element: ' + elem)
 *                 self[_make_isotope_string(elem, int(isotope) if isotope else 0)             # <<<<<<<<<<<<<<
 *                         ] += int(number) if number else 1
 * 
 */
      __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_self), __pyx_t_1, __pyx_t_3) < 0)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "glypy/composition/ccomposition.pyx":363
 *             raise ChemicalCompositionError('Invalid formula: ' + formula)
 *         else:
 *             for elem, isotope, number in re.findall(_atom, formula):             # <<<<<<<<<<<<<<
 *                 if not elem in mass_data:
 *                     raise ChemicalCompositionError('Unknown chemical element: ' + elem)
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_L3:;

  /* "glypy/composition/ccomposition.pyx":355
 * 
 *     @cython.boundscheck(False)
 *     cpdef _from_formula(self, str formula, dict mass_data):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str elem, isotope, number
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition._from_formula", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_elem);
  __Pyx_XDECREF(__pyx_v_isotope);
  __Pyx_XDECREF(__pyx_v_number);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_39_from_formula(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_39_from_formula(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_formula = 0;
  PyObject *__pyx_v_mass_data = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_from_formula (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_formula,&__pyx_n_s_mass_data,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formula)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mass_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_from_formula", 1, 2, 2, 1); __PYX_ERR(0, 355, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_from_formula") < 0)) __PYX_ERR(0, 355, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_formula = ((PyObject*)values[0]);
    __pyx_v_mass_data = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_from_formula", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 355, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition._from_formula", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_formula), (&PyString_Type), 1, "formula", 1))) __PYX_ERR(0, 355, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mass_data), (&PyDict_Type), 1, "mass_data", 1))) __PYX_ERR(0, 355, __pyx_L1_error)
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_38_from_formula(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), __pyx_v_formula, __pyx_v_mass_data);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_38_from_formula(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_formula, PyObject *__pyx_v_mass_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_from_formula", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_11composition_12ccomposition_12CComposition__from_formula(__pyx_v_self, __pyx_v_formula, __pyx_v_mass_data, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition._from_formula", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":370
 * 
 *     @cython.boundscheck(True)
 *     def _from_formula_parens(self, formula, mass_data):             # <<<<<<<<<<<<<<
 *         # Parsing a formula backwards.
 *         cdef:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_41_from_formula_parens(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_41_from_formula_parens(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_formula = 0;
  PyObject *__pyx_v_mass_data = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_from_formula_parens (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_formula,&__pyx_n_s_mass_data,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formula)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mass_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_from_formula_parens", 1, 2, 2, 1); __PYX_ERR(0, 370, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_from_formula_parens") < 0)) __PYX_ERR(0, 370, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_formula = values[0];
    __pyx_v_mass_data = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_from_formula_parens", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 370, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition._from_formula_parens", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_40_from_formula_parens(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), __pyx_v_formula, __pyx_v_mass_data);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_40_from_formula_parens(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_formula, PyObject *__pyx_v_mass_data) {
  Py_ssize_t __pyx_v_prev_chem_symbol_start;
  Py_ssize_t __pyx_v_i;
  int __pyx_v_seek_mode;
  int __pyx_v_group_coef;
  PyObject *__pyx_v_parse_stack = 0;
  PyObject *__pyx_v_resolve_stack = 0;
  PyObject *__pyx_v_num_atoms = NULL;
  PyObject *__pyx_v_brace_pos = NULL;
  PyObject *__pyx_v_isotope_num = NULL;
  int __pyx_v_element_found;
  PyObject *__pyx_v_element_name = NULL;
  PyObject *__pyx_v_isotope_string = NULL;
  PyObject *__pyx_v_ch = NULL;
  PyObject *__pyx_v_chunk = NULL;
  PyObject *__pyx_v_elem = NULL;
  PyObject *__pyx_v_cnt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *(*__pyx_t_15)(PyObject *);
  Py_ssize_t __pyx_t_16;
  int __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_from_formula_parens", 0);

  /* "glypy/composition/ccomposition.pyx":379
 * 
 * 
 *         prev_chem_symbol_start = len(formula)             # <<<<<<<<<<<<<<
 *         i = len(formula) - 1
 * 
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_formula); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 379, __pyx_L1_error)
  __pyx_v_prev_chem_symbol_start = __pyx_t_1;

  /* "glypy/composition/ccomposition.pyx":380
 * 
 *         prev_chem_symbol_start = len(formula)
 *         i = len(formula) - 1             # <<<<<<<<<<<<<<
 * 
 *         seek_mode = 0
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_formula); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 380, __pyx_L1_error)
  __pyx_v_i = (__pyx_t_1 - 1);

  /* "glypy/composition/ccomposition.pyx":382
 *         i = len(formula) - 1
 * 
 *         seek_mode = 0             # <<<<<<<<<<<<<<
 *         parse_stack = ""
 *         resolve_stack = []
 */
  __pyx_v_seek_mode = 0;

  /* "glypy/composition/ccomposition.pyx":383
 * 
 *         seek_mode = 0
 *         parse_stack = ""             # <<<<<<<<<<<<<<
 *         resolve_stack = []
 *         group_coef = 1
 */
  __Pyx_INCREF(__pyx_kp_s__3);
  __pyx_v_parse_stack = __pyx_kp_s__3;

  /* "glypy/composition/ccomposition.pyx":384
 *         seek_mode = 0
 *         parse_stack = ""
 *         resolve_stack = []             # <<<<<<<<<<<<<<
 *         group_coef = 1
 * 
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_resolve_stack = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glypy/composition/ccomposition.pyx":385
 *         parse_stack = ""
 *         resolve_stack = []
 *         group_coef = 1             # <<<<<<<<<<<<<<
 * 
 *         while i >= 0:
 */
  __pyx_v_group_coef = 1;

  /* "glypy/composition/ccomposition.pyx":387
 *         group_coef = 1
 * 
 *         while i >= 0:             # <<<<<<<<<<<<<<
 *             if seek_mode < 1:
 *                 if (formula[i] == ")"):
 */
  while (1) {
    __pyx_t_3 = ((__pyx_v_i >= 0) != 0);
    if (!__pyx_t_3) break;

    /* "glypy/composition/ccomposition.pyx":388
 * 
 *         while i >= 0:
 *             if seek_mode < 1:             # <<<<<<<<<<<<<<
 *                 if (formula[i] == ")"):
 *                     seek_mode += 1
 */
    __pyx_t_3 = ((__pyx_v_seek_mode < 1) != 0);
    if (__pyx_t_3) {

      /* "glypy/composition/ccomposition.pyx":389
 *         while i >= 0:
 *             if seek_mode < 1:
 *                 if (formula[i] == ")"):             # <<<<<<<<<<<<<<
 *                     seek_mode += 1
 *                     if i + 1 == prev_chem_symbol_start:
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_formula, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_kp_s__5, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 389, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_3) {

        /* "glypy/composition/ccomposition.pyx":390
 *             if seek_mode < 1:
 *                 if (formula[i] == ")"):
 *                     seek_mode += 1             # <<<<<<<<<<<<<<
 *                     if i + 1 == prev_chem_symbol_start:
 *                         group_coef = 1
 */
        __pyx_v_seek_mode = (__pyx_v_seek_mode + 1);

        /* "glypy/composition/ccomposition.pyx":391
 *                 if (formula[i] == ")"):
 *                     seek_mode += 1
 *                     if i + 1 == prev_chem_symbol_start:             # <<<<<<<<<<<<<<
 *                         group_coef = 1
 *                     elif formula[i + 1].isdigit():
 */
        __pyx_t_3 = (((__pyx_v_i + 1) == __pyx_v_prev_chem_symbol_start) != 0);
        if (__pyx_t_3) {

          /* "glypy/composition/ccomposition.pyx":392
 *                     seek_mode += 1
 *                     if i + 1 == prev_chem_symbol_start:
 *                         group_coef = 1             # <<<<<<<<<<<<<<
 *                     elif formula[i + 1].isdigit():
 *                         group_coef = int(formula[i + 1:prev_chem_symbol_start])
 */
          __pyx_v_group_coef = 1;

          /* "glypy/composition/ccomposition.pyx":391
 *                 if (formula[i] == ")"):
 *                     seek_mode += 1
 *                     if i + 1 == prev_chem_symbol_start:             # <<<<<<<<<<<<<<
 *                         group_coef = 1
 *                     elif formula[i + 1].isdigit():
 */
          goto __pyx_L7;
        }

        /* "glypy/composition/ccomposition.pyx":393
 *                     if i + 1 == prev_chem_symbol_start:
 *                         group_coef = 1
 *                     elif formula[i + 1].isdigit():             # <<<<<<<<<<<<<<
 *                         group_coef = int(formula[i + 1:prev_chem_symbol_start])
 *                     i -= 1
 */
        __pyx_t_1 = (__pyx_v_i + 1);
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_formula, __pyx_t_1, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 393, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_isdigit); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 393, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 393, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_3) {

          /* "glypy/composition/ccomposition.pyx":394
 *                         group_coef = 1
 *                     elif formula[i + 1].isdigit():
 *                         group_coef = int(formula[i + 1:prev_chem_symbol_start])             # <<<<<<<<<<<<<<
 *                     i -= 1
 *                     continue
 */
          __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_formula, (__pyx_v_i + 1), __pyx_v_prev_chem_symbol_start, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 394, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_group_coef = __pyx_t_6;

          /* "glypy/composition/ccomposition.pyx":393
 *                     if i + 1 == prev_chem_symbol_start:
 *                         group_coef = 1
 *                     elif formula[i + 1].isdigit():             # <<<<<<<<<<<<<<
 *                         group_coef = int(formula[i + 1:prev_chem_symbol_start])
 *                     i -= 1
 */
        }
        __pyx_L7:;

        /* "glypy/composition/ccomposition.pyx":395
 *                     elif formula[i + 1].isdigit():
 *                         group_coef = int(formula[i + 1:prev_chem_symbol_start])
 *                     i -= 1             # <<<<<<<<<<<<<<
 *                     continue
 *                 # Read backwards until a non-number character is met.
 */
        __pyx_v_i = (__pyx_v_i - 1);

        /* "glypy/composition/ccomposition.pyx":396
 *                         group_coef = int(formula[i + 1:prev_chem_symbol_start])
 *                     i -= 1
 *                     continue             # <<<<<<<<<<<<<<
 *                 # Read backwards until a non-number character is met.
 *                 if (formula[i].isdigit() or formula[i] == '-'):
 */
        goto __pyx_L3_continue;

        /* "glypy/composition/ccomposition.pyx":389
 *         while i >= 0:
 *             if seek_mode < 1:
 *                 if (formula[i] == ")"):             # <<<<<<<<<<<<<<
 *                     seek_mode += 1
 *                     if i + 1 == prev_chem_symbol_start:
 */
      }

      /* "glypy/composition/ccomposition.pyx":398
 *                     continue
 *                 # Read backwards until a non-number character is met.
 *                 if (formula[i].isdigit() or formula[i] == '-'):             # <<<<<<<<<<<<<<
 *                     i -= 1
 *                     continue
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_formula, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_isdigit); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (!__pyx_t_7) {
      } else {
        __pyx_t_3 = __pyx_t_7;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_formula, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_kp_s__6, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_3 = __pyx_t_7;
      __pyx_L9_bool_binop_done:;
      if (__pyx_t_3) {

        /* "glypy/composition/ccomposition.pyx":399
 *                 # Read backwards until a non-number character is met.
 *                 if (formula[i].isdigit() or formula[i] == '-'):
 *                     i -= 1             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
        __pyx_v_i = (__pyx_v_i - 1);

        /* "glypy/composition/ccomposition.pyx":400
 *                 if (formula[i].isdigit() or formula[i] == '-'):
 *                     i -= 1
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
        goto __pyx_L3_continue;

        /* "glypy/composition/ccomposition.pyx":398
 *                     continue
 *                 # Read backwards until a non-number character is met.
 *                 if (formula[i].isdigit() or formula[i] == '-'):             # <<<<<<<<<<<<<<
 *                     i -= 1
 *                     continue
 */
      }

      /* "glypy/composition/ccomposition.pyx":404
 *                 else:
 *                     # If the number of atoms is omitted then it is 1.
 *                     if i + 1 == prev_chem_symbol_start:             # <<<<<<<<<<<<<<
 *                         num_atoms = 1
 *                     else:
 */
      /*else*/ {
        __pyx_t_3 = (((__pyx_v_i + 1) == __pyx_v_prev_chem_symbol_start) != 0);
        if (__pyx_t_3) {

          /* "glypy/composition/ccomposition.pyx":405
 *                     # If the number of atoms is omitted then it is 1.
 *                     if i + 1 == prev_chem_symbol_start:
 *                         num_atoms = 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         try:
 */
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_XDECREF_SET(__pyx_v_num_atoms, __pyx_int_1);

          /* "glypy/composition/ccomposition.pyx":404
 *                 else:
 *                     # If the number of atoms is omitted then it is 1.
 *                     if i + 1 == prev_chem_symbol_start:             # <<<<<<<<<<<<<<
 *                         num_atoms = 1
 *                     else:
 */
          goto __pyx_L11;
        }

        /* "glypy/composition/ccomposition.pyx":407
 *                         num_atoms = 1
 *                     else:
 *                         try:             # <<<<<<<<<<<<<<
 *                             num_atoms = int(formula[i + 1:prev_chem_symbol_start])
 *                         except ValueError:
 */
        /*else*/ {
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
            __Pyx_XGOTREF(__pyx_t_8);
            __Pyx_XGOTREF(__pyx_t_9);
            __Pyx_XGOTREF(__pyx_t_10);
            /*try:*/ {

              /* "glypy/composition/ccomposition.pyx":408
 *                     else:
 *                         try:
 *                             num_atoms = int(formula[i + 1:prev_chem_symbol_start])             # <<<<<<<<<<<<<<
 *                         except ValueError:
 *                             raise ChemicalCompositionError(
 */
              __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_formula, (__pyx_v_i + 1), __pyx_v_prev_chem_symbol_start, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 408, __pyx_L12_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_4 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 408, __pyx_L12_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_XDECREF_SET(__pyx_v_num_atoms, __pyx_t_4);
              __pyx_t_4 = 0;

              /* "glypy/composition/ccomposition.pyx":407
 *                         num_atoms = 1
 *                     else:
 *                         try:             # <<<<<<<<<<<<<<
 *                             num_atoms = int(formula[i + 1:prev_chem_symbol_start])
 *                         except ValueError:
 */
            }
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            goto __pyx_L19_try_end;
            __pyx_L12_error:;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

            /* "glypy/composition/ccomposition.pyx":409
 *                         try:
 *                             num_atoms = int(formula[i + 1:prev_chem_symbol_start])
 *                         except ValueError:             # <<<<<<<<<<<<<<
 *                             raise ChemicalCompositionError(
 *                                 'Badly-formed number of atoms: %s' % formula)
 */
            __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
            if (__pyx_t_6) {
              __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition._from_formula_parens", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_2) < 0) __PYX_ERR(0, 409, __pyx_L14_except_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GOTREF(__pyx_t_2);

              /* "glypy/composition/ccomposition.pyx":410
 *                             num_atoms = int(formula[i + 1:prev_chem_symbol_start])
 *                         except ValueError:
 *                             raise ChemicalCompositionError(             # <<<<<<<<<<<<<<
 *                                 'Badly-formed number of atoms: %s' % formula)
 * 
 */
              __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 410, __pyx_L14_except_error)
              __Pyx_GOTREF(__pyx_t_12);

              /* "glypy/composition/ccomposition.pyx":411
 *                         except ValueError:
 *                             raise ChemicalCompositionError(
 *                                 'Badly-formed number of atoms: %s' % formula)             # <<<<<<<<<<<<<<
 * 
 *                     # Read isotope number if specified, else it is undefined (=0).
 */
              __pyx_t_13 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Badly_formed_number_of_atoms_s, __pyx_v_formula); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 411, __pyx_L14_except_error)
              __Pyx_GOTREF(__pyx_t_13);
              __pyx_t_14 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
                __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_12);
                if (likely(__pyx_t_14)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                  __Pyx_INCREF(__pyx_t_14);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_12, function);
                }
              }
              __pyx_t_11 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13);
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 410, __pyx_L14_except_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_Raise(__pyx_t_11, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __PYX_ERR(0, 410, __pyx_L14_except_error)
            }
            goto __pyx_L14_except_error;
            __pyx_L14_except_error:;

            /* "glypy/composition/ccomposition.pyx":407
 *                         num_atoms = 1
 *                     else:
 *                         try:             # <<<<<<<<<<<<<<
 *                             num_atoms = int(formula[i + 1:prev_chem_symbol_start])
 *                         except ValueError:
 */
            __Pyx_XGIVEREF(__pyx_t_8);
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
            goto __pyx_L1_error;
            __pyx_L19_try_end:;
          }
        }
        __pyx_L11:;

        /* "glypy/composition/ccomposition.pyx":414
 * 
 *                     # Read isotope number if specified, else it is undefined (=0).
 *                     if formula[i] == ']':             # <<<<<<<<<<<<<<
 *                         brace_pos = formula.rfind('[', 0, i)
 *                         if brace_pos == -1:
 */
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_formula, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_kp_s_, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_3) {

          /* "glypy/composition/ccomposition.pyx":415
 *                     # Read isotope number if specified, else it is undefined (=0).
 *                     if formula[i] == ']':
 *                         brace_pos = formula.rfind('[', 0, i)             # <<<<<<<<<<<<<<
 *                         if brace_pos == -1:
 *                             raise ChemicalCompositionError(
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_formula, __pyx_n_s_rfind); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 415, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 415, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_11 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[4] = {__pyx_t_11, __pyx_kp_s__2, __pyx_int_0, __pyx_t_4};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[4] = {__pyx_t_11, __pyx_kp_s__2, __pyx_int_0, __pyx_t_4};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 415, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_11) {
              __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_11); __pyx_t_11 = NULL;
            }
            __Pyx_INCREF(__pyx_kp_s__2);
            __Pyx_GIVEREF(__pyx_kp_s__2);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_kp_s__2);
            __Pyx_INCREF(__pyx_int_0);
            __Pyx_GIVEREF(__pyx_int_0);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_int_0);
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_6, __pyx_t_4);
            __pyx_t_4 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF_SET(__pyx_v_brace_pos, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "glypy/composition/ccomposition.pyx":416
 *                     if formula[i] == ']':
 *                         brace_pos = formula.rfind('[', 0, i)
 *                         if brace_pos == -1:             # <<<<<<<<<<<<<<
 *                             raise ChemicalCompositionError(
 *                                 'Badly-formed isotope number: %s' % formula)
 */
          __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_brace_pos, __pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 416, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(__pyx_t_3)) {

            /* "glypy/composition/ccomposition.pyx":417
 *                         brace_pos = formula.rfind('[', 0, i)
 *                         if brace_pos == -1:
 *                             raise ChemicalCompositionError(             # <<<<<<<<<<<<<<
 *                                 'Badly-formed isotope number: %s' % formula)
 *                         try:
 */
            __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 417, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);

            /* "glypy/composition/ccomposition.pyx":418
 *                         if brace_pos == -1:
 *                             raise ChemicalCompositionError(
 *                                 'Badly-formed isotope number: %s' % formula)             # <<<<<<<<<<<<<<
 *                         try:
 *                             isotope_num = int(formula[brace_pos + 1:i])
 */
            __pyx_t_12 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Badly_formed_isotope_number_s, __pyx_v_formula); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 418, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_12);
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 417, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_Raise(__pyx_t_2, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __PYX_ERR(0, 417, __pyx_L1_error)

            /* "glypy/composition/ccomposition.pyx":416
 *                     if formula[i] == ']':
 *                         brace_pos = formula.rfind('[', 0, i)
 *                         if brace_pos == -1:             # <<<<<<<<<<<<<<
 *                             raise ChemicalCompositionError(
 *                                 'Badly-formed isotope number: %s' % formula)
 */
          }

          /* "glypy/composition/ccomposition.pyx":419
 *                             raise ChemicalCompositionError(
 *                                 'Badly-formed isotope number: %s' % formula)
 *                         try:             # <<<<<<<<<<<<<<
 *                             isotope_num = int(formula[brace_pos + 1:i])
 *                         except ValueError:
 */
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
            __Pyx_XGOTREF(__pyx_t_10);
            __Pyx_XGOTREF(__pyx_t_9);
            __Pyx_XGOTREF(__pyx_t_8);
            /*try:*/ {

              /* "glypy/composition/ccomposition.pyx":420
 *                                 'Badly-formed isotope number: %s' % formula)
 *                         try:
 *                             isotope_num = int(formula[brace_pos + 1:i])             # <<<<<<<<<<<<<<
 *                         except ValueError:
 *                             raise ChemicalCompositionError(
 */
              __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_brace_pos, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L24_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_formula, 0, __pyx_v_i, &__pyx_t_2, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 420, __pyx_L24_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L24_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_XDECREF_SET(__pyx_v_isotope_num, __pyx_t_2);
              __pyx_t_2 = 0;

              /* "glypy/composition/ccomposition.pyx":419
 *                             raise ChemicalCompositionError(
 *                                 'Badly-formed isotope number: %s' % formula)
 *                         try:             # <<<<<<<<<<<<<<
 *                             isotope_num = int(formula[brace_pos + 1:i])
 *                         except ValueError:
 */
            }
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            goto __pyx_L31_try_end;
            __pyx_L24_error:;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

            /* "glypy/composition/ccomposition.pyx":421
 *                         try:
 *                             isotope_num = int(formula[brace_pos + 1:i])
 *                         except ValueError:             # <<<<<<<<<<<<<<
 *                             raise ChemicalCompositionError(
 *                                 'Badly-formed isotope number: %s' % formula)
 */
            __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
            if (__pyx_t_6) {
              __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition._from_formula_parens", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_5, &__pyx_t_12) < 0) __PYX_ERR(0, 421, __pyx_L26_except_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GOTREF(__pyx_t_12);

              /* "glypy/composition/ccomposition.pyx":422
 *                             isotope_num = int(formula[brace_pos + 1:i])
 *                         except ValueError:
 *                             raise ChemicalCompositionError(             # <<<<<<<<<<<<<<
 *                                 'Badly-formed isotope number: %s' % formula)
 *                         i = brace_pos - 1
 */
              __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 422, __pyx_L26_except_error)
              __Pyx_GOTREF(__pyx_t_11);

              /* "glypy/composition/ccomposition.pyx":423
 *                         except ValueError:
 *                             raise ChemicalCompositionError(
 *                                 'Badly-formed isotope number: %s' % formula)             # <<<<<<<<<<<<<<
 *                         i = brace_pos - 1
 *                     else:
 */
              __pyx_t_13 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Badly_formed_isotope_number_s, __pyx_v_formula); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 423, __pyx_L26_except_error)
              __Pyx_GOTREF(__pyx_t_13);
              __pyx_t_14 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
                __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_11);
                if (likely(__pyx_t_14)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                  __Pyx_INCREF(__pyx_t_14);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_11, function);
                }
              }
              __pyx_t_4 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_13);
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 422, __pyx_L26_except_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_Raise(__pyx_t_4, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __PYX_ERR(0, 422, __pyx_L26_except_error)
            }
            goto __pyx_L26_except_error;
            __pyx_L26_except_error:;

            /* "glypy/composition/ccomposition.pyx":419
 *                             raise ChemicalCompositionError(
 *                                 'Badly-formed isotope number: %s' % formula)
 *                         try:             # <<<<<<<<<<<<<<
 *                             isotope_num = int(formula[brace_pos + 1:i])
 *                         except ValueError:
 */
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_8);
            __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
            goto __pyx_L1_error;
            __pyx_L31_try_end:;
          }

          /* "glypy/composition/ccomposition.pyx":424
 *                             raise ChemicalCompositionError(
 *                                 'Badly-formed isotope number: %s' % formula)
 *                         i = brace_pos - 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         isotope_num = 0
 */
          __pyx_t_12 = __Pyx_PyInt_SubtractObjC(__pyx_v_brace_pos, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 424, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_1 = __Pyx_PyIndex_AsSsize_t(__pyx_t_12); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 424, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_v_i = __pyx_t_1;

          /* "glypy/composition/ccomposition.pyx":414
 * 
 *                     # Read isotope number if specified, else it is undefined (=0).
 *                     if formula[i] == ']':             # <<<<<<<<<<<<<<
 *                         brace_pos = formula.rfind('[', 0, i)
 *                         if brace_pos == -1:
 */
          goto __pyx_L22;
        }

        /* "glypy/composition/ccomposition.pyx":426
 *                         i = brace_pos - 1
 *                     else:
 *                         isotope_num = 0             # <<<<<<<<<<<<<<
 * 
 *                     # Match the element name to the mass_data.
 */
        /*else*/ {
          __Pyx_INCREF(__pyx_int_0);
          __Pyx_XDECREF_SET(__pyx_v_isotope_num, __pyx_int_0);
        }
        __pyx_L22:;

        /* "glypy/composition/ccomposition.pyx":429
 * 
 *                     # Match the element name to the mass_data.
 *                     element_found = False             # <<<<<<<<<<<<<<
 *                     # Sort the keys from longest to shortest to workaround
 *                     # the overlapping keys issue
 */
        __pyx_v_element_found = 0;

        /* "glypy/composition/ccomposition.pyx":432
 *                     # Sort the keys from longest to shortest to workaround
 *                     # the overlapping keys issue
 *                     for element_name in sorted(mass_data, key=len, reverse=True):             # <<<<<<<<<<<<<<
 *                         if formula.endswith(element_name, 0, i + 1):
 *                             isotope_string = _make_isotope_string(
 */
        __pyx_t_12 = PyTuple_New(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 432, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_v_mass_data);
        __Pyx_GIVEREF(__pyx_v_mass_data);
        PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_mass_data);
        __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 432, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_key, __pyx_t_2) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_reverse, Py_True) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_12, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
          __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5); __pyx_t_1 = 0;
          __pyx_t_15 = NULL;
        } else {
          __pyx_t_1 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 432, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_15 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 432, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        for (;;) {
          if (likely(!__pyx_t_15)) {
            if (likely(PyList_CheckExact(__pyx_t_5))) {
              if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_5)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_2); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 432, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            } else {
              if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_2); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 432, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            }
          } else {
            __pyx_t_2 = __pyx_t_15(__pyx_t_5);
            if (unlikely(!__pyx_t_2)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 432, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_2);
          }
          __Pyx_XDECREF_SET(__pyx_v_element_name, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "glypy/composition/ccomposition.pyx":433
 *                     # the overlapping keys issue
 *                     for element_name in sorted(mass_data, key=len, reverse=True):
 *                         if formula.endswith(element_name, 0, i + 1):             # <<<<<<<<<<<<<<
 *                             isotope_string = _make_isotope_string(
 *                                 element_name, isotope_num)
 */
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_formula, __pyx_n_s_endswith); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 433, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_i + 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 433, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_11 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[4] = {__pyx_t_11, __pyx_v_element_name, __pyx_int_0, __pyx_t_4};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[4] = {__pyx_t_11, __pyx_v_element_name, __pyx_int_0, __pyx_t_4};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 433, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_11) {
              __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_11); __pyx_t_11 = NULL;
            }
            __Pyx_INCREF(__pyx_v_element_name);
            __Pyx_GIVEREF(__pyx_v_element_name);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_element_name);
            __Pyx_INCREF(__pyx_int_0);
            __Pyx_GIVEREF(__pyx_int_0);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_int_0);
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_6, __pyx_t_4);
            __pyx_t_4 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 433, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (__pyx_t_3) {

            /* "glypy/composition/ccomposition.pyx":435
 *                         if formula.endswith(element_name, 0, i + 1):
 *                             isotope_string = _make_isotope_string(
 *                                 element_name, isotope_num)             # <<<<<<<<<<<<<<
 *                             self[isotope_string] += num_atoms
 *                             i -= len(element_name)
 */
            if (!(likely(PyString_CheckExact(__pyx_v_element_name))||((__pyx_v_element_name) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_element_name)->tp_name), 0))) __PYX_ERR(0, 435, __pyx_L1_error)
            __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_isotope_num); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 435, __pyx_L1_error)

            /* "glypy/composition/ccomposition.pyx":434
 *                     for element_name in sorted(mass_data, key=len, reverse=True):
 *                         if formula.endswith(element_name, 0, i + 1):
 *                             isotope_string = _make_isotope_string(             # <<<<<<<<<<<<<<
 *                                 element_name, isotope_num)
 *                             self[isotope_string] += num_atoms
 */
            __pyx_t_2 = __pyx_f_5glypy_11composition_12ccomposition__make_isotope_string(((PyObject*)__pyx_v_element_name), __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_XDECREF_SET(__pyx_v_isotope_string, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "glypy/composition/ccomposition.pyx":436
 *                             isotope_string = _make_isotope_string(
 *                                 element_name, isotope_num)
 *                             self[isotope_string] += num_atoms             # <<<<<<<<<<<<<<
 *                             i -= len(element_name)
 *                             prev_chem_symbol_start = i + 1
 */
            __Pyx_INCREF(__pyx_v_isotope_string);
            __pyx_t_2 = __pyx_v_isotope_string;
            __pyx_t_12 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_t_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 436, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = PyNumber_InPlaceAdd(__pyx_t_12, __pyx_v_num_atoms); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 436, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_self), __pyx_t_2, __pyx_t_13) < 0)) __PYX_ERR(0, 436, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "glypy/composition/ccomposition.pyx":437
 *                                 element_name, isotope_num)
 *                             self[isotope_string] += num_atoms
 *                             i -= len(element_name)             # <<<<<<<<<<<<<<
 *                             prev_chem_symbol_start = i + 1
 *                             element_found = True
 */
            __pyx_t_16 = PyObject_Length(__pyx_v_element_name); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 437, __pyx_L1_error)
            __pyx_v_i = (__pyx_v_i - __pyx_t_16);

            /* "glypy/composition/ccomposition.pyx":438
 *                             self[isotope_string] += num_atoms
 *                             i -= len(element_name)
 *                             prev_chem_symbol_start = i + 1             # <<<<<<<<<<<<<<
 *                             element_found = True
 *                             break
 */
            __pyx_v_prev_chem_symbol_start = (__pyx_v_i + 1);

            /* "glypy/composition/ccomposition.pyx":439
 *                             i -= len(element_name)
 *                             prev_chem_symbol_start = i + 1
 *                             element_found = True             # <<<<<<<<<<<<<<
 *                             break
 * 
 */
            __pyx_v_element_found = 1;

            /* "glypy/composition/ccomposition.pyx":440
 *                             prev_chem_symbol_start = i + 1
 *                             element_found = True
 *                             break             # <<<<<<<<<<<<<<
 * 
 *                     if not element_found:
 */
            goto __pyx_L35_break;

            /* "glypy/composition/ccomposition.pyx":433
 *                     # the overlapping keys issue
 *                     for element_name in sorted(mass_data, key=len, reverse=True):
 *                         if formula.endswith(element_name, 0, i + 1):             # <<<<<<<<<<<<<<
 *                             isotope_string = _make_isotope_string(
 *                                 element_name, isotope_num)
 */
          }

          /* "glypy/composition/ccomposition.pyx":432
 *                     # Sort the keys from longest to shortest to workaround
 *                     # the overlapping keys issue
 *                     for element_name in sorted(mass_data, key=len, reverse=True):             # <<<<<<<<<<<<<<
 *                         if formula.endswith(element_name, 0, i + 1):
 *                             isotope_string = _make_isotope_string(
 */
        }
        __pyx_L35_break:;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "glypy/composition/ccomposition.pyx":442
 *                             break
 * 
 *                     if not element_found:             # <<<<<<<<<<<<<<
 *                         raise ChemicalCompositionError(
 *                             'Unknown chemical element in the formula: %s' % formula)
 */
        __pyx_t_3 = ((!(__pyx_v_element_found != 0)) != 0);
        if (unlikely(__pyx_t_3)) {

          /* "glypy/composition/ccomposition.pyx":443
 * 
 *                     if not element_found:
 *                         raise ChemicalCompositionError(             # <<<<<<<<<<<<<<
 *                             'Unknown chemical element in the formula: %s' % formula)
 *             else:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 443, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);

          /* "glypy/composition/ccomposition.pyx":444
 *                     if not element_found:
 *                         raise ChemicalCompositionError(
 *                             'Unknown chemical element in the formula: %s' % formula)             # <<<<<<<<<<<<<<
 *             else:
 *                 ch = formula[i]
 */
          __pyx_t_13 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Unknown_chemical_element_in_the, __pyx_v_formula); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 444, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_12 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          __pyx_t_5 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_12, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_13);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 443, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_Raise(__pyx_t_5, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __PYX_ERR(0, 443, __pyx_L1_error)

          /* "glypy/composition/ccomposition.pyx":442
 *                             break
 * 
 *                     if not element_found:             # <<<<<<<<<<<<<<
 *                         raise ChemicalCompositionError(
 *                             'Unknown chemical element in the formula: %s' % formula)
 */
        }
      }

      /* "glypy/composition/ccomposition.pyx":388
 * 
 *         while i >= 0:
 *             if seek_mode < 1:             # <<<<<<<<<<<<<<
 *                 if (formula[i] == ")"):
 *                     seek_mode += 1
 */
      goto __pyx_L5;
    }

    /* "glypy/composition/ccomposition.pyx":446
 *                             'Unknown chemical element in the formula: %s' % formula)
 *             else:
 *                 ch = formula[i]             # <<<<<<<<<<<<<<
 *                 parse_stack += ch
 *                 i -= 1
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_formula, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 446, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_ch, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "glypy/composition/ccomposition.pyx":447
 *             else:
 *                 ch = formula[i]
 *                 parse_stack += ch             # <<<<<<<<<<<<<<
 *                 i -= 1
 *                 if(ch == "("):
 */
      __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_parse_stack, __pyx_v_ch); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (!(likely(PyString_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_parse_stack, ((PyObject*)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "glypy/composition/ccomposition.pyx":448
 *                 ch = formula[i]
 *                 parse_stack += ch
 *                 i -= 1             # <<<<<<<<<<<<<<
 *                 if(ch == "("):
 *                     seek_mode -= 1
 */
      __pyx_v_i = (__pyx_v_i - 1);

      /* "glypy/composition/ccomposition.pyx":449
 *                 parse_stack += ch
 *                 i -= 1
 *                 if(ch == "("):             # <<<<<<<<<<<<<<
 *                     seek_mode -= 1
 *                     if seek_mode == 0:
 */
      __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_ch, __pyx_kp_s__4, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 449, __pyx_L1_error)
      if (__pyx_t_3) {

        /* "glypy/composition/ccomposition.pyx":450
 *                 i -= 1
 *                 if(ch == "("):
 *                     seek_mode -= 1             # <<<<<<<<<<<<<<
 *                     if seek_mode == 0:
 * 
 */
        __pyx_v_seek_mode = (__pyx_v_seek_mode - 1);

        /* "glypy/composition/ccomposition.pyx":451
 *                 if(ch == "("):
 *                     seek_mode -= 1
 *                     if seek_mode == 0:             # <<<<<<<<<<<<<<
 * 
 *                         resolve_stack.append(Composition(
 */
        __pyx_t_3 = ((__pyx_v_seek_mode == 0) != 0);
        if (__pyx_t_3) {

          /* "glypy/composition/ccomposition.pyx":453
 *                     if seek_mode == 0:
 * 
 *                         resolve_stack.append(Composition(             # <<<<<<<<<<<<<<
 *                                              # Omit the last character, then reverse the parse
 *                                              # stack string.
 */
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Composition_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 453, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);

          /* "glypy/composition/ccomposition.pyx":456
 *                                              # Omit the last character, then reverse the parse
 *                                              # stack string.
 *                                              formula=parse_stack[:-1][::-1],             # <<<<<<<<<<<<<<
 *                                              mass_data=mass_data)
 *                                              * group_coef)
 */
          __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (unlikely(__pyx_v_parse_stack == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 456, __pyx_L1_error)
          }
          __pyx_t_13 = PySequence_GetSlice(__pyx_v_parse_stack, 0, -1L); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 456, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_t_13, __pyx_slice__7); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 456, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_formula, __pyx_t_12) < 0) __PYX_ERR(0, 456, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

          /* "glypy/composition/ccomposition.pyx":457
 *                                              # stack string.
 *                                              formula=parse_stack[:-1][::-1],
 *                                              mass_data=mass_data)             # <<<<<<<<<<<<<<
 *                                              * group_coef)
 *                         prev_chem_symbol_start = i + 1
 */
          if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_mass_data, __pyx_v_mass_data) < 0) __PYX_ERR(0, 456, __pyx_L1_error)

          /* "glypy/composition/ccomposition.pyx":453
 *                     if seek_mode == 0:
 * 
 *                         resolve_stack.append(Composition(             # <<<<<<<<<<<<<<
 *                                              # Omit the last character, then reverse the parse
 *                                              # stack string.
 */
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 453, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "glypy/composition/ccomposition.pyx":458
 *                                              formula=parse_stack[:-1][::-1],
 *                                              mass_data=mass_data)
 *                                              * group_coef)             # <<<<<<<<<<<<<<
 *                         prev_chem_symbol_start = i + 1
 *                         seek_mode = False
 */
          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_group_coef); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 458, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = PyNumber_Multiply(__pyx_t_12, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 458, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "glypy/composition/ccomposition.pyx":453
 *                     if seek_mode == 0:
 * 
 *                         resolve_stack.append(Composition(             # <<<<<<<<<<<<<<
 *                                              # Omit the last character, then reverse the parse
 *                                              # stack string.
 */
          __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_resolve_stack, __pyx_t_5); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 453, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "glypy/composition/ccomposition.pyx":459
 *                                              mass_data=mass_data)
 *                                              * group_coef)
 *                         prev_chem_symbol_start = i + 1             # <<<<<<<<<<<<<<
 *                         seek_mode = False
 *                         parse_stack = ""
 */
          __pyx_v_prev_chem_symbol_start = (__pyx_v_i + 1);

          /* "glypy/composition/ccomposition.pyx":460
 *                                              * group_coef)
 *                         prev_chem_symbol_start = i + 1
 *                         seek_mode = False             # <<<<<<<<<<<<<<
 *                         parse_stack = ""
 *                 elif(formula[i] == ")"):
 */
          __pyx_v_seek_mode = 0;

          /* "glypy/composition/ccomposition.pyx":461
 *                         prev_chem_symbol_start = i + 1
 *                         seek_mode = False
 *                         parse_stack = ""             # <<<<<<<<<<<<<<
 *                 elif(formula[i] == ")"):
 *                     seek_mode += 1
 */
          __Pyx_INCREF(__pyx_kp_s__3);
          __Pyx_DECREF_SET(__pyx_v_parse_stack, __pyx_kp_s__3);

          /* "glypy/composition/ccomposition.pyx":451
 *                 if(ch == "("):
 *                     seek_mode -= 1
 *                     if seek_mode == 0:             # <<<<<<<<<<<<<<
 * 
 *                         resolve_stack.append(Composition(
 */
        }

        /* "glypy/composition/ccomposition.pyx":449
 *                 parse_stack += ch
 *                 i -= 1
 *                 if(ch == "("):             # <<<<<<<<<<<<<<
 *                     seek_mode -= 1
 *                     if seek_mode == 0:
 */
        goto __pyx_L38;
      }

      /* "glypy/composition/ccomposition.pyx":462
 *                         seek_mode = False
 *                         parse_stack = ""
 *                 elif(formula[i] == ")"):             # <<<<<<<<<<<<<<
 *                     seek_mode += 1
 *                 else:
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_formula, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_kp_s__5, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 462, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_3) {

        /* "glypy/composition/ccomposition.pyx":463
 *                         parse_stack = ""
 *                 elif(formula[i] == ")"):
 *                     seek_mode += 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     # continue to accumulate tokens
 */
        __pyx_v_seek_mode = (__pyx_v_seek_mode + 1);

        /* "glypy/composition/ccomposition.pyx":462
 *                         seek_mode = False
 *                         parse_stack = ""
 *                 elif(formula[i] == ")"):             # <<<<<<<<<<<<<<
 *                     seek_mode += 1
 *                 else:
 */
        goto __pyx_L38;
      }

      /* "glypy/composition/ccomposition.pyx":466
 *                 else:
 *                     # continue to accumulate tokens
 *                     pass             # <<<<<<<<<<<<<<
 * 
 *         # Unspool the resolve stack, adding together the chunks
 */
      /*else*/ {
      }
      __pyx_L38:;
    }
    __pyx_L5:;
    __pyx_L3_continue:;
  }

  /* "glypy/composition/ccomposition.pyx":471
 *         # at this level. __add__ operates immutably, so must manually
 *         # loop through each chunk.
 *         for chunk in resolve_stack:             # <<<<<<<<<<<<<<
 *             for elem, cnt in chunk.items():
 *                 self[elem] += cnt
 */
  __pyx_t_5 = __pyx_v_resolve_stack; __Pyx_INCREF(__pyx_t_5); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_5)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_2); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 471, __pyx_L1_error)
    #else
    __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_chunk, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "glypy/composition/ccomposition.pyx":472
 *         # loop through each chunk.
 *         for chunk in resolve_stack:
 *             for elem, cnt in chunk.items():             # <<<<<<<<<<<<<<
 *                 self[elem] += cnt
 * 
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_chunk, __pyx_n_s_items); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_2 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_12);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_12 = __pyx_t_2; __Pyx_INCREF(__pyx_t_12); __pyx_t_16 = 0;
      __pyx_t_15 = NULL;
    } else {
      __pyx_t_16 = -1; __pyx_t_12 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 472, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_15 = Py_TYPE(__pyx_t_12)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 472, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_15)) {
        if (likely(PyList_CheckExact(__pyx_t_12))) {
          if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_12)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_12, __pyx_t_16); __Pyx_INCREF(__pyx_t_2); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 472, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_12, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_12)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_12, __pyx_t_16); __Pyx_INCREF(__pyx_t_2); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 472, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_12, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_15(__pyx_t_12);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 472, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
        PyObject* sequence = __pyx_t_2;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 472, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_13 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_13 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_4);
        #else
        __pyx_t_13 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_11 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_18 = Py_TYPE(__pyx_t_11)->tp_iternext;
        index = 0; __pyx_t_13 = __pyx_t_18(__pyx_t_11); if (unlikely(!__pyx_t_13)) goto __pyx_L44_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_13);
        index = 1; __pyx_t_4 = __pyx_t_18(__pyx_t_11); if (unlikely(!__pyx_t_4)) goto __pyx_L44_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_18(__pyx_t_11), 2) < 0) __PYX_ERR(0, 472, __pyx_L1_error)
        __pyx_t_18 = NULL;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L45_unpacking_done;
        __pyx_L44_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_18 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 472, __pyx_L1_error)
        __pyx_L45_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_elem, __pyx_t_13);
      __pyx_t_13 = 0;
      __Pyx_XDECREF_SET(__pyx_v_cnt, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "glypy/composition/ccomposition.pyx":473
 *         for chunk in resolve_stack:
 *             for elem, cnt in chunk.items():
 *                 self[elem] += cnt             # <<<<<<<<<<<<<<
 * 
 *     cpdef _from_dict(self, comp):
 */
      __Pyx_INCREF(__pyx_v_elem);
      __pyx_t_2 = __pyx_v_elem;
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 473, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = PyNumber_InPlaceAdd(__pyx_t_4, __pyx_v_cnt); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 473, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_self), __pyx_t_2, __pyx_t_13) < 0)) __PYX_ERR(0, 473, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "glypy/composition/ccomposition.pyx":472
 *         # loop through each chunk.
 *         for chunk in resolve_stack:
 *             for elem, cnt in chunk.items():             # <<<<<<<<<<<<<<
 *                 self[elem] += cnt
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "glypy/composition/ccomposition.pyx":471
 *         # at this level. __add__ operates immutably, so must manually
 *         # loop through each chunk.
 *         for chunk in resolve_stack:             # <<<<<<<<<<<<<<
 *             for elem, cnt in chunk.items():
 *                 self[elem] += cnt
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "glypy/composition/ccomposition.pyx":370
 * 
 *     @cython.boundscheck(True)
 *     def _from_formula_parens(self, formula, mass_data):             # <<<<<<<<<<<<<<
 *         # Parsing a formula backwards.
 *         cdef:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition._from_formula_parens", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_parse_stack);
  __Pyx_XDECREF(__pyx_v_resolve_stack);
  __Pyx_XDECREF(__pyx_v_num_atoms);
  __Pyx_XDECREF(__pyx_v_brace_pos);
  __Pyx_XDECREF(__pyx_v_isotope_num);
  __Pyx_XDECREF(__pyx_v_element_name);
  __Pyx_XDECREF(__pyx_v_isotope_string);
  __Pyx_XDECREF(__pyx_v_ch);
  __Pyx_XDECREF(__pyx_v_chunk);
  __Pyx_XDECREF(__pyx_v_elem);
  __Pyx_XDECREF(__pyx_v_cnt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":475
 *                 self[elem] += cnt
 * 
 *     cpdef _from_dict(self, comp):             # <<<<<<<<<<<<<<
 *         '''
 *         Directly overwrite this object's keys with the values in
 */

static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_43_from_dict(PyObject *__pyx_v_self, PyObject *__pyx_v_comp); /*proto*/
static PyObject *__pyx_f_5glypy_11composition_12ccomposition_12CComposition__from_dict(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_comp, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_from_dict", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_from_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_43_from_dict)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_comp) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_comp);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/composition/ccomposition.pyx":480
 *         `comp` without checking their type.
 *         '''
 *         PyDict_Update(self, comp)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double calc_mass(self, int average=False, charge=None, dict mass_data=nist_mass) except -1:
 */
  __pyx_t_5 = PyDict_Update(((PyObject *)__pyx_v_self), __pyx_v_comp); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 480, __pyx_L1_error)

  /* "glypy/composition/ccomposition.pyx":475
 *                 self[elem] += cnt
 * 
 *     cpdef _from_dict(self, comp):             # <<<<<<<<<<<<<<
 *         '''
 *         Directly overwrite this object's keys with the values in
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition._from_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_43_from_dict(PyObject *__pyx_v_self, PyObject *__pyx_v_comp); /*proto*/
static char __pyx_doc_5glypy_11composition_12ccomposition_12CComposition_42_from_dict[] = "\n        Directly overwrite this object's keys with the values in\n        `comp` without checking their type.\n        ";
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_43_from_dict(PyObject *__pyx_v_self, PyObject *__pyx_v_comp) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_from_dict (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_42_from_dict(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), ((PyObject *)__pyx_v_comp));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_42_from_dict(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_comp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_from_dict", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5glypy_11composition_12ccomposition_12CComposition__from_dict(__pyx_v_self, __pyx_v_comp, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition._from_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":482
 *         PyDict_Update(self, comp)
 * 
 *     cpdef double calc_mass(self, int average=False, charge=None, dict mass_data=nist_mass) except -1:             # <<<<<<<<<<<<<<
 *         cdef object mdid
 *         mdid = id(mass_data)
 */

static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_45calc_mass(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_5glypy_11composition_12ccomposition_12CComposition_calc_mass(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass *__pyx_optional_args) {
  int __pyx_v_average = ((int)0);
  PyObject *__pyx_v_charge = ((PyObject *)Py_None);
  PyObject *__pyx_v_mass_data = __pyx_k__8;
  PyObject *__pyx_v_mdid = 0;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  double __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_mass", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_average = __pyx_optional_args->average;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_charge = __pyx_optional_args->charge;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_mass_data = __pyx_optional_args->mass_data;
        }
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_calc_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_45calc_mass)) {
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_average); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 482, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_3, __pyx_v_charge, __pyx_v_mass_data};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 482, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_3, __pyx_v_charge, __pyx_v_mass_data};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 482, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 482, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
          __Pyx_INCREF(__pyx_v_charge);
          __Pyx_GIVEREF(__pyx_v_charge);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_charge);
          __Pyx_INCREF(__pyx_v_mass_data);
          __Pyx_GIVEREF(__pyx_v_mass_data);
          PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_v_mass_data);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 482, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 482, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_8;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glypy/composition/ccomposition.pyx":484
 *     cpdef double calc_mass(self, int average=False, charge=None, dict mass_data=nist_mass) except -1:
 *         cdef object mdid
 *         mdid = id(mass_data)             # <<<<<<<<<<<<<<
 *         if self._mass_args is not None and average is self._mass_args[0]\
 *                 and charge == self._mass_args[1] and mdid == self._mass_args[2]:
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_mass_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_mdid = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":485
 *         cdef object mdid
 *         mdid = id(mass_data)
 *         if self._mass_args is not None and average is self._mass_args[0]\             # <<<<<<<<<<<<<<
 *                 and charge == self._mass_args[1] and mdid == self._mass_args[2]:
 *             return self._mass
 */
  __pyx_t_10 = (__pyx_v_self->_mass_args != ((PyObject*)Py_None));
  __pyx_t_11 = (__pyx_t_10 != 0);
  if (__pyx_t_11) {
  } else {
    __pyx_t_9 = __pyx_t_11;
    goto __pyx_L4_bool_binop_done;
  }

  /* "glypy/composition/ccomposition.pyx":486
 *         mdid = id(mass_data)
 *         if self._mass_args is not None and average is self._mass_args[0]\
 *                 and charge == self._mass_args[1] and mdid == self._mass_args[2]:             # <<<<<<<<<<<<<<
 *             return self._mass
 *         else:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_average); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "glypy/composition/ccomposition.pyx":485
 *         cdef object mdid
 *         mdid = id(mass_data)
 *         if self._mass_args is not None and average is self._mass_args[0]\             # <<<<<<<<<<<<<<
 *                 and charge == self._mass_args[1] and mdid == self._mass_args[2]:
 *             return self._mass
 */
  if (unlikely(__pyx_v_self->_mass_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 485, __pyx_L1_error)
  }
  __pyx_t_11 = (__pyx_t_1 == PyTuple_GET_ITEM(__pyx_v_self->_mass_args, 0));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = (__pyx_t_11 != 0);
  if (__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L4_bool_binop_done;
  }

  /* "glypy/composition/ccomposition.pyx":486
 *         mdid = id(mass_data)
 *         if self._mass_args is not None and average is self._mass_args[0]\
 *                 and charge == self._mass_args[1] and mdid == self._mass_args[2]:             # <<<<<<<<<<<<<<
 *             return self._mass
 *         else:
 */
  if (unlikely(__pyx_v_self->_mass_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 486, __pyx_L1_error)
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_charge, PyTuple_GET_ITEM(__pyx_v_self->_mass_args, 1), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L4_bool_binop_done;
  }
  if (unlikely(__pyx_v_self->_mass_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 486, __pyx_L1_error)
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_mdid, PyTuple_GET_ITEM(__pyx_v_self->_mass_args, 2), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = __pyx_t_10;
  __pyx_L4_bool_binop_done:;

  /* "glypy/composition/ccomposition.pyx":485
 *         cdef object mdid
 *         mdid = id(mass_data)
 *         if self._mass_args is not None and average is self._mass_args[0]\             # <<<<<<<<<<<<<<
 *                 and charge == self._mass_args[1] and mdid == self._mass_args[2]:
 *             return self._mass
 */
  if (__pyx_t_9) {

    /* "glypy/composition/ccomposition.pyx":487
 *         if self._mass_args is not None and average is self._mass_args[0]\
 *                 and charge == self._mass_args[1] and mdid == self._mass_args[2]:
 *             return self._mass             # <<<<<<<<<<<<<<
 *         else:
 *             self._mass_args = (average, charge, mdid)
 */
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_v_self->_mass); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 487, __pyx_L1_error)
    __pyx_r = __pyx_t_8;
    goto __pyx_L0;

    /* "glypy/composition/ccomposition.pyx":485
 *         cdef object mdid
 *         mdid = id(mass_data)
 *         if self._mass_args is not None and average is self._mass_args[0]\             # <<<<<<<<<<<<<<
 *                 and charge == self._mass_args[1] and mdid == self._mass_args[2]:
 *             return self._mass
 */
  }

  /* "glypy/composition/ccomposition.pyx":489
 *             return self._mass
 *         else:
 *             self._mass_args = (average, charge, mdid)             # <<<<<<<<<<<<<<
 *             self._mass = calculate_mass(composition=self, average=average, charge=charge, mass_data=mass_data)
 *             return self._mass
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_average); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_charge);
    __Pyx_GIVEREF(__pyx_v_charge);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_charge);
    __Pyx_INCREF(__pyx_v_mdid);
    __Pyx_GIVEREF(__pyx_v_mdid);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_mdid);
    __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_mass_args);
    __Pyx_DECREF(__pyx_v_self->_mass_args);
    __pyx_v_self->_mass_args = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glypy/composition/ccomposition.pyx":490
 *         else:
 *             self._mass_args = (average, charge, mdid)
 *             self._mass = calculate_mass(composition=self, average=average, charge=charge, mass_data=mass_data)             # <<<<<<<<<<<<<<
 *             return self._mass
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_calculate_mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_composition, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 490, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_average); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_average, __pyx_t_4) < 0) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_charge, __pyx_v_charge) < 0) __PYX_ERR(0, 490, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_mass_data, __pyx_v_mass_data) < 0) __PYX_ERR(0, 490, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->_mass);
    __Pyx_DECREF(__pyx_v_self->_mass);
    __pyx_v_self->_mass = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "glypy/composition/ccomposition.pyx":491
 *             self._mass_args = (average, charge, mdid)
 *             self._mass = calculate_mass(composition=self, average=average, charge=charge, mass_data=mass_data)
 *             return self._mass             # <<<<<<<<<<<<<<
 * 
 *     property mass:
 */
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_v_self->_mass); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 491, __pyx_L1_error)
    __pyx_r = __pyx_t_8;
    goto __pyx_L0;
  }

  /* "glypy/composition/ccomposition.pyx":482
 *         PyDict_Update(self, comp)
 * 
 *     cpdef double calc_mass(self, int average=False, charge=None, dict mass_data=nist_mass) except -1:             # <<<<<<<<<<<<<<
 *         cdef object mdid
 *         mdid = id(mass_data)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.calc_mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1.0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mdid);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_45calc_mass(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_45calc_mass(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_average;
  PyObject *__pyx_v_charge = 0;
  PyObject *__pyx_v_mass_data = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_mass (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_average,&__pyx_n_s_charge,&__pyx_n_s_mass_data,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)Py_None);
    values[2] = __pyx_k__8;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_average);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mass_data);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_mass") < 0)) __PYX_ERR(0, 482, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_average = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_average == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 482, __pyx_L3_error)
    } else {
      __pyx_v_average = ((int)0);
    }
    __pyx_v_charge = values[1];
    __pyx_v_mass_data = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_mass", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 482, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.calc_mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mass_data), (&PyDict_Type), 1, "mass_data", 1))) __PYX_ERR(0, 482, __pyx_L1_error)
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_44calc_mass(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), __pyx_v_average, __pyx_v_charge, __pyx_v_mass_data);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_44calc_mass(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, int __pyx_v_average, PyObject *__pyx_v_charge, PyObject *__pyx_v_mass_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_mass", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.average = __pyx_v_average;
  __pyx_t_2.charge = __pyx_v_charge;
  __pyx_t_2.mass_data = __pyx_v_mass_data;
  __pyx_t_1 = __pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition->calc_mass(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 482, __pyx_L1_error)
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.calc_mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":494
 * 
 *     property mass:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.calc_mass()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_4mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_4mass_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_4mass___get__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_12CComposition_4mass___get__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "glypy/composition/ccomposition.pyx":495
 *     property mass:
 *         def __get__(self):
 *             return self.calc_mass()             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, args=None, formula=None, dict mass_data=None, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->calc_mass(__pyx_v_self, 0, NULL); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 495, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":494
 * 
 *     property mass:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.calc_mass()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":497
 *             return self.calc_mass()
 * 
 *     def __init__(self, args=None, formula=None, dict mass_data=None, **kwargs):             # <<<<<<<<<<<<<<
 *         #dict.__init__(self)
 *         if mass_data is None:
 */

/* Python wrapper */
static int __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_47__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_47__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_formula = 0;
  PyObject *__pyx_v_mass_data = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_formula,&__pyx_n_s_mass_data,0};
    PyObject* values[3] = {0,0,0};
    values[0] = ((PyObject *)Py_None);
    values[1] = ((PyObject *)Py_None);
    values[2] = ((PyObject*)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formula);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mass_data);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 497, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_args = values[0];
    __pyx_v_formula = values[1];
    __pyx_v_mass_data = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 497, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mass_data), (&PyDict_Type), 1, "mass_data", 1))) __PYX_ERR(0, 497, __pyx_L1_error)
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_46__init__(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self), __pyx_v_args, __pyx_v_formula, __pyx_v_mass_data, __pyx_v_kwargs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5glypy_11composition_12ccomposition_12CComposition_46__init__(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_formula, PyObject *__pyx_v_mass_data, PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_mass_data);

  /* "glypy/composition/ccomposition.pyx":499
 *     def __init__(self, args=None, formula=None, dict mass_data=None, **kwargs):
 *         #dict.__init__(self)
 *         if mass_data is None:             # <<<<<<<<<<<<<<
 *             mass_data = nist_mass
 * 
 */
  __pyx_t_1 = (__pyx_v_mass_data == ((PyObject*)Py_None));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":500
 *         #dict.__init__(self)
 *         if mass_data is None:
 *             mass_data = nist_mass             # <<<<<<<<<<<<<<
 * 
 *         if formula is not None:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_nist_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyDict_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_mass_data, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "glypy/composition/ccomposition.pyx":499
 *     def __init__(self, args=None, formula=None, dict mass_data=None, **kwargs):
 *         #dict.__init__(self)
 *         if mass_data is None:             # <<<<<<<<<<<<<<
 *             mass_data = nist_mass
 * 
 */
  }

  /* "glypy/composition/ccomposition.pyx":502
 *             mass_data = nist_mass
 * 
 *         if formula is not None:             # <<<<<<<<<<<<<<
 *             self._from_formula(formula, mass_data)
 * 
 */
  __pyx_t_2 = (__pyx_v_formula != Py_None);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "glypy/composition/ccomposition.pyx":503
 * 
 *         if formula is not None:
 *             self._from_formula(formula, mass_data)             # <<<<<<<<<<<<<<
 * 
 *         elif args:
 */
    if (!(likely(PyString_CheckExact(__pyx_v_formula))||((__pyx_v_formula) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_formula)->tp_name), 0))) __PYX_ERR(0, 503, __pyx_L1_error)
    __pyx_t_3 = ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->_from_formula(__pyx_v_self, ((PyObject*)__pyx_v_formula), __pyx_v_mass_data, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 503, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "glypy/composition/ccomposition.pyx":502
 *             mass_data = nist_mass
 * 
 *         if formula is not None:             # <<<<<<<<<<<<<<
 *             self._from_formula(formula, mass_data)
 * 
 */
    goto __pyx_L4;
  }

  /* "glypy/composition/ccomposition.pyx":505
 *             self._from_formula(formula, mass_data)
 * 
 *         elif args:             # <<<<<<<<<<<<<<
 *             if isinstance(args, dict):
 *                 self._from_dict(args)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_args); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 505, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "glypy/composition/ccomposition.pyx":506
 * 
 *         elif args:
 *             if isinstance(args, dict):             # <<<<<<<<<<<<<<
 *                 self._from_dict(args)
 *             elif isinstance(args, str):
 */
    __pyx_t_1 = PyDict_Check(__pyx_v_args); 
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "glypy/composition/ccomposition.pyx":507
 *         elif args:
 *             if isinstance(args, dict):
 *                 self._from_dict(args)             # <<<<<<<<<<<<<<
 *             elif isinstance(args, str):
 *                 try:
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->_from_dict(__pyx_v_self, __pyx_v_args, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 507, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "glypy/composition/ccomposition.pyx":506
 * 
 *         elif args:
 *             if isinstance(args, dict):             # <<<<<<<<<<<<<<
 *                 self._from_dict(args)
 *             elif isinstance(args, str):
 */
      goto __pyx_L5;
    }

    /* "glypy/composition/ccomposition.pyx":508
 *             if isinstance(args, dict):
 *                 self._from_dict(args)
 *             elif isinstance(args, str):             # <<<<<<<<<<<<<<
 *                 try:
 *                     self._from_formula(args, mass_data)
 */
    __pyx_t_2 = PyString_Check(__pyx_v_args); 
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "glypy/composition/ccomposition.pyx":509
 *                 self._from_dict(args)
 *             elif isinstance(args, str):
 *                 try:             # <<<<<<<<<<<<<<
 *                     self._from_formula(args, mass_data)
 *                 except ChemicalCompositionError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_6);
        /*try:*/ {

          /* "glypy/composition/ccomposition.pyx":510
 *             elif isinstance(args, str):
 *                 try:
 *                     self._from_formula(args, mass_data)             # <<<<<<<<<<<<<<
 *                 except ChemicalCompositionError:
 *                     raise ChemicalCompositionError(
 */
          if (!(likely(PyString_CheckExact(__pyx_v_args))||((__pyx_v_args) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_args)->tp_name), 0))) __PYX_ERR(0, 510, __pyx_L6_error)
          __pyx_t_3 = ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->_from_formula(__pyx_v_self, ((PyObject*)__pyx_v_args), __pyx_v_mass_data, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "glypy/composition/ccomposition.pyx":509
 *                 self._from_dict(args)
 *             elif isinstance(args, str):
 *                 try:             # <<<<<<<<<<<<<<
 *                     self._from_formula(args, mass_data)
 *                 except ChemicalCompositionError:
 */
        }
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L11_try_end;
        __pyx_L6_error:;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "glypy/composition/ccomposition.pyx":511
 *                 try:
 *                     self._from_formula(args, mass_data)
 *                 except ChemicalCompositionError:             # <<<<<<<<<<<<<<
 *                     raise ChemicalCompositionError(
 *                         'Could not create a Composition object from '
 */
        __Pyx_ErrFetch(&__pyx_t_3, &__pyx_t_7, &__pyx_t_8);
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 511, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_t_9);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_ErrRestore(__pyx_t_3, __pyx_t_7, __pyx_t_8);
        __pyx_t_3 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0;
        if (__pyx_t_10) {
          __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_3) < 0) __PYX_ERR(0, 511, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GOTREF(__pyx_t_3);

          /* "glypy/composition/ccomposition.pyx":512
 *                     self._from_formula(args, mass_data)
 *                 except ChemicalCompositionError:
 *                     raise ChemicalCompositionError(             # <<<<<<<<<<<<<<
 *                         'Could not create a Composition object from '
 *                         'string: "{}": not a '
 */
          __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 512, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_11);

          /* "glypy/composition/ccomposition.pyx":515
 *                         'Could not create a Composition object from '
 *                         'string: "{}": not a '
 *                         'formula'.format(args))             # <<<<<<<<<<<<<<
 *         else:
 *             self._from_dict(kwargs)
 */
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Could_not_create_a_Composition_o, __pyx_n_s_format); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 515, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_14 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
            }
          }
          __pyx_t_12 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_v_args) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_v_args);
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 515, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_13 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          __pyx_t_9 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_13, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_12);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 512, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_Raise(__pyx_t_9, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __PYX_ERR(0, 512, __pyx_L8_except_error)
        }
        goto __pyx_L8_except_error;
        __pyx_L8_except_error:;

        /* "glypy/composition/ccomposition.pyx":509
 *                 self._from_dict(args)
 *             elif isinstance(args, str):
 *                 try:             # <<<<<<<<<<<<<<
 *                     self._from_formula(args, mass_data)
 *                 except ChemicalCompositionError:
 */
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
        goto __pyx_L1_error;
        __pyx_L11_try_end:;
      }

      /* "glypy/composition/ccomposition.pyx":508
 *             if isinstance(args, dict):
 *                 self._from_dict(args)
 *             elif isinstance(args, str):             # <<<<<<<<<<<<<<
 *                 try:
 *                     self._from_formula(args, mass_data)
 */
    }
    __pyx_L5:;

    /* "glypy/composition/ccomposition.pyx":505
 *             self._from_formula(formula, mass_data)
 * 
 *         elif args:             # <<<<<<<<<<<<<<
 *             if isinstance(args, dict):
 *                 self._from_dict(args)
 */
    goto __pyx_L4;
  }

  /* "glypy/composition/ccomposition.pyx":517
 *                         'formula'.format(args))
 *         else:
 *             self._from_dict(kwargs)             # <<<<<<<<<<<<<<
 *         self._mass = None
 *         self._mass_args = None
 */
  /*else*/ {
    __pyx_t_3 = ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->__pyx_vtab)->_from_dict(__pyx_v_self, __pyx_v_kwargs, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L4:;

  /* "glypy/composition/ccomposition.pyx":518
 *         else:
 *             self._from_dict(kwargs)
 *         self._mass = None             # <<<<<<<<<<<<<<
 *         self._mass_args = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_mass);
  __Pyx_DECREF(__pyx_v_self->_mass);
  __pyx_v_self->_mass = Py_None;

  /* "glypy/composition/ccomposition.pyx":519
 *             self._from_dict(kwargs)
 *         self._mass = None
 *         self._mass_args = None             # <<<<<<<<<<<<<<
 * 
 * Composition = CComposition
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_mass_args);
  __Pyx_DECREF(__pyx_v_self->_mass_args);
  __pyx_v_self->_mass_args = ((PyObject*)Py_None);

  /* "glypy/composition/ccomposition.pyx":497
 *             return self.calc_mass()
 * 
 *     def __init__(self, args=None, formula=None, dict mass_data=None, **kwargs):             # <<<<<<<<<<<<<<
 *         #dict.__init__(self)
 *         if mass_data is None:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("glypy.composition.ccomposition.CComposition.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mass_data);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":524
 * 
 * 
 * cpdef CComposition composition_sum(list compositions):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_3composition_sum(PyObject *__pyx_self, PyObject *__pyx_v_compositions); /*proto*/
static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_5glypy_11composition_12ccomposition_composition_sum(PyObject *__pyx_v_compositions, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_accumulator = 0;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_current = 0;
  PyObject *__pyx_v_elem = 0;
  long __pyx_v_cnt;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pvalue;
  Py_ssize_t __pyx_v_ppos;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_t_5;
  long __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("composition_sum", 0);

  /* "glypy/composition/ccomposition.pyx":532
 *         PyObject *pkey
 *         PyObject *pvalue
 *         Py_ssize_t ppos = 0             # <<<<<<<<<<<<<<
 * 
 *     accumulator = CComposition()
 */
  __pyx_v_ppos = 0;

  /* "glypy/composition/ccomposition.pyx":534
 *         Py_ssize_t ppos = 0
 * 
 *     accumulator = CComposition()             # <<<<<<<<<<<<<<
 *     for i in range(len(compositions)):
 *         current = compositions[i]
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5glypy_11composition_12ccomposition_CComposition)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_accumulator = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":535
 * 
 *     accumulator = CComposition()
 *     for i in range(len(compositions)):             # <<<<<<<<<<<<<<
 *         current = compositions[i]
 *         ppos = 0
 */
  if (unlikely(__pyx_v_compositions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 535, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_GET_SIZE(__pyx_v_compositions); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 535, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glypy/composition/ccomposition.pyx":536
 *     accumulator = CComposition()
 *     for i in range(len(compositions)):
 *         current = compositions[i]             # <<<<<<<<<<<<<<
 *         ppos = 0
 * 
 */
    if (unlikely(__pyx_v_compositions == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 536, __pyx_L1_error)
    }
    if (!(likely(((PyList_GET_ITEM(__pyx_v_compositions, __pyx_v_i)) == Py_None) || likely(__Pyx_TypeTest(PyList_GET_ITEM(__pyx_v_compositions, __pyx_v_i), __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(0, 536, __pyx_L1_error)
    __pyx_t_1 = PyList_GET_ITEM(__pyx_v_compositions, __pyx_v_i);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_current, ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glypy/composition/ccomposition.pyx":537
 *     for i in range(len(compositions)):
 *         current = compositions[i]
 *         ppos = 0             # <<<<<<<<<<<<<<
 * 
 *         while(PyDict_Next(current, &ppos, &pkey, &pvalue)):
 */
    __pyx_v_ppos = 0;

    /* "glypy/composition/ccomposition.pyx":539
 *         ppos = 0
 * 
 *         while(PyDict_Next(current, &ppos, &pkey, &pvalue)):             # <<<<<<<<<<<<<<
 *             elem = <str>pkey
 *             cnt = accumulator.getitem(elem)
 */
    while (1) {
      __pyx_t_5 = (PyDict_Next(((PyObject *)__pyx_v_current), (&__pyx_v_ppos), (&__pyx_v_pkey), (&__pyx_v_pvalue)) != 0);
      if (!__pyx_t_5) break;

      /* "glypy/composition/ccomposition.pyx":540
 * 
 *         while(PyDict_Next(current, &ppos, &pkey, &pvalue)):
 *             elem = <str>pkey             # <<<<<<<<<<<<<<
 *             cnt = accumulator.getitem(elem)
 *             accumulator.setitem(elem, cnt + PyInt_AsLong(<object>pvalue))
 */
      __pyx_t_1 = ((PyObject *)__pyx_v_pkey);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_elem, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "glypy/composition/ccomposition.pyx":541
 *         while(PyDict_Next(current, &ppos, &pkey, &pvalue)):
 *             elem = <str>pkey
 *             cnt = accumulator.getitem(elem)             # <<<<<<<<<<<<<<
 *             accumulator.setitem(elem, cnt + PyInt_AsLong(<object>pvalue))
 *     return accumulator
 */
      __pyx_v_cnt = ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_accumulator->__pyx_vtab)->getitem(__pyx_v_accumulator, __pyx_v_elem);

      /* "glypy/composition/ccomposition.pyx":542
 *             elem = <str>pkey
 *             cnt = accumulator.getitem(elem)
 *             accumulator.setitem(elem, cnt + PyInt_AsLong(<object>pvalue))             # <<<<<<<<<<<<<<
 *     return accumulator
 * 
 */
      __pyx_t_6 = PyInt_AsLong(((PyObject *)__pyx_v_pvalue)); if (unlikely(__pyx_t_6 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 542, __pyx_L1_error)
      ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_accumulator->__pyx_vtab)->setitem(__pyx_v_accumulator, __pyx_v_elem, (__pyx_v_cnt + __pyx_t_6));
    }
  }

  /* "glypy/composition/ccomposition.pyx":543
 *             cnt = accumulator.getitem(elem)
 *             accumulator.setitem(elem, cnt + PyInt_AsLong(<object>pvalue))
 *     return accumulator             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_accumulator));
  __pyx_r = __pyx_v_accumulator;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":524
 * 
 * 
 * cpdef CComposition composition_sum(list compositions):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.composition_sum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_accumulator);
  __Pyx_XDECREF((PyObject *)__pyx_v_current);
  __Pyx_XDECREF(__pyx_v_elem);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_3composition_sum(PyObject *__pyx_self, PyObject *__pyx_v_compositions); /*proto*/
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_3composition_sum(PyObject *__pyx_self, PyObject *__pyx_v_compositions) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("composition_sum (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_compositions), (&PyList_Type), 1, "compositions", 1))) __PYX_ERR(0, 524, __pyx_L1_error)
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_2composition_sum(__pyx_self, ((PyObject*)__pyx_v_compositions));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_2composition_sum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compositions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("composition_sum", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5glypy_11composition_12ccomposition_composition_sum(__pyx_v_compositions, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glypy.composition.ccomposition.composition_sum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glypy/composition/ccomposition.pyx":548
 * @cython.wraparound(False)
 * @cython.boundscheck(False)
 * cpdef double calculate_mass(CComposition composition=None, str formula=None, int average=False, charge=None, mass_data=None) except -1:             # <<<<<<<<<<<<<<
 *     """Calculates the monoisotopic mass of a chemical formula or CComposition object.
 * 
 */

static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_5calculate_mass(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_5glypy_11composition_12ccomposition_calculate_mass(CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass *__pyx_optional_args) {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None);
  PyObject *__pyx_v_formula = ((PyObject*)Py_None);
  int __pyx_v_average = ((int)0);
  PyObject *__pyx_v_charge = ((PyObject *)Py_None);
  PyObject *__pyx_v_mass_data = ((PyObject *)Py_None);
  int __pyx_v_old_charge;
  int __pyx_v_isotope_num;
  int __pyx_v_isotope;
  int __pyx_v_quantity;
  double __pyx_v_mass;
  double __pyx_v_isotope_mass;
  double __pyx_v_isotope_frequency;
  long __pyx_v__charge;
  PyObject *__pyx_v_isotope_string = 0;
  PyObject *__pyx_v_element_name = 0;
  PyObject *__pyx_v_mass_provider = 0;
  PyObject *__pyx_v_key_list = 0;
  Py_ssize_t __pyx_v_iter_pos;
  PyObject *__pyx_v_interim;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  long __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_t_13;
  double __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calculate_mass", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_composition = __pyx_optional_args->composition;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_formula = __pyx_optional_args->formula;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_average = __pyx_optional_args->average;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_charge = __pyx_optional_args->charge;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_mass_data = __pyx_optional_args->mass_data;
            }
          }
        }
      }
    }
  }
  __Pyx_INCREF((PyObject *)__pyx_v_composition);
  __Pyx_INCREF(__pyx_v_charge);

  /* "glypy/composition/ccomposition.pyx":581
 *         list key_list
 *         PyObject* interm
 *         Py_ssize_t iter_pos = 0             # <<<<<<<<<<<<<<
 * 
 *     if mass_data is None:
 */
  __pyx_v_iter_pos = 0;

  /* "glypy/composition/ccomposition.pyx":583
 *         Py_ssize_t iter_pos = 0
 * 
 *     if mass_data is None:             # <<<<<<<<<<<<<<
 *         mass_provider = nist_mass
 *     else:
 */
  __pyx_t_1 = (__pyx_v_mass_data == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":584
 * 
 *     if mass_data is None:
 *         mass_provider = nist_mass             # <<<<<<<<<<<<<<
 *     else:
 *         mass_provider = mass_data
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_nist_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyDict_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 584, __pyx_L1_error)
    __pyx_v_mass_provider = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glypy/composition/ccomposition.pyx":583
 *         Py_ssize_t iter_pos = 0
 * 
 *     if mass_data is None:             # <<<<<<<<<<<<<<
 *         mass_provider = nist_mass
 *     else:
 */
    goto __pyx_L3;
  }

  /* "glypy/composition/ccomposition.pyx":586
 *         mass_provider = nist_mass
 *     else:
 *         mass_provider = mass_data             # <<<<<<<<<<<<<<
 * 
 *     if composition is None:
 */
  /*else*/ {
    if (!(likely(PyDict_CheckExact(__pyx_v_mass_data))||((__pyx_v_mass_data) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_mass_data)->tp_name), 0))) __PYX_ERR(0, 586, __pyx_L1_error)
    __pyx_t_3 = __pyx_v_mass_data;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_mass_provider = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "glypy/composition/ccomposition.pyx":588
 *         mass_provider = mass_data
 * 
 *     if composition is None:             # <<<<<<<<<<<<<<
 *         if formula is not None:
 *             composition = CComposition(formula)
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_composition) == Py_None);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "glypy/composition/ccomposition.pyx":589
 * 
 *     if composition is None:
 *         if formula is not None:             # <<<<<<<<<<<<<<
 *             composition = CComposition(formula)
 *         else:
 */
    __pyx_t_1 = (__pyx_v_formula != ((PyObject*)Py_None));
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (likely(__pyx_t_2)) {

      /* "glypy/composition/ccomposition.pyx":590
 *     if composition is None:
 *         if formula is not None:
 *             composition = CComposition(formula)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ChemicalCompositionError("Must provide a composition or formula argument")
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5glypy_11composition_12ccomposition_CComposition), __pyx_v_formula); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_composition, ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "glypy/composition/ccomposition.pyx":589
 * 
 *     if composition is None:
 *         if formula is not None:             # <<<<<<<<<<<<<<
 *             composition = CComposition(formula)
 *         else:
 */
      goto __pyx_L5;
    }

    /* "glypy/composition/ccomposition.pyx":592
 *             composition = CComposition(formula)
 *         else:
 *             raise ChemicalCompositionError("Must provide a composition or formula argument")             # <<<<<<<<<<<<<<
 *     else:
 *         if formula is not None:
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 592, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_s_Must_provide_a_composition_or_fo) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s_Must_provide_a_composition_or_fo);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 592, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 592, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "glypy/composition/ccomposition.pyx":588
 *         mass_provider = mass_data
 * 
 *     if composition is None:             # <<<<<<<<<<<<<<
 *         if formula is not None:
 *             composition = CComposition(formula)
 */
    goto __pyx_L4;
  }

  /* "glypy/composition/ccomposition.pyx":594
 *             raise ChemicalCompositionError("Must provide a composition or formula argument")
 *     else:
 *         if formula is not None:             # <<<<<<<<<<<<<<
 *             raise ChemicalCompositionError("Must provide a composition or formula argument, but not both")
 * 
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_formula != ((PyObject*)Py_None));
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (unlikely(__pyx_t_1)) {

      /* "glypy/composition/ccomposition.pyx":595
 *     else:
 *         if formula is not None:
 *             raise ChemicalCompositionError("Must provide a composition or formula argument, but not both")             # <<<<<<<<<<<<<<
 * 
 *     # Get charge.
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 595, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_s_Must_provide_a_composition_or_fo_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s_Must_provide_a_composition_or_fo_2);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 595, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 595, __pyx_L1_error)

      /* "glypy/composition/ccomposition.pyx":594
 *             raise ChemicalCompositionError("Must provide a composition or formula argument")
 *     else:
 *         if formula is not None:             # <<<<<<<<<<<<<<
 *             raise ChemicalCompositionError("Must provide a composition or formula argument, but not both")
 * 
 */
    }
  }
  __pyx_L4:;

  /* "glypy/composition/ccomposition.pyx":598
 * 
 *     # Get charge.
 *     if charge is None:             # <<<<<<<<<<<<<<
 *         charge = composition.getitem('H+')
 *     else:
 */
  __pyx_t_1 = (__pyx_v_charge == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":599
 *     # Get charge.
 *     if charge is None:
 *         charge = composition.getitem('H+')             # <<<<<<<<<<<<<<
 *     else:
 *         if charge != 0 and composition.getitem('H+') != 0:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_composition->__pyx_vtab)->getitem(__pyx_v_composition, __pyx_kp_s_H)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_charge, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "glypy/composition/ccomposition.pyx":598
 * 
 *     # Get charge.
 *     if charge is None:             # <<<<<<<<<<<<<<
 *         charge = composition.getitem('H+')
 *     else:
 */
    goto __pyx_L7;
  }

  /* "glypy/composition/ccomposition.pyx":601
 *         charge = composition.getitem('H+')
 *     else:
 *         if charge != 0 and composition.getitem('H+') != 0:             # <<<<<<<<<<<<<<
 *             raise ChemicalCompositionError("Charge is specified both by the number of protons and parameters")
 *     _charge = PyInt_AsLong(charge)
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_v_charge, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_composition->__pyx_vtab)->getitem(__pyx_v_composition, __pyx_kp_s_H) != 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L9_bool_binop_done:;
    if (unlikely(__pyx_t_2)) {

      /* "glypy/composition/ccomposition.pyx":602
 *     else:
 *         if charge != 0 and composition.getitem('H+') != 0:
 *             raise ChemicalCompositionError("Charge is specified both by the number of protons and parameters")             # <<<<<<<<<<<<<<
 *     _charge = PyInt_AsLong(charge)
 *     old_charge = composition.getitem('H+')
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_s_Charge_is_specified_both_by_the) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s_Charge_is_specified_both_by_the);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 602, __pyx_L1_error)

      /* "glypy/composition/ccomposition.pyx":601
 *         charge = composition.getitem('H+')
 *     else:
 *         if charge != 0 and composition.getitem('H+') != 0:             # <<<<<<<<<<<<<<
 *             raise ChemicalCompositionError("Charge is specified both by the number of protons and parameters")
 *     _charge = PyInt_AsLong(charge)
 */
    }
  }
  __pyx_L7:;

  /* "glypy/composition/ccomposition.pyx":603
 *         if charge != 0 and composition.getitem('H+') != 0:
 *             raise ChemicalCompositionError("Charge is specified both by the number of protons and parameters")
 *     _charge = PyInt_AsLong(charge)             # <<<<<<<<<<<<<<
 *     old_charge = composition.getitem('H+')
 *     composition.setitem('H+', charge)
 */
  __pyx_t_6 = PyInt_AsLong(__pyx_v_charge); if (unlikely(__pyx_t_6 == ((long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 603, __pyx_L1_error)
  __pyx_v__charge = __pyx_t_6;

  /* "glypy/composition/ccomposition.pyx":604
 *             raise ChemicalCompositionError("Charge is specified both by the number of protons and parameters")
 *     _charge = PyInt_AsLong(charge)
 *     old_charge = composition.getitem('H+')             # <<<<<<<<<<<<<<
 *     composition.setitem('H+', charge)
 * 
 */
  __pyx_v_old_charge = ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_composition->__pyx_vtab)->getitem(__pyx_v_composition, __pyx_kp_s_H);

  /* "glypy/composition/ccomposition.pyx":605
 *     _charge = PyInt_AsLong(charge)
 *     old_charge = composition.getitem('H+')
 *     composition.setitem('H+', charge)             # <<<<<<<<<<<<<<
 * 
 *     # Calculate mass.
 */
  __pyx_t_6 = __Pyx_PyInt_As_long(__pyx_v_charge); if (unlikely((__pyx_t_6 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 605, __pyx_L1_error)
  ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_composition->__pyx_vtab)->setitem(__pyx_v_composition, __pyx_kp_s_H, __pyx_t_6);

  /* "glypy/composition/ccomposition.pyx":608
 * 
 *     # Calculate mass.
 *     mass = 0.0             # <<<<<<<<<<<<<<
 *     key_list = PyDict_Keys(composition)
 *     for iter_pos in range(len(key_list)):
 */
  __pyx_v_mass = 0.0;

  /* "glypy/composition/ccomposition.pyx":609
 *     # Calculate mass.
 *     mass = 0.0
 *     key_list = PyDict_Keys(composition)             # <<<<<<<<<<<<<<
 *     for iter_pos in range(len(key_list)):
 *         isotope_string = <str>PyList_GET_ITEM(key_list, iter_pos)
 */
  __pyx_t_3 = PyDict_Keys(((PyObject *)__pyx_v_composition)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_key_list = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "glypy/composition/ccomposition.pyx":610
 *     mass = 0.0
 *     key_list = PyDict_Keys(composition)
 *     for iter_pos in range(len(key_list)):             # <<<<<<<<<<<<<<
 *         isotope_string = <str>PyList_GET_ITEM(key_list, iter_pos)
 *         # element_name, isotope_num = _parse_isotope_string(isotope_string)
 */
  if (unlikely(__pyx_v_key_list == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 610, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_key_list); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 610, __pyx_L1_error)
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_iter_pos = __pyx_t_9;

    /* "glypy/composition/ccomposition.pyx":611
 *     key_list = PyDict_Keys(composition)
 *     for iter_pos in range(len(key_list)):
 *         isotope_string = <str>PyList_GET_ITEM(key_list, iter_pos)             # <<<<<<<<<<<<<<
 *         # element_name, isotope_num = _parse_isotope_string(isotope_string)
 *         element_name = _parse_isotope_string(isotope_string, &isotope_num)
 */
    __pyx_t_10 = PyList_GET_ITEM(__pyx_v_key_list, __pyx_v_iter_pos);
    __pyx_t_3 = ((PyObject *)__pyx_t_10);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_isotope_string, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "glypy/composition/ccomposition.pyx":613
 *         isotope_string = <str>PyList_GET_ITEM(key_list, iter_pos)
 *         # element_name, isotope_num = _parse_isotope_string(isotope_string)
 *         element_name = _parse_isotope_string(isotope_string, &isotope_num)             # <<<<<<<<<<<<<<
 * 
 *         # Calculate average mass if required and the isotope number is
 */
    __pyx_t_3 = __pyx_f_5glypy_11composition_12ccomposition__parse_isotope_string(__pyx_v_isotope_string, (&__pyx_v_isotope_num)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_element_name, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "glypy/composition/ccomposition.pyx":617
 *         # Calculate average mass if required and the isotope number is
 *         # not specified.
 *         if (not isotope_num) and average:             # <<<<<<<<<<<<<<
 *             for isotope in mass_provider[element_name]:
 *                 if isotope != 0:
 */
    __pyx_t_1 = ((!(__pyx_v_isotope_num != 0)) != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_1 = (__pyx_v_average != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L14_bool_binop_done:;
    if (__pyx_t_2) {

      /* "glypy/composition/ccomposition.pyx":618
 *         # not specified.
 *         if (not isotope_num) and average:
 *             for isotope in mass_provider[element_name]:             # <<<<<<<<<<<<<<
 *                 if isotope != 0:
 *                     quantity = <int>composition.getitem(element_name)
 */
      if (unlikely(__pyx_v_mass_provider == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 618, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_mass_provider, __pyx_v_element_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
        __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_11 = 0;
        __pyx_t_12 = NULL;
      } else {
        __pyx_t_11 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 618, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_12 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 618, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      for (;;) {
        if (likely(!__pyx_t_12)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 618, __pyx_L1_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 618, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 618, __pyx_L1_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 618, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_12(__pyx_t_4);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 618, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_isotope = __pyx_t_13;

        /* "glypy/composition/ccomposition.pyx":619
 *         if (not isotope_num) and average:
 *             for isotope in mass_provider[element_name]:
 *                 if isotope != 0:             # <<<<<<<<<<<<<<
 *                     quantity = <int>composition.getitem(element_name)
 *                     isotope_mass = <double>mass_provider[element_name][isotope][0]
 */
        __pyx_t_2 = ((__pyx_v_isotope != 0) != 0);
        if (__pyx_t_2) {

          /* "glypy/composition/ccomposition.pyx":620
 *             for isotope in mass_provider[element_name]:
 *                 if isotope != 0:
 *                     quantity = <int>composition.getitem(element_name)             # <<<<<<<<<<<<<<
 *                     isotope_mass = <double>mass_provider[element_name][isotope][0]
 *                     isotope_frequency = <double>mass_provider[element_name][isotope][1]
 */
          __pyx_v_quantity = ((int)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_composition->__pyx_vtab)->getitem(__pyx_v_composition, __pyx_v_element_name));

          /* "glypy/composition/ccomposition.pyx":621
 *                 if isotope != 0:
 *                     quantity = <int>composition.getitem(element_name)
 *                     isotope_mass = <double>mass_provider[element_name][isotope][0]             # <<<<<<<<<<<<<<
 *                     isotope_frequency = <double>mass_provider[element_name][isotope][1]
 * 
 */
          if (unlikely(__pyx_v_mass_provider == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 621, __pyx_L1_error)
          }
          __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_mass_provider, __pyx_v_element_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 621, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_isotope, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 621, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 621, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 621, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_v_isotope_mass = ((double)__pyx_t_14);

          /* "glypy/composition/ccomposition.pyx":622
 *                     quantity = <int>composition.getitem(element_name)
 *                     isotope_mass = <double>mass_provider[element_name][isotope][0]
 *                     isotope_frequency = <double>mass_provider[element_name][isotope][1]             # <<<<<<<<<<<<<<
 * 
 *                     mass += quantity * isotope_mass * isotope_frequency
 */
          if (unlikely(__pyx_v_mass_provider == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 622, __pyx_L1_error)
          }
          __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_mass_provider, __pyx_v_element_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_isotope, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 622, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 622, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_v_isotope_frequency = ((double)__pyx_t_14);

          /* "glypy/composition/ccomposition.pyx":624
 *                     isotope_frequency = <double>mass_provider[element_name][isotope][1]
 * 
 *                     mass += quantity * isotope_mass * isotope_frequency             # <<<<<<<<<<<<<<
 *         else:
 *             interim = PyDict_GetItem(mass_provider, element_name)
 */
          __pyx_v_mass = (__pyx_v_mass + ((__pyx_v_quantity * __pyx_v_isotope_mass) * __pyx_v_isotope_frequency));

          /* "glypy/composition/ccomposition.pyx":619
 *         if (not isotope_num) and average:
 *             for isotope in mass_provider[element_name]:
 *                 if isotope != 0:             # <<<<<<<<<<<<<<
 *                     quantity = <int>composition.getitem(element_name)
 *                     isotope_mass = <double>mass_provider[element_name][isotope][0]
 */
        }

        /* "glypy/composition/ccomposition.pyx":618
 *         # not specified.
 *         if (not isotope_num) and average:
 *             for isotope in mass_provider[element_name]:             # <<<<<<<<<<<<<<
 *                 if isotope != 0:
 *                     quantity = <int>composition.getitem(element_name)
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "glypy/composition/ccomposition.pyx":617
 *         # Calculate average mass if required and the isotope number is
 *         # not specified.
 *         if (not isotope_num) and average:             # <<<<<<<<<<<<<<
 *             for isotope in mass_provider[element_name]:
 *                 if isotope != 0:
 */
      goto __pyx_L13;
    }

    /* "glypy/composition/ccomposition.pyx":626
 *                     mass += quantity * isotope_mass * isotope_frequency
 *         else:
 *             interim = PyDict_GetItem(mass_provider, element_name)             # <<<<<<<<<<<<<<
 *             interim = PyDict_GetItem(<dict>interim, isotope_num)
 *             isotope_mass = PyFloat_AsDouble(<object>PyTuple_GetItem(<tuple>interim, 0))
 */
    /*else*/ {
      __pyx_v_interim = PyDict_GetItem(__pyx_v_mass_provider, __pyx_v_element_name);

      /* "glypy/composition/ccomposition.pyx":627
 *         else:
 *             interim = PyDict_GetItem(mass_provider, element_name)
 *             interim = PyDict_GetItem(<dict>interim, isotope_num)             # <<<<<<<<<<<<<<
 *             isotope_mass = PyFloat_AsDouble(<object>PyTuple_GetItem(<tuple>interim, 0))
 * 
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_isotope_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 627, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_interim = PyDict_GetItem(((PyObject *)__pyx_v_interim), __pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "glypy/composition/ccomposition.pyx":628
 *             interim = PyDict_GetItem(mass_provider, element_name)
 *             interim = PyDict_GetItem(<dict>interim, isotope_num)
 *             isotope_mass = PyFloat_AsDouble(<object>PyTuple_GetItem(<tuple>interim, 0))             # <<<<<<<<<<<<<<
 * 
 *             mass += (composition.getitem(isotope_string) * isotope_mass)
 */
      __pyx_t_10 = PyTuple_GetItem(((PyObject *)__pyx_v_interim), 0); if (unlikely(__pyx_t_10 == ((PyObject *)NULL))) __PYX_ERR(0, 628, __pyx_L1_error)
      __pyx_t_14 = PyFloat_AsDouble(((PyObject *)__pyx_t_10)); if (unlikely(__pyx_t_14 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 628, __pyx_L1_error)
      __pyx_v_isotope_mass = __pyx_t_14;

      /* "glypy/composition/ccomposition.pyx":630
 *             isotope_mass = PyFloat_AsDouble(<object>PyTuple_GetItem(<tuple>interim, 0))
 * 
 *             mass += (composition.getitem(isotope_string) * isotope_mass)             # <<<<<<<<<<<<<<
 * 
 *     # Calculate m/z if required.
 */
      __pyx_v_mass = (__pyx_v_mass + (((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_composition->__pyx_vtab)->getitem(__pyx_v_composition, __pyx_v_isotope_string) * __pyx_v_isotope_mass));
    }
    __pyx_L13:;
  }

  /* "glypy/composition/ccomposition.pyx":633
 * 
 *     # Calculate m/z if required.
 *     if _charge != 0:             # <<<<<<<<<<<<<<
 *         mass /= abs(_charge)
 * 
 */
  __pyx_t_2 = ((__pyx_v__charge != 0) != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":634
 *     # Calculate m/z if required.
 *     if _charge != 0:
 *         mass /= abs(_charge)             # <<<<<<<<<<<<<<
 * 
 *     if old_charge != 0:
 */
    __pyx_t_6 = labs(__pyx_v__charge); if (unlikely(__pyx_t_6 == ((long)-1))) __PYX_ERR(0, 634, __pyx_L1_error)
    if (unlikely(__pyx_t_6 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 634, __pyx_L1_error)
    }
    __pyx_v_mass = (__pyx_v_mass / __pyx_t_6);

    /* "glypy/composition/ccomposition.pyx":633
 * 
 *     # Calculate m/z if required.
 *     if _charge != 0:             # <<<<<<<<<<<<<<
 *         mass /= abs(_charge)
 * 
 */
  }

  /* "glypy/composition/ccomposition.pyx":636
 *         mass /= abs(_charge)
 * 
 *     if old_charge != 0:             # <<<<<<<<<<<<<<
 *         composition.setitem('H+', old_charge)
 *     else:
 */
  __pyx_t_2 = ((__pyx_v_old_charge != 0) != 0);
  if (__pyx_t_2) {

    /* "glypy/composition/ccomposition.pyx":637
 * 
 *     if old_charge != 0:
 *         composition.setitem('H+', old_charge)             # <<<<<<<<<<<<<<
 *     else:
 *         PyDict_DelItem(composition, "H+")
 */
    ((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_composition->__pyx_vtab)->setitem(__pyx_v_composition, __pyx_kp_s_H, __pyx_v_old_charge);

    /* "glypy/composition/ccomposition.pyx":636
 *         mass /= abs(_charge)
 * 
 *     if old_charge != 0:             # <<<<<<<<<<<<<<
 *         composition.setitem('H+', old_charge)
 *     else:
 */
    goto __pyx_L20;
  }

  /* "glypy/composition/ccomposition.pyx":639
 *         composition.setitem('H+', old_charge)
 *     else:
 *         PyDict_DelItem(composition, "H+")             # <<<<<<<<<<<<<<
 *     return mass
 */
  /*else*/ {
    __pyx_t_13 = PyDict_DelItem(((PyObject *)__pyx_v_composition), __pyx_kp_s_H); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 639, __pyx_L1_error)
  }
  __pyx_L20:;

  /* "glypy/composition/ccomposition.pyx":640
 *     else:
 *         PyDict_DelItem(composition, "H+")
 *     return mass             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_mass;
  goto __pyx_L0;

  /* "glypy/composition/ccomposition.pyx":548
 * @cython.wraparound(False)
 * @cython.boundscheck(False)
 * cpdef double calculate_mass(CComposition composition=None, str formula=None, int average=False, charge=None, mass_data=None) except -1:             # <<<<<<<<<<<<<<
 *     """Calculates the monoisotopic mass of a chemical formula or CComposition object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glypy.composition.ccomposition.calculate_mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1.0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_isotope_string);
  __Pyx_XDECREF(__pyx_v_element_name);
  __Pyx_XDECREF(__pyx_v_mass_provider);
  __Pyx_XDECREF(__pyx_v_key_list);
  __Pyx_XDECREF((PyObject *)__pyx_v_composition);
  __Pyx_XDECREF(__pyx_v_charge);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_5calculate_mass(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5glypy_11composition_12ccomposition_4calculate_mass[] = "Calculates the monoisotopic mass of a chemical formula or CComposition object.\n\n    Parameters\n    ----------\n    composition : CComposition\n        A Composition object with the elemental composition of a substance. Exclusive with `formula`\n    formula: str\n        A string describing a chemical composition. Exclusive with `composition`\n    average : bool, optional\n        If :py:const:`True` then the average mass is calculated. Note that mass\n        is not averaged for elements with specified isotopes. Default is\n        :py:const:`False`.\n    charge : int, optional\n        If not 0 then m/z is calculated: the mass is increased\n        by the corresponding number of proton masses and divided\n        by z.\n    mass_data : dict, optional\n        A dict with the masses of the chemical elements (the default\n        value is :py:data:`nist_mass`).\n\n    Returns\n    -------\n        mass : float\n    ";
static PyObject *__pyx_pw_5glypy_11composition_12ccomposition_5calculate_mass(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_composition = 0;
  PyObject *__pyx_v_formula = 0;
  int __pyx_v_average;
  PyObject *__pyx_v_charge = 0;
  PyObject *__pyx_v_mass_data = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calculate_mass (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_composition,&__pyx_n_s_formula,&__pyx_n_s_average,&__pyx_n_s_charge,&__pyx_n_s_mass_data,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[0] = (PyObject *)((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None);
    values[1] = ((PyObject*)Py_None);
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_composition);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formula);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_average);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mass_data);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calculate_mass") < 0)) __PYX_ERR(0, 548, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)values[0]);
    __pyx_v_formula = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_average = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_average == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 548, __pyx_L3_error)
    } else {
      __pyx_v_average = ((int)0);
    }
    __pyx_v_charge = values[3];
    __pyx_v_mass_data = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calculate_mass", 0, 0, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 548, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glypy.composition.ccomposition.calculate_mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_composition), __pyx_ptype_5glypy_11composition_12ccomposition_CComposition, 1, "composition", 0))) __PYX_ERR(0, 548, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_formula), (&PyString_Type), 1, "formula", 1))) __PYX_ERR(0, 548, __pyx_L1_error)
  __pyx_r = __pyx_pf_5glypy_11composition_12ccomposition_4calculate_mass(__pyx_self, __pyx_v_composition, __pyx_v_formula, __pyx_v_average, __pyx_v_charge, __pyx_v_mass_data);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5glypy_11composition_12ccomposition_4calculate_mass(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_composition, PyObject *__pyx_v_formula, int __pyx_v_average, PyObject *__pyx_v_charge, PyObject *__pyx_v_mass_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calculate_mass", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 5;
  __pyx_t_2.composition = __pyx_v_composition;
  __pyx_t_2.formula = __pyx_v_formula;
  __pyx_t_2.average = __pyx_v_average;
  __pyx_t_2.charge = __pyx_v_charge;
  __pyx_t_2.mass_data = __pyx_v_mass_data;
  __pyx_t_1 = __pyx_f_5glypy_11composition_12ccomposition_calculate_mass(0, &__pyx_t_2); if (unlikely(__pyx_t_1 == ((double)-1.0))) __PYX_ERR(0, 548, __pyx_L1_error)
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glypy.composition.ccomposition.calculate_mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition __pyx_vtable_5glypy_11composition_12ccomposition_CComposition;

static PyObject *__pyx_tp_new_5glypy_11composition_12ccomposition_CComposition(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *p;
  PyObject *o = (&PyDict_Type)->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)o);
  p->__pyx_vtab = __pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition;
  p->_mass = Py_None; Py_INCREF(Py_None);
  p->_mass_args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5glypy_11composition_12ccomposition_CComposition(PyObject *o) {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *p = (struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_mass);
  Py_CLEAR(p->_mass_args);
  PyObject_GC_Track(o);
  (&PyDict_Type)->tp_dealloc(o);
}

static int __pyx_tp_traverse_5glypy_11composition_12ccomposition_CComposition(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *p = (struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)o;
  if (!(&PyDict_Type)->tp_traverse); else { e = (&PyDict_Type)->tp_traverse(o,v,a); if (e) return e; }
  if (p->_mass) {
    e = (*v)(p->_mass, a); if (e) return e;
  }
  if (p->_mass_args) {
    e = (*v)(p->_mass_args, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5glypy_11composition_12ccomposition_CComposition(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *p = (struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)o;
  if (!(&PyDict_Type)->tp_clear); else (&PyDict_Type)->tp_clear(o);
  tmp = ((PyObject*)p->_mass);
  p->_mass = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_mass_args);
  p->_mass_args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static int __pyx_mp_ass_subscript_5glypy_11composition_12ccomposition_CComposition(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_31__setitem__(o, i, v);
  }
  else {
    if ((&PyDict_Type)->tp_as_mapping && (&PyDict_Type)->tp_as_mapping->mp_ass_subscript)
      return (&PyDict_Type)->tp_as_mapping->mp_ass_subscript(o, i, v);
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_tp_richcompare_5glypy_11composition_12ccomposition_CComposition(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_23__eq__(o1, o2);
    }
    case Py_NE: {
      return __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_25__ne__(o1, o2);
    }
    default: {
      if (likely((&PyDict_Type)->tp_richcompare)) return (&PyDict_Type)->tp_richcompare(o1, o2, op);
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_5glypy_11composition_12ccomposition_12CComposition_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_4mass_1__get__(o);
}

static PyMethodDef __pyx_methods_5glypy_11composition_12ccomposition_CComposition[] = {
  {"__reduce__", (PyCFunction)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_13__reduce__, METH_NOARGS, 0},
  {"__getstate__", (PyCFunction)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_15__getstate__, METH_NOARGS, 0},
  {"__setstate__", (PyCFunction)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_17__setstate__, METH_O, 0},
  {"__missing__", (PyCFunction)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_29__missing__, METH_O, 0},
  {"copy", (PyCFunction)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_33copy, METH_NOARGS, 0},
  {"clone", (PyCFunction)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_35clone, METH_NOARGS, 0},
  {"update", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_37update, METH_VARARGS|METH_KEYWORDS, 0},
  {"_from_formula", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_39_from_formula, METH_VARARGS|METH_KEYWORDS, 0},
  {"_from_formula_parens", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_41_from_formula_parens, METH_VARARGS|METH_KEYWORDS, 0},
  {"_from_dict", (PyCFunction)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_43_from_dict, METH_O, __pyx_doc_5glypy_11composition_12ccomposition_12CComposition_42_from_dict},
  {"calc_mass", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5glypy_11composition_12ccomposition_12CComposition_45calc_mass, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5glypy_11composition_12ccomposition_CComposition[] = {
  {(char *)"mass", __pyx_getprop_5glypy_11composition_12ccomposition_12CComposition_mass, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyNumberMethods __pyx_tp_as_number_CComposition = {
  __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_7__add__, /*nb_add*/
  __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_11__sub__, /*nb_subtract*/
  __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_19__mul__, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_27__neg__, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  0, /*nb_nonzero*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_5__iadd__, /*nb_inplace_add*/
  __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_9__isub__, /*nb_inplace_subtract*/
  __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_21__imul__, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PyMappingMethods __pyx_tp_as_mapping_CComposition = {
  0, /*mp_length*/
  0, /*mp_subscript*/
  __pyx_mp_ass_subscript_5glypy_11composition_12ccomposition_CComposition, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_5glypy_11composition_12ccomposition_CComposition = {
  PyVarObject_HEAD_INIT(0, 0)
  "glypy.composition.ccomposition.CComposition", /*tp_name*/
  sizeof(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5glypy_11composition_12ccomposition_CComposition, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_3__repr__, /*tp_repr*/
  &__pyx_tp_as_number_CComposition, /*tp_as_number*/
  0, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_CComposition, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_1__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "A Composition object stores a chemical composition of a\n    substance. Basically it is a dict object, in which keys are the names\n    of chemical elements and values contain integer numbers of\n    corresponding atoms in a substance.\n\n    The main improvement over dict is that Composition objects allow\n    addition and subtraction.\n\n    If ``formula`` is not specified, the constructor will look at the first\n    positional argument and try to build the object from it. Without\n    positional arguments, a Composition will be constructed directly from\n    keyword arguments.\n\n    Parameters\n    ----------\n    formula : str, optional\n        A string with a chemical formula. All elements must be present in\n        `mass_data`.\n    mass_data : dict, optional\n        A dict with the masses of chemical elements (the default\n        value is :py:data:`nist_mass`). It is used for formulae parsing only.\n        ", /*tp_doc*/
  __pyx_tp_traverse_5glypy_11composition_12ccomposition_CComposition, /*tp_traverse*/
  __pyx_tp_clear_5glypy_11composition_12ccomposition_CComposition, /*tp_clear*/
  __pyx_tp_richcompare_5glypy_11composition_12ccomposition_CComposition, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5glypy_11composition_12ccomposition_CComposition, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5glypy_11composition_12ccomposition_CComposition, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5glypy_11composition_12ccomposition_12CComposition_47__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5glypy_11composition_12ccomposition_CComposition, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"composition_sum", (PyCFunction)__pyx_pw_5glypy_11composition_12ccomposition_3composition_sum, METH_O, 0},
  {"calculate_mass", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5glypy_11composition_12ccomposition_5calculate_mass, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5glypy_11composition_12ccomposition_4calculate_mass},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_ccomposition(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_ccomposition},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "ccomposition",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 0},
  {&__pyx_kp_s_A_Z_a_z_d, __pyx_k_A_Z_a_z_d, sizeof(__pyx_k_A_Z_a_z_d), 0, 0, 1, 0},
  {&__pyx_kp_s_A_Z_a_z_d_d, __pyx_k_A_Z_a_z_d_d, sizeof(__pyx_k_A_Z_a_z_d_d), 0, 0, 1, 0},
  {&__pyx_kp_s_Badly_formed_isotope_number_s, __pyx_k_Badly_formed_isotope_number_s, sizeof(__pyx_k_Badly_formed_isotope_number_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Badly_formed_number_of_atoms_s, __pyx_k_Badly_formed_number_of_atoms_s, sizeof(__pyx_k_Badly_formed_number_of_atoms_s), 0, 0, 1, 0},
  {&__pyx_n_s_CComposition, __pyx_k_CComposition, sizeof(__pyx_k_CComposition), 0, 0, 1, 1},
  {&__pyx_kp_s_Cannot_multiply_Composition_by_n, __pyx_k_Cannot_multiply_Composition_by_n, sizeof(__pyx_k_Cannot_multiply_Composition_by_n), 0, 0, 1, 0},
  {&__pyx_kp_s_Charge_is_specified_both_by_the, __pyx_k_Charge_is_specified_both_by_the, sizeof(__pyx_k_Charge_is_specified_both_by_the), 0, 0, 1, 0},
  {&__pyx_n_s_ChemicalCompositionError, __pyx_k_ChemicalCompositionError, sizeof(__pyx_k_ChemicalCompositionError), 0, 0, 1, 1},
  {&__pyx_kp_s_Composition, __pyx_k_Composition, sizeof(__pyx_k_Composition), 0, 0, 1, 0},
  {&__pyx_n_s_Composition_2, __pyx_k_Composition_2, sizeof(__pyx_k_Composition_2), 0, 0, 1, 1},
  {&__pyx_kp_s_Could_not_create_a_Composition_o, __pyx_k_Could_not_create_a_Composition_o, sizeof(__pyx_k_Could_not_create_a_Composition_o), 0, 0, 1, 0},
  {&__pyx_kp_s_H, __pyx_k_H, sizeof(__pyx_k_H), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_formula, __pyx_k_Invalid_formula, sizeof(__pyx_k_Invalid_formula), 0, 0, 1, 0},
  {&__pyx_kp_s_Must_provide_a_composition_or_fo, __pyx_k_Must_provide_a_composition_or_fo, sizeof(__pyx_k_Must_provide_a_composition_or_fo), 0, 0, 1, 0},
  {&__pyx_kp_s_Must_provide_a_composition_or_fo_2, __pyx_k_Must_provide_a_composition_or_fo_2, sizeof(__pyx_k_Must_provide_a_composition_or_fo_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Unknown_chemical_element, __pyx_k_Unknown_chemical_element, sizeof(__pyx_k_Unknown_chemical_element), 0, 0, 1, 0},
  {&__pyx_kp_s_Unknown_chemical_element_in_the, __pyx_k_Unknown_chemical_element_in_the, sizeof(__pyx_k_Unknown_chemical_element_in_the), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 0},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_kp_s__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 0},
  {&__pyx_kp_s__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 1, 0},
  {&__pyx_kp_s__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 0, 1, 0},
  {&__pyx_kp_s__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 0, 1, 0},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_atom, __pyx_k_atom, sizeof(__pyx_k_atom), 0, 0, 1, 1},
  {&__pyx_n_s_average, __pyx_k_average, sizeof(__pyx_k_average), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_calc_mass, __pyx_k_calc_mass, sizeof(__pyx_k_calc_mass), 0, 0, 1, 1},
  {&__pyx_n_s_calculate_mass, __pyx_k_calculate_mass, sizeof(__pyx_k_calculate_mass), 0, 0, 1, 1},
  {&__pyx_n_s_charge, __pyx_k_charge, sizeof(__pyx_k_charge), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_clone, __pyx_k_clone, sizeof(__pyx_k_clone), 0, 0, 1, 1},
  {&__pyx_n_s_compile, __pyx_k_compile, sizeof(__pyx_k_compile), 0, 0, 1, 1},
  {&__pyx_n_s_composition, __pyx_k_composition, sizeof(__pyx_k_composition), 0, 0, 1, 1},
  {&__pyx_n_s_composition_factory, __pyx_k_composition_factory, sizeof(__pyx_k_composition_factory), 0, 0, 1, 1},
  {&__pyx_n_s_elem, __pyx_k_elem, sizeof(__pyx_k_elem), 0, 0, 1, 1},
  {&__pyx_n_s_endswith, __pyx_k_endswith, sizeof(__pyx_k_endswith), 0, 0, 1, 1},
  {&__pyx_n_s_findall, __pyx_k_findall, sizeof(__pyx_k_findall), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_formula, __pyx_k_formula, sizeof(__pyx_k_formula), 0, 0, 1, 1},
  {&__pyx_n_s_formula_2, __pyx_k_formula_2, sizeof(__pyx_k_formula_2), 0, 0, 1, 1},
  {&__pyx_n_s_formula_pattern, __pyx_k_formula_pattern, sizeof(__pyx_k_formula_pattern), 0, 0, 1, 1},
  {&__pyx_n_s_from_dict, __pyx_k_from_dict, sizeof(__pyx_k_from_dict), 0, 0, 1, 1},
  {&__pyx_n_s_from_formula, __pyx_k_from_formula, sizeof(__pyx_k_from_formula), 0, 0, 1, 1},
  {&__pyx_n_s_from_formula_parens, __pyx_k_from_formula_parens, sizeof(__pyx_k_from_formula_parens), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_glypy_composition_ccomposition, __pyx_k_glypy_composition_ccomposition, sizeof(__pyx_k_glypy_composition_ccomposition), 0, 0, 1, 1},
  {&__pyx_kp_s_glypy_composition_ccomposition_p, __pyx_k_glypy_composition_ccomposition_p, sizeof(__pyx_k_glypy_composition_ccomposition_p), 0, 0, 1, 0},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_isdigit, __pyx_k_isdigit, sizeof(__pyx_k_isdigit), 0, 0, 1, 1},
  {&__pyx_n_s_iso, __pyx_k_iso, sizeof(__pyx_k_iso), 0, 0, 1, 1},
  {&__pyx_n_s_isotope_pattern, __pyx_k_isotope_pattern, sizeof(__pyx_k_isotope_pattern), 0, 0, 1, 1},
  {&__pyx_n_s_isotope_string, __pyx_k_isotope_string, sizeof(__pyx_k_isotope_string), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_label, __pyx_k_label, sizeof(__pyx_k_label), 0, 0, 1, 1},
  {&__pyx_n_s_len, __pyx_k_len, sizeof(__pyx_k_len), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mass_data, __pyx_k_mass_data, sizeof(__pyx_k_mass_data), 0, 0, 1, 1},
  {&__pyx_n_s_mass_dict, __pyx_k_mass_dict, sizeof(__pyx_k_mass_dict), 0, 0, 1, 1},
  {&__pyx_n_s_match, __pyx_k_match, sizeof(__pyx_k_match), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_nist_mass, __pyx_k_nist_mass, sizeof(__pyx_k_nist_mass), 0, 0, 1, 1},
  {&__pyx_n_s_parse_isotope_string, __pyx_k_parse_isotope_string, sizeof(__pyx_k_parse_isotope_string), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_re, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_reverse, __pyx_k_reverse, sizeof(__pyx_k_reverse), 0, 0, 1, 1},
  {&__pyx_n_s_rfind, __pyx_k_rfind, sizeof(__pyx_k_rfind), 0, 0, 1, 1},
  {&__pyx_n_s_round, __pyx_k_round, sizeof(__pyx_k_round), 0, 0, 1, 1},
  {&__pyx_kp_s_s_d, __pyx_k_s_d, sizeof(__pyx_k_s_d), 0, 0, 1, 0},
  {&__pyx_n_s_sorted, __pyx_k_sorted, sizeof(__pyx_k_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_builtin_round = __Pyx_GetBuiltinName(__pyx_n_s_round); if (!__pyx_builtin_round) __PYX_ERR(0, 320, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 409, __pyx_L1_error)
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_n_s_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 432, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(0, 484, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "glypy/composition/ccomposition.pyx":456
 *                                              # Omit the last character, then reverse the parse
 *                                              # stack string.
 *                                              formula=parse_stack[:-1][::-1],             # <<<<<<<<<<<<<<
 *                                              mass_data=mass_data)
 *                                              * group_coef)
 */
  __pyx_slice__7 = PySlice_New(Py_None, Py_None, __pyx_int_neg_1); if (unlikely(!__pyx_slice__7)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__7);
  __Pyx_GIVEREF(__pyx_slice__7);

  /* "glypy/composition/ccomposition.pyx":67
 *     return element_name
 * 
 * def parse_isotope_string(str label):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int iso
 */
  __pyx_tuple__10 = PyTuple_Pack(3, __pyx_n_s_label, __pyx_n_s_iso, __pyx_n_s_elem); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_glypy_composition_ccomposition_p, __pyx_n_s_parse_isotope_string, 67, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_5glypy_11composition_12ccomposition__atom = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_5glypy_11composition_12ccomposition__formula = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_5glypy_11composition_12ccomposition__isotope_string = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_5glypy_11composition_12ccomposition_isotope_pattern = Py_None; Py_INCREF(Py_None);
  __pyx_v_5glypy_11composition_12ccomposition_formula_pattern = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_atom, (void *)&__pyx_v_5glypy_11composition_12ccomposition__atom, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_formula_2, (void *)&__pyx_v_5glypy_11composition_12ccomposition__formula, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_isotope_string, (void *)&__pyx_v_5glypy_11composition_12ccomposition__isotope_string, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_isotope_pattern, (void *)&__pyx_v_5glypy_11composition_12ccomposition_isotope_pattern, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_formula_pattern, (void *)&__pyx_v_5glypy_11composition_12ccomposition_formula_pattern, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("_parse_isotope_string", (void (*)(void))__pyx_f_5glypy_11composition_12ccomposition__parse_isotope_string, "PyObject *(PyObject *, int *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_make_isotope_string", (void (*)(void))__pyx_f_5glypy_11composition_12ccomposition__make_isotope_string, "PyObject *(PyObject *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("calculate_mass", (void (*)(void))__pyx_f_5glypy_11composition_12ccomposition_calculate_mass, "double (int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition = &__pyx_vtable_5glypy_11composition_12ccomposition_CComposition;
  __pyx_vtable_5glypy_11composition_12ccomposition_CComposition.clone = (struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, int __pyx_skip_dispatch))__pyx_f_5glypy_11composition_12ccomposition_12CComposition_clone;
  __pyx_vtable_5glypy_11composition_12ccomposition_CComposition.calc_mass = (double (*)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass *__pyx_optional_args))__pyx_f_5glypy_11composition_12ccomposition_12CComposition_calc_mass;
  __pyx_vtable_5glypy_11composition_12ccomposition_CComposition._from_formula = (PyObject *(*)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5glypy_11composition_12ccomposition_12CComposition__from_formula;
  __pyx_vtable_5glypy_11composition_12ccomposition_CComposition._from_dict = (PyObject *(*)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5glypy_11composition_12ccomposition_12CComposition__from_dict;
  __pyx_vtable_5glypy_11composition_12ccomposition_CComposition.getitem = (long (*)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *))__pyx_f_5glypy_11composition_12ccomposition_12CComposition_getitem;
  __pyx_vtable_5glypy_11composition_12ccomposition_CComposition.setitem = (void (*)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, long))__pyx_f_5glypy_11composition_12ccomposition_12CComposition_setitem;
  __pyx_vtable_5glypy_11composition_12ccomposition_CComposition.add_from = (struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *))__pyx_f_5glypy_11composition_12ccomposition_12CComposition_add_from;
  __pyx_vtable_5glypy_11composition_12ccomposition_CComposition.subtract_from = (struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *))__pyx_f_5glypy_11composition_12ccomposition_12CComposition_subtract_from;
  __pyx_vtable_5glypy_11composition_12ccomposition_CComposition._create = (struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *))__pyx_f_5glypy_11composition_12ccomposition_12CComposition__create;
  __pyx_type_5glypy_11composition_12ccomposition_CComposition.tp_base = (&PyDict_Type);
  if (PyType_Ready(&__pyx_type_5glypy_11composition_12ccomposition_CComposition) < 0) __PYX_ERR(0, 85, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5glypy_11composition_12ccomposition_CComposition.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5glypy_11composition_12ccomposition_CComposition.tp_dictoffset && __pyx_type_5glypy_11composition_12ccomposition_CComposition.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5glypy_11composition_12ccomposition_CComposition.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_5glypy_11composition_12ccomposition_CComposition.tp_dict, __pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition) < 0) __PYX_ERR(0, 85, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CComposition, (PyObject *)&__pyx_type_5glypy_11composition_12ccomposition_CComposition) < 0) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_ptype_5glypy_11composition_12ccomposition_CComposition = &__pyx_type_5glypy_11composition_12ccomposition_CComposition;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(2, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(3, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initccomposition(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initccomposition(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_ccomposition(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_ccomposition(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_ccomposition(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'ccomposition' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_ccomposition(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("ccomposition", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_glypy__composition__ccomposition) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "glypy.composition.ccomposition")) {
      if (unlikely(PyDict_SetItemString(modules, "glypy.composition.ccomposition", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely(__Pyx_modinit_variable_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "glypy/composition/ccomposition.pyx":4
 * 
 * # Credit to Pyteomics - http://pythonhosted.org/pyteomics - for majority of design
 * import re             # <<<<<<<<<<<<<<
 * from .mass_dict import nist_mass
 * from .base import ChemicalCompositionError, composition_factory
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_re, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_re, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":5
 * # Credit to Pyteomics - http://pythonhosted.org/pyteomics - for majority of design
 * import re
 * from .mass_dict import nist_mass             # <<<<<<<<<<<<<<
 * from .base import ChemicalCompositionError, composition_factory
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_nist_mass);
  __Pyx_GIVEREF(__pyx_n_s_nist_mass);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_nist_mass);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_mass_dict, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_nist_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nist_mass, __pyx_t_1) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glypy/composition/ccomposition.pyx":6
 * import re
 * from .mass_dict import nist_mass
 * from .base import ChemicalCompositionError, composition_factory             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ChemicalCompositionError);
  __Pyx_GIVEREF(__pyx_n_s_ChemicalCompositionError);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ChemicalCompositionError);
  __Pyx_INCREF(__pyx_n_s_composition_factory);
  __Pyx_GIVEREF(__pyx_n_s_composition_factory);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_composition_factory);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_base, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ChemicalCompositionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ChemicalCompositionError, __pyx_t_2) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_composition_factory); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_composition_factory, __pyx_t_2) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glypy/composition/ccomposition.pyx":31
 * # Forward Declaration
 * cdef:
 *     str _atom = r'([A-Z][a-z+]*)(?:\[(\d+)\])?([+-]?\d+)?'             # <<<<<<<<<<<<<<
 *     str _formula = r'^({})*$'.format(_atom)
 *     str _isotope_string = r'^([A-Z][a-z+]*)(?:\[(\d+)\])?$'
 */
  __Pyx_INCREF(__pyx_kp_s_A_Z_a_z_d_d);
  __Pyx_XGOTREF(__pyx_v_5glypy_11composition_12ccomposition__atom);
  __Pyx_DECREF_SET(__pyx_v_5glypy_11composition_12ccomposition__atom, __pyx_kp_s_A_Z_a_z_d_d);
  __Pyx_GIVEREF(__pyx_kp_s_A_Z_a_z_d_d);

  /* "glypy/composition/ccomposition.pyx":32
 * cdef:
 *     str _atom = r'([A-Z][a-z+]*)(?:\[(\d+)\])?([+-]?\d+)?'
 *     str _formula = r'^({})*$'.format(_atom)             # <<<<<<<<<<<<<<
 *     str _isotope_string = r'^([A-Z][a-z+]*)(?:\[(\d+)\])?$'
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s__9, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_5glypy_11composition_12ccomposition__atom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_5glypy_11composition_12ccomposition__formula);
  __Pyx_DECREF_SET(__pyx_v_5glypy_11composition_12ccomposition__formula, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glypy/composition/ccomposition.pyx":33
 *     str _atom = r'([A-Z][a-z+]*)(?:\[(\d+)\])?([+-]?\d+)?'
 *     str _formula = r'^({})*$'.format(_atom)
 *     str _isotope_string = r'^([A-Z][a-z+]*)(?:\[(\d+)\])?$'             # <<<<<<<<<<<<<<
 * 
 *     object isotope_pattern = re.compile(_isotope_string)
 */
  __Pyx_INCREF(__pyx_kp_s_A_Z_a_z_d);
  __Pyx_XGOTREF(__pyx_v_5glypy_11composition_12ccomposition__isotope_string);
  __Pyx_DECREF_SET(__pyx_v_5glypy_11composition_12ccomposition__isotope_string, __pyx_kp_s_A_Z_a_z_d);
  __Pyx_GIVEREF(__pyx_kp_s_A_Z_a_z_d);

  /* "glypy/composition/ccomposition.pyx":35
 *     str _isotope_string = r'^([A-Z][a-z+]*)(?:\[(\d+)\])?$'
 * 
 *     object isotope_pattern = re.compile(_isotope_string)             # <<<<<<<<<<<<<<
 *     object formula_pattern = re.compile(_formula)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compile); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_5glypy_11composition_12ccomposition__isotope_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_v_5glypy_11composition_12ccomposition_isotope_pattern);
  __Pyx_DECREF_SET(__pyx_v_5glypy_11composition_12ccomposition_isotope_pattern, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glypy/composition/ccomposition.pyx":36
 * 
 *     object isotope_pattern = re.compile(_isotope_string)
 *     object formula_pattern = re.compile(_formula)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compile); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_5glypy_11composition_12ccomposition__formula); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_v_5glypy_11composition_12ccomposition_formula_pattern);
  __Pyx_DECREF_SET(__pyx_v_5glypy_11composition_12ccomposition_formula_pattern, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glypy/composition/ccomposition.pyx":67
 *     return element_name
 * 
 * def parse_isotope_string(str label):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int iso
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5glypy_11composition_12ccomposition_1parse_isotope_string, NULL, __pyx_n_s_glypy_composition_ccomposition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_isotope_string, __pyx_t_2) < 0) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glypy/composition/ccomposition.pyx":482
 *         PyDict_Update(self, comp)
 * 
 *     cpdef double calc_mass(self, int average=False, charge=None, dict mass_data=nist_mass) except -1:             # <<<<<<<<<<<<<<
 *         cdef object mdid
 *         mdid = id(mass_data)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nist_mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 482, __pyx_L1_error)
  __pyx_k__8 = ((PyObject*)__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nist_mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 482, __pyx_L1_error)
  __pyx_k__8 = ((PyObject*)__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glypy/composition/ccomposition.pyx":521
 *         self._mass_args = None
 * 
 * Composition = CComposition             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Composition_2, ((PyObject *)__pyx_ptype_5glypy_11composition_12ccomposition_CComposition)) < 0) __PYX_ERR(0, 521, __pyx_L1_error)

  /* "glypy/composition/ccomposition.pyx":1
 * # cython: boundscheck=False             # <<<<<<<<<<<<<<
 * 
 * # Credit to Pyteomics - http://pythonhosted.org/pyteomics - for majority of design
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init glypy.composition.ccomposition", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init glypy.composition.ccomposition");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* StringJoin */
#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
